

//#include <cvc4/cvc4.h> // use this after CVC4 is properly installed
#include <cvc4/cvc4.h>
#include <sicstus/sicstus.h>
#include <unordered_map>
#include <vector>


using namespace std;
using namespace CVC4;

static ExprManager* em;
static SmtEngine* smtt;

vector<Expr> prolog_list_to_expr_vector(SP_term_ref list);
Type mk_type(SP_term_ref type);
void print_assertions(void);

extern "C" {
    static unordered_map<string, Kind> op_table;
    static unordered_map<string, Type> type_table;

    static unordered_map<SP_integer, Expr> *id_table;
    static SP_integer id_counter = 2;

    void init(void) {
        em = new ExprManager();
        smtt = new SmtEngine(em);
        id_counter = 2;

        smtt->setOption("produce-models", true);

        // use finite model finding,
        // use instantiation engine,
        // fully rate quantifiers if needed,
        // enumerate bounded integers,
        // perform first-order propagation on quantifiers.
        smtt->setOption("finite-model-find", true);
        smtt->setOption("fmf-inst-engine", true);
        smtt->setOption("full-saturate-quant", true);
        smtt->setOption("fmf-bound-int", true);
        // TODO: has this option been removed?
        // smtt->setOption("fo-prop-quant",true);

        // counter-example based instantiation
        // TODO: these options break our tests on linux64
        // smtt->setOption("enable-cbqi",true);
        // smtt->setOption("cbqi-recurse",true);

        // interactive mode = produce assertions for printing
        // smtt->setOption("interactive-mode", true);

        // rewrite certain modulus operations (helps to generate models)
        smtt->setOption("rewrite-divk",true);

        smtt->setLogic("ALL_SUPPORTED"); // Set the logic
        // smtt->setLogic("UFCLIAFS"); // Set the logic

        // AST output language is the easiest to parse
        smtt->setOption("output-language","LANG_AST");

        id_table = new unordered_map<SP_integer, Expr>();
        id_table->insert(make_pair<SP_integer,Expr>(0,em->mkConst(false)));
        id_table->insert(make_pair<SP_integer,Expr>(1,em->mkConst(true)));


        // types
        type_table["integer"] = em->integerType();
        type_table["boolean"] = em->booleanType();
        type_table["string"] = em->stringType();

        // predicates
        op_table["negation"] = kind::NOT;
        op_table["equal"] = kind::EQUAL;
        op_table["less"] = kind::LT;
        op_table["less_equal"] = kind::LEQ;
        op_table["greater"] = kind::GT;
        op_table["greater_equal"] = kind::GEQ;

        // arithmetic
        op_table["unary_minus"] = kind::UMINUS;
        op_table["add"] = kind::PLUS;
        op_table["minus"] = kind::MINUS;
        op_table["multiplication"] = kind::MULT;
        op_table["power_of"] = kind::POW;
        op_table["modulo"] = kind::INTS_MODULUS;

        // sets
        op_table["member"] = kind::MEMBER;
        op_table["subset"] = kind::SUBSET;
        op_table["union"] = kind::UNION;
        op_table["intersection"] = kind::INTERSECTION;
        op_table["set_subtraction"] = kind::SETMINUS;

        // couples
        op_table["couple"] = kind::TUPLE;

        // logic
        op_table["conjunct"] = kind::AND;
        op_table["disjunct"] = kind::OR;
        op_table["equivalence"] = kind::IFF;
        op_table["implication"] = kind::IMPLIES;
        op_table["if_then_else"] = kind::ITE;
        op_table["distinct"] = kind::DISTINCT;

        // quantifiers
        op_table["forall"] = kind::FORALL;
        op_table["exists"] = kind::EXISTS;

        // functions
        op_table["function"] = kind::APPLY_UF;
    }

    void reset(void) {
        //op_table.clear();
        //type_table.clear();
        //id_table.clear();
        //id_counter = 2;
        //em = new ExprManager();
        //smtt = new SmtEngine(em);
        init();
    }

    static SP_integer insert_id(Expr x) {
        SP_integer id = id_counter;
        id_table->insert(std::pair<SP_integer, Expr>( id,  x));
        id_counter++;

        //cout << id << ": " << x << endl;
        return id;
    }

    SP_integer mk_var(SP_term_ref type, char *varname) {
        assert(SP_term_type(type) != SP_TYPE_VARIABLE);
        Type t = mk_type(type);
        Expr x = em->mkVar(varname, t);
        return insert_id(x);
    }

    SP_integer mk_bounded_var(SP_term_ref type, char *varname) {
        assert(SP_term_type(type) != SP_TYPE_VARIABLE);
        Type t = mk_type(type);
        Expr x = em->mkBoundVar(varname, t);
        return insert_id(x);
    }

    SP_integer mk_empty_set(SP_term_ref type) {
        assert(SP_term_type(type) != SP_TYPE_VARIABLE);
        Type t = mk_type(type);
        Expr emptyset = em->mkConst(EmptySet(t));
        return insert_id(emptyset);
    }

    SP_integer mk_int_const(int val) {
        Expr x = em->mkConst(Rational(val));
        return insert_id(x);
    }

    SP_integer mk_bool_const(char *val) {
        if (0 == strcmp(val, "boolean_true")) {
            return 1;
        } else if (0 == strcmp(val, "boolean_false")) {
            return 0;
        } else {
            // explode
            assert(0);
        }
    }

    SP_integer mk_string_const(SP_atom content) {
        const char *str = SP_string_from_atom(content);
        Expr x = em->mkConst(String(str));
        return insert_id(x);
    }

    SP_integer mk_record_const(SP_term_ref type, SP_term_ref fields) {
        assert(SP_term_type(type) != SP_TYPE_VARIABLE);
        RecordType record_type = mk_type(type);
        Record record = record_type.getRecord();

        assert(SP_is_list(fields));
        vector<Expr> fields_expr = prolog_list_to_expr_vector(fields);

        Expr x = em->mkExpr(kind::RECORD, em->mkConst(record), fields_expr);
        return insert_id(x);
    }

    SP_integer mk_record_field(SP_integer record, SP_integer field_nr, SP_atom field_name) {
        Expr record_expr = id_table->at(record);

        Expr record_select = em->mkConst(RecordSelect(string(SP_string_from_atom(field_name))));
        Expr x = em->mkExpr(kind::RECORD_SELECT,record_select,record_expr);
        return insert_id(x);
    }

    SP_integer mk_op(char *opname, SP_integer arg1, SP_integer arg2) {
        Expr a = id_table->at(arg1);
        Expr b = id_table->at(arg2);

        string name = string(opname);
        if(op_table.count(name) > 0) {
            Kind op = op_table[name];
            Expr x = em->mkExpr(op, a, b);
            return insert_id(x);
        } else {
            // special cases that have to be Constructed
            // div in B does not equal div in SMT-LIB
            if("div" == name) {
                Expr ite_if = em->mkExpr(kind::GT, a, em->mkConst(Rational(0)));
                Expr a_div_b = em->mkExpr(kind::INTS_DIVISION, a, b);
                Expr ite_else_if = em->mkExpr(kind::GT, b, em->mkConst(Rational(0)));
                Expr a_div_b_plus_1 = em->mkExpr(kind::PLUS, a_div_b, em->mkConst(Rational(1)));
                Expr a_div_b_minus_1 = em->mkExpr(kind::MINUS, a_div_b, em->mkConst(Rational(1)));
                Expr ite_then = em->mkExpr(kind::ITE, ite_else_if, a_div_b_plus_1, a_div_b_minus_1);
                Expr x = em->mkExpr(kind::ITE, ite_if, a_div_b, ite_then);
                return insert_id(x);
            }

            // we should not reach this point
            assert(false);
            return 0;
        }
    }


    SP_integer mk_couple(SP_term_ref type, SP_integer arg1, SP_integer arg2) {
        /* introduced to handle special case of z3 / might occur here as well? */
        return mk_op("couple", arg1, arg2);
    }

    SP_integer mk_quantifier(char *opname, SP_term_ref list, SP_integer expr) {
        assert(SP_is_list(list));
        vector<Expr> args = prolog_list_to_expr_vector(list);

        Kind op = op_table[string(opname)];
        Expr var_list = em->mkExpr(kind::BOUND_VAR_LIST, args);
        Expr x = em->mkExpr(op,var_list,id_table->at(expr));

        return insert_id(x);
    }

    SP_integer mk_op_arglist(char *opname, SP_term_ref list) {
        assert(SP_is_list(list));
        vector<Expr> args = prolog_list_to_expr_vector(list);

        Kind op = op_table[string(opname)];
        Expr x = em->mkExpr(op, args);

        return insert_id(x);
    }

    SP_term_ref add_interpreter_constraint(SP_integer spe, SP_atom name, SP_integer timeout) {
        SP_term_ref return_value = SP_new_term_ref();
        try{
            // receives a new constraint with a given name,
            // adds it to the smt solver,
            // checks for a contradiction
            // returns unsat(ListOfNamesInUnsatCore), sat or unknown

            Expr x = id_table->at(spe);
            // cout << "expr x is " << x << endl;
            smtt->assertFormula(x);

            smtt->setTimeLimit(timeout);
            Result r = smtt->checkSat();

            if(r == Result::SAT) {
                SP_put_atom(return_value, SP_atom_from_string("sat"));
                return return_value;
            } else if(r == Result::UNSAT) {
                SP_term_ref list = SP_new_term_ref();

                int ret = SP_cons_functor(return_value, SP_atom_from_string("unsat"), 1, list);
                assert(ret);
                return return_value;

        }
    } catch(CVC4::LogicException& e) {
        SP_term_ref eterm = SP_new_term_ref();
        SP_term_ref arg = SP_new_term_ref();

        int ret = SP_put_atom(arg,SP_atom_from_string(e.getMessage().c_str()));
        assert(ret);
        ret = SP_cons_functor(eterm, SP_atom_from_string("logic_exception"),
                        1, arg);
        assert(ret);

        SP_raise_exception(eterm);
        return SP_atom_from_string("unknown"); // to make the compiler happy
    }
    catch(CVC4::TypeCheckingException& e) {
        SP_term_ref eterm = SP_new_term_ref();
        SP_term_ref msg = SP_new_term_ref();
        SP_term_ref expr = SP_new_term_ref();

        int ret = SP_put_atom(msg,SP_atom_from_string(e.getMessage().c_str()));
        assert(ret);
        ret = SP_put_atom(msg,SP_atom_from_string(e.getExpression().toString().c_str()));
        assert(ret);
        ret = SP_cons_functor(eterm, SP_atom_from_string("type_checking_exception"),
                        1, msg, expr);
        assert(ret);

        SP_raise_exception(eterm);
        return SP_atom_from_string("unknown"); // to make the compiler happy
    }
        SP_put_atom(return_value, SP_atom_from_string("unknown"));
        return return_value;
    }


    SP_atom query(SP_integer expr, SP_integer timeout) {
        Expr x = id_table->at(expr);
        // cout << "querying expr " << x << endl;
        try {
            // print_assertions();
            smtt->setTimeLimit(timeout);
            Result r = smtt->checkSat(x);

            if(r.isSat() == Result::SAT) {
                return SP_atom_from_string("sat");
            } else if(r.isSat() == Result::UNSAT) {
                return SP_atom_from_string("unsat");
            }
            return SP_atom_from_string("unknown");
        } catch(CVC4::LogicException& e) {
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.getMessage().c_str()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("logic_exception"),
                            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            return SP_atom_from_string("unknown"); // to make the compiler happy
        }
        catch(CVC4::TypeCheckingException& e) {
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref msg = SP_new_term_ref();
            SP_term_ref expr = SP_new_term_ref();

            int ret = SP_put_atom(msg,SP_atom_from_string(e.getMessage().c_str()));
            assert(ret);
            ret = SP_put_atom(msg,SP_atom_from_string(e.getExpression().toString().c_str()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("type_checking_exception"),
                            1, msg, expr);
            assert(ret);

            SP_raise_exception(eterm);
            return SP_atom_from_string("unknown"); // to make the compiler happy
        }
    }

    SP_atom get_string(SP_integer expr) {
        Expr x = id_table->at(expr);
        assert(x.isVariable());
        Expr xx = smtt->getValue(x);
        // cout << xx.toString() << "######" << endl;
        return SP_atom_from_string(xx.toString().c_str());
    }

    SP_integer mk_set(SP_term_ref list) {
        assert(SP_is_list(list));
        vector<Expr> exprs = prolog_list_to_expr_vector(list);

        if(exprs.size() == 1) {
            return insert_id(em->mkExpr(kind::SINGLETON,exprs.back()));
        } else {
            Expr last = em->mkExpr(kind::SINGLETON,exprs.back());

            exprs.pop_back();
            exprs.push_back(last);

            Expr x = em->mkExpr(kind::INSERT, exprs);
            return insert_id(x);
        }
    }

    void mk_sort(SP_atom name) {
        std::string setname = string(SP_string_from_atom(name));
        if(type_table.count(setname) == 0) {
            SortType type = em->mkSort(setname);
            type_table[setname] = type;
        }
    }

    void mk_sort_with_cardinality(SP_atom name, SP_integer card) {
        // the cardinality constraint is set up once the identifiers are given
        // we do not set it here (correct only for enumerated sets but not used elsewhere)
        mk_sort(name);
    }

    void push_frame(void) {
        smtt->push();
    }

    void pop_frame(void) {
        smtt->pop();
    }

    void setup_enumerated_set(SP_term_ref list) {
        // the set identifiers are distinct and define the cardinality
        assert(SP_is_list(list));
        vector<Expr> ids = prolog_list_to_expr_vector(list);

        if(ids.size() > 1) {
            smtt->assertFormula(em->mkExpr(kind::DISTINCT,ids));
        }
        smtt->assertFormula(em->mkExpr(kind::CARDINALITY_CONSTRAINT,ids.front(),em->mkConst(Rational(ids.size()))));

        // each is enforced to be equal to a corresponding uninterpreted constant
        int num = 0;

        for (vector<Expr>::iterator it = ids.begin(); it != ids.end(); ++it) {
            Expr uc = em->mkConst(UninterpretedConstant(em->getType(*it),num++));
            Expr equal = em->mkExpr(kind::EQUAL,*it, uc);
            smtt->assertFormula(equal);
        }
    }

    SP_integer conjoin_negated_state(SP_term_ref list, SP_integer old_expr) {
        assert(SP_is_list(list));
        vector<Expr> exprs = prolog_list_to_expr_vector(list);
        vector<Expr> negated_exprs;

        for (vector<Expr>::iterator it = exprs.begin(); it != exprs.end(); ++it) {
            Expr equal = em->mkExpr(kind::EQUAL,*it, smtt->getValue(*it));
            Expr negated = em->mkExpr(kind::NOT, equal);
            negated_exprs.push_back(negated);
        }

        if(negated_exprs.size() == 1) {
            Expr new_expr = em->mkExpr(kind::AND, negated_exprs.front(), id_table->at(old_expr));
            return insert_id(new_expr);
        } else {
            Expr negstate = em->mkExpr(kind::OR, negated_exprs);
            Expr new_expr = em->mkExpr(kind::AND, negstate, id_table->at(old_expr));
            return insert_id(new_expr);
        }
    }
}

Type mk_type(SP_term_ref type) {
    int ret;
    // simple atom types 'integer' and 'boolean'
    if(SP_is_atom(type)) {
        SP_atom as_atom;
        ret = SP_get_atom(type,&as_atom);
        assert(ret);
        const char *typestr = SP_string_from_atom(as_atom);
        Type to_return = type_table[string(typestr)];
        if(type_table.count(typestr) == 0) {
            SP_printf("Type missing in table: %s\n",typestr);
        } else {
            return to_return;
        }

    }

    SP_atom name;
    int arity;
    ret = SP_get_functor(type,&name,&arity);
    assert(ret);

    // sets
    const char *atom = SP_string_from_atom(name);
    if(strcmp(atom, "set") == 0) {
        SP_term_ref arg = SP_new_term_ref();
        ret = SP_get_arg(1,type,arg);
        assert(ret);
        assert(SP_term_type(arg) != SP_TYPE_VARIABLE);
        Type subtype = mk_type(arg);
        return em->mkSetType(subtype);
    }
    if(strcmp(atom, "global") == 0) {
        SP_term_ref arg = SP_new_term_ref();
        ret = SP_get_arg(1,type,arg);
        assert(ret);
        SP_atom as_atom;
        ret = SP_get_atom(arg,&as_atom);
        assert(ret);
        const char *typestr = SP_string_from_atom(as_atom);
        return type_table[string(typestr)];
    }
    if(strcmp(atom, "couple") == 0) {
        SP_term_ref arg1 = SP_new_term_ref();
        ret = SP_get_arg(1,type,arg1);
        assert(ret);
        SP_term_ref arg2 = SP_new_term_ref();
        ret = SP_get_arg(2,type,arg2);
        assert(ret);
        assert(SP_term_type(arg1) != SP_TYPE_VARIABLE);
        assert(SP_term_type(arg2) != SP_TYPE_VARIABLE);

        Type subtype1 = mk_type(arg1);
        Type subtype2 = mk_type(arg2);
        return em->mkTupleType({subtype1,subtype2});
    }
    if(strcmp(atom, "record") == 0) {
        // first argument is a list of field(Name,Type)
        SP_term_ref list_of_fields = SP_new_term_ref();
        ret = SP_get_arg(1,type,list_of_fields);
        assert(ret);

        // iterate over the list and place fields in vector
        vector<pair<string,Type>> fields;
        SP_term_ref head = SP_new_term_ref();
        SP_term_ref tail = SP_new_term_ref();
        while (SP_get_list(list_of_fields, head, tail)) {
            SP_term_ref name = SP_new_term_ref();
            ret = SP_get_arg(1,head,name);
            assert(ret);
            ret = SP_get_arg(2,head,type);
            assert(ret);

            assert(SP_term_type(type) != SP_TYPE_VARIABLE);

            Type field_type = mk_type(type);

            SP_atom as_atom;
            ret = SP_get_atom(name,&as_atom);
            assert(ret);
            const char *namestr = SP_string_from_atom(as_atom);
            string field_name = string(namestr);

            fields.push_back(make_pair(field_name,field_type));

            list_of_fields = tail;
        }

        return em->mkRecordType(Record(fields));
    }

    // TODO: further types?
    printf("Unsupported type: %s\n",atom);
    assert(false);
}

void print_assertions(void) {
    vector<Expr> assertions = smtt->getAssertions();
    for (vector<Expr>::iterator it = assertions.begin(); it != assertions.end(); ++it) {
        cout << *it << endl;
    }
}

vector<Expr> prolog_list_to_expr_vector(SP_term_ref list) {
    assert(SP_is_list(list));

    vector<Expr> args;
    SP_term_ref head = SP_new_term_ref();
    SP_term_ref tail = SP_new_term_ref();
    while (SP_get_list(list, head, tail)) {
        assert(SP_is_integer(head));
        SP_integer id;
        SP_get_integer(head, &id);

        Expr x = id_table->at(id);
        args.push_back(x);
        list = tail;
    }
    return args;
}
