#include <stdlib.h>
#include <stdio.h>
#include "lib/include/zmq.h"
#include "lib/include/czmq.h"
#include "helpers.h"

void drop_frame(zmsg_t *msg) {
    zframe_t *frame = zmsg_pop(msg);
    assert(frame);
    zframe_destroy(&frame);
}

int get_number(zmsg_t *msg) {
    char *nr_s = zmsg_popstr(msg);
    int nr;
    int rc = sscanf(nr_s,"%d",&nr);
    assert(rc==1);
    free(nr_s);

    return nr;
}

Chunk get_chunk(zmsg_t *msg) {
    Chunk res;

    zframe_t *chunk_f = zmsg_pop(msg);
    int chunk_size = zframe_size(chunk_f);
    char *chunk = zframe_strdup(chunk_f);
    zframe_destroy(&chunk_f);
    
    res.data = chunk;
    res.size = chunk_size;

    return res;
}

ChunkArray get_chunk_array(zmsg_t *msg) {
    ChunkArray res;
    res.size = get_number(msg);
    res.chunks = calloc(sizeof(Chunk), res.size);
    
    int i;
    for (i = 0; i < res.size; i++) {
        res.chunks[i] = get_chunk(msg);
    }

    return res;
}

State get_state(zmsg_t *msg) {
    return get_chunk_array(msg);
}


InitialResponse get_init_response(zmsg_t *msg) {
    InitialResponse res = {0};
    drop_frame(msg); // message type
    drop_frame(msg); // ID

    char *submsg_header;
    while ((submsg_header = zmsg_popstr(msg)) != NULL) {
        enum SubMessageType header = atoi(submsg_header);
        free(submsg_header);

        int len = get_number(msg);

        switch (header) {
            case initial_state:
                res.initial_state = get_state(msg); break;
            case transition_group_list:
                res.transition_groups = get_chunk_array(msg); break;
            case variables_name_list:
                res.variables = get_chunk_array(msg); break;
            case variables_type_list:
                res.variable_types = get_chunk_array(msg); break;
            case state_label_matrix:
                res.state_labels = get_matrix(msg); break;
            case may_write_matrix:
                res.may_write = get_matrix(msg); break;
            case must_write_matrix:
                res.must_write = get_matrix(msg); break;
            case reads_action_matrix:
                res.reads_action = get_matrix(msg); break;
            case reads_guard_matrix:
                res.reads_guard = get_matrix(msg); break;
            default:
                for (; len > 0; len--) {
                    drop_frame(msg);
                }
        }
    }
    

    return res;
}


MatrixRow get_row(zmsg_t *msg) {
    MatrixRow res;
    res.transition_group = get_chunk(msg);
    res.variables = get_chunk_array(msg);

    return res;
}


Matrix get_matrix(zmsg_t *msg) {
    Matrix m;
    m.nr_rows = get_number(msg);

    m.rows = (MatrixRow*) calloc(m.nr_rows, sizeof(MatrixRow));

    int i;
    for (i = 0; i < m.nr_rows; i++) {
        m.rows[i] = get_row(msg);
    }

    return m;
}

void put_state(zmsg_t *msg, State s) {
    zmsg_addstrf(msg,"%ld",s.size);
    int i;
    for (i=0; i < s.size; i++) {
        zframe_t *frame = zframe_new(s.chunks[i].data, s.chunks[i].size);
        zmsg_append(msg,&frame);
    }
}


void destroy_chunk_array(ChunkArray *arr) {
    int i;
    for (i = 0; i < arr->size; i++) {
        free(arr->chunks[i].data);
    }
    free(arr->chunks);
    arr->chunks = NULL;
    *arr = (ChunkArray) {0};
}

void destroy_state(State *s) {
    destroy_chunk_array(s);
}

void destroy_chunk(Chunk *norris) {
    free(norris->data);
    *norris = (Chunk) {0};
}

void destroy_matrix_row(MatrixRow row) {
    destroy_chunk(&(row.transition_group));
    destroy_chunk_array(&(row.variables));
}

void destroy_matrix(Matrix *m) {
    int i;
    for (i = 0; i < m->nr_rows; i++) {
        destroy_matrix_row(m->rows[i]);
    }
    free(m->rows);
    *m = (Matrix) {0};
}

void destroy_initial_response(InitialResponse *resp) {
    destroy_state(&(resp->initial_state));

    destroy_chunk_array(&(resp->transition_groups));
    destroy_chunk_array(&(resp->variables));

    destroy_matrix(&(resp->state_labels));

    destroy_matrix(&resp->may_write);
    destroy_matrix(&resp->must_write);
    *resp = (InitialResponse) {0};
}
