:- module(prettyprinter, [pretty_print_error/1]).
:- use_module(plspec_logger).


pretty_print_error(fail(postcondition_violated(matched_pre(Pre),
                                               violated_post(Post),
                                               value(Val)))) :- !,
    log(error, 'A postcondition was violated!', []),
    log(error, 'The matched precondition was "~w".', [Pre]),
    log(error, 'However, the postcondition "~w" does not hold.', [Post]),
    translate_term_into_atom_with_max_depth(Val,ValLimited),
    log(error, 'The offending value was: ~p.', [ValLimited]),
    (debug_argument(Post,'$root',[Val]) -> true ; true).
pretty_print_error(fail(prespec_violated(specs(PreSpecs), values(Vals),
                                         location(Functor)))) :- !,
    log(error, 'No precondition was matched in ~w.', [Functor]),
    log(error, 'Specified preconditions were: ~w.', [PreSpecs]),
    translate_term_into_atom_with_max_depth(Vals,ValsLimited),
    log(error, 'However, none of these is matched by: ~p.', [ValsLimited]),
    (debug_arguments(PreSpecs,Functor,Vals) -> true ; true).
pretty_print_error(fail(spec_violated(spec(T), value(V), location(Location)))) :- !,
    log(error, 'An invariant was violated in ~w.', [Location]),
    log(error, 'The spec was: ~w.', [T]),
    log(error, 'However, the value was bound to: ~w.', [V]).
pretty_print_error(fail(spec_not_found(spec(Spec)))) :- !,
    %% TODO: not all failures include a location
    log(error, 'Spec "~w" was not found.', [Spec]).
pretty_print_error(fail(spec_not_found(spec(Spec), location(Location)))) :- !,
    log(error, '~nA spec for ~w was not found.', [Location]),
    log(error, 'Spec "~w" was not found.', [Spec]).
pretty_print_error(X) :-
    log(error, 'plspec raised an error term that is unhandled.', []),
    log(error, '~p.', [X]).

:- use_module(validator,[valid/2,spec_indirection/2]).
:- use_module(library(lists),[include/3]).


% debug a single argument and a single specification
debug_argument(SpecDef,Position,Arg) :-
    spec_indirection(SpecDef,SpecRHS),
    debug_argument(SpecRHS,Position,Arg).
debug_argument(one_of(List),Position,Arg) :- !,
    include(prettyprinter:spec_top_level_match(Arg),List,Matches),
    (Matches==[]
       -> get_functor(Arg,FA,NA),
          log(error,'Value at ~w with functor ~w/~w does not match any case of ~w.',[Position,FA,NA,List])
     ; Matches = [OneMatch]
       -> debug_argument(OneMatch,Position,Arg)
     ; fail).
debug_argument(compound(Spec),Position,Arg) :- !, functor(Spec,F,N),
     get_functor(Arg,FA,NA),
     ( F=FA,N=NA
      -> Spec =.. [F|Specs], Arg =.. [F|Args],
         l_debug_arguments(Specs,F/N,1,Args)
      ; log(error,'Value at ~w with functor ~w/~w does not match functor ~w/~w of ~w.',[Position,FA,NA,F,N,Spec])
     ).
debug_argument(list(Spec),Position,Arg) :- !,
     ( Arg = [] -> fail
       ; Arg = [H|T] ->
          (debug_argument(Spec,Position,H) -> true
           ; debug_argument(list(Spec),list(Position),T)
          )
      ; get_functor(Arg,FA,NA),
        log(error,'Value at ~w with functor ~w/~w ist not a list.',[Position,FA,NA])
     ).
debug_argument(atom(X),Position,Arg) :- !,
     Arg \== X,
     get_functor(Arg,FA,NA),
     log(error,'Value at ~w with functor ~w/~w does not match atom ~w.',[Position,FA,NA,X]).
debug_argument(ground,Position,Arg) :- !,
     term_variables(Arg,Vars), Vars \==[],
     get_functor(Arg,FA,NA),
     log(error,'Value at ~w with functor ~w/~w does have variables ~w and is not ground.',[Position,FA,NA,Vars]).
debug_argument(Basic,Position,Arg) :- basic_type(Basic), !,
     \+ valid(Basic, Arg),
     get_functor(Arg,FA,NA),
     log(error,'Value at ~w with functor ~w/~w ist not ~w.',[Position,FA,NA,Basic]).


% basic plspec types:
basic_type(atom).
basic_type(atomic).
basic_type(float).
basic_type(integer).
basic_type(nonvar).
basic_type(number).
basic_type(var).

% check at the top-level whether a spec potentially matches an argument
spec_top_level_match(Arg,R) :- var(Arg),!,R==var.
spec_top_level_match(X,atom(X)) :- !.
spec_top_level_match(Arg,compound(Spec)) :- functor(Spec,F,N), functor(Arg,F,N).
spec_top_level_match(X,T) :- basic_type(T), valid(T,X),!.


% debug a list of argument and a list of specification alternatives
debug_arguments([SingleSpec],Position,Args) :-
    % currently we only support debugging a single list
    l_debug_arguments(SingleSpec,Position,1,Args).

% debug a list of specs and a list of (supposedly) matching arguments
l_debug_arguments([Spec1|TSpec],Position,ArgNr,[Arg1|TArgs]) :-
     (valid(Spec1, Arg1)
       -> A1 is ArgNr+1,
          l_debug_arguments(TSpec,Position,A1,TArgs)
       ;  get_functor(Arg1,FA,NA),
          log(error,'Argument ~w of ~w with functor ~w/~w does not match spec ~w.',[ArgNr,Position,FA,NA,Spec1]),
          \+ basic_type(Spec1), % otherwise we will just repeat the error message
          debug_argument(Spec1,Position,Arg1)
     ).
% TO DO: treat things like one_of pattern matches

% a safe version of functor/3 which also works with variables
get_functor(X,F,N) :- var(X),!,F='$VAR',N = -1.
get_functor(X,F,N) :- functor(X,F,N).

:- use_module(library(codesio), [write_term_to_codes/3]).
translate_term_into_atom_with_max_depth(Term,Atom) :-
               translate_term_into_atom_with_max_depth(Term,5,Atom).
translate_term_into_atom_with_max_depth(Term,_,Atom) :- atomic(Term),!,Atom=Term.
translate_term_into_atom_with_max_depth(Term,Limit,Atom) :-
                write_term_to_codes(Term,Temp,[quoted(true),numbervars(true),max_depth(Limit)]),
                atom_codes(Atom,Temp).
