:- module(b_to_cnf, [b_to_cnf/3]).

:- use_module(library(atts)).

:- use_module(probsrc(bsyntaxtree),[get_texpr_expr/2,create_texpr/4]).
:- use_module(probsrc(error_manager), [add_error_fail/3]).
:- use_module(probsrc(translate), [translate_bexpression/2]).

% the sat solver can not just work on prob's own variables,
% as it relies on replacing them by integers using unification
% this is later undone, however coroutines will be triggered on the way
% to avoid side effects, we instead attach a dedicated variables
% to each B value variable using an attribute
:- attribute corresponding_sat_var/1.
get_corresponding_var(BVar,CorrespondingVar) :-
  get_atts(BVar,+(corresponding_sat_var(CorrespondingVar))), !.
get_corresponding_var(BVar,CorrespondingVar) :-
  put_atts(BVar,+(corresponding_sat_var(CorrespondingVar))).

verify_attributes(_, Value, []) :-
  Value == pred_true, !.
verify_attributes(_, Value, []) :-
  Value == pred_false, !.
verify_attributes(_, Value, []) :-
  add_error_fail(b_to_cnf_verify_attributes,'Tried to bind variable with attached SAT variable to non-boolean value', Value).

b_to_cnf(In,State,Out) :-
  get_texpr_expr(In,Expr),
  b_to_cnf_aux(Expr,State,Out), !.
b_to_cnf(In,_,_) :-
  translate_bexpression(In,PPIn), !,
  add_error_fail(b_to_cnf,'CNF conversion failed', PPIn).

b_to_cnf_aux(identifier(Name),State,[[Res]]) :-
  member(bind(Name,Res),State).
b_to_cnf_aux(value(X),_,[[1]]) :- X==pred_true.
b_to_cnf_aux(value(X),_,[[2]]) :- X==pred_false.
b_to_cnf_aux(value(X),_,[[VarForSatSolver]]) :-
  var(X), get_corresponding_var(X,VarForSatSolver),
  when(ground(VarForSatSolver),(integer(VarForSatSolver) -> true ; X=VarForSatSolver)).

b_to_cnf_aux(equal(A,B),State,CNF) :-
  ground(B),
  get_texpr_expr(B,value(pred_true)),
  b_to_cnf(A,State,CNF).
b_to_cnf_aux(equal(A,B),State,CNF) :-
  ground(B),
  get_texpr_expr(B,value(pred_false)),
  create_texpr(value(pred_true),boolean,[],True),
  create_texpr(equal(A,True),pred,[],EqualTrue),
  b_to_cnf_aux(negation(EqualTrue),State,CNF).

b_to_cnf_aux(negation(A),State,[[Res]]) :-
  b_to_cnf(A,State,[[V]]),
  (var(V)     -> Res = neg(V) ;
   V = neg(X) -> Res = X). % negated literal

b_to_cnf_aux(implication(A,B),State,CNF) :-
  create_texpr(negation(A),pred,[],NotA),
  b_to_cnf_aux(disjunct(NotA,B),State,CNF).

b_to_cnf_aux(disjunct(D1,D2),State,Res) :-
  get_texpr_expr(D1,conjunct(A,B)), !,
  create_texpr(disjunct(A,D2),pred,[],DJ1),
  create_texpr(disjunct(B,D2),pred,[],DJ2),
  b_to_cnf_aux(conjunct(DJ1,DJ2),State,Res).
b_to_cnf_aux(disjunct(D1,D2),State,Res) :-
  get_texpr_expr(D2,conjunct(A,B)), !,
  create_texpr(disjunct(D1,A),pred,[],DJ1),
  create_texpr(disjunct(D1,B),pred,[],DJ2),
  b_to_cnf_aux(conjunct(DJ1,DJ2),State,Res).

b_to_cnf_aux(disjunct(D1,D2),State,[Res]) :-
  b_to_cnf(D1,State,[ResD1]),
  b_to_cnf(D2,State,[ResD2]),
  append(ResD1,ResD2,Res).

b_to_cnf_aux(conjunct(A,B),State,Res) :-
  b_to_cnf(A,State,RA),
  b_to_cnf(B,State,RB),
  append(RA,RB,Res).
