# Makefile to support generation of extensions to ProB that are shared libraries

# Before importing this file, you should set the following variables:
#
# EXTNAME (mandatory)         The name of the extension.
#                             It is e.g. used for the file names or SICStus' resource name
# csrcs (optional)            List of C source files that are compiled into the extension
#                             and have no dependency to SICSTUS
# cppsrcs (optional)          List of C++ source files without dependency to SICSTUS
# csrcs_sicsdep (optional)    List of C source files with dependency to SICSTUS
# cppsrcs_sicsdep (optional)  List of C++ source files with dependency to SICSTUS
# shobjs (optional)           List of additional objects that should be compiled into the share lib
#
# You can use the following variables after the import (e.g. to add further dependencies or
# set additional compile flags):
#
# shlib              Name of the generated library
# target_lib         Generated file in the target directory (ProB's lib directory)
# prolog_base_file   Name of the Prolog file that defines the extension's interface to SICSTUS
# SICSCFLAGS         Additional compiler flags used to compile objects with SICSTUS dependencies
#                    for this library
# shcobjs            List of object files that correspond to C files
# shcppobjs          List of object files that correspond to C++ files
# shcobjs_sicsdep    List of object files that correspond to C files with SICSTUS dependencies
# shcppobjs_sicsdep  List of object files that correspond to C++ files with SICSTUS dependencies
# shobjs_sicsdep     List of object files with SICSTUS dependencies
# allshobjs          List of all object files that are compiled into the shared library

# MakefileCommons is in the same directory as this file but this file
# is imported by makefiles in sub-directories and the path is relative
# to the importer.
include ../MakefileCommons

################################################################################
# some common targets to compile shared libraries (without DLLs,
# for DLLs, see cross-compilation below

shlib=$(EXTNAME).$(shext)

target_lib = $(target_dir)/$(shlib)
prolog_base_file = $(EXTNAME).pl

SICSCFLAGS=-DSP_RESNAME=$(EXTNAME) $(SICSTUSCFLAGS)
#SPLFRFLAGS=--keep --verbose
MAPFILE=$(EXTNAME)_glue.mapfile
# required on MacPro for C++ projects counter, myheap:; make clean solved issue
#STDLIBOPT=-stdlib=libstdc++

sh_glue_obj = $(EXTNAME)_glue.o
shcobjs = $(csrcs:.c=.o)
shcobjs_sicsdep = $(csrcs_sicsdep:.c=.o) $(sh_glue_obj)

# C++ files might use either .cpp or .cc as extensions
tshcppobjs = $(cppsrcs:.cpp=.o)
shcppobjs = $(tshcppobjs:.cc=.o)
tshcppobjs_sicsdep = $(cppsrcs_sicsdep:.cpp=.o)
shcppobjs_sicsdep = $(tshcppobjs_sicsdep:.cc=.o)

shobjs_sicsdep = $(shcobjs_sicsdep) $(shcppobjs_sicsdep)
all_compiled_shobjs = $(shcobjs) $(shcppobjs) $(shobjs_sicsdep)
allshobjs = $(all_compiled_shobjs) $(shobjs)

.PHONY: install clean clean_orig clean_so clean_dll

install: $(target_lib)
$(target_lib): $(shlib)
	mkdir -p $(target_dir)
	cp -f $(shlib) $(target_lib)

# use splfr to link all object files
$(shlib): $(MAPFILE) $(allshobjs)
	$(CC) $(SICSTUSCFLAGS) $(SHOPT) --shared -Wl$(call VERSION_SCRIPT,$(MAPFILE)),-lm $(allshobjs) $(spld_lopts) -lstdc++ $(STDLIBOPT) -o $(shlib)

# rule to generate the *_glue files for Prolog extensions
$(EXTNAME)_glue.c $(EXTNAME)_glue.h $(MAPFILE): $(prolog_base_file)
	$(checksicstusdir)
	$(splfr) $(SPLFRFLAGS) --nocompile --resource=$(EXTNAME) --source=$(prolog_base_file) --cflags=-Wall

# The source files with SICSTUS dependencies are compiled by the standard implicit rules
# with additional compile flags
$(shcobjs_sicsdep): CFLAGS+=$(SICSCFLAGS) -Wall
$(shcppobjs_sicsdep): CXXFLAGS+=$(SICSCFLAGS) -Wall

# each object file depends on all header files. This is not optimal but safe
$(allshobjs): $(header)
# all objects for shared libraries must be position independent
$(allshobjs): CFLAGS+= $(FPIC)
$(allshobjs): CXXFLAGS+= $(FPIC)

# clean files that are generated by the shared library build
clean_so:
	rm -f *.bundle *.so
	find . -name "*.o" -print0 | xargs -0 rm -f

# default clean: do clean_so and clean_dll
# adding intermediate rule clean_orig to allow easy overwriting by
# referencing clean_orig as dependency
clean_orig: clean_so clean_dll
clean: clean_orig


################################################################################
# cross-compilation of Windows-DLLs in Linux with mingw
# - you must set the SICSTUSDIR environment variable to use this
# - you must install mingw
#
# To use this, the variables EXTNAME and csrc or cppsrc must be set.
# Conventions / prerequisites:
#  - name of the DLL to generate: "EXTNAME.dll"
#  - name of the extension's Prolog file where the foreign interface is defined: "EXTNAME.pl"
#  - resource name of the extension: "EXTNAME"
#  - "csrcs" consists of the C files that must be compiled for the DLL
#  - "cppsrcs" consists of the C++ files that must be compiled for the DLL
#  - additional object files can be specified in "dllobjects"
#  - "dllprerequesite" can contain optional prerequisites for the DLL
#  - "DLLCOPTS" resp. "DLLCPPOPTS" can contain optionally C resp. C++ compiler flags

GLUEOBJFILE=$(EXTNAME)_glue.obj
DLLFILE=$(EXTNAME).dll

dllcobjs = $(csrcs:.c=.obj)
dllcobjs_sicsdep = $(csrcs_sicsdep:.c=.obj)

# C++ files might use either .cpp or .cc as extensions
tdllcppobjs = $(cppsrcs:.cpp=.obj)
dllcppobjs = $(tdllcppobjs:.cc=.obj)
tdllcppobjs_sicsdep = $(cppsrcs_sicsdep:.cpp=.obj)
dllcppobjs_sicsdep = $(tdllcppobjs_sicsdep:.cc=.obj)

dllobjs_sicsdep = $(dllcobjs_sicsdep) $(dllcppobjs_sicsdep)
alldllobjs = $(dllcobjs) $(dllcppobjs) $(dllobjs_sicsdep) $(dllobjects)


$(DLLFILE): $(MAPFILE) $(alldllobjs) $(GLUEOBJFILE)
	$(checksicstusdir)
	$(MINGWCPPC) $(SICSTUSCFLAGS) --shared -Wl$(call VERSION_SCRIPT,$(MAPFILE)),-lm $(alldllobjs) $(GLUEOBJFILE) -static-libgcc -static-libstdc++ -o $(DLLFILE)

# rule to compile C++ files for DLLs (with dependency to SICStus)
$(dllcppobjs_sicsdep): %.obj: %.cpp
	$(checksicstusdir)
	$(MINGWCPPC) $(SICSCFLAGS) $(DLLCPPOPTS) -c $(@:.obj=.cpp) -o $(@)

# rule to compile C++ files for DLLs (without dependency to SICStus)
$(dllcppobjs):  %.obj: %.cpp
	$(MINGWCPPC) $(DLLCPPOPTS) -c $(@:.obj=.cpp) -o $(@)

# rule to compile C files for DLLs (with dependency to SICStus)
$(dllcobjs_sicsdep) $(GLUEOBJFILE): %.obj: %.c
	$(checksicstusdir)
	$(MINGWCC) $(SICSCFLAGS) $(DLLCOPTS) -c $(@:.obj=.c) -o $(@)

# rule to compile C files for DLLs (with dependency to SICStus)
$(dllcobjs): %.obj: %.c
	$(MINGWCC) $(DLLCOPTS) -c $(@:.obj=.c) -o $(@)

.PHONY: crosscompile
crosscompile: $(DLLFILE) $(crosscompile_artifacts)
	cp $(DLLFILE) $(crosscompile_artifacts) $(target_dir)
crosscompile: shext=dll
crosscompile: objext=obj

clean_dll:
	rm -f $(alldllobjs) $(DLLFILE)
	find . \( -name "$(EXTNAME)_glue.*" -o -name "*.obj" \) -delete
