#include <assert.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hash.h"
#include "hashtrie.h"
#include "shm_common.h"
#include "shm_srv.h"

#include "assert_helper.h"


#ifndef UPDATE_INFOS
static void update_infos(void *infos, void const * const new_infos) {
    char *old = (char*) infos;
    char const * const new = (char const * const) new_infos;
    int i;
    for (i = 0; i < INFO_LEN; i++) {
        old[i] |= new[i];
    }
}
#else
extern void update_infos(void *infos, void const * const new_infos);
#endif


hashtrie get_trie(const struct semap_env senv,
                  const key_t start) {
    // the start keys are chosen somewhat arbitrarily
    struct shm_env *env = init_shm_common(senv,
                                          10000000,     // size_t elements_per_nodes_block
                                          10000000,     // size_t elements_per_digests_block
                                          start,        // key_t meta_info
                                          start+1,      // key_t digests_start
                                          start+20000); // key_t nodes_start
    return (hashtrie) {env};
}

void teardown_hashtrie(hashtrie *t,
                       const int server,
                       const int id) {
    if (server) {
        shm_srv_cleanup(&(t->shm->semaphores), id);
    }
    shm_cleanup(t->shm, server);
    free(t->shm);

    *t = (hashtrie) {0};
}


static int calculate_pos(char const * const kkey, const int depth) {
    unsigned char const * const key = (unsigned char const * const) kkey;
    // pattern repeats after 8 times
    int offset = depth / 8;
    int moddepth = depth % 8;

    unsigned char const * const offi = key + 5 * offset;

    uint64_t a = 0;
    int i;
    for (i = 0; i < 5; i++) {
       a |= ((uint64_t) offi[i]) << 8*i;
    }

    uint64_t b = a >> (moddepth * 5);
    int x = b & 31;
    return x;
}


// TODO
// this should be > 0
// however, this means refactoring all code calling this
// and depending on the implication that digests include
// empty positions
static int is_digest(tNode const * const node, const int pos) {
    return node->indices[pos] >= 0;
}



static int64_t skip_first_layers(const char key[HASH_LEN], int *branch) {
    *branch = calculate_pos(key, 0);

    int64_t index = *branch;
    int i;
    for (i = 1; i < DEPTH_START; i++) {
        index |= calculate_pos(key, i) << i * 5;
    }
    
    return index;
}


size_t trie_get_checked_states_count(const hashtrie t) {
    struct hash_meta_infos x = get_meta_infos(t.shm);
    return x.checked_states;
}

size_t trie_count_elements(const hashtrie t) {
    struct hash_meta_infos x = get_meta_infos(t.shm);
    return x.pos_hashes;
}

char *trie_iterate(const hashtrie t, size_t *blocklen) {
    struct hash_meta_infos x = get_meta_infos(t.shm);
    lock_digests_read(t.shm);
    char *res = get_shm_digests(t.shm, x.pos_hashes);
    release_digests_read(t.shm);
    *blocklen = x.pos_hashes * HASH_INFO;
    return res;
}

int get_max_depth(const hashtrie t) {
    struct hash_meta_infos x = get_meta_infos(t.shm);
    return x.max_depth;
}

static inline int is_empty_pos(tNode const * const node, const int pos) {
    // if it's empty then everything is at max value
    return !(node->indices[pos]);
    //return node->indices[pos] == 0xffffffffffffffff;
}

static size_t h_iterate(const hashtrie t, tNode const * const node) {
    size_t i, sum = 0;
    for (i = 0; i < 32; i++) {
        if (is_empty_pos(node, i))
            continue;
        if (is_digest(node, i)) {
            sum++;
        } else {
            sum += h_iterate(t, lookup_node(t.shm, node->indices[i]));
        }
    }
    return sum;
}

/* might be useful for debugging */
size_t trie_count_via_iterate(const hashtrie t) {
    size_t res = 0;
    int i;
    for (i = 0; i < PRE_INIT_NODES; i++) {
        res += h_iterate(t, lookup_node(t.shm, i));
    }
    return res;
}

inline static int is_checked(void const * const infos) {
    unsigned char *ptr = (unsigned char*) infos;
    // TODO: implementation
    // could be: first bit indicates checked/unchecked status
    return (ptr[0] & 0xff) == 0xff;
}


static int64_t h_lookup_index(const hashtrie t,
                              const int64_t current_node_index,
                              char const * const key,
                              const int depth,
                              const int branchpos) {
    tNode *node = lookup_node(t.shm, current_node_index);
    int pos = calculate_pos(key, depth);
    if (!is_digest(node, pos)) {
        return h_lookup_index(t, node->indices[pos], key, depth+1, branchpos);
    } else {
        lock_nodes_read(t.shm, branchpos);
        if (is_digest(node, pos)) {
            
            if (is_empty_pos(node, pos)) {
                release_nodes_read(t.shm, branchpos);
                return 0;
            } else {
                char *ref = lookup_digest(t.shm, node->indices[pos]);
                release_nodes_read(t.shm, branchpos);
                if (memcmp(key, ref, HASH_LEN) != 0) {
                    return 0;
                } else {
                    return node->indices[pos];
                }
            }
        } else {
            release_nodes_read(t.shm, branchpos);
            return h_lookup_index(t, node->indices[pos], key, depth+1, branchpos);
        }
    }
}

int64_t lookup_index(const hashtrie t,
                     char const * const key) {
    int branch;
    int64_t node = skip_first_layers(key, &branch);
    return h_lookup_index(t, node, key, DEPTH_START, branch);
}

static void *h_get(const hashtrie t, const int64_t id, char const * const key, const int depth, const int branchpos) {
    tNode *node = lookup_node(t.shm, id);
    int pos = calculate_pos(key, depth);
    if (!is_digest(node, pos)) {
        return h_get(t, node->indices[pos], key, depth+1, branchpos);
    } else {
        lock_nodes_read(t.shm, branchpos);
        if (is_digest(node, pos)) {
            if (is_empty_pos(node, pos)) {
                release_nodes_read(t.shm, branchpos);
                return NULL;
            } else {
                char *ref = lookup_digest(t.shm, node->indices[pos]);
                release_nodes_read(t.shm, branchpos);
                if (memcmp(key, ref, HASH_LEN) != 0) {
                    return NULL;
                } else {
                    return ref + HASH_LEN;
                }
            }
        } else {
            release_nodes_read(t.shm, branchpos);
            return h_get(t, node->indices[pos], key, depth+1, branchpos);
        }
    }
}

void *get(const hashtrie t, const char key[HASH_LEN]) {
    int branch;
    int64_t node = skip_first_layers(key, &branch);
    return h_get(t, node, key, DEPTH_START, branch);
}

int trie_contains(const hashtrie t, const char key[HASH_LEN]) {
    void *resptr = (void*) get(t, key);
    if (resptr == NULL) {
        return 0;
    } else if (is_checked(resptr)) {
        return 1;
    } else {
        return -1;
    }
}

int trie_contains_processed(const hashtrie t, const char key[HASH_LEN]) {
    return trie_contains(t, key) == 1;
}


static int64_t transform_with(const hashtrie t, const int64_t digest_id, const int newdepth);

static struct shm_node init_node(const hashtrie t) {
    struct shm_node res = alloc_node(t.shm);
    memset(res.nodeptr->indices, 0x00, 32 * sizeof(int64_t));
    return res;
}

static void set_digest(tNode * const node, const int pos) {
    if (!is_digest(node, pos)) {
        node->indices[pos] *= -1;
    }
}

static void set_node(tNode * const node, const int pos) {
    if (is_digest(node, pos)) {
        node->indices[pos] *= -1;
    }
}


static int64_t insert(const hashtrie t, char const * const digest, void const * const infos) {
    lock_digests_write(t.shm);
    struct shm_digest x = alloc_digest(t.shm);
    char *target = x.ptr;
    memcpy(target, digest, HASH_LEN);
    memcpy(target + HASH_LEN, infos, INFO_LEN);

    release_digests_write(t.shm);

    int checked = is_checked(infos);
    if (checked) {
        inc_checked_states(t.shm);
    }
    return x.index;
}


static void add(const hashtrie t, const int64_t index, const char key[HASH_LEN], void const * const infos, const int depth, const int branchpos) {
    tNode *node = lookup_node(t.shm, index);
    assert_esf(depth < 31, "Somehow, the hashtrie broke. You cannot fix this. Please report this error along with the model you were checking. Depth was %d.", depth);
    int pos = calculate_pos(key, depth);

    if (!is_digest(node, pos)) {
        add(t, node->indices[pos], key, infos, depth+1, branchpos);
    } else {
        lock_nodes_write(t.shm, branchpos);
        if (is_digest(node, pos)) {
            if (is_empty_pos(node, pos)) {
                node->indices[pos] = insert(t, key, infos);
                set_digest(node, pos);
                release_nodes_write(t.shm, branchpos);
                update_max_depth(t.shm, depth);
            } else {
                char *digest = lookup_digest(t.shm, node->indices[pos]);
                if (memcmp(digest, key, HASH_LEN) == 0) {
                    release_nodes_write(t.shm, branchpos);
                    // TODO: this line seems incorect, might throw off the counting
                    if (is_checked(infos) && !is_checked(digest + HASH_LEN))
                        inc_checked_states(t.shm);
                    update_infos(digest + HASH_LEN, infos);
                    /* TODO: move this code
                    int i;
                    for (i = 0; i < INFO_LEN; i++) {
                        digest[HASH_LEN + i] |= infos[i];
                    }
                    */
                } else {
                    int64_t succ = transform_with(t, node->indices[pos], depth+1);
                    node = lookup_node(t.shm, index); // memory block may be moved; recalculate the pointer is necessary
                    node->indices[pos] = succ;
                    set_node(node, pos);
                    release_nodes_write(t.shm, branchpos);
                    add(t, node->indices[pos], key, infos, depth+1, branchpos);
                }
            }
        } else {
            release_nodes_write(t.shm, branchpos);
            add(t, node->indices[pos], key, infos, depth+1, branchpos);
        }
    }
}

// MAYMOVE shm node structure
// TODO: MREMAP_MAYMOVE was used with POSIX shared memory, which was not
//       compatible with OS X.
//       Maybe the locking can be more finegrained now?
static int64_t transform_with(const hashtrie t, const int64_t digest_index, const int newdepth) {
    struct shm_node r = init_node(t);
    tNode *node = r.nodeptr;
    char *digest = lookup_digest(t.shm, digest_index);

    int pos = calculate_pos(digest, newdepth);
    assert_esf(is_empty_pos(node, pos), "Transforming a hash trie node failed. Position %d of the newly allocated node %ld should be empty but is not.", pos, r.index);
    node->indices[pos] = digest_index;
    set_digest(node, pos);

    return r.index;
}

void trie_put(const hashtrie t, const char digest[HASH_LEN], void const * const infos) {
    //assert(digest[0] == 0 || digest[0] == 1);
    int branch;
    int64_t node = skip_first_layers(digest, &branch);
    add(t, node, digest, infos, DEPTH_START, branch);
}

