#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/msg_headers.h"
#include "../common/assert_helper.h"

#include "../common/get_timestamp.h"

#include "../common/sequence_numbers.h"
#include "mstats.h"

extern size_t published_message_count;

int extract_proxy_id(char const * const id_s) {
    return atoi(id_s);
}

/* NOTE:
 * possible stored headers are:
 * PSTATS_HAS_PMS
 * PSTATS_HAS_FULL
 * PSTATS_HAS_EMPTY
 * and "empty" that's a zero (from proxies that did not yet send a statistic) */

/* termination
 * #hashes_known == #hashes_checked does not work
 * if only proxies are able to write into the hash trie.
 * hence, we should check whether all stats-digests say
 * that all workers per node are empty,
 * maybe also wait for the amount of state condition to hold.
 * however, we have to ensure that we only terminate if the global view of the
 * queues is consistent, e.g. not if:
 * proxy 1 sent work to (empty) proxy 2;
 * proxy 1 runs empty;
 * all queues are empty;
 * proxy 2 updates "I have work" */
int stats_can_terminate(mStats const * const stats) {
    int i;
    for (i = 1; i <= stats->len; i++) {
        if (stats->pdigests[i].header != PSTATS_COMPLETELY_EMPTY 
                && stats->pdigests[i].header != 0) {
            return 0;
        }
        
    }
    return 1;
}

static void extend_mstats(mStats * const digests, const int id) {
    if (id > digests->len) {
        size_t old_len = digests->len;
        digests->len = id;
        digests->pdigests = realloc(digests->pdigests, (id + 1) * sizeof(pDigest));
        int i;
        for (i = old_len + 1; i <= id; i++) {
            digests->pdigests[i] = (pDigest) {0};
        }
    }
}

void print_mstats(mStats const * const digests, int fd_logfile) {
    size_t len = digests->len;
    pDigest *arr = digests->pdigests;

    int i;
    dprintf(fd_logfile, "Header | Amount | ACKs (recv'd) | ACKs (expected) | Delta\n");
    for (i = 1; i <= len; i++) {
        switch(arr[i].header) {
            case PSTATS_HAS_PMS:
                dprintf(fd_logfile, "   PMS | "); break;
            case PSTATS_HAS_FULL:
                dprintf(fd_logfile, "  FULL | "); break;
            case PSTATS_HAS_EMPTY:
                dprintf(fd_logfile, " EMPTY | "); break;
            case PSTATS_COMPLETELY_EMPTY:
                dprintf(fd_logfile, "CEMPTY | "); break;
            case P_TRANSFER_DONE:
                dprintf(fd_logfile, "T-DONE | "); break;
            default:
                dprintf(fd_logfile, " DAFUQ[%d] | ", arr[i].header);
        }
        dprintf(fd_logfile, "    %2d | ", arr[i].full);
        dprintf(fd_logfile, "           %2d | ", arr[i].acks_received);
        dprintf(fd_logfile, "             %2d | ", arr[i].acks_expected);
        dprintf(fd_logfile, "%2d\n", arr[i].acks_expected - arr[i].acks_received);
    }
    dprintf(fd_logfile, "------------------------\n");
}


/* updates a new digest-message into the buffer of digests. 
 * new_digest was sent with given header. 
 * function takes ownership of the message. */
void update_queue_digests(mStats * const digests, const int header, zmsg_t *new_digest, int fd_logfile) {
    //zmsg_print(new_digest);
    char *id_s = zmsg_popstr(new_digest);
    int id = extract_proxy_id(id_s);
    free(id_s);
    extend_mstats(digests, id);
    pDigest * const pdig = digests->pdigests;

    assert(pdig[id].acks_expected - pdig[id].acks_received >= 0);
    if ((pdig[id].acks_expected - pdig[id].acks_received != 0)
          && (header != P_TRANSFER_DONE)) {
        zmsg_destroy(&new_digest);
        return;
    }

    int old_header = pdig[id].header;
    pdig[id].header = header;

    switch(header) {
        case PSTATS_HAS_PMS:
            zmsg_destroy(&new_digest);
            break;
        case PSTATS_HAS_FULL: {
            char *full_s = zmsg_popstr(new_digest);
            pdig[id].full = atoi(full_s);
            free(full_s);
            zmsg_destroy(&new_digest);
            break;
        }
        case PSTATS_COMPLETELY_EMPTY:
        case PSTATS_HAS_EMPTY: {
            char *empty_s = zmsg_popstr(new_digest);
            pdig[id].empty = atoi(empty_s);
            free(empty_s);

            if (pdig[id].need_msg) {
                zmsg_destroy(&(pdig[id].need_msg));
            }
            pdig[id].need_msg = new_digest;
            break;
        }
        case P_TRANSFER_DONE: {
            // unfreeze
            char *timestamp = zmsg_popstr(new_digest);
            dprintf(fd_logfile, "id %d acked at timestamp %s, (expected %d more)\n", id, timestamp, pdig[id].acks_expected - pdig[id].acks_received);
            free(timestamp);
            pdig[id].acks_received++;
            //printf("id: %d, EXPECTED: %d, RECEIVED: %d\n", id, pdig[id].acks_expected, pdig[id].acks_received);
            assert (pdig[id].acks_expected - pdig[id].acks_received >= 0);
            if (pdig[id].acks_expected - pdig[id].acks_received >= 1) {
                pdig[id].header = old_header;
                zmsg_destroy(&new_digest);
                print_mstats(digests, fd_logfile);
                return;
            }
            char *header_s = zmsg_popstr(new_digest);
            int new_header;
            int rc = sscanf(header_s, "%d", &new_header);
            assert_esf(rc == 1, "Protocol error: expected a decimal number, received %s.", header_s);
            free(header_s);

            update_queue_digests(digests, new_header, new_digest, fd_logfile);
            print_mstats(digests, fd_logfile);
            break;
        }
    }
}

/* returns the possible amount of global transfers */
int m_can_balance(mStats const * const digests) {
    int i;
    pDigest *pdigests = digests->pdigests;

    int has_count = 0;
    int need_count = 0;
    for (i = 1; i <= digests->len; i++) {
        if (pdigests[i].header == PSTATS_HAS_FULL) {
            has_count += pdigests[i].full;
        } else if (pdigests[i].header == PSTATS_HAS_EMPTY
                      || pdigests[i].header == PSTATS_COMPLETELY_EMPTY) {
            need_count += pdigests[i].empty;
        }
    }
    return (has_count > need_count) ? need_count : has_count;
}


const char *lookup_proxy(char const * const idmap[], const int id) {
    return idmap[id];
}

// returns new sequence number
int m_do_balance(mStats const * const digests,
                  int seq_num,
                  zsock_t * const pub,
                  char const * const idmap[],
                  int fd_logfile) {
    int nmbr_balances = m_can_balance(digests);
    if (!nmbr_balances) {
        return seq_num;
    }
    size_t timestamp = get_sizettime();

    dprintf(fd_logfile, "%d times balance party: %lu\n", nmbr_balances, timestamp);
    print_mstats(digests, fd_logfile);
    pDigest * const pdig = digests->pdigests;

    int last_has  = 1;
    int last_need = 1;

    int i;
    zmsg_t *msg = zmsg_new();
    int bal = 0;
    while (bal < nmbr_balances) {
        assert(msg);
        // find proxy that has work
        for (i = last_has; i <= digests->len; i++) {
            if (pdig[i].header == PSTATS_HAS_FULL) {
                last_has = i;
                break;
            }
        }

        // find proxy that needs work
        for (i = last_need; i <= digests->len; i++) {
            if (pdig[i].header == PSTATS_HAS_EMPTY
                    || pdig[i].header == PSTATS_COMPLETELY_EMPTY) {
                last_need = i;
                break;
            }
        }

        assert (last_has != last_need);

        int min;
        if (pdig[last_has].full < pdig[last_need].empty) {
            min = pdig[last_has].full;
        } else {
            min = pdig[last_need].empty;
        }

        for (i = 0; i < min; i++) {
            const char *ip = lookup_proxy(idmap, last_need);
            char *port = zmsg_popstr(pdig[last_need].need_msg);

            zmsg_addstrf(msg, "tcp://%s:%s", ip, port);
            free(port);

            pdig[last_has].full--;
            pdig[last_has].acks_expected++;
            pdig[last_need].empty--;
            pdig[last_need].acks_expected++;

            bal++;
        }

        if (pdig[last_need].empty == 0) {
            pdig[last_need].header = PSTATS_HAS_PMS;
        }

        if (pdig[last_has].full == 0) {
            pdig[last_has].header = PSTATS_HAS_PMS;
            dprintf(fd_logfile, "{:type :transfer :proxy_from %d, :at %ld}\n", last_has, timestamp);
            seq_num = next_seq(seq_num);
            int rc = zsock_send(pub, "iii8m", CTRL_TRANSFER_TO, seq_num, last_has, timestamp, msg);
            assert(rc == 0);
            published_message_count++;
            zmsg_destroy(&msg);
            assert(msg == NULL);
            if (bal != nmbr_balances) {
                // there are more transfers
                msg = zmsg_new();
            } else {
                // we are done with all transfers
                // we can't flag them every time,
                // because we might reuse the empty proxy later on
                pdig[last_need].header = PSTATS_HAS_PMS;
            }
        }
    }

    // if the full proxy still has work, but no more empty proxies are available
    if (msg) {
        pdig[last_has].header = PSTATS_HAS_PMS;
        pdig[last_need].header = PSTATS_HAS_PMS;
        dprintf(fd_logfile, "{:type :transfer :proxy_from %d, :at %ld}\n", last_has, timestamp);
        seq_num = next_seq(seq_num);
        int rc = zsock_send(pub, "iii8m", CTRL_TRANSFER_TO, seq_num, last_has, timestamp, msg);
        assert(rc == 0);
        published_message_count++;
        zmsg_destroy(&msg);
    }
    print_mstats(digests, fd_logfile);
    return seq_num;
}


void teardown_mstats(mStats * const digests) {
    pDigest *x = digests->pdigests;
    int i;
    for (i = 1; i <= digests->len; i++) {
        if (x[i].need_msg) {
            zmsg_destroy(&(x[i].need_msg));
        }
    }
    free(x);
    digests->pdigests = NULL;
    digests->len = 0;
}
