#include <assert.h>
#include <stdlib.h>

#include <zmq.h>
#include <czmq.h>

#include "../master/mstats.h"
#include "../proxy/pstats.h"
#include "../proxy/p_pull_push.h"
#include "../common/msg_headers.h"

#include "test_helpers.h"

static void test_update_queue_digests(zsock_t * const push, zsock_t * const pull) {
    mStats stats = (mStats) {0};

    pStats ps = {0};
    update_qs(&ps, 1, 0, 100);
    update_qs(&ps, 2, 300, 100);

    psend_stats(push, 2, &ps, 5000, 0);
    free(ps.queuesizes);

    int header = -1;
    zmsg_t *msg = NULL;
    int rc = zsock_recv(pull, "im", &header, &msg);
    assert(rc == 0);


    update_queue_digests(&stats, header, msg);
    assert(stats.pdigests);
    assert(stats.len == 2);
    
    // this is undefined
//    assert(stats.pdigests[0].full == 0);
//    assert(stats.pdigests[0].empty == 0);
//    assert(stats.pdigests[0].need_msg == NULL);

    
    assert(stats.pdigests[1].header == 0);
    assert(stats.pdigests[1].empty == 0);
    assert(stats.pdigests[1].need_msg == NULL);

    assert(stats.pdigests[2].header == PSTATS_HAS_PMS);

    teardown_mstats(&stats);
    printf(".");
}

static void test_can_balance_all_full(void) {
    mStats stats = {0};

    size_t len = 3;

    pDigest digests[len + 1];
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 2);
    digests[2] = get_pdigest(PSTATS_HAS_FULL, 3);
    digests[3] = get_pdigest(PSTATS_HAS_FULL, 1);

    stats.len = len;
    stats.pdigests = digests;

    int rc = m_can_balance(&stats);
    assert(rc == 0);
    printf(".");
}

static void test_can_balance2(void) {
    mStats stats = {0};

    size_t len = 3;

    pDigest digests[len + 1];
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 2);
    digests[2] = get_pdigest(PSTATS_HAS_FULL, 3);
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 1);

    stats.len = len;
    stats.pdigests = digests;

    int rc = m_can_balance(&stats);
    assert(rc == 1);
    printf(".");
}

static void test_can_balance_sophisticated(void) {
    mStats stats = {0};

    size_t len = 3;

    pDigest digests[len + 1];
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 2);
    digests[2] = get_pdigest(PSTATS_HAS_FULL, 3);
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 4);
    // no message needed

    stats.len = len;
    stats.pdigests = digests;

    int rc = m_can_balance(&stats);
    assert(rc == 4);
    printf(".");
}

static void test_can_balance_all_empty(void) {
    mStats stats = {0};

    size_t len = 3;

    pDigest digests[len + 1];
    digests[1] = get_pdigest(PSTATS_HAS_EMPTY, 2);
    digests[2] = get_pdigest(PSTATS_HAS_EMPTY, 1);
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 4);

    stats.len = len;
    stats.pdigests = digests;

    int rc = m_can_balance(&stats);
    assert(rc == 0);
    printf(".");
}

static void test_can_balance_with_pms(void) {
    mStats stats = {0};

    size_t len = 3;

    pDigest digests[len + 1];
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 1);
    digests[2] = get_pdigest(PSTATS_HAS_PMS, 1);
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 4);

    stats.len = len;
    stats.pdigests = digests;

    int rc = m_can_balance(&stats);
    assert(rc == 1);
    printf(".");
}

// tests: single transfer from one source to one destination
static void test_do_balance(zsock_t * const pub, zsock_t * const sub) {
    mStats stats = {0};

    size_t len = 3;

    pDigest *digests = malloc(sizeof(pDigest) * (len + 1));;
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 2);
    digests[2] = get_pdigest(PSTATS_HAS_FULL, 3);
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 1);
    zmsg_t *msg = zmsg_new();
    zmsg_addstrf(msg, "%d", 5013);
    digests[3].need_msg = msg;

    stats.len = len;
    stats.pdigests = digests;


    const char *idmap[] = {"", "10.0.0.19", "10.0.0.20", "10.0.0.21"};
    
    m_do_balance(&stats, pub, idmap);

    int header;
    char *id;
    zmsg_t *endpoints;
    int rc = zsock_recv(sub, "ism",  &header, &id, &endpoints);
    assert(rc == 0);
    
    assert(zmsg_size(endpoints) == 1);
    assert(header == CTRL_TRANSFER_TO);
    assert(0 == strcmp(id, "1"));

    char *endpoint1 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint1, "tcp://10.0.0.21:5013"));


    free(id);
    free(endpoint1);

    zmsg_destroy(&endpoints);


    assert(digests[1].header == PSTATS_HAS_PMS);
    assert(digests[1].full == 1);
    assert(digests[1].frozen == 1);

    assert(digests[2].header == PSTATS_HAS_FULL);
    assert(digests[2].full == 3);
    assert(digests[2].frozen == 0);

    assert(digests[3].header == PSTATS_HAS_PMS);
    assert(digests[3].empty == 0);
    assert(digests[3].frozen == 1);

    teardown_mstats(&stats);

    printf(".");
}
                            

// tests: two transfers from one source to one destination
static void test_do_balance2(zsock_t * const pub, zsock_t * const sub) {
    mStats stats = {0};

    size_t len = 3;

    pDigest *digests = malloc(sizeof(pDigest) * (len + 1));;
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 2);
    digests[2] = get_pdigest(PSTATS_HAS_FULL, 3);
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 2);
    zmsg_t *msg = zmsg_new();
    zmsg_addstrf(msg, "%d", 5013);
    zmsg_addstrf(msg, "%d", 5014);
    digests[3].need_msg = msg;

    stats.len = len;
    stats.pdigests = digests;


    const char *idmap[] = {"", "10.0.0.19", "10.0.0.20", "10.0.0.21"};
    
    m_do_balance(&stats, pub, idmap);

    int header;
    char *id;
    zmsg_t *endpoints;
    int rc = zsock_recv(sub, "ism",  &header, &id, &endpoints);
    assert(rc == 0);
    
    assert(zmsg_size(endpoints) == 2);
    assert(header == CTRL_TRANSFER_TO);
    assert(0 == strcmp(id, "1"));

    char *endpoint1 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint1, "tcp://10.0.0.21:5013"));

    char *endpoint2 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint2, "tcp://10.0.0.21:5014"));

    free(id);
    free(endpoint1);
    free(endpoint2);

    zmsg_destroy(&endpoints);


    assert(digests[1].header == PSTATS_HAS_PMS);
    assert(digests[1].full == 0);
    assert(digests[1].frozen == 2);

    assert(digests[2].header == PSTATS_HAS_FULL);
    assert(digests[2].full == 3);
    assert(digests[2].frozen == 0);

    assert(digests[3].header == PSTATS_HAS_PMS);
    assert(digests[3].empty == 0);
    assert(digests[3].frozen == 2);

    teardown_mstats(&stats);

    printf(".");
}

// tests: three transfers from two sources to one destination
static void test_do_balance3(zsock_t * const pub, zsock_t * const sub) {
    mStats stats = {0};

    size_t len = 3;

    pDigest *digests = malloc(sizeof(pDigest) * (len + 1));;
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 2);
    digests[2] = get_pdigest(PSTATS_HAS_FULL, 3);
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 3);
    zmsg_t *msg = zmsg_new();
    zmsg_addstrf(msg, "%d", 5013);
    zmsg_addstrf(msg, "%d", 5014);
    zmsg_addstrf(msg, "%d", 5015);
    digests[3].need_msg = msg;

    stats.len = len;
    stats.pdigests = digests;


    const char *idmap[] = {"", "10.0.0.19", "10.0.0.20", "10.0.0.21"};
    
    m_do_balance(&stats, pub, idmap);

    // first proxy
    int header;
    char *id;
    zmsg_t *endpoints;
    int rc = zsock_recv(sub, "ism",  &header, &id, &endpoints);
    assert(rc == 0);
    
    assert(zmsg_size(endpoints) == 2);
    assert(header == CTRL_TRANSFER_TO);
    assert(0 == strcmp(id, "1"));

    char *endpoint1 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint1, "tcp://10.0.0.21:5013"));

    char *endpoint2 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint2, "tcp://10.0.0.21:5014"));

    free(id);
    free(endpoint1);
    free(endpoint2);

    zmsg_destroy(&endpoints);


    // second proxy
    rc = zsock_recv(sub, "ism",  &header, &id, &endpoints);
    assert(rc == 0);
    
    assert(zmsg_size(endpoints) == 1);
    assert(header == CTRL_TRANSFER_TO);
    assert(0 == strcmp(id, "2"));

    char *endpoint3 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint3, "tcp://10.0.0.21:5015"));
    free(endpoint3);

    free(id);
    zmsg_destroy(&endpoints);



    assert(digests[1].header == PSTATS_HAS_PMS);
    assert(digests[1].full == 0);
    assert(digests[1].frozen == 2);

    assert(digests[2].header == PSTATS_HAS_PMS);
    assert(digests[2].full == 2);
    assert(digests[2].frozen == 1);

    assert(digests[3].header == PSTATS_HAS_PMS);
    assert(digests[3].empty == 0);
    assert(digests[3].frozen == 3);

    teardown_mstats(&stats);

    printf(".");
}

// tests: four transfers from two sources to two destinations
static void test_do_balance4(zsock_t * const pub, zsock_t * const sub) {
    mStats stats = {0};

    size_t len = 4;

    pDigest *digests = malloc(sizeof(pDigest) * (len + 1));;
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 2);
    digests[2] = get_pdigest(PSTATS_HAS_FULL, 3);
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 3);
    digests[4] = get_pdigest(PSTATS_HAS_EMPTY, 1);
    zmsg_t *msg3 = zmsg_new();
    zmsg_addstrf(msg3, "%d", 5013);
    zmsg_addstrf(msg3, "%d", 5014);
    zmsg_addstrf(msg3, "%d", 5015);
    digests[3].need_msg = msg3;
    zmsg_t *msg4 = zmsg_new();
    zmsg_addstrf(msg4, "%d", 5013);
    digests[4].need_msg = msg4;

    stats.len = len;
    stats.pdigests = digests;


    const char *idmap[] = {"", "10.0.0.19", "10.0.0.20", "10.0.0.21", "10.0.0.22"};
    
    m_do_balance(&stats, pub, idmap);

    // first proxy
    int header;
    char *id;
    zmsg_t *endpoints;
    int rc = zsock_recv(sub, "ism",  &header, &id, &endpoints);
    assert(rc == 0);
    
    assert(zmsg_size(endpoints) == 2);
    assert(header == CTRL_TRANSFER_TO);
    assert(0 == strcmp(id, "1"));

    char *endpoint1 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint1, "tcp://10.0.0.21:5013"));

    char *endpoint2 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint2, "tcp://10.0.0.21:5014"));

    free(id);
    free(endpoint1);
    free(endpoint2);

    zmsg_destroy(&endpoints);


    // second proxy
    rc = zsock_recv(sub, "ism",  &header, &id, &endpoints);
    assert(rc == 0);
    
    assert(zmsg_size(endpoints) == 2);
    assert(header == CTRL_TRANSFER_TO);
    assert(0 == strcmp(id, "2"));

    char *endpoint3 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint3, "tcp://10.0.0.21:5015"));
    free(endpoint3);

    char *endpoint4 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint4, "tcp://10.0.0.22:5013"));
    free(endpoint4);

    free(id);
    zmsg_destroy(&endpoints);



    assert(digests[1].header == PSTATS_HAS_PMS);
    assert(digests[1].full == 0);
    assert(digests[1].frozen == 2);

    assert(digests[2].header == PSTATS_HAS_PMS);
    assert(digests[2].full == 1);
    assert(digests[2].frozen == 2);

    assert(digests[3].header == PSTATS_HAS_PMS);
    assert(digests[3].empty == 0);
    assert(digests[3].frozen == 3);

    assert(digests[4].header == PSTATS_HAS_PMS);
    assert(digests[4].empty == 0);
    assert(digests[4].frozen == 1);

    teardown_mstats(&stats);

    printf(".");
}


// tests: two transfers from one source to two destination
static void test_do_balance5(zsock_t * const pub, zsock_t * const sub) {
    mStats stats = {0};

    size_t len = 3;

    pDigest *digests = malloc(sizeof(pDigest) * (len + 1));;
    digests[1] = get_pdigest(PSTATS_HAS_FULL, 2);
    digests[2] = get_pdigest(PSTATS_HAS_EMPTY, 1);
    zmsg_t *msg2 = zmsg_new();
    zmsg_addstrf(msg2, "%d", 5013);
    digests[2].need_msg = msg2;
    digests[3] = get_pdigest(PSTATS_HAS_EMPTY, 2);
    zmsg_t *msg3 = zmsg_new();
    zmsg_addstrf(msg3, "%d", 5013);
    zmsg_addstrf(msg3, "%d", 5014);
    digests[3].need_msg = msg3;

    stats.len = len;
    stats.pdigests = digests;


    const char *idmap[] = {"", "10.0.0.19", "10.0.0.20", "10.0.0.21"};
    
    m_do_balance(&stats, pub, idmap);

    int header;
    char *id;
    zmsg_t *endpoints;
    int rc = zsock_recv(sub, "ism",  &header, &id, &endpoints);
    assert(rc == 0);
    
    assert(zmsg_size(endpoints) == 2);
    assert(header == CTRL_TRANSFER_TO);
    assert(0 == strcmp(id, "1"));

    char *endpoint1 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint1, "tcp://10.0.0.20:5013"));

    char *endpoint2 = zmsg_popstr(endpoints);
    assert(0 == strcmp(endpoint2, "tcp://10.0.0.21:5013"));

    free(id);
    free(endpoint1);
    free(endpoint2);

    zmsg_destroy(&endpoints);


    assert(digests[1].header == PSTATS_HAS_PMS);
    assert(digests[1].full == 0);
    assert(digests[1].frozen == 2);

    assert(digests[2].header == PSTATS_HAS_PMS);
    assert(digests[2].empty == 0);
    assert(digests[2].frozen == 1);

    assert(digests[3].header == PSTATS_HAS_PMS);
    assert(digests[3].empty == 1);
    assert(digests[3].frozen == 1);

    teardown_mstats(&stats);

    printf(".");
}


int main(int argc, char *argv[]) {
    zsock_t *pull = zsock_new_pull("inproc://foo");
    zsock_t *push = zsock_new_push("inproc://foo");
    assert(pull);
    assert(push);

    test_update_queue_digests(push, pull);

    zsock_destroy(&pull);
    zsock_destroy(&push);

    test_can_balance_all_full();
    test_can_balance2();
    test_can_balance_sophisticated();
    test_can_balance_all_empty();
    test_can_balance_with_pms();


    zsock_t *pub = zsock_new_pub("inproc://foo-pubsub");
    zsock_t *sub = zsock_new_sub("inproc://foo-pubsub", "");
    assert(pub);
    assert(sub);

    test_do_balance(pub, sub);
    test_do_balance2(pub, sub);
    test_do_balance3(pub, sub);
    test_do_balance4(pub, sub);
    test_do_balance5(pub, sub);

    zsock_destroy(&pub);
    zsock_destroy(&sub);


    puts("");

    return 0;
}
