#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <zmq.h>
#include <czmq.h>

#include <sicstus/sicstus.h>

#include "../common/queue.h"
#include "../common/shm_cln.h"
#include "../common/magic_constants.h"
#include "../common/hashtrie.h"


#include "wid.h"
#include "wpush.h"
#include "wctrl.h"
#include "wprocessor.h"
#include "wsink.h"

/* use these global variables only for callbacks from SICStus;
 * avoid usage whenever possible */
int g_worker_id = 0;
/* NOT thread-safe zocket. Use this only when you know what you're doing. */
zsock_t *g_zocket_push = NULL; 
wQueue *g_queue = NULL;
/* seems like we need a to store the trie here, too */
hashtrie g_trie = (hashtrie) {0};
int g_generate_successors = 1;

static int get_strategy(char const * const mode) {
    if (strcmp(mode, "mixed") == 0) {
        return 50;
    } else if (strcmp(mode, "breadth_first") == 0) {
        return 0;
    } else {
        return 100;
    }
}


static SP_pred_ref init_predicate(const char *name, int arity, const char *module) {
    SP_pred_ref pred = SP_predicate(name,arity,module);
    if (pred == NULL) {
        fprintf(stderr, "Unable to find callback module %s:%s/%i for zmq worker, exiting.\n",
                module, name, arity);
        exit(1);
    }
    return pred;
}

static void load_model(zchunk_t * const model_chunk) {
    char *model = (char*) zchunk_data(model_chunk);
    SP_pred_ref load_cmd = init_predicate("load", 1, "worker");

    SP_term_ref model_ref = SP_new_term_ref();
    SP_put_address(model_ref, model);
    SP_query_cut_fail(load_cmd, model_ref);
}


static SP_pred_ref process_pred = NULL;
static zpoller_t *poller_sub = NULL;
static zpoller_t *poller_rep = NULL;
static zsock_t *zocket_sub = NULL;
static zsock_t *zocket_rep = NULL;
static zsock_t *zocket_push = NULL;
static int initd = 0;
static int fd_logfile = -1;
static hashtrie trie = {0};
static wQueue *q = NULL;
static int proxy_nmbr = -1;

void setup_worker(const int proxy_number,
                  const int port_start,
                  char const * const strategy_s,
                  char const * const logfile_name,
                  char const * const tmpdir) {
    (void) tmpdir;
    fd_logfile = open(logfile_name, O_CREAT | O_RDWR | O_TRUNC, 0666);
    assert(fd_logfile != -1);
    dprintf(fd_logfile, "{:status :online}\n");

    const int strategy = get_strategy(strategy_s);

    // manage ID and model
    struct id_response resp = query_proxy_for_id(proxy_number);
    g_worker_id = resp.id;
    dprintf(fd_logfile, "{:type :info, :recv-id %d}\n", resp.id);
    load_model(resp.model);
    dprintf(fd_logfile, "{:status :model-loaded}\n");
    zchunk_destroy(&resp.model);

    q = init_queue(strategy);
    g_queue = q;


    // TODO: work on hash trie, shared memory etc., enforce read-only access
    struct semap_env senv = open_semaphores(proxy_number);
    trie = get_trie(senv, HASHTRIE_KEY_START + 40000 * proxy_number);
    g_trie = trie;

    zocket_push = init_wpush(proxy_number);
    g_zocket_push = zocket_push;

    zocket_sub = init_wsub(proxy_number);
    zocket_rep = init_work_sink(port_start, g_worker_id, proxy_number);
    proxy_nmbr = proxy_number;

    // send an initial stats package
    send_small_stats(zocket_push, g_worker_id, 0);


    zsys_catch_interrupts(); // this is for internal use

    process_pred = init_predicate("zmq_check", 3, "worker");
    poller_sub = zpoller_new(zocket_sub, NULL);
    poller_rep = zpoller_new(zocket_rep, NULL);
    initd = 1;
}


// another fancy reactor
int work_reactor(void) {
    assert(initd);
    if (zsys_interrupted) {
        return -1;
    }
    if (zpoller_wait(poller_sub, 0)) {
        int rc = wrecv_ctrl(zocket_sub, zocket_push, q, g_worker_id, fd_logfile);
        if (rc == -1) { // this is the only socket that may make us shut down
            return 1;
        }
    }
    if (zpoller_wait(poller_rep, 0)) {
        handle_wsink(zocket_rep, q, zocket_push, g_worker_id, fd_logfile);
    }
    processor(q, zocket_push, g_worker_id, process_pred, trie, fd_logfile);
    return 0;
}

void teardown_worker(void) {
    zpoller_destroy(&poller_sub);
    zpoller_destroy(&poller_rep);

    zsock_destroy(&zocket_rep);
    zsock_destroy(&zocket_sub);
    zsock_destroy(&zocket_push);

    teardown_hashtrie(&trie, 0, proxy_nmbr);

    teardown_queue(&q);
}


