% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(b_trace_checking, [tcltk_check_state_sequence_from_file/1,
                             tcltk_check_sequence_from_file/3,
                             check_default_trace_for_specfile/1, get_default_trace_file/2,
                             print_trace_as_fdr_check/0,
                             tcltk_save_history_as_trace_file/2, %print_trace_for_replay/1,
                             perform_initialisation/0,
                             find_successor_state/4
                            ]).

:- use_module(library(lists)).

:- use_module(tools).
:- use_module(module_information,[module_info/2]).
:- module_info(group,testing).
:- module_info(description,'Replay saved traces; raise error if not possible. Used in regression testing.').

:- use_module(bsyntaxtree, [get_texpr_type/2, conjunct_predicates/2]).
:- use_module(bmachine,[b_machine_name/1, b_is_operation_name/1, b_get_machine_operation_typed_parameters/2]).
:- use_module(b_global_sets,[lookup_global_constant/2]).
:- use_module(error_manager).
:- use_module(debug).
:- use_module(preferences,[get_preference/2,preference/2,
                           temporary_set_preference/3,reset_temporary_preference/2, get_prob_application_type/1]).
:- use_module(specfile,[b_mode/0, csp_mode/0, get_operation_name/2, get_operation_return_values_and_arguments/3]).
:- use_module(translate,[translate_event/2]).
:- use_module(state_space,[current_options/1, current_state_id/1, current_expression/2,
   max_reached_for_node/1, not_all_transitions_added/1,time_out_for_node/3]).
:- use_module(junit_tests).
:- use_module(tools_printing,[print_red/1, print_green/1, format_with_colour/4]).
% --------------------------------

% replay trace with state predicates; one predicate per line

% b_trace_checking:tcltk_check_state_sequence_from_file('~/Desktop/scheduler6.ptrace').

tcltk_check_state_sequence_from_file(File) :-
   open(File,read,Stream,[encoding('UTF-8')]),
   state_space:current_state_id(ID),
   format(user_output,'Opened predicate list file: ~w~n Starting from state ~w.~n',[File,ID]),
   call_cleanup(tcltk_check_state_sequence_from_file_aux(File,0,Stream,ID),
                (format(user_output,'Finished processing list file: ~w~n',[File]),
                 close(Stream))).

tcltk_check_state_sequence_from_file_aux(File,LineNr,Stream,ID) :- read_line(Stream,Codes),
   Codes \= end_of_file,
   L1 is LineNr+1,
   !, %print(codes(Codes)),nl,
   process_line(File,L1,Stream,Codes,ID).
tcltk_check_state_sequence_from_file_aux(_,LineNr,_,_) :- printsilent('reached EOF at line: '), printsilent(LineNr),nls.

:- use_module(bmachine,[b_parse_wo_type_machine_predicate_from_codes_to_raw_expr/2,
                        b_type_check_raw_expr/4]).
:- use_module(translate_keywords).

% TO DO: also allow operation names, parameters to be specified?
process_line(File,LineNr,Stream,Codes,ID) :-
   % insert newlines to ensure error message correct:
    insert_nl_prefix(LineNr,Codes,CodesNL),
    b_parse_wo_type_machine_predicate_from_codes_to_raw_expr(CodesNL,Parsed),
    % in case we did translate keywords (e.g., in Event-B mode); translate them back:
    translate_keywords:raw_backtranslate_keyword_ids(Parsed,Parsed2),
    % now type check the raw expression:
    b_type_check_raw_expr(Parsed2,[prob_ids(all),variables],Pred,closed),
    % bmachine:b_parse_machine_predicate_from_codes(CodesNL,[prob_ids(all),variables],Pred),
    !,
    format('Line ~w : ~w --?--> ',[LineNr,ID]),translate:print_bexpr(Pred),nl,
    find_successor_state(ID,Pred,SuccID,LineNr),
    tcltk_check_state_sequence_from_file_aux(File,LineNr,Stream,SuccID).
process_line(_File,LineNr,_Stream,Codes,_ID) :-
    atom_codes(Atom,Codes),
    add_error(tcltk_check_state_sequence_from_file,'Could not parse state predicate: ',Atom),
    add_error(tcltk_check_state_sequence_from_file,'Line: ',LineNr),
    %add_error(tcltk_check_state_sequence_from_file,'File: ',File),
    fail.

insert_nl_prefix(N,Acc,Acc) :- N =< 1,!.
insert_nl_prefix(N,Acc,[10,13|Res]) :- N1 is N-1, insert_nl_prefix(N1,Acc,Res).

:- use_module(b_global_sets,[add_prob_deferred_set_elements_to_store/3]).
:- use_module(specfile,[expand_const_and_vars_to_full_store/2]).
:- use_module(state_space,[visited_expression/2,
    max_reached_for_node/1, max_reached_or_timeout_for_node/1, time_out_for_node/1]).
:- use_module(library(codesio),[format_to_codes/3]).
:- use_module(eventhandling, [announce_event/1]).

find_successor_state(_ID,Pred,NewID,_) :-
  % transition(ID,ActionOpAsTerm,OpID,SuccID),
   tcltk_interface:tcltk_get_options(_),
   current_options(Options), % print(Options),nl,
   member( (_,Action,ActionOpAsTerm,NewID), Options),
   state_space:visited_expression(NewID,State),
   expand_const_and_vars_to_full_store(State,EState),
   % print(try(NewID,State)),nl,
   add_prob_deferred_set_elements_to_store(EState,DEState,all),
   announce_event(start_solving),
   b_interpreter:b_test_boolean_expression_wf(Pred,[],DEState), % deferred
   announce_event(end_solving),
   debug_println(19,successor_found(NewID)),
   print(' --> '),my_print_event(ActionOpAsTerm,_),nl,nl,
   my_perform_action_as_string(Action,ActionOpAsTerm,NewID),
   !.
find_successor_state(ID,Pred,NewID,_LineNr) :-
   max_reached_or_timeout_for_node(ID), % we have not computed all options
   get_op_name_for_state(ID,OpName), print(try_providing_custom_predicate(OpName)),nl,
   tcltk_interface:tcltk_add_user_executed_operation_typed(OpName,Pred,NewID), % The Pred can only talk about parameters and the state before; thus this only works for setup_constants and initialise_machine
   (silent_mode(on) -> true
     ; print_green(successor_found(NewID)),nl,
       print(' --> '),print(OpName),nl),
   !.
find_successor_state(ID,Pred,_,LineNr) :-
   % we could try and execute operation by predicate here; in case max_reached_for_node(ID) or  time_out_for_node(ID)
   format('State ID = ~w; line number = ~w~n',[ID,LineNr]),
   (max_reached_for_node(ID)
    -> format_to_codes('Try increasing the MAX_OPERATIONS preference. No successor found satisfying predicate at step ~w of replay: ',[LineNr],Codes)
     ; time_out_for_node(ID) -> format_to_codes('TIME-OUT occurred. No successor found satisfying predicate at step ~w of replay: ',[LineNr],Codes)
     ; format_to_codes('No successor found satisfying predicate at step ~w of replay: ',[LineNr],Codes)
   ),
   atom_codes(Msg,Codes),
   translate:translate_bexpression(Pred,TP),
   add_error(tcltk_check_state_sequence_from_file,Msg,TP),
   fail.

:- use_module(bmachine,[b_machine_has_constants_or_properties/0]).
get_op_name_for_state(root,OP) :- !,
 (b_machine_has_constants_or_properties -> OP='$setup_constants' ; OP='$initialise_machine').
get_op_name_for_state(ID,'$initialise_machine') :-
   visited_expression(ID,concrete_constants(_)),!.
%get_op_name_for_state(ID,OpName) :- bmachine:b_top_level_operation(OpName). % not supported yet

% --------------------------------

check_default_trace_for_specfile(File) :- get_default_trace_file(File,TFile),
   tcltk_check_sequence_from_file(prolog,TFile,backtrack).

get_default_trace_file(File,TraceFile) :-
   split_filename(File,FN,_Ext),
   string_concatenate(FN,'.trace',TraceFile).

:- use_module(tools_printing,[print_error/1]).
tcltk_check_sequence_from_file(Style,File,deterministic_trace_replay) :- !,
    temporary_set_preference(deterministic_trace_replay,true,CHNG),
    format('Performing deterministic replay (lower memory usage, but no backtracking in case replay fails) for file: ~w~n',[File]),
    (tcltk_check_sequence_from_file2(Style,File)
     -> reset_temporary_preference(deterministic_trace_replay,CHNG)
     ;  reset_temporary_preference(deterministic_trace_replay,CHNG)).
tcltk_check_sequence_from_file(Style,File,_DetOrBacktrack) :-
   tcltk_check_sequence_from_file2(Style,File).

tcltk_check_sequence_from_file2(Style,File) :-
   read_trace_file(Style,File,MachineName,Trace),
   length(Trace,Len),
   (b_mode -> b_machine_name(OurName) ; true),
   ((OurName=MachineName
    ; MachineName = 'dummy(uses)') % old trace files have these
     -> printsilent_message('Checking file for machine: '), printsilent_message(OurName)
      ; print_error('### Trace File has been generated for different machine: '),
        print_error(MachineName),
        print_error('### Our machine: '), print_error(OurName)
   ),
   formatsilent('Starting trace check of length ~w~n',[Len]),
   statistics(walltime,[Time1,_]),
   (perform_sequence_of_operations(Trace)
     -> nls,print_green('Trace checking successful !'),nl
     ;   nls,
         length(Trace,Len), max_nr_replayed(Replayed),
         ajoin(['Trace Checking was not successful for ', MachineName,
                ', replayed ',Replayed,'/',Len,' operations from:'],Msg),
         add_error(trace_checking_fails,Msg,File)
        %% ,throw(trace_checking_failed(MachineName))  % comment in if you want to stop immediately upon such an error
   ),
   statistics(walltime,[Time2,_]), Time is Time2-Time1,
   formatsilent('Walltime: ~w ms~n',[Time]),
   (csp_mode -> print_trace_as_fdr_check ; true),

   (junit_mode(_) -> (get_error(trace_checking,Errors) -> V=error([Errors])
                                           ; otherwise -> V=pass),
                     create_and_print_junit_result(File,'Trace_checking', MachineName, Time, V)
   ; otherwise -> true).

:- use_module(json,[json_parse_file/2]).
read_trace_file(prolog,File,MachineName,Trace) :-
   formatsilent('% Opening trace file: ~w~n',[File]),
   my_see(File),
   parse_trace_file(MachineName,Trace),!,
   seen.
read_trace_file(json,File,MachineName,Trace) :-
   MachineName = 'dummy(uses)', % name not stored in JSON files
   json_parse_file(File,Term),
   !,
   (translate_json_term(Term,Trace) -> true
     ; add_error(trace_checking_fails,'Could not translate JSON transitionList: ',Term),
       %trace, translate_json_term(Term,Trace),
       Trace = []).
read_trace_file(_,File,_,_) :-
   add_error(trace_checking_fails,'Could not read trace file: ',File),fail.

my_see(File) :- on_exception(error(existence_error(_,_),_),
                   see(File), add_error_fail(my_see,'File does not exist: ',File)).

parse_trace_file(MName,Trace) :-
   safe_read(Term),!,
   (Term = end_of_file -> (Trace = [])
     ; (Term = machine(MName)
        -> Trace = T
        ;  Trace = [Term|T]
        ),
        parse_trace_file(MName,T)
    ).

safe_read(T) :- on_exception(E,read(T),
         ( add_error(safe_read,'Exception while reading input: ~w ~n',[E]), T=end_of_file)).


translate_json_term(obj(ObjList),Trace) :-
  member(pair(transitionList,List),ObjList),
  !,
  temporary_set_preference(repl_cache_parsing,true,CHNG),
  maplist(translate_json_operation,List,Trace),
  reset_temporary_preference(repl_cache_parsing,CHNG).

:- use_module(bsyntaxtree, [def_get_texpr_ids/2]).
translate_json_operation(obj([pair(name,string(OpNameC))|RestPair]), Operation ) :- % new style
    %print(rest(RestPair)),nl,
    get_parameters(RestPair,Paras),
    !,
    atom_codes(OpName,OpNameC), print(json_translate_operation(OpName)),nl,
    (bmachine:b_get_machine_operation_parameter_names(OpName,OpParameterNames) -> true
      ; add_error(translate_json_operation,'Unknown operation: ',OpName),fail),
    maplist(translate_json_para,Paras,Bindings),
    order_paras(OpParameterNames,OpName,Bindings,BValues), % put the parameters into the order expected
    Op =.. [OpName|BValues],
    (bmachine:b_get_machine_operation(OpName,OpResults,_P,_Body,_OType,_OpPos),
     OpResults \= [] % we have return values
         -> Operation = '-->'(Op,BResValues),
            (member(pair(results,obj(ResParas)),RestPair) % we could provide this as an option: check results or not
              -> maplist(translate_json_para,ResParas,Bindings2),
                 def_get_texpr_ids(OpResults,OpResultNames),
                 order_paras(OpResultNames,OpName,Bindings2,BResValues)
              ; true)
          ; Operation = Op). %, print(translated(Op)),nl.

get_parameters(RestPair,Paras) :- member(pair(params,obj(Paras)),RestPair),!.
%get_parameters(RestPair,Paras) :- member(pair(parameters,Paras),RestPair), !. % old style
get_parameters(_,[]). % e.g., for initialise machine

order_paras([],OpName,Rest,[]) :- (Rest=[] -> true ; format('*** Unknown parameters for ~w: ~w~n',[OpName,Rest])).
order_paras([ID|T],OpName,Env,[Val|TVal]) :-
   (select(bind(ID,Val),Env,Env2) -> order_paras(T,OpName,Env2,TVal)
    ; %add_error(translate_json_operation,'Parameter not specified: ',ID:Env), % keep Val a variable
      format('*** Parameter not specified for ~w: ~w~n',[OpName,ID]),
      order_paras(T,OpName,Env,TVal)
   ).

% TO DO: using eval_strings is very ugly, use a better API predicate
translate_json_para(pair(Name,string(C)),bind(Name,Value)) :- !,
   eval_strings:eval_expression_codes(C,_Res,_EnumWarning,_LocalState,_Typed,_TypeInfo),
   eval_strings:last_expression(_Type,_Expr,Value).
%translate_json_para(string(C),Value) :- !, % old style
%   eval_strings:eval_expression_codes(C,_Res,_EnumWarning,_LocalState,_Typed,_TypeInfo),
%   eval_strings:last_expression(_Type,_Expr,Value).
%   %print(eval(S,Value)),nl.
translate_json_para(Para,_) :- add_error(translate_json_para,'Unknown JSON para:',Para),fail.

is_a_reset_operation(start_cspm_MAIN).
is_a_reset_operation(start_csp_MAIN).
is_a_reset_operation(Op) :- functor(Op,'$setup_constants',_).
is_a_reset_operation(Op) :- functor(Op,'$initialise_machine',_),
                            b_mode, \+ b_machine_has_constants_or_properties.

:- use_module(bmachine, [b_machine_has_constants_or_properties/0]).
perform_initialisation :- b_mode, b_machine_has_constants_or_properties,!,
   perform_sequence_of_operations(['$setup_constants','$initialise_machine']).
perform_initialisation :- perform_sequence_of_operations(['$initialise_machine']).

:- dynamic max_nr_replayed/1.
update_max_nr_replayed(Ident) :-
   (max_nr_replayed(Nr),Nr >= Ident -> true
    ; retract(max_nr_replayed(_)), assert(max_nr_replayed(Ident))).

perform_sequence_of_operations(Trace) :- retractall(max_nr_replayed(_)),
   assert(max_nr_replayed(0)),
   perform_sequence_of_operations2(Trace,0).

perform_sequence_of_operations2([],Ident) :-
   (silent_mode(on) -> true ; ident(Ident),print('SUCCESS'),nl).
perform_sequence_of_operations2([Op|T],Ident) :-
   get_error_context(Context),
   state_space:current_state_id(ID),get_operation_name(Op,FOP),
   set_error_context(operation(FOP,ID)),
   perform_one_operation_in_sequence(Op,Ident),
   update_max_nr_replayed(Ident),
   restore_error_context(Context),
   Ident1 is Ident+1,
   (get_preference(deterministic_trace_replay,true) -> ! ; true),
   perform_sequence_of_operations2(T,Ident1).

:- use_module(store,[lookup_value_for_existing_id/3]).

perform_one_operation_in_sequence('$non_det_modified_vars'(ActionAsTerm,NonDetVars),Ident) :- !,
   current_expression(_CurID,State),
   current_options(Options),
   member( (_Id,Action,ActionOpAsTerm,NewID), Options),
   unify_action_as_term(ActionAsTerm,ActionOpAsTerm,State),
   %ident(Ident), format('checking non_det_modified_vars for ~w in ~w~n',[ActionAsTerm,NewID]),nl,
   check_non_det_vars(NewID,NonDetVars),
   !,
   perform_action_tclk(Action,ActionOpAsTerm,NewID,Ident).
perform_one_operation_in_sequence('$non_det_modified_vars'(Op,_),Ident) :- !,
   ident(Ident), format('ignoring non_det_modified_vars for ~w~n',[Op]), % TO DO: use constraint-based finding of successor
   perform_one_operation_in_sequence(Op,Ident).
perform_one_operation_in_sequence('$check_value'(ID,VAL),Ident) :- !,
   ident(Ident), format('checking value of ~w ',[ID]),
   state_space:current_expression(_,State),
   expand_const_and_vars_to_full_store(State,EState),
   lookup_value_for_existing_id(ID,EState,StoredVal),
   translate:print_bvalue(StoredVal),
   (unify(StoredVal,VAL) -> print_green(' ok'),nl
     ; print_red(' failed'),nl, %print(StoredVal),nl, print(VAL),nl,
       fail).
perform_one_operation_in_sequence(Op,Ident) :-
   (silent_mode(on) -> true
    ; ident(Ident),print('<- '),print(Ident), print(': '), my_print_event(Op,OpStr),
      print(' :: '),
      state_space:current_state_id(CurID), print(CurID),nl
    ),
   ((functor(Op,'$initialise_machine',_),\+(state_space:current_expression(_,concrete_constants(_)))
      ; is_a_reset_operation(Op))
     ->  tcltk_interface:tcltk_goto_state(reset,root),
         tcltk_interface:tcltk_get_options(_)
     ;   true),
   convert_action(Op,COp),
   if(perform_single_operation(COp,Ident),true,
      (silent_mode(off),
       ident(Ident), print_red(' **** FAIL ****'),nl,
       %% trace, perform_single_operation(COp,Ident),
       fail)),
   (get_preference(deterministic_trace_replay,true) -> ! ; true),
   (silent_mode(on) -> true
    ; ident(Ident), print(' | '), print(Ident), print(': '), print(OpStr),
      %functor(Op,OpName,_), print(OpName),
      %((OpName = '-->', arg(1,Op,A1), functor(A1,RealOpName,_)) -> print(RealOpName) ; true),
      print_green(' success -->'),
      state_space:current_state_id(NewCurID),
      statistics(walltime,[Tot,Delta]),
      format('~w  [~w (Delta ~w) ms]~n',[NewCurID,Tot,Delta])
   ).

my_print_event(Op,OpStrTA) :-
        temporary_set_preference(expand_avl_upto,4,CHNG),
        translate_event(Op,OpStr),
        truncate_atom(OpStr,100,OpStrTA),
        reset_temporary_preference(expand_avl_upto,CHNG),
        write(OpStrTA).

/* convert global constants such as b with S={a,b,c} into fd(2,'S') */
convert_action(V,R) :- var(V),!,R=V.
convert_action(io(V,Ch),R) :- !, strip_dots(V,SV),R=io(SV,Ch,_SRCSPAN).
convert_action(io(V,Ch,_),R) :- !, strip_dots(V,SV),R=io(SV,Ch,_SRCSPAN).
convert_action(tick(_SRCSPAN),R) :- !, R = tick(_).
convert_action(X,Res) :- nonvar(X),X=..[F|Args],!,
    l_convert_term(Args,CArgs),Res=..[F|CArgs].
convert_action(X,X).

convert_term(V,R) :- var(V),!,R=V.
convert_term(closure(A,B,C),R) :- !, R=closure(A,B,C). % we may mess with identifiers; see test 1575
convert_term(term(bool(0)),R) :- !, R=pred_false /* bool_false */.
convert_term(term(bool(1)),R) :- !, R=pred_true /* bool_true */.
convert_term(bool_true,pred_true /* bool_true */) :- !.
convert_term(bool_false,pred_false /* bool_false */) :- !.
convert_term(string(S),R) :- !, R=string(S).
convert_term(X,Representation) :- atomic(X), lookup_global_constant(X,Representation),!.
convert_term(X,Res) :- nonvar(X),X=..[F|Args],!,
    l_convert_term(Args,CArgs),Res=..[F|CArgs].
convert_term(X,X).

l_convert_term([],[]).
l_convert_term([H|T],[CH|CT]) :- convert_term(H,CH), l_convert_term(T,CT).

/* newer version of ProB no longer has dot() constructor */
strip_dots([],[]).
strip_dots([H|T],[X|ST]) :- (H=dot(X) -> true ; X=H), strip_dots(T,ST).
strip_dots(tail_in(X),[in(XR)]) :-
   (X=dotTuple(R) -> XR=tuple(R) ; XR = X). /* we also no longer use tail_in */

get_state(concrete_constants(X),R) :- !, R=X.
get_state(const_and_vars(_,X),R) :- !, R=X.
get_state(X,X).

% check whether state NewID contains all the bindings in NonDetVars
check_non_det_vars(NewID,NonDetVars) :-
   visited_expression(NewID,State),
   get_state(State,VarState),
   maplist(check_non_det_bind(VarState),NonDetVars).

check_non_det_bind(State,bind(Var,Val)) :- member(bind(Var,StoredVal),State), unify(StoredVal,Val).

:- use_module(state_space,[current_state_id/1, current_expression/2, visited_expression_id/1]).
perform_single_operation('$jump'(TO,FROM),Ident) :- !,
   current_state_id(ID),
   print_red('Warning: trace file contains JUMP!'),nl,
   (ID=FROM -> true ; ident(Ident),format('Cannot perform jump from ~w (to ~w)~n',[FROM,TO])),
   (visited_expression_id(TO) -> tcltk_interface:tcltk_goto_state(jump,TO)
     ; add_error(trace_checking_fails,'State ID does not exist for jump:',TO),fail
   ).
perform_single_operation(ActionAsTerm,Ident) :-
   % tools_printing:print_term_summary(perform_single_op(ActionAsTerm)),
   current_expression(_CurID,State),
   current_options(Options),
   ( if( find_action_in_options(ActionAsTerm,Options,State, Action,ActionOpAsTerm,NewID),
         perform_action_tclk(Action,ActionOpAsTerm,NewID,Ident),
         if(perform_single_operation_retry(Options,ActionAsTerm,Action,NewID,State),
              perform_action_tclk(Action,_AlternateActionOpAsTerm,NewID,Ident),
              if(perform_custom_operation_retry(ActionAsTerm,_NewID),true,
                 (replace_result_by_variable(ActionAsTerm,A2),
                  find_action_in_options(A2,Options,State, Action,_,NewID),
                  format_with_colour(user_output,[red],'Result of operation call in trace does not match, e.g., ~w leading to state ~w~n',[Action,NewID]),
                  fail
                 )
                )
            )

        )
    ;
        single_operation_skip(ActionAsTerm,Ident)
   ),
   (get_preference(deterministic_trace_replay,true) -> ! ; true).
perform_single_operation(ActionAsTerm,Ident) :-
   current_options(Options),
   member( (_Id,PActionS,'*permute*',NewID), Options),  /* permute action from symmetry reduction */
   perform_action_tclk(PActionS,'*permute*',NewID,Ident), !,
   (silent_mode(on) -> true
    ; ident(Ident), print(' | ** PERMUTING **'),nl),
   perform_single_operation(ActionAsTerm,Ident).
perform_single_operation(ActionAsTerm,Ident) :-
   %%user:current_expression(CurID,State),
   \+ functor(ActionAsTerm,tau,1),
   current_options(Options), %print(options(Options)),nl,
   member( (_Id,AX,tau(X),NewID), Options),
   (silent_mode(on) -> true
    ; ident(Ident), print(' | ** SKIPPING ADDITIONAL TAU in SPEC. **'),nl),
   perform_action_tclk(AX,tau(X),NewID,Ident),
   perform_single_operation(ActionAsTerm,Ident).
%perform_single_operation(ActionAsTerm,Ident) :-
%   ident(Ident),print(unsuccessful(ActionAsTerm)),nl,fail.


perform_action_tclk(Action,ActionOpAsTerm,NewID,_Ident) :-
    my_perform_action_as_string(Action,ActionOpAsTerm,NewID),
    tcltk_interface:tcltk_get_options(_).
perform_action_tclk(_Action,_ActionOpAsTerm,_NewID,Ident) :- % undo action
    state_space:current_state_id(FromID),
    tcltk_interface:tcltk_backtrack, tcltk_interface:tcltk_get_options(_),
    state_space:current_state_id(ToID),
    silent_mode(off),
    ident(Ident),
    print(backtrack_from_to(FromID,ToID)),nl,
    fail.

% a more flexible version: allow string representations of ProB to evolve compared to when test suites where stored
my_perform_action_as_string(ActionStr,ActionOpAsTerm,NewID) :-
    tcltk_interface:tcltk_perform_action_string(ActionStr,ActionOpAsTerm,NewID).

% succeeds if we have an operation with results and then replaces results by fresh variable
replace_result_by_variable('-->'(Op,_),'-->'(Op,_)).

find_action_in_options(ActionAsTerm,Options,State, Action,ActionOpAsTerm,NewID) :-
    member( (_Id,Action,ActionOpAsTerm,NewID), Options),
    unify_action_as_term(ActionAsTerm,ActionOpAsTerm,State).

:- use_module(library(avl)).
unify_action_as_term(A,A,_) :- !.
unify_action_as_term('-->'(Op,Results),'-->'(Op2,Results2),State) :- !,
    unify_action_as_term(Op,Op2,State),
    l_unify(Results,Results2).
unify_action_as_term(A,B,State) :- functor(A,F,N), functor(B,F2,N2),
    unify_functor(F,F2,State),
    (N=N2 -> true
     ; formatsilent('*** SAME EVENT WITH VARYING ARITY: ~w : ~w vs ~w arguments ***~n',[F,N,N2]),
       % Probably due to show_eventb_any_arguments differently set ?
       fail),
    A=..[F|AA], B=..[F2|BB],
    l_unify(AA,BB).
l_unify([],[]).
l_unify([H|T],[I|S]) :- unify(H,I), l_unify(T,S).


unify_functor('initialise_machine','$initialise_machine',_) :- !. % old-style format
unify_functor('setup_constants','$setup_constants',root) :- !. % old-style format
unify_functor(X,X,_).

%:- use_module(custom_explicit_sets,[equal_explicit_sets/4]).
:- use_module(debug).
unify(A,A) :- !.
unify(A,avl_set(B)) :- %print(exp(A,B)),nl,
    expand_explicit_set(A,AA),
    %nl,translate:print_bvalue(AA),nl, translate:print_bvalue(avl_set(B)),nl, % print(chck(A,AA,B)),nl,
   AA = avl_set(B),!.
%unify(A,B) :- print(unify(A,B)),nl,fail.
unify(C1,C2) :- is_a_set(C1),(C2=closure(_P,_T,_B) ; C2=global_set(_)),
   %print(unify_closure(C1,C2)),nl,
   (C1=[_|_] -> custom_explicit_sets:convert_to_avl(C1,C11) ; C11=C1), % TO DO: replace by proper expansion predicate
   (unify_explicit_sets(C11,C2)
    -> true
     ; print(ko_unify_closure),nl, translate:print_bvalue(C11),nl, translate:print_bvalue(C2),nl,
       %terms:term_hash((C1,C2),H), print(hash(H)),nl, (H=140674321 -> trace, unify_explicit_sets(C11,C2)),
       fail),!.
 %avl_domain(B,R), print(chck(A,B,R)),nl,l_unify(A,R).
unify(C1,C2) :- %tools_printing:print_term_summary(fail(C1,C2)),nl,
  debug_println(9,unify_failed(C1,C2)),fail.

is_a_set(closure(_,_,_)).
is_a_set([]).
is_a_set([_|_]).
is_a_set(global_set(_)).
is_a_set(avl_set(_)).

expand_explicit_set([H|T],AA) :- !, custom_explicit_sets:convert_to_avl([H|T],AA).
expand_explicit_set(A,AA) :-
   on_enumeration_warning(custom_explicit_sets:try_expand_and_convert_to_avl_if_smaller_than(A,AA,200), fail).
unify_explicit_sets(C1,C2) :-
   ALLOW=no_expansion, %(C1 = avl_set(_) -> ALLOW=allow_expansion ; ALLOW=no_expansion),
   on_enumeration_warning(custom_explicit_sets:equal_explicit_sets4(C1,C2,ALLOW,no_wf_available), fail).


perform_single_operation_retry(Options,ActionAsTerm,Action,NewID,root) :-
   functor(ActionAsTerm,'setup_constants',_), !, % rename to new format
   nl,print('DEPRECATED TRACE FILE'),nl,nl,nl,
   ActionAsTerm =.. [_|Args], ActionAsTerm2 =..['$setup_constants'|Args],
   perform_single_operation_retry(Options,ActionAsTerm2,Action,NewID).
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID,State) :-
   (State=root ; State=concrete_constants(_)),
   functor(ActionAsTerm,'initialise_machine',_), !, % rename to new format
   nl,print('DEPRECATED TRACE FILE'),nl,nl,nl,
   ActionAsTerm =.. [_|Args], ActionAsTerm2 =..['$initialise_machine'|Args],
   perform_single_operation_retry(Options,ActionAsTerm2,Action,NewID).
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID,_State) :-
   perform_single_operation_retry(Options,ActionAsTerm,Action,NewID).


% check that parameters provided for setup_constants in trace file can be found in State
%check_initialised_args([],_).
%check_initialised_args([V|T],State) :-
%   select(bind(_,V2),State,Rest),
%   unify(V,V2),
%   %print('SELECTED: '),translate:print_bvalue(V2),nl,
%   !,
%   check_initialised_args(T,Rest).
%check_initialised_args([V|T],State) :-
%    print('MISMATCH'),nl, %print(mismatch(V,C)),nl,
%    print('COULD NOT FIND: '), translate:print_bvalue(V),nl,
%    tools_printing:print_term_summary(V),nl,
%    %print(V),nl, print(State),nl,
%    V=closure(_,_,_), memberchk(bind(_,closure(_,_,_)),State),
%    check_initialised_args(T,State).

perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,'$setup_constants',_), !,
   (get_prob_application_type(tcltk) -> OTHER = ['$partial_setup_constants'] ; OTHER = []),
   perform_alternative_op_with_same_functor(Options, ['$setup_constants' | OTHER],
                                            'set up constants', 'initialisation of constants',
                                            ActionAsTerm, Action, NewID),
   %visited_expression(NewID,concrete_constants(C)), ActionAsTerm =.. [_|Args],
   % check that state NewID corresponds to parameters provided for setup_constants in trace file
   %print(check_args),nl, check_initialised_args(Args,C), print('argsok'),nl.  % TO DO: try and enable this for all tests
   true.
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,'$initialise_machine',_),!,
   perform_alternative_op_with_same_functor(Options, ['$initialise_machine'],
                                            'initialise', 'initialisation',
                                            ActionAsTerm, Action, NewID),
     true. %visited_expression(NewID,S), (S=const_and_vars(_,VS) -> true ; VS=S), ActionAsTerm =.. [_|Args],
     %check_initialised_args(Args,VS), print('initargsok'),nl.
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,tau,_), arg(1,ActionAsTerm,Arg), nonvar(Arg),!,
   perform_alternative_op_with_same_functor(Options, [tau],
                                            tau, tau,
                                            ActionAsTerm, Action, NewID).
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,F,0), preference(show_eventb_any_arguments,true),
    /* then allow also the same operation name but with more arguments */
   member( (_Id,Action,ActionAsTerm2,NewID), Options),
   functor(ActionAsTerm2,F,FArity), FArity>0,
   formatsilent('% Allowing extra Event-B style ANY arguments.~n',[]).
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,F,N),N>0, preference(show_eventb_any_arguments,false),
    /* then allow also the same operation name but with more arguments */
   member( (_Id,Action,F,NewID), Options),
   formatsilent('% Allowing operation without Event-B style ANY arguments.~n',[]).
%perform_single_operation_retry(Options,_ActionsAsTerm,_Action,_NewID) :-
%   print_message('No more options: '), print(Options),nl,fail.

:- use_module(runtime_profiler,[profile_single_call/3]).
:- use_module(tools,[start_ms_timer/1, stop_ms_timer_with_msg/2]).
:- use_module(tcltk_interface,[tcltk_add_user_executed_operation_typed/3, tcltk_get_options/1]).

perform_custom_operation_retry(OpName,NewID) :-
   (OpName = '$initialise_machine' ; OpName = '$setup_constants'),
    !,
    state_space:current_state_id(CurID),
    (max_reached_for_node(CurID) ; not_all_transitions_added(CurID)),
    preferences:get_preference(maxNrOfInitialisations,0),
    tcltk_add_user_executed_operation_typed(OpName,b(truth,pred,[]),NewID),
    (get_preference(deterministic_trace_replay,true) -> ! ; true),
    print_green(init_found(NewID)),nl.
perform_custom_operation_retry(ActionAsTerm,NewID) :-
    (ActionAsTerm = '-->'(_,_) % for operations with return types we cannot provide a predicate to solve yet
      -> preferences:get_preference(maxNrOfEnablingsPerOperation,0)
      ;  true),
    state_space:current_state_id(CurID),
    get_operation_name(ActionAsTerm,OpName),
    (max_reached_for_node(CurID) ; time_out_for_node(CurID,OpName,_) ; not_all_transitions_added(CurID)),
     debug_println(9,try_custom(CurID,ActionAsTerm)), % Execute operation by predicate
    % (state_space:transition(CurID,AA,BB,CC),format('Successor of ~w: ~w --~w--> ~w~n',[CurID,AA,BB,CC]),fail ; true),
    get_operation_return_values_and_arguments(ActionAsTerm,ExpectedResults,OpArgs),
    % TO DO: optionally check return values
    (OpArgs = [] -> Pred = b(truth,pred,[])
     ; \+ b_is_operation_name(OpName)
       -> add_error(trace_checking_fails,'Operation does not exist in loaded specification:',OpName),fail
     ; b_get_machine_operation_typed_parameters(OpName,TArgParas),
       l_generate_predicate(OpArgs,TArgParas,OpName,Conjuncts),
       conjunct_predicates(Conjuncts,Pred)
    ),
    !,
    %print('PRED: '),translate:print_bexpr(Pred),nl,
    start_ms_timer(Timer),
    profile_single_call(OpName,CurID,
           tcltk_interface:tcltk_add_user_executed_operation_typed(OpName,OpTerm,Pred,NewID) % The Pred can only talk about parameters and the state before; thus this only works for setup_constants and initialise_machine
    ),
    (get_preference(deterministic_trace_replay,true) -> ! ; true), % by preventing backtracking we can reduce memory consumption
    print_green(successor_found(NewID, OpName)),nl,
    get_operation_return_values_and_arguments(OpTerm,ActualResults,_),
    (l_unify(ExpectedResults,ActualResults) -> true
      ;  ajoin(['Result values for ',OpName,' do not match for transition from ',CurID,' to ',NewID,':'],Msg),
         translate_bvalue(ExpectedResults,ER),
         translate_bvalue(ActualResults,AR),
         add_warning(b_trace_checking,Msg,[expected(ER),actual(AR)])
    ),
    (debug_mode(on) -> stop_ms_timer_with_msg(Timer,'Custom: ') ; true),
    garbage_collect, % important to keep memory consumption down for long traces e.g. for Innotrans VBF traces
    (debug_mode(on) -> print_memory_used_wo_gc,nl ; true),
    tcltk_get_options(_).

l_generate_predicate([],[],_,[]) :- !.
l_generate_predicate([],_,Op,[]) :- !, add_warning(b_trace_checking,'Trace file contains too few parameters for: ',Op).
l_generate_predicate(_,[],Op,[]) :- !, add_warning(b_trace_checking,'Trace file contains too many parameters for: ',Op).
l_generate_predicate([ArgVal|TArg],[TypedPara|TPar],OpName,[Pred|TPred]) :-
    generate_predicate(ArgVal,TypedPara,Pred),
    l_generate_predicate(TArg,TPar,OpName,TPred).
generate_predicate(ArgVal,TypedPara,b(equal(TypedPara,TVal),pred,[])) :-
    get_texpr_type(TypedPara,Type),
    TVal = b(value(ArgVal),Type,[]).

perform_alternative_op_with_same_functor(Options,Functors,ActionMsg,AltMsg,ActionAsTerm,Action,NewID) :-
    \+ member( (_,_,ActionAsTerm,_), Options),
    (silent_mode(on) -> true
      ; ajoin(['Could not ', ActionMsg, ' with parameters from trace file.'], Msg1),
        ajoin(['Will attempt any possible ', AltMsg, '.'], Msg2),
        print_message(Msg1), print_message(Msg2)
    ),
    %print(ActionAsTerm),nl, print(Options),nl,nl,
    member( (_,ActionStr,MachineActionAsTerm,NewID), Options),
    functor(MachineActionAsTerm,Functor,_),
    member(Functor,Functors),
    (Action=ActionStr -> true ; print(cannot_match(Action,ActionStr)),nl,fail).

single_operation_skip(tau(_),Ident) :-
   (silent_mode(on) -> true
    ; ident(Ident), print(' | *** TRYING TO SKIP TAU IN TRACE FILE'),nl).



ident(0).
ident(N) :- N > 40, !, N1 is N mod 40, ident(N1).
ident(N) :- N>0, print('- '), N1 is N-1, ident(N1).


/* ------------------------------------------------------------------ */

:- use_module(state_space,[get_action_trace/1]).
print_trace_as_fdr_check :-
   print('-- Trace Check Generated by ProB:'),nl,
   get_action_trace(T),
   reverse(T,RT),
   print('PROB_TEST_TRACE = '),print_fdr(RT),
   nl,
   print('assert MAIN [T= PROB_TEST_TRACE'),nl.

print_fdr([]) :- print('STOP'),nl.
print_fdr([jump|T]) :- print_fdr(T).
print_fdr([action(Str,Term)|T]) :-
   (Term = i(_)
      -> (nl,print('   ( STOP /\\ '))
    ; (Term = tick(_)
       -> print('SKIP ; ')
       ;  ((Term = tau(_) ; Term = 'start_cspm_MAIN')
           -> true
           ;  (print(Str), print(' -> '))
          )
   )),
   print_fdr(T),
   (Term = i(_)
      -> print('   )')
      ;  true
   ).

% -----------------------

tcltk_save_history_as_trace_file(Style,File) :-
   formatsilent('% Saving history (~w format) to: ~w~n',[Style,File]),
   tell(File),
   (specfile:b_or_z_mode
    -> b_machine_name(OurName)
     ; specfile:currently_opened_file(OurName)
   ),
   (Style=prolog -> print_quoted(machine(OurName)), print('.'),nl ; true),
   call_cleanup(print_trace_for_replay(Style),
                (told,print_message(done))).

:- use_module(state_space,[transition/4, op_trace_ids/1]). % transition(CurID,Term,TransId,DestID)

% print the animator's history trace for later replay
% prolog style is currently used by ProB Tcl/Tk
% json style is used by ProB2 UI
print_trace_for_replay(Style) :-
   op_trace_ids(OpIds),
   reverse(OpIds,Trace),
   print_trace_for_replay(Style,Trace).

print_trace_for_replay(prolog,Trace) :- !, print_transition_list_prolog(Trace).
print_trace_for_replay(json,Trace) :- !, print_transition_list_json(Trace).
print_trace_for_replay(Style,_) :-
   add_internal_error('Style not supported: ',print_trace_for_replay(Style,_)).

print_transition_list_prolog(Trace) :-
   member(OpId,Trace),
   transition(PrevId,Op,OpId,SuccId),
   (atomic(Op),
    ( transition(PrevId,Op,_,SuccId2), SuccId2 \= SuccId -> true), % another transition with same label Op exists
      translate:get_non_det_modified_vars_in_target_id(Op,SuccId,NonDetVars)
    ->  print_quoted('$non_det_modified_vars'(Op,NonDetVars))
     ;  print_quoted(Op)
   ),
   write('.'),nl,
   fail.
print_transition_list_prolog(_).

:- use_module(version, [version_str/1, revision/1]). % , lastchangeddate/1
:- use_module(library(system),[ datime/1]).
:- use_module(specfile,[currently_opened_file/1]).

% print list in JSON format for ProB2 UI:
print_transition_list_json(Trace) :-
   format('{~n  "transitionList": [~n',[]),
   print_json_list(Trace,root),
   format('  ]~n }{~n',[]),
   datime(datime(Yr,Mon,Day,Hr,Min,_Sec)),
   format('  "Creation Date": " ~w/~w/~w ~w:~w",~n',[Day,Mon,Yr,Hr,Min]),
   version_str(Version),
   format('  "ProB CLI Version": "~w",~n',[Version]),
   revision(Rev),
   format('  "ProB CLI Revision": "~w",~n',[Rev]),
   currently_opened_file(File),
   format('  "Model": "~w"~n',[File]),
   format('}~n',[]).

print_json_list([],_).
print_json_list([H|T],FromID) :-
   (H=jump(NewID) -> true
     ; print_json_opid(H,FromID,NewID) -> true
     ; add_error(print_json_list,'Writing operation failed: ',H), NewID=FromID),
   (T==[] -> nl ; write(','),nl, print_json_list(T,NewID)).

:- use_module(specfile,[get_operation_internal_name/2,
                        get_operation_return_values_and_arguments/3,
                        get_possible_language_specific_top_level_event/3]).

% print an operation id (transition id) for a JSON Trace
print_json_opid(OpId,PrevId,SuccId) :- %print(Op),nl, trace,
   (transition(PrevId,Op,OpId,SuccId) -> PrevId1=PrevId
     ; add_warning(print_json_list,'Cannot find operation id from given source id: ',OpId:from(PrevId):to(SuccId)),
       transition(PrevId1,Op,OpId,SuccId)
   ),
   get_operation_internal_name(Op,OpName), % TO DO: currently this is INITIALISATION instead  of $initialise_machine
   (get_possible_language_specific_top_level_event(OpName,ResultNames,ParaNames) -> true
     ; ResultNames=unknown, ParaNames=unknown),
   get_operation_return_values_and_arguments(Op,ReturnValues,Paras),
   % TO DO: extract variables and constants for $initialise_machine, $setup_constants
   format('   {~n',[]),
   format('   "name": "~w",~n',[OpName]),
   format('   "params": {~n',[]),
   (get_preference(show_eventb_any_arguments,true),
    \+ same_length(Paras,ParaNames)
    -> print_json_paras(Paras,1,unknown)
     ; print_json_paras(Paras,1,ParaNames)
    ),
   format('     },~n',[]),
   format('   "results": {~n',[]),
   print_json_paras(ReturnValues,1,ResultNames),
   format('     },~n',[]),
   format('   "destState": {~n',[]),
   get_change_list(PrevId1,SuccId,ChangedVarNames,ChangedValues),
   print_json_paras(ChangedValues,1,ChangedVarNames),
   format('     },~n',[]),
   format('   "destStateNotChanged": [~n',[]),
   get_unchanged_list(PrevId1,SuccId,UnChangedVarNames),
   print_json_names(UnChangedVarNames),
   format('     ],~n',[]),
   format('   "preds": null~n',[]),
   format('   }',[]).

:- use_module(translate, [translate_bvalue/2]).
:- use_module(tools,[b_escape_string_atom/2]).
mytranslate(H,ES) :- translate_bvalue(H,S), b_escape_string_atom(S,ES).

% print_json_paras(ListOfValues,ParameterNr,ListOfParameterNames)
print_json_paras([],_,_) :- !.
print_json_paras([H|T],Nr,[HName|NT]) :- !, Nr1 is Nr+1,
      b_escape_string_atom(HName,EN),
      mytranslate(H,ES),
      format('        "~w": "~w"',[EN,ES]),
      (T==[] -> nl ; write(','),nl, print_json_paras(T,Nr1,NT)).
print_json_paras([H|T],Nr,unknown) :- Nr1 is Nr+1,
      mytranslate(H,ES),
      !,
      format('        "para~w": "~w"',[Nr,ES]),
      (T==[] -> nl ; write(','),nl, print_json_paras(T,Nr1,unknown)).
print_json_paras(P,Nr,PN) :- add_internal_error('Cannot print JSON parameters:',print_json_paras(P,Nr,PN)).

% print a list of identifier names for JSON
print_json_names([]).
print_json_names([HName|T]) :-
      b_escape_string_atom(HName,EN),
      (T=[] -> format('        "~w"~n',[EN])
            ;  format('        "~w",~n',[EN])),
      print_json_names(T).

% find change bindings in the state
changed_binding(FromId,ToId,Bind) :-
      visited_expression(FromId,FromState),
      visited_expression(ToId,ToState),
      get_id_binding(ToState,Bind),
      \+ get_id_binding(FromState,Bind).

% find variables which have not been modified in the state
unchanged_binding(FromId,ToId,Bind) :-
      visited_expression(FromId,FromState), % slightly inefficient to expand states again
      visited_expression(ToId,ToState),
      get_id_binding(ToState,Bind),
      get_id_binding(FromState,Bind).

get_change_list(FromId,ToId,ChangedVarNames,ChangedValues) :-
     findall(Bind,changed_binding(FromId,ToId,Bind),List),
     maplist(decompose_binding,List,ChangedVarNames,ChangedValues).
get_unchanged_list(FromId,ToId,UnChangedVarNames) :-
     findall(Bind,unchanged_binding(FromId,ToId,Bind),List),
     maplist(decompose_binding,List,UnChangedVarNames,_).

decompose_binding(bind(ID,Val),ID,Val).


get_id_binding(const_and_vars(_,List),Bind) :- !, member(Bind,List).
get_id_binding(concrete_constants(List),Bind) :- !, member(Bind,List).
get_id_binding(List,Bind) :- !, member(Bind,List).

