% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(btypechecker,
          [btype/7,                % typecheck B syntax node
           btype_l/6,        % typecheck list of B syntax nodes
           btype_ground/6,   % typecheck and check if all new identifier have ground types
           btype_ground_dl/7, % typecheck and check if all new identifier have ground types; difference-list version
           btype_ground_dl_in_section/8, % version of above with additional context name parameter
           check_ground_types_dl/4, % check if a list of identifiers has only ground types
           unify_types_werrors/5, % unify two types, return list of errors in case of type mismatch
           unify_types_werrors_l/5, % unify two list of types
           unify_types_strict/2, % unify two types, just fail in case of type mismatch

           ext2int/7, % convert a (raw) syntax element into a typed element
           remove_pos/2, % remove position info from a (raw) syntax element
           % TODO[DP, 15.03.2013]: In case of a re-structuring of modules,
           % maybe it is a good idea to move the type environment predicates to a module of its own
           env_empty/1,  % create an empty type environment
           env_lookup_type/3, % lookup a type of an identifier in the environment
           env_lookup_infos/3, % lookup additional infos of an identifier in the environment
           env_store/5, % store the type of an identifier in the environment
           env_store_operator/4, % store a theory operator into the environment
           env_store_definition/3, % store an untyped definition_decl/6 ast in the environment
           env_identifiers/2,
           add_identifiers_to_environment/3,
           operation_infos/1, % give a list of (uninstantiated) infos that operations need for type-checking
           compute_accessed_vars_infos_for_operation/8, % (re)-compute reads/write infos for operation

           couplise_list/2,     % [a,b,c] -> couple(couple(a,b),c)
           prime_identifiers/2, prime_identifier/2,
           prime_identifiers0/2, prime_identifier0/2, prime_atom0/2,

           fasttype/2, % manually create a syntax tree with type infos

           openenv/2,  % create an "open" type environment, where unknown identifiers will be added on demand
           openenv_identifiers/2, % get the introduced identifiers of an open type environment
           opentype/5, % type-check an expression without having every identifier declared in advance

           reset_typechecker/0,
           machine_string/1, % succeeds for every explicit string seen so far by the type checker
           store_typecheck_error/4
          ]).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'ProBs type checker and type inference module.').

:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(library(avl)).
:- use_module(library(ordsets)).

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(kernel_records,[normalise_record_type/2, record_has_multiple_field_names/2]).

:- use_module(translate).

:- use_module(bsyntaxtree).
:- use_module(b_ast_cleanup).
:- use_module(debug).
:- use_module(input_syntax_tree,[raw_replace/3]).
:- use_module(preferences,[get_preference/2, preference/2]).
:- use_module(b_read_write_info,[get_accessed_vars/5]).

% operator for type inference rules
:- op(800,xfx,::).

%:- dynamic op_modifies_reads_cache/3. % store modifies and reads information for operations
:- dynamic machine_string/1. /* keep a list of all explicit strings seen in the machine so far */
/* Probably better: do it in a separeate traversal; talk to Daniel about this */
reset_typechecker :- reset_machine_strings. %, retractall(op_modifies_reads_cache(_,_,_)).
reset_machine_strings :- retractall(machine_string(_)).
assert_machine_string(S) :- machine_string(S) -> true ; assert(machine_string(S)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% miniature B type checker

btype_ground(Exprs,Env,NonGroundExceptions,Types,TExprs,Errors) :-
    btype_ground_dl(Exprs,Env,NonGroundExceptions,Types,TExprs,Errors,[]).

:- dynamic current_type_checking_context/1.
get_current_context(C) :-
  (current_type_checking_context(SecName) -> ajoin([' in ',SecName],C) ; C='').

% a version of btype_ground_dl which stores current context
btype_ground_dl_in_section(SectionName,Exprs,Env,NonGroundExceptions,Types,TExprs,EIn,EOut) :-
    retractall(current_type_checking_context(_)),
    assert(current_type_checking_context(SectionName)),
    call_cleanup(btype_ground_dl(Exprs,Env,NonGroundExceptions,Types,TExprs,EIn,EOut),
       retractall(current_type_checking_context(_))).
% dl: difference list version
btype_ground_dl(Exprs,Env,NonGroundExceptions,Types,TExprs,EIn,EOut) :-
    ~~mnf(same_length(Exprs,Types)), ~~mnf(same_length(Exprs,TExprs)),
    ~~mnf(get_texpr_types(TExprs,Types)),
    ~~mnf(do_typecheck_dl(Exprs,Env,[ground(NonGroundExceptions)],TExprs,EIn,EOut)).

% do_typecheck type-checks the expression and checks if the types
% of introduced variables (like in quantifiers) are ground
do_typecheck(Exprs,Env,Options,TExprs,Errors) :-
    do_typecheck_dl(Exprs,Env,Options,TExprs,Errors,[]).

do_typecheck_dl(Exprs,Env,Options,TExprs,ErrorsIn,ErrorsOut) :-
    typecheck_result_init(ErrorsIn,Delayed,Identifiers,TRStart),
    % type-check the expressions
    btype_l(Exprs,Env,_Type,TExprs,TRStart,TR),
    typecheck_result_get_errors(TR,E1),
    typecheck_result_get_delayed_rules(TR,[]),
    typecheck_result_get_identifiers(TR,[]),
    foldl(trigger_delayed_rule,Delayed,E1,E2),
    % get the newly introduced identifiers and check if their types are
    % ground
    ( memberchk(ground(NonGroundExceptions),Options) ->
        check_ground_types_dl(Identifiers,NonGroundExceptions,E2,ErrorsOut)
    ; otherwise ->
        E2=ErrorsOut).

trigger_delayed_rule(dr(Trigger,Errors),Ein,Eout) :-
    % ground trigger, enforce blocked rule
    Trigger = 0,
    % by now Errors should have given a ground value by the co-routine
    add_all(Errors,Ein,Eout).
add_all([]) --> !.
add_all([E|Rest]) --> [E], add_all(Rest).


% Utilities for tr(.) DCG argument for type-checking
type_error_or_warning_occurred(TRin,TRout) :-
% TO DO ? : also check for warnings
    TRout=tr(Errors1,_,_,_),
    TRin =tr(Errors2,_,_,_), Errors1 \== Errors2.

typecheck_result_init(Errors,DR,Ids,tr(Errors,DR,Ids,[error_count(0,50,unknown)])). % 50: maximum number of errors reported; note: the error manager has itself an upper-bound on the number of errors reported

% typecheck_result_add_error(Error,TRin,TRout) usually this will be called with TRout a variable
typecheck_result_add_error(Error,tr(Errors,DR,Ids,Flags),tr(RestErrors,DR,Ids,NewFlags)) :-
   (select(error_count(N,Max,PrevPos),Flags,NF)
    -> N1 is N+1, NewFlags = [error_count(N1,Max,NewLastPos)|NF],
       (Error = error(_,Pos) -> NewLastPos = Pos ; NewLastPos = PrevPos),
       (N<Max -> Errors = [Error|RestErrors]
         ; Errors = RestErrors % we have reached the maximum: do not store error; we could throw an exception to stop typechecking altogether,  ,throw(tr_exceeded(Errors,DR,Ids,Flags))
        )
    ;  NewFlags = Flags, Errors = [Error|RestErrors]
    ).

typecheck_result_add_delayed_rule(Trigger,RuleErrors,
           tr(Errors,[dr(Trigger,RuleErrors)|Rules],Ids,Flags),tr(Errors,Rules,Ids,Flags)) :- !.
typecheck_result_add_delayed_rule(Trigger,RuleErrors,In,Out) :-
    add_internal_error('Could not add delayed rule: ',typecheck_result_add_delayed_rule(Trigger,RuleErrors,In,Out)),
    Out=In.

typecheck_result_add_identifier(Id,tr(Errors,Rules,[Id|Ids],Flags),tr(Errors,Rules,Ids,Flags)).

%typecheck_max_errors_exceeded(tr(_,_,_,Flags)) :-
%    member(error_count(N,Max,_),Flags),number(N),number(Max),N>Max.

typecheck_result_get_errors(tr(FullErrors,_DR,_Ids,Flags),RestErrors) :-
    ((member(error_count(N,Max,LastPos),Flags),number(N),number(Max),N>Max)
     -> tools:ajoin(['Maximum number (',Max,') of type errors exceeded: ',N],Msg),
        ExcErr=error(Msg,LastPos),
        %append(Errors,[ExcErr],FullErrors)
        FullErrors = [ExcErr|RestErrors]
     ; FullErrors = RestErrors).

typecheck_result_get_delayed_rules(tr(_Errors,Rules,_Ids,_Flags),Rules).

typecheck_result_get_identifiers(tr(_Errors,_Rules,Ids,_Flags),Ids).

%typecheck_result_finalize(tr([],[],[])). % not used at the moment

typecheck_result_get_flags(tr(_,_,_,Flags),Flags).

typecheck_add_flag(Flag,tr(Errors,Rules,Ids,Flags),tr(Errors,Rules,Ids,NewFlags)) :-
    (member(Flag,Flags) -> NewFlags=Flags ; NewFlags=[Flag|Flags]).

% btype/7 is the central predicate of the type-checker
% type-checking is done in three "stages":
%  btype/5:   extracts the obligatory position information from the input syntax element
%                decouples the type in the argument and the type of the expression by
%                  an explicit unify_types/5 call
%                calls btype1
%  btype1/7:  calls btype_rewrite for optional rewriting rules
%                if no more rewriting rules can be applied, btype2 is called
%  btype2/7:  does the actual type-checking

% btype(+Src,+Ext,+Env,Type,-Result,TRin,TRout):
%  Src:    Where the call comes from
%  Ext:    The AST of the expression to type check as it comes from the parser
%  Env:    The type environment (see the env_... predicates)
%  Type:   The type of the expression (could be as input or output)
%  Result: The type checked AST
%  TRin/TRout: typecheck_result data structure, see typecheck_result_... predicates
btype(Src,Ext,Env,Type,Result,TRin,TRout) :-
    (var(Ext)
      -> add_internal_error('btype Ext is variable: ',Src:Ext)
       ; true),
    if(btype_aux(Ext,Env,Type,Result,TRin,TRout),true,
       add_error_and_fail(btype,'btype failed: ',Src:Ext)
       %(add_error(btype,'btype failed: ',Src:Ext),trace, btype_aux(Ext,Env,Type,Result,TRin,TRout))
       ),
    (var(TRout) -> add_error(btype,'TRout is variable: ',Src:Ext) ; true).

btype_aux(Ext,Env,Type,Result,TRin,TRout) :-
    % convert the external representation Ext (first argument is the position)
    % into the typed representation
    ext2int(btype,Ext,Expr,Pos,Type,TExpr,Infos,Result0),
    % type-check the expression
    btype1(Expr,Pos,Env,TExpr,LType,Infos,NewPos,TRin,TR1),
    % an explicit unify decouples the expected and inferred type
    (Pos\=NewPos
      -> %debug_println(19,updating_position_info(Pos,NewPos)),
         update_pos(Result0,NewPos,Result)
      ;  Result = Result0),
    unify_types(Type,LType,Pos,TExpr,TR1,TRout).

update_pos(b(R0,T,Infos),NewPos,b(R0,T,[nodeid(NewPos)|Infos0])) :-
   delete(Infos,node_id(_),Infos0).

% btype1/7 tries to apply rewriting rules until no more rules
% are applyable, then btype2/7 is called
% btype1(+Expr,+Pos,+Env,-TExpr,-Type,-Infos,-NewPos,+TRin,-TRout):
% NewPos: potentially updated position information when definitions are expanded
btype1(Expr,Pos,Env,TExpr,Type,Infos,NewPos,TRin,TRout) :-
    catch( btype_try_rewrite(Expr,Pos,Env,TExpr,Type,Infos,NewPos,TRin,TRout),
           E,
           handle_rewrite_exception(E,Pos,TExpr,Infos,TRin,TRout)
         ).
btype_try_rewrite(Expr,Pos,Env,TExpr,Type,Infos,NewPos2,TRin,TRout) :-
    btype_rewrite(Expr,Env,Pos,NewExpr,NewPos),!,
    btype1(NewExpr,NewPos,Env,TExpr,Type,Infos,NewPos2,TRin,TRout).
btype_try_rewrite(Expr,Pos,Env,TExpr,Type,Infos,Pos,TRin,TRout2) :-
    btype2(Expr,Pos,Env,TExpr,Type,Infos,TRin,TRout1),
    btype_static_check(TExpr,Pos,Env,TRout1,TRout2).

% a few static checking rules
btype_static_check(let(IDList,Equalities,_Body),Pos,Env) -->
    btype_static_check_let(IDList,Equalities,Pos,no_reuse,Env). % here Atelier-B prohibits reuse
btype_static_check(let_expression(IDList,Equalities,_Body),Pos,Env) -->
    btype_static_check_let(IDList,Equalities,Pos,allow_reuse,Env). % here we allow reuse
btype_static_check(let_predicate(IDList,Equalities,_Body),Pos,Env) -->
    btype_static_check_let(IDList,Equalities,Pos,allow_reuse,Env).
btype_static_check(_,_,_) --> [].

btype_static_check_let(IDList,Equalities,Pos,AllowReuse,Env,TRin,TRout) :-
    def_get_texpr_ids(IDList,Ids), sort(Ids,SIds),
    check_let_equalities(Equalities,Pos,AllowReuse,SIds,[],NotDefined,_Defined,Env,TRin,TRout1),
    (NotDefined==[] -> TRout = TRout1
      ; type_error_or_warning_occurred(TRin,TRout1) ->
                format(user_error,'*** Let does not define variables: ~w~n',[NotDefined]),
                TRout1 = TRout % we have already raised errors
     ; %print(tr(TRin)),nl,print(tr(TRout1)),nl,
       ajoin_with_sep(['Let does not define variables:'|NotDefined],' ',Msg),
       store_typecheck_error(Msg,Pos,TRout1,TRout)).

% TO DO: we should maybe run these after the type checker has run
% AllowReuse: allow let identifiers to be re-used in later RHS sides; normally not allowed by Atelier-B
check_let_equalities(b(P,T,I),Pos,AllowReuse,ToDefine,AlreadyDefined,T2,A2,Env) -->
    check_let_equalities_aux(P,T,I,Pos,AllowReuse,ToDefine,AlreadyDefined,T2,A2,Env).
check_let_equalities_aux(P,Type,Info,Pos,AR,_,_,_,_,_) -->
    {var(P)},!,
    {add_internal_error('Illegal variable:',check_let_equalities_aux(P,Type,Info,Pos,AR,_,_,_,_,_))}.
check_let_equalities_aux(conjunct(L,R),_,_,Pos,AllowReuse,ToDefine,AlreadyDefined,T2,A2,Env) --> !,
   check_let_equalities(L,Pos,AllowReuse,ToDefine,AlreadyDefined,T1,Already1,Env),
   check_let_equalities(R,Pos,AllowReuse,T1,Already1,T2,A2,Env).
check_let_equalities_aux(equal(LHS,RHS),_,_,_Pos,AllowReuse,ToDefine,AlreadyDefined,T2,A2,Env) -->
   {get_texpr_id(LHS,LID)},
   %{print(equal(LID,ToDefine,AlreadyDefined,T2,A2)),nl},
    {select(LID,ToDefine,T2)}, % OK: we define an introduced ID
    !,
    {ord_add_element(AlreadyDefined,LID,A2)},
    {find_identifier_uses(RHS,[],Used)}, % can we already call this here ?? TO DO Check non-ground typing does not cause problem
    check_let_rhs(RHS,Used,ToDefine,AlreadyDefined,AllowReuse,Env).
check_let_equalities_aux(equal(LHS,RHS),Type,Info,Pos,_AllowReuse,T,A,T,A,_Env,TrIn,TrOut) :-
    translate_bexpression(b(equal(LHS,RHS),Type,Info),IStr), % are the types guaranteed to be instantiated ?
    (get_texpr_id(LHS,LID)
     -> (member(LID,A) -> Msg1 = 'LET redefines identifier: '
         ;  Msg1 = 'LET defines undeclared identifier: '
         )
      ;  Msg1 = 'LET contains an equality whose left-hand-side is not an identifier: '),
     ajoin([Msg1,IStr],Msg),
     (get_texpr_pos(LHS,PosM) -> true ; PosM=Pos),
    store_typecheck_error(Msg,PosM,TrIn,TrOut).
check_let_equalities_aux(IllegalPred,Type,Info,_Pos,_AllowReuse,T,A,T,A,_Env) -->
    {translate_bexpression(b(IllegalPred,Type,Info),IStr),
    ajoin(['LET contains a predicate which is not an equality: ',IStr],Msg)},
    store_typecheck_error(Msg,Info).

:- use_module(specfile,[z_mode/0]).
check_let_rhs(_RHS,Used,ToDefine,AlreadyDefined,_AllowReuse,Env) --> {z_mode}, !,
     % We do not seem to call this code in Z Mode !
     ({ord_union(ToDefine,AlreadyDefined,All),
       possible_ambiguity(Used,All,Env,ClashIds)}
       -> {format('*** Warning: possible clash in LET: ~w~n',[ClashIds])}
       ;  []
     ).
check_let_rhs(RHS,Used,ToDefine,AlreadyDefined,no_reuse,_Env) -->
    {get_preference(allow_let_to_reuse_introduced_ids,false)},
    {ord_union(ToDefine,AlreadyDefined,All), ord_intersection(Used,All,Inter)},
    {Inter\=[]},
    !,
    % potentially cyclic let
    {ajoin_with_sep(['LET right-hand-side uses left-hand introduced ids (set ALLOW_COMPLEX_LETS to TRUE to allow this):'|Inter],' ',Msg),
     get_texpr_info(RHS,Info)},
    store_typecheck_error(Msg,Info).
check_let_rhs(RHS,Used,ToDefine,_AlreadyDefined,allow_reuse,_Env) -->
    {ord_intersection(Used,ToDefine,Inter), Inter \= []}, % we use left-hand side ID not yet introduced
    !,
    % potentially cyclic let
    {ajoin_with_sep(['LET right-hand-side uses not-yet-defined identifiers:'|Inter],' ',Msg),
     get_texpr_info(RHS,Info)},
    store_typecheck_error(Msg,Info).
check_let_rhs(RHS,Used,_ToDefine,AlreadyDefined,allow_reuse,Env) -->
    {possible_ambiguity(Used,AlreadyDefined,Env,ClashIds)},
    % we use variables which are visible and introduced: could be very confusing
    !,
    {ajoin_with_sep(['Possible ambiguity: LET right-hand-side re-uses visible identifiers:'|ClashIds],' ',Msg),
     get_texpr_info(RHS,Info)},
    store_typecheck_error(Msg,Info).
check_let_rhs(_,_,_,_,_,_) --> [].

possible_ambiguity(Used,AlreadyDefined,Env,ClashIds) :-
     env_get_visible_ids(Env,Visible),
     ord_intersection(Used,AlreadyDefined,Inter1),
     ord_intersection(Inter1,Visible,ClashIds),
     ClashIds \= [].

handle_rewrite_exception(rewrite_exception(Msg,Name),Pos,TExpr,Infos,TRin,TRout) :- !,
    store_typecheck_error(Msg,Pos,TRin,TRout),
    % Type remains unbound, Infos are empty
    Infos = [],
    TExpr = identifier(Name).
handle_rewrite_exception(E,_Pos,_TExpr,_Infos,_TRin,_TRout) :- throw(E).

% btype2/8 encodes the type checking rules
btype2(identifier(I1), Pos, Env, identifier(I), Type, Infos, TRin, TRout) :-
    !, % In case the identifier is a list, join it with '.' (e.g. ['M',x] -> 'M.x')
    ( is_list(I1) -> ajoin_with_sep(I1,'.',I)
    ; otherwise   -> I1=I),
    % get the identifier's type from the type environment
    lookup_type(I,Env,Pos,Type,TRin,TR1),
    % copy also additional information from the environment,
    lookup_infos(I,Env,AllInfos),
    % but not all (e.g. no position information)
    extract_id_information(AllInfos,Infos),
    % an access to this identifier might be illegal, check this
    ( member(error(Error),AllInfos) -> store_typecheck_error(Error, Pos, TR1, TRout)
    ; otherwise -> TR1=TRout).
btype2(primed_identifier(Id,N), Pos, Env, identifier(FullId), Type, Infos, TRin, TRout) :-
    !, % a primed identifier is of the form x$0, we just convert it to
    % a normal identifier called "x$0" and type-check it.
    number_suffix(Id,N,FullId),
    btype(primed_identifier(Id,N),identifier(Pos,FullId),Env,Type,TId,TRin,TRout),
    get_texpr_info(TId,Infos).
btype2(forall(Ids,EImplication), _, Env, forall(TIdentifiers,TDomain,TPred), pred, [], TRin, TRout) :-
    !,add_ext_variables_with_info(Ids,Env,[introduced_by(forall)],TIdentifiers,Subenv,TRin,TR1),
    ext2int(btype2,EImplication,Implication,IPos,_,_,_,_),
    (btype_rewrite(Implication,Env,IPos,Implication2,IPos2) -> true % in case we have a definition we need to expand it now to transform it into an implication
     ; Implication2=Implication, IPos2=IPos),
    % we expect an implication as predicate, if not, generate an error
    ( Implication2=implication(Domain,Pred) ->
        btype(forall_domain(Ids),Domain,Subenv,pred,TDomain,TR1,TR2),
        btype(forall_pred(Ids),Pred,Subenv,pred,TPred,TR2,TRout)
    ; otherwise ->
        store_typecheck_error('expected implication in universal quantification ',IPos2,TR1,TR2),
        create_texpr(truth,pred,[],TDomain),
        btype(foralle_impl(Ids),EImplication,Subenv,pred,TPred,TR2,TRout)
    ).
btype2(external_function_call(FunName,FunParams,_FunDef,TypeParams,Decl), Pos, Env,
       external_function_call(FunName,TFunParams), OutputType , [], TRin, TRout) :-
       /* note: position info is automatically added to info field; hence [] ok */
       (FunParams = []  -> InputType = empty_type, FullType = set(OutputType)
         ; FullType = set(set(couple(InputType,OutputType)))),
      !, btype_external_call(FunName,TypeParams,FunParams,Decl,FullType,
                           InputType,Pos,Env,TFunParams,TRin,TRout).
btype2(external_pred_call(FunName,FunParams,_FunDef,TypeParams,Decl), Pos, Env,
       external_pred_call(FunName,TFunParams), pred, [], TRin, TRout) :-
    !, btype_external_call(FunName,TypeParams,FunParams,Decl,set(InputType),InputType,
                           Pos,Env,TFunParams,TRin,TRout).
btype2(external_subst_call(FunName,FunParams,_FunDef,TypeParams,Decl), Pos, Env,
       external_subst_call(FunName,TFunParams), subst, [], TRin, TRout) :-
    !, btype_external_call(FunName,TypeParams,FunParams,Decl,set(InputType),InputType,
                           Pos,Env,TFunParams,TRin,TRout).
btype2(integer(I), _, _, integer(I), integer, [], TR, TR) :- !.
btype2(integer_set(S), _, _, integer_set(S), set(integer), [], TR, TR) :- !.
btype2(string(I), _, _, string(I), string, [], TR, TR) :- !, assert_machine_string(I).
btype2(assign(IdFuns,Exprs), Pos, Env, assign(TIdFuns,TExprs), subst, [], TRin, TRout) :-
    !,
    common_prefix(IdFuns,Exprs,IdFuns1,Exprs1,OK),
    btype_l(IdFuns1,Env,Types,TIdFuns,TRin,TR1),
    btype_l(Exprs1, Env,Types,TExprs, TR1, TR2),
    foldl(check_assign_lhs(Env),TIdFuns,TR2,TR3),
    ( OK=same_length -> TRout=TR3
    ; otherwise ->
        store_typecheck_error('different number of variables on left and right side of a substitution ',Pos,TR3,TRout)
    ).
btype2(becomes_element_of(Ids,Expr), _, Env, becomes_element_of(TIds,TExpr), subst, [], TRin, TRout) :-
    !,btype_l(Ids,Env,Types,TIds,TRin,TR1),
    foldl(check_assign_lhs(Env),TIds,TR1,TR2),
    couplise_list(Types,Type),
    btype(becomes_element_of(Ids),Expr,Env,set(Type),TExpr,TR2, TRout).
btype2(becomes_such(Ids,Pred), _, Env, becomes_such(TIds,TPred), subst, [], TRin, TRout) :-
    !,btype_l(Ids,Env,_,TIds,TRin,TR1),
    foldl(check_assign_lhs(Env),TIds,TR1,TR2),
    % for each LHS identifier x (in Ids), we add an identifier of the form
    % x$0 with the same type to the environment
    foldl(add_primed_id,TIds,Env,Subenv),
    btype(becomes_such,Pred,Subenv,pred,TPred,TR2,TRout).
btype2(evb2_becomes_such(Ids,Pred), _, Env, evb2_becomes_such(TIds,TPred), subst, [], TRin, TRout) :-
    !,check_evb_becomes_such(evb2_becomes_such,Ids,Pred,Env,TIds,TPred,TRin,TRout).
btype2(operation_call(Op,Res,Args), Pos, Env, TOp, subst, Infos, TRin, TRout) :-
    %print(opcall(Op)),nl, portray_env(Env),nl,
    !,type_operation_call(Op,Res,Args,Pos,Env,TOp,Infos,TRin,TRout).
btype2(operation_call_in_expr(Op,Args), Pos, Env, operation_call_in_expr(TOp,TArgs), Type, Infos, TRin, TRout) :-
    !, ext2int(btype2_opcall,Op,identifier(Opid),IPos,_,_,_,_),
    btype2(identifier(op(Opid)),IPos,Env,Identifier,op(ArgTypes,ResTypes),OpInfos,TRin,TR1),
    Identifier = identifier(_), % Identifier = identifier(op(Opid))
    create_texpr(Identifier,subst,OpInfos,TOp),
    btype_l(Args,Env,ArgTypes,TArgs,TR1,TR2),
    couplise_list(ResTypes,Type),
    get_operation_infos(OpInfos,Infos),
    % print(op_call_in_expr(Opid,Infos)),
    ( memberchk(modifies(Modifies),Infos) ->
      (Modifies==[] -> TR2=TRout
        ; var(Modifies) -> TR2=TRout,
          % the info has not yet been computed; add a co-routine (which must however raise error immediately)
          when(nonvar(Modifies),(Modifies==[] -> true
           ; add_error(btype2,'query-operation expected when calling operation in expressions ',Opid,Pos)))
       ; otherwise -> print(error(Infos)),nl, store_typecheck_error('query-operation expected when calling operation in expressions ',Pos,TR2,TRout)
      )
      ; add_internal_error(btypechecker,'Missing modifies info field:',operation_call_in_expr(Op,Args),Pos),TR2=TRout
    ).
btype2(case(Expr,Eithers,Else), _, Env, case(TExpr,TEithers,TElse), subst, [], TRin, TRout) :-
    !,btype(case,Expr,Env,Type,TExpr,TRin,TR1),
    foldl(btype_caseor(Env,Type),Eithers,TEithers,LExprs,TR1,TR2),
    append(LExprs,Exprs),
    btype(case_else,Else,Env,subst,TElse,TR2,TR3),
    % The next check could be made dependend on a preference (non-literal values allowed in
    % CASE statements), in that case add TR3=TRout as an alternative
    check_case_expressions(Exprs,TR3,TRout).
btype2(struct(Fields), Pos, Env, struct(TRecord), set(RType), [], TRin, TRout) :-
    !, btype(struct,rec(Pos,Fields),Env,_,TRecord,TRin,TR1),
    get_texpr_expr(TRecord,rec(TFields)),
    foldl(extract_field_type,TFields,FieldTypes,TR1,TR2),
    normalise_record_type(record(FieldTypes),RType),
    TR2=TRout.
    % (record_has_multiple_field_names(FieldTypes,FieldName) %% error already raised by rec(.) clause below
btype2(rec(Fields), Pos, Env, rec(TFields), RType, [], TRin, TRout) :-
    !, foldl(btype_field_rec(Env),Fields,FTypes,TFields1,TRin,TR2),
    normalise_record_type(record(FTypes),RType),
    % we sort the field arguments here, so we do not have to normalise them every
    % time in the interpreter
    RType = record(NormedFieldTypes),
    sort_record_fields(NormedFieldTypes,TFields1,TFields),
    (record_has_multiple_field_names(NormedFieldTypes,FieldName)
      -> ajoin(['Field ',FieldName,' used multiple times in record '],Msg),
         store_typecheck_error(Msg,Pos,TR2,TRout)
      ; TR2=TRout
     ).
btype2(record_field(Record,I), Pos, Env, record_field(TRecord,Id), Type, [], TRin, TRout) :-
    !,
    btype(record_field,Record, Env, RType, TRecord, TRin, TR1),
    remove_pos(I,identifier(Id)),
    unify_types(record(Fields),RType,Pos,TRecord,TR1,TR2),
    check_field_type(Fields,Id,Type,Pos,TR2,TRout).
btype2(refined_operation(Id,Results,Args,_RefinesID,Body), Pos, Env, TExpr, Type, Infos, TRin, TRout) :-
    !, % TO DO: do we need to treat RefinesID ?
    btype2(operation(Id,Results,Args,Body), Pos, Env, TExpr, Type, Infos, TRin, TRout).
btype2(operation(Id,Results,Args,Body), Pos, Env, operation(TId,TResults,TArgs,TBody), Type,
         Infos, TRin, TRout) :-
    %print(checking_op(Id)),nl,
    !,  %debug_stats(checking_operation(Id,Args)),
    % type-check the identifier to look up the operation's type
    % in the environment.
    % we call ext2int and btype2/7 directly instead of going over
    % btype/5, because we use op(Name) instead of Name to look up
    % the type. This is done to seperate operation and variable name
    % spaces (Note: that was a stupid and ugly hack)
    ext2int(btype2_operation,Id,identifier(Opid),IPos,Type,Topid,IdInfos,TId),
    btype2(identifier(op(Opid)),IPos,Env,Topid,Type,IdInfos,TRin,TR0),
    % introduce a sub-environment Subenv by adding the argument and
    % result variables
    check_for_duplicate_raw_ids(Results,[],'results for operation ',Opid,FilteredResults,TR0,TR1),
    add_ext_variables_with_info(FilteredResults,Env,[introduced_by(operation)],TResults,Subenv1,TR1,TR2),
    check_for_duplicate_raw_ids(Args,[],'arguments for operation ',Opid,FilteredArgs,TR2,TR2b),
    add_ext_variables_with_info(FilteredArgs,Subenv1,[readonly,introduced_by(operation)],TArgs,Subenv,TR2b,TR3),
    % get their types and determine the resulting operation's type
    get_texpr_types(TResults,LResultTypes),
    get_texpr_types(TArgs,LArgTypes),
    unify_types(op(LArgTypes,LResultTypes),Type,Pos,TId,TR3,TR4),
    % type the operation's body in the sub-environment
    btype(operation(Id),Body, Subenv, subst, TBody, TR4, TRout),
    % the list of modified machine variables is the list of changed
    % variables by the substitutions in the body minus the result variables
    compute_accessed_vars_infos_for_operation(Id,TResults,TArgs,TBody,Modifies,Reads,NonDetModified,Infos),
    %assert(op_modifies_reads_cache(Opid,Modifies,Reads)), % cache information for operation_calls; possibly temporary solution until we understand the code for operation_call [Note: we can have multiple versions ! at different refinement levels !!]
    % make sure that the List of modified variables of this operation
    % is also up-to-date in the type environment
    (memberchk(modifies(EnvModifies),IdInfos),
     memberchk(reads(EnvReads),IdInfos),
     memberchk(non_det_modifies(EnvNonDetModifies),IdInfos) ->
        % Check if the environment variables and the local variables are unifiable.
        % If not, an operation with the same name must have been defined before
        % We can ignore this, later an error message will be raised.
        ( EnvModifies=Modifies -> true; add_message(btypechecker,'Conflict in modifies info for ',Id,IPos)),
        ( EnvReads=Reads -> true; add_message(btypechecker,'Conflict in reads info for ',Id,IPos)),
        ( EnvNonDetModifies=NonDetModified -> true; add_message(btypechecker,'Conflict in non_det_modifies info for ',Id,IPos))
      ; add_error(btypechecker,'Cannot store modifies/reads info for: ',Opid,IPos)
    ),
    debug_format(9,'Finished checking operation ~w, reads:~w, writes=~w~n',[Id,Reads,Modifies]).
btype2(while(Pred,Body,Inv,Var), _, Env, while(TPred,TBody,TInv,TVar), subst, FullInfo, TRin, TRout) :- !,
    btype(while,Pred,Env,pred,TPred,TRin,TR1),
    btype(while_body,Body,Env,subst,TBody,TR1,TR2),
    % In contrast to the other substitutions, in the invariant and variant
    % part of a while loop it is possible to reference abstract variables,
    % as they are not used to influence the behaviour of the substitution.
    % We create a new environment InvEnv where those access restrictions are
    % removed.
    allow_access_to_abstract_vars(Env,InvEnv),
    add_primed_old_value_variables(InvEnv,PInvEnv),
    btype(while_inv,Inv,PInvEnv,pred,TInv,TR2,TR3),
    btype(while_variant,Var,InvEnv,integer,TVar,TR3,TRout),
    % tag this statement with an appropiate info if a reference to the original value
    % of a variable is made in the invariant.
    % This info is used in the clean_up (see b_ast_cleanup) to insert a LET statement.
    check_for_old_state_references(TInv,Info),
    FullInfo = [modifies(Modifies),reads(AllReads)|Info],
    % TO DO: ideally we should compute those in one go for all substitutions:
    %  here we compute information again which we have already computed for operations
    get_accessed_vars(b(while(TPred,TBody,TInv,TVar),subst,Info),[],Modifies,_NonDetModifies,Reads),
    (debug:debug_mode(off) -> true
     ; format('WHILE Reads : ~w~n   Modifies : ~w~n',[Reads,Modifies]),
        print('VARIANT: '),translate:print_bexpr(TVar),nl
    ),
    (member(refers_to_old_state(SortedRefs),Info)
      -> findall(Primed,member(oref(Primed,_,_),SortedRefs),PrimedDollarVars), sort(PrimedDollarVars,SPDV),
         ord_union(SPDV,Reads,AllReads)  %, print(add(SPDV,Reads)),nl
      ;  AllReads = Reads).

btype2(parallel(Substs),Pos,Env,parallel(TSubsts), subst, [], TRin, TRout) :-
    !, btype_l(Substs,Env,_,TSubsts,TRin,TR1),
    check_no_double_assignment(TSubsts,Pos,TR1,TRout).
btype2(minus_or_set_subtract(A,B), Pos, Env, minus_or_set_subtract(TA,TB), Type, [], TRin, TRout) :-
    !,btype_l([A,B],Env,[TypeA,TypeB],[TA,TB],TRin,TR1),
    delayed_type_minus(TypeA,TypeB,Type,Pos,A,B,TR1,TRout).
btype2(concat(A,B), Pos, Env, concat(TA,TB), Type, [], TRin, TRout) :-
    get_preference(allow_sequence_operators_on_strings,true),
    btype_l([A,B],Env,[TypeA,TypeB],[TA,TB],TRin,TR1),
    TypeA \== integer, % otherwise trigger error_rewrite below for exponentiation
    !,
    delayed_type_string_or_seq(TypeA,TypeB,Type,Pos,A,B,TR1,TRout).
btype2(rev(A), Pos, Env, rev(TA), Type, [], TRin, TRout) :-
    get_preference(allow_sequence_operators_on_strings,true),
    Type = TypeA,
    btype_l([A],Env,[TypeA],[TA],TRin,TR1),
    !,
    delayed_type_string_or_seq(TypeA,TypeA,TypeA,Pos,A,A,TR1,TRout).
btype2(size(A), Pos, Env, size(TA), Type, [], TRin, TRout) :-
    get_preference(allow_sequence_operators_on_strings,true),
    Type = integer,
    btype_l([A],Env,[TypeA],[TA],TRin,TR1),
    !,
    delayed_type_string_or_seq(TypeA,TypeA,TypeA,Pos,A,A,TR1,TRout).
btype2(general_concat(A), Pos, Env, general_concat(TA), Type, [], TRin, TRout) :-
    get_preference(allow_sequence_operators_on_strings,true),
    btype_l([A],Env,[TypeA],[TA],TRin,TR1),
    !,
    delayed_type_seq_string_or_seq_seq(TypeA,Type,Pos,A,TR1,TRout).
btype2(mult_or_cart(A,B), Pos, Env, mult_or_cart(TA,TB), Type, [], TRin, TRout) :-
    !,btype_l([A,B],Env,[TypeA,TypeB],[TA,TB],TRin,TR1),
    delayed_type_times(TypeA,TypeB,Type,Pos,A,B,TR1,TRout).
btype2(substitution(Subst,Pred), _Pos, Env, TSubst, pred, [], TRin, TRout) :-
    !, type_substitution_expression(Subst,Pred,Env,TSubst,TRin,TRout).
btype2(extended_expr(Id,Args,Args2), Pos, Env, Result, Type, [was_operator(Id)|Info], TRin, TRout) :-
    !, lookup_eventb_operator(Id,Args,Args2,Pos,Env,expr,Result,Type,Info,TRin,TRout).
btype2(extended_pred(Id,Args,Args2), Pos, Env, Result, Type, [was_operator(Id)|Info], TRin, TRout) :-
    !, lookup_eventb_operator(Id,Args,Args2,Pos,Env,pred,Result,Type,Info,TRin,TRout).
btype2(recursive_let(Id,Set),_Pos,Env,recursive(TId,TSet),set(Type),[],TRin,TRout) :- !,
    add_ext_variables([Id], Env, [TId], SubEnv, TRin, TR1),
    get_texpr_type(TId,set(Type)),
    btype(recursive,Set,SubEnv,set(Type),TSet,TR1,TRout).
btype2(set_extension(RawIntegers),_Pos,_Env,value(V),set(Type),[],TR,TR) :-
    % optimisation: is this a set of integers or pair of integers?
    % -> skip type checking the content and just generate an AVL set directly
    simple_set_extension_type(RawIntegers,Type),
    maplist(read_raw_integers(Type),RawIntegers,Integers),!,
    custom_explicit_sets:convert_to_avl(Integers,V).
btype2(typeset,_Pos,_Env,typeset,set(_),[],TR,TR) :- !.
btype2(typeof(InnerExpr,TypeExpr),_Pos,Env,Expr,Type,Info,TRin,TRout) :-
    !, btype(typeof,TypeExpr,Env,set(Type),_TTypeExpr,TRin,TR1),
    btype(typeof,InnerExpr,Env,Type,TExpr,TR1,TRout),
    get_texpr_expr(TExpr,Expr),get_texpr_info(TExpr,Info).
btype2(Expr, Pos, Env, TExpr, Type, [], TRin, TRout) :-
    % this is the most simple case of a type-check
    safe_functor(btype2_expr,Expr,Functor,Arity),
    safe_functor(btype2_lookup,Lookup,Functor,Arity),
    Lookup :: Type,
    !,
    safe_functor(btype2_texpr,TExpr,Functor,Arity),
    type_and_unify_args(1,Arity,Pos,Env,Expr,Lookup,TExpr,TRin,TRMid),
    (\+ type_error_or_warning_occurred(TRin,TRMid)
     -> TRMid = TRout
      ; typecheck_result_get_flags(TRMid,Flags),
        nonmember(try_rewrite,Flags),
        typecheck_add_flag(try_rewrite,TRMid,TRMid0), % avoid an exponential blow-up in rewrites; we only do one rewrite at a time
        %print(try_rewrite(Expr)),nl,
        error_rewrite(Expr,RewrittenExpr,RewriteMsg), % a common user error may be present
        % print(rewrite(RewrittenExpr)),nl,
        % Copy the code from above; we assume
        copy_term(TRMid0,TRMid1), % copy error variable; avoid instantiating TRMid with error msgs from RewrittenExpr
        btype2(RewrittenExpr,Pos,Env,_,_,[],TRMid1,TRMid2),
        %safe_functor(btype2_expr,RewrittenExpr,Functor2,Arity2),
        %safe_functor(btype2_lookup,Lookup2,Functor2,Arity2),
        %Lookup :: Type, print(rewritten_t(Lookup,Type)),nl,
       % safe_functor(btype2_texpr,TExpr2,Functor2,Arity2),
        %type_and_unify_args(1,Arity2,Pos,Env,RewrittenExpr,Lookup2,TExpr2,TRMid1,TRMid1),
        %  print(rewrite_ok(TRMid1)),nl,
        \+ type_error_or_warning_occurred(TRMid1,TRMid2) % the rewrite removed the type error
        ->  ajoin(['Did you ',RewriteMsg,' ?'],Msg),
            store_typecheck_error(Msg,Pos,TRMid,TRout)
    ;  %print(no_rule_applicable),nl,
       TRMid=TRout).

% merge the pragma ast nodes into the info fields
btype2(label(Label,Node), _Pos, Env, TExpr, Type, InfosOut, TRin, TRout) :- !,
    btype(label, Node, Env, Type, Result, TRin, TRout),
    get_texpr_expr(Result,TExpr),
    get_texpr_info(Result,Infos),
    add_label_to_infos(Infos,Label,InfosOut).
btype2(description(Desc,Node), _Pos, Env, TExpr, Type, [description(Desc)|Infos], TRin, TRout) :- !,
    btype(description, Node, Env, Type, Result, TRin, TRout),
    get_texpr_expr(Result,TExpr),
    get_texpr_info(Result,Infos).
btype2(symbolic_comprehension_set(A,B), Pos, Env, TExpr, Type, [prob_annotation('SYMBOLIC')|Infos], TRin, TRout) :- !,
    btype2(comprehension_set(A,B), Pos, Env, TExpr, Type, Infos, TRin, TRout).
btype2(symbolic_quantified_union(A,B,C), Pos, Env, TExpr, Type, [prob_annotation('SYMBOLIC')|Infos], TRin, TRout) :- !,
    btype2(quantified_union(A,B,C), Pos, Env, TExpr, Type, Infos, TRin, TRout).
btype2(symbolic_lambda(A,B,C), Pos, Env, TExpr, Type, [prob_annotation('SYMBOLIC')|Infos], TRin, TRout) :- !,
    btype2(lambda(A,B,C), Pos, Env, TExpr, Type, Infos, TRin, TRout).
btype2(symbolic_composition(A,B), Pos, Env, TExpr, Type, [prob_annotation('SYMBOLIC')|Infos], TRin, TRout) :- !,
    btype2(composition(A,B), Pos, Env, TExpr, Type, Infos, TRin, TRout).
btype2(conversion(Node), _Pos, Env, TExpr, Type, [conversion|Infos], TRin, TRout) :- !,
    btype(conversion, Node, Env, Type, Result, TRin, TRout),
    get_texpr_expr(Result,TExpr),
    get_texpr_info(Result,Infos).
btype2(let_predicate(Ids,Equalities,Pred), Pos, Env, let_predicate(TIdsOut,TExprsOut,TPred), pred, [], TRin, TRout) :- !,
    check_for_duplicate_raw_ids(Ids,[],'LET arguments','',FIds,TRin,TR0),
    add_ext_variables_with_info(FIds,Env,[introduced_by(let_predicate)],TIds,Subenv,TR0,TR1),
    btype(let_predicate, Equalities, Subenv, pred, TEqualities, TR1, TR2), % should we introduced variables only one by one ?
    btype(let_predicate, Pred, Subenv, pred, TPred, TR2, TR3),
    btype_static_check_let(TIds,TEqualities,Pos,allow_reuse,Env,TR3,TR4),
    split_let_equalities_into_ids_and_expressions(Ids,TEqualities,TIdsOut,TExprsOut,Pos,TR4,TRout).
btype2(let_expression(Ids,Equalities,Expr), Pos, Env, let_expression(TIdsOut,TExprsOut,TExpr), Type, [], TRin, TRout) :- !,
    check_for_duplicate_raw_ids(Ids,[],'LET arguments','',FIds,TRin,TR0),
    add_ext_variables_with_info(FIds,Env,[introduced_by(let_expression)],TIds,Subenv,TR0,TR1),
    btype(let_expression, Equalities, Subenv, pred, TEqualities, TR1, TR2),
    btype(let_expression, Expr, Subenv, Type, TExpr, TR2, TR3),
    btype_static_check_let(TIds,TEqualities,Pos,allow_reuse,Env,TR3,TR4),
    split_let_equalities_into_ids_and_expressions(Ids,TEqualities,TIdsOut,TExprsOut,Pos,TR4,TRout).
% error case. should not be reached
btype2(Expr, _Pos, _, _, _, _, TR, TR) :-
    safe_functor(btype2_uncovered,Expr,F,Arity),
    add_internal_error('Uncovered construct in type checker: ',F/Arity),
    fail.


compute_accessed_vars_infos_for_operation(_Id,TResults,TArgs,TBody,Modifies,Reads,NonDetModified,Infos) :-
    append(TResults,TArgs,TLocals),
    get_texpr_ids(TLocals,Locals), list_to_ord_set(Locals,SLocals),
    get_accessed_vars(TBody,[],AllModifies,AllNonDetModified,AllReads), % we now no longer pass Locals, we filter out later
    %print(modified_reads(Opid,AllModifies,AllReads)),nl,
    %print(nondet(Id,AllNonDetModified,SLocals,IdInfos)),nl,
    ord_subtract(AllModifies,SLocals,Modifies),
    ord_subtract(AllReads,SLocals,Reads),
    ord_subtract(AllNonDetModified,SLocals,NonDetModified),
    ord_intersection(AllModifies,SLocals,LocalModifies),
    ord_intersection(AllReads,SLocals,LocalReads),
    Infos = [modifies(Modifies),reads(Reads),non_det_modifies(NonDetModified),
          modifies_locals(LocalModifies),reads_locals(LocalReads)].

split_let_equalities_into_ids_and_expressions(RawIds,EqualitiesPred,IDs,Expressions,_,TR,TR) :-
    conjunction_to_list(EqualitiesPred,ListOfEqualities),
    split_let_equality_into_id_and_expression2(ListOfEqualities,IDs,Expressions,RawIds),
    !.
split_let_equalities_into_ids_and_expressions(_,EQs,[],[],Pos,TRIn,TROut) :-
    translate_bexpression(EQs,PPEqs),
    ajoin_with_sep(['Let equalities could not be split into ids and expressions:',PPEqs],' ',Msg),
    store_typecheck_error(Msg,Pos,TRIn,TROut).

split_let_equality_into_id_and_expression2([],[],[],_).
split_let_equality_into_id_and_expression2([Equality|TE],[TID|TI],[Expression|TEs],RawIds) :-
    get_texpr_expr(Equality,equal(TID,Expression)),
    get_texpr_id(TID,ID), select(identifier(_,ID),RawIds,RawIds2),
    !,
    split_let_equality_into_id_and_expression2(TE,TI,TEs,RawIds2).
split_let_equality_into_id_and_expression2([_|TE],TI,TEs,RawIds) :-
    split_let_equality_into_id_and_expression2(TE,TI,TEs,RawIds).

add_label_to_infos([],Label,[label([Label])]).
add_label_to_infos([label(LA)|T],Label,[label([Label|LA])|T]) :- !.
add_label_to_infos([F|T],Label,[F|TN]) :-
    add_label_to_infos(T,Label,TN).

% check if a set extension is of type integer or couple(integer,...) by inspecting
% the first element
simple_set_extension_type([First,_|_],Type) :- % At least two elements
    simple_set_extension_type2(First,Type).    % of type integer or couple of integers
simple_set_extension_type2(integer(_,_),integer).
simple_set_extension_type2(couple(Pos,List),Type) :-
    couplise_list_pos(List,Pos,Couple),
    simple_set_extension_type2(Couple,Type).
simple_set_extension_type2(couple(_Pos,A,B),couple(TA,TB)) :-
    simple_set_extension_type2(A,TA),
    simple_set_extension_type2(B,TB).
% convert a raw integer or pairs of integers into a value
read_raw_integers(integer,integer(_Pos,I),int(I)).
read_raw_integers(couple(TA,TB),couple(_Pos,List),(VA,VB)) :-
    read_raw_integer_couple(TA,VA,List,Rest),
    read_raw_integer_couple(TB,VB,Rest,[]).
read_raw_integer_couple(integer,int(I)) --> [integer(_Pos,I)].
read_raw_integer_couple(couple(TA,TB),(VA,VB)) -->
    read_raw_integer_couple(TA,VA),
    read_raw_integer_couple(TB,VB).

% find common prefix of two lists (assigned variables, assigned expressions)
common_prefix([],[],[],[],Res) :- !, Res=same_length.
common_prefix([H1|T1],[H2|T2],[H1|TT1],[H2|TT2],OK) :- !,
    common_prefix(T1,T2,TT1,TT2,OK).
common_prefix(_,_,[],[],not_same_length).

check_no_double_assignment(TSubsts,Pos,TRin,TRout) :-
    maplist(modified_var,TSubsts,Mods),
    ( (preference(allow_simultaneous_assignments,false),
       find_common_variables(Mods,Var)) ->
        ajoin(['Simultaneous assignment to variable ',Var],Msg),
        store_typecheck_error(Msg,Pos,TRin,TRout)
    ; otherwise ->
        TRin = TRout).
modified_var(TSubst,Modified) :-
    get_accessed_vars(TSubst,[],Modified,_NDModified,_Read).
find_common_variables([Vars|Rest],Var) :-
    find_common_variables2(Rest,Vars,Var).
find_common_variables2([Vars|Vrest],Prior,Var) :-
    ord_intersection(Vars,Prior,FoundVars),
    ( memberchk(Var,FoundVars) -> true
    ; otherwise ->
        ord_union(Vars,Prior,NewVars),
        find_common_variables2(Vrest,NewVars,Var)).

type_substitution_expression(Subst,Pred,Env,
                             let_predicate(TypedLhs,TypedRhs,TPred),TRin,TRout) :-
    ( remove_pos(Subst,assign(Lhs,Rhs)) ->
        ( same_length(Lhs,Rhs) ->
            /* Typecheck the right side of the substitution */
            btype_l(Rhs,Env,Types,TypedRhs,TRin,TR1),
            /* The left site of the substitution introduces new variables which have the
               same type as the corrresponding element of the RHS */
            add_ext_variables_with_info(Lhs,Env,[introduced_by(let)],TypedLhs,SubEnv,TR1,TR2),
            get_texpr_types(TypedLhs,Types),
            /* Type the predicates in the sub-environment */
            btype(substitution_expression,Pred,SubEnv,pred,TPred,TR2,TRout)
        ; otherwise /* LHS hat not the same number of elements as RHS */ ->
            ext2int(Subst,_,Pos,_,_,_,_),
            TypedLhs=[], TypedRhs=[], create_texpr(truth,pred,[],TPred),
            store_typecheck_error('Expected same number of elements on both sides of an assignment',
                        Pos,TRin,TRout))
    ; otherwise /* Substitution is no assignment (the parser should not allow this) */ ->
        add_internal_error('Expected assignment in substitution expression, but was: ',Subst),
        fail).

delayed_type_minus(TypeA,TypeB,ResType,Pos,ExprA,ExprB,TRin,TRout) :-
    arg(1,ExprA,PosA), arg(1,ExprB,PosB),
    % In case of A-B=C we know that A, B and C have the same type
    unify_types_l([ResType,ResType],[TypeA,TypeB],[PosA,PosB],TRin,TR1),
    % but not if it is integer or set(_)
    typecheck_result_add_delayed_rule(Trigger,Errors,TR1,TRout),
    delayed_type_minus2(ResType,Trigger,Pos,Errors).
:- block delayed_type_minus2(-,-,?,?).
delayed_type_minus2(Type,_Trigger,Pos,Errors) :-
    ( var(Type) ->
        Errors = [error('Ambiguous expression (arithmetic minus or set minus?)',Pos)]
    ; Type = integer ->
        Errors = []
    ; otherwise ->
        unify_types_werrors_l([set(_)],[Type],[Pos],Errors,[])
    ).

% for concat: allow a types to be either string or seq(A)
delayed_type_string_or_seq(TypeA,TypeB,ResType,Pos,ExprA,ExprB,TRin,TRout) :-
    arg(1,ExprA,PosA), arg(1,ExprB,PosB),
    % In case of A^B=C we know that A, B and C have the same type
    unify_types_l([ResType,ResType],[TypeA,TypeB],[PosA,PosB],TRin,TR1),
    % but not if it is integer or set(_)
    typecheck_result_add_delayed_rule(Trigger,Errors,TR1,TRout),
    delayed_type_string_or_seq2(ResType,Trigger,Pos,Errors).
:- block delayed_type_string_or_seq2(-,-,?,?).
delayed_type_string_or_seq2(Type,_Trigger,Pos,Errors) :-
    ( %var(Type) ->  Errors = [error('Ambiguous expression (seq(_) or string operator?)',Pos)] ;
      Type == string ->
        Errors = []
    ; otherwise ->
        unify_types_werrors_l([seq(_)],[Type],[Pos],Errors,[])
    ).

% for general_concat(seq(Type)) -> Type, where Type = string or seq(T)
delayed_type_seq_string_or_seq_seq(TypeA,ResType,Pos,ExprA,TRin,TRout) :-
    arg(1,ExprA,PosA),
    unify_types_l([seq(ResType)],[TypeA],[PosA],TRin,TR1),
    % but not if it is integer or set(_)
    typecheck_result_add_delayed_rule(Trigger,Errors,TR1,TRout),
    delayed_type_string_or_seq2(ResType,Trigger,Pos,Errors).

% delayed_type_times(TypeA,TypeB,ResType,Pos,ExprA,ExprB,TRin,TRout):
%   "A*B=R"
%   TypeA:   type of A, TypeB: type of B, ResType: type of R
%   Pos:     position info for error messages
%   ExprA/B: expressions to generate pretty error messages
delayed_type_times(TypeA,TypeB,ResType,Pos,ExprA,ExprB,TRin,TRout) :-
    arg(1,ExprA,PosA), arg(1,ExprB,PosB),
    ( maplist(var,[TypeA,TypeB,ResType]) ->
        typecheck_result_add_delayed_rule(Trigger,Errors,TRin,TRout),
        delayed_type_times2(TypeA,TypeB,ResType,Trigger,Pos,PosA,PosB,Errors)
    ; contains_integer_type([TypeA,TypeB,ResType]) ->
        % One type is an integer
        unify_types_l([integer,integer,integer],[ResType,TypeA,TypeB],[Pos,PosA,PosB],TRin,TRout)
    ; otherwise ->
        % A type is known, no need for a delay
        unify_types_l([set(couple(X,Y)),set(X),set(Y)],[ResType,TypeA,TypeB],[Pos,PosA,PosB],TRin,TRout)
    ).
:- block delayed_type_times2(-,-,-,-,?,?,?,?).
delayed_type_times2(A,B,Res,_Trigger,Pos,PosA,PosB,Errors) :-
    ( maplist(var,[A,B,Res]) ->
        Errors = [error('Ambiguous expression (multiplication or cartesian product?)',Pos)]
    ; contains_integer_type([A,B,Res]) ->
        unify_types_werrors_l([integer,integer,integer],[Res,A,B],[Pos,PosA,PosB],Errors,[])
    ; otherwise ->
        unify_types_werrors_l([set(couple(X,Y)),set(X),set(Y)],[Res,A,B],[Pos,PosA,PosB],Errors,[])
    ).

contains_integer_type([T|_Rest]) :- T==integer,!.
contains_integer_type([_T|Rest]) :- contains_integer_type(Rest).

type_operation_call(Op,Res,Args,Pos,Env,operation_call(TOp,TRes,TArgs),Infos,TRin,TRout) :-
    ext2int(btype2_opcall,Op,identifier(Opid),IPos,_,_,_,_), % succeeds if operation name is a proper identifier
    !,
    btype2(identifier(op(Opid)),IPos,Env,Identifier,op(ArgTypes1,ResTypes1),OpInfos,TRin,TR1),
    Identifier = identifier(OpIdW), % OpIdW = op(_)
    create_texpr(Identifier,subst,OpInfos,TOp),
    btype_l(Args,Env,_ArgTypes,TArgs,TR1,TR2),
    btype_l(Res,Env,_ResTypes,TRes,TR2,TR3),
    foldl(check_assign_lhs(Env),TRes,TR3,TR4),
    ( same_length(Args,ArgTypes1) ->
        ( same_length(Res,ResTypes1) ->
            % number of arguments resp. results ok
            type_operation_call2(OpIdW,Env,Pos,TRes,TArgs,ArgTypes1,ResTypes1,Infos,TR4,TRout)
        ; otherwise ->
            % wrong number of result identifiers given
            add_operation_call_error(ResTypes1,Res,Pos,Opid,'result identifiers(s)',Infos,TR4,TRout))
    ; otherwise ->
        % wrong number of arguments given
        add_operation_call_error(ArgTypes1,Args,Pos,Opid,'arguments(s)',Infos,TR4,TRout)).
type_operation_call(Op,_Res,_Args,Pos,_Env,skip,[],TRin,TRout) :-
    % operation name is not an id but e.g. a function call
    functor(Op,Func,_),
    ajoin(['Expected an identifier for operation call but got "',Func,'" operator'],Msg),
    store_typecheck_error(Msg,Pos,TRin,TRout).
type_operation_call2(OpId,Env,Pos,TRes,TArgs,ArgTypes,ResTypes,Infos,TRin,TRout) :-
    check_arguments(TArgs,ArgTypes,TRin,TR1), % check the arguments
    check_arguments(TRes,ResTypes,TR1,TR2),   % check the result

    % if the called operation is not found in the environment,
    % an error was stored above, and we skip the following section
    ( env_lookup_infos(OpId,Env,EnvInfos) ->
        % we currently return the modifies/1 information of the called operation
        % the only information about this node. This information is needed by
        % get_modified_vars/3.
        get_operation_infos(EnvInfos,Infos),
        % check if we are in a section where we can only call operations
        % that make no changes to the state ("inquiry")
        ( (memberchk(dontcall,EnvInfos), get_preference(allow_local_operation_calls,false)) ->
            % it is not allowed do call the operation: this is the case if a local
            % operation is used (only operations of included/used/seen machines are allowed to call)
            % TO DO: we are allowed to call LOCAL_OPERATIONS from other OPERATIONS
            store_typecheck_error('a local operation must not be called',Pos,TR2,TRout)
        ; memberchk(inquiry,EnvInfos) ->
            % check whether we call a query-operation, if the target it marked "inquiry only"
            ( memberchk(modifies([]),Infos) -> TR2=TRout
            ; otherwise -> store_typecheck_error('query-operation expected ',Pos,TR2,TRout))
        ; otherwise ->
            % ok, we may call even state-changing operations
            TR2=TRout
        )
    ; otherwise ->
        % operation not found, skip additional checks above and
        % state (maybe incorrectly, but errors are reported anyway),
        % that the operation call does not modify any variables
        Infos = [modifies([]),reads([]),non_det_modifies([])],
        TR2=TRout).
check_arguments(TArgs,ArgTypes1,TRin,TRout) :-
    get_texpr_types(TArgs,ArgTypes), maplist(get_texpr_pos,TArgs,ArgPos),
    unify_types_l(ArgTypes1,ArgTypes,ArgPos,TRin,TRout).

% Generator an error for the wrong numbers of arguments resp. result identifiers
add_operation_call_error(ExpTypes,GivenExpr,Pos,Opid,Title,Infos,TRin,TRout) :-
    Infos = [modifies([]),reads([])],     % Info field must be set, even in case of an error
    length(ExpTypes,Expected),length(GivenExpr,Given),
    ajoin(['Expected ',Expected,' ',Title,' for operation ',Opid,
           ', but ',Given,' ',Title,' given'],Msg),
    store_typecheck_error(Msg,Pos,TRin,TRout).


% common typing patterns:
% type a list of expressions, return the list of types and typed expressions
btype_l([],_Env,[],[],TR,TR).
btype_l([Expr|Rest],Env,[Type|TRest],[TExpr|TERest],TRin,TRout) :-
    btype(btype_l,Expr,Env,Type,TExpr,TRin,TR1),
    btype_l(Rest,Env,TRest,TERest,TR1,TRout).

% type a list of expressions whose type is the same
btype_same([],_Env,_Type,[],TR,TR).
btype_same([Expr|ERest],Env,Type,[TExpr|TERest],TRin,TRout) :-
    btype(btype_same,Expr, Env, ElemType, TExpr, TRin, TR1),
    get_texpr_pos(TExpr,Pos),
    unify_types(Type,ElemType,Pos,TExpr,TR1,TR2),
    btype_same(ERest,Env,Type,TERest,TR2,TRout).

% for a list of identifiers, create a subenvironment with
% their primed versions. I.e., for each identifier "x", introduce
% a variable "x$0" into the environment.
% The type and additional information is copied from the original
% identifier, aditionally information about what the original
% identifier was ist stored.
add_primed_id(TId,InEnv,SubEnv) :-
    def_get_texpr_id(TId,Id),
    get_texpr_type(TId,Type),
    get_texpr_info(TId,Infos),
    number_suffix(Id,0,FullId),
    % the additional information is used in the interpreter to store the
    % values before the substitution under the alternative names
    %env_store(FullId,Type,[before_substitution(Id,FullId)|Infos],InEnv,SubEnv).
    env_store(FullId,Type,Infos,InEnv,SubEnv).



% special cases

check_evb_becomes_such(Label,Ids,Pred,Env,TIds,TPred,TRin,TRout) :-
    btype_l(Ids,Env,_,TIds,TRin,TR1),
    foldl(check_assign_lhs(Env),TIds,TR1,TR2),
    prime_identifiers(TIds,TPrimed),
    add_identifiers_to_environment(TPrimed,Env,Subenv),
    btype(Label,Pred,Subenv,pred,TPred,TR2,TRout).

% CASE ... EITHER .. OR ... END
btype_caseor(Env,Type,CaseOr,TCaseOr,TExprs,TRin,TRout) :-
    ext2int(btype_caseor,CaseOr,case_or(Exprs,Subst),_,pred,case_or(TExprs,TSubst),[],TCaseOr),
    % a case expresiion can have multiple expressions that may match, all of the same type "Type"
    btype_same(Exprs,Env,Type,TExprs,TRin,TR1),
    btype(btype_caseor,Subst,Env,subst,TSubst,TR1,TRout).
    % TO DO: check that Field1 of case_or(Field1,_) is a literal value
check_case_expressions(Exprs,TRin,TRout) :-
    maplist(is_literal,Exprs),!,  % only literals allowed
    ( contains_duplicate_literal(Exprs,Duplicate) ->
        translate_bexpression(Duplicate,DStr),
        get_texpr_pos(Duplicate,Pos),
        ajoin(['Duplicate case in CASE statement: ',DStr],Msg),
        store_typecheck_error(Msg,Pos,TRin,TRout)
    ; otherwise ->              % ok, all literals, no duplicates
        TRin=TRout
    ).
check_case_expressions(Exprs,TRin,TRout) :-
    % there are elements that are not literals
    exclude(is_literal,Exprs,NonLiterals),
    foldl(non_literal_error,NonLiterals,TRin,TRout).
is_literal(TExpr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,Info),
    is_literal2(Expr,Info).
is_literal2(identifier(_),Info) :-
    memberchk(enumerated_set_element,Info).
is_literal2(integer(_),_Info).
is_literal2(boolean_true,_Info).
is_literal2(boolean_false,_Info).
is_literal2(string(_),_Info).
non_literal_error(TExpr,TRin,TRout) :-
    translate_bexpression(TExpr,Str),
    get_texpr_pos(TExpr,Pos),
    ajoin(['CASE argument \'',Str,'\' is not a literal value.'],Msg),
    store_typecheck_error(Msg,Pos,TRin,TRout).
contains_duplicate_literal(TExprs,Duplicate) :-
    % This can be done, because we know that TExprs are literals, so there are no
    % sub-expressions with positions information or similar problems
    append(_Prefix,[Dup|Rest],TExprs),
    get_texpr_expr(Dup,Expr),
    get_texpr_expr(Duplicate,Expr),
    memberchk(Duplicate,Rest),!.

btype_field_rec(Env,Ext,field(Id,Type),field(Id,TExpr),TRin,TRout) :-
    remove_pos(Ext,rec_entry(Identifier,Expr)),
    (remove_pos(Identifier,identifier(Id)) -> true
      ; add_error_and_fail(btype_field_rec,'Invalid record field Identifier',Identifier)),
    btype(field,Expr,Env,Type,TExpr,TRin,TRout).

extract_field_type(field(Id,TExpr),field(Id,Type),TRin,TRout) :-
    get_texpr_type(TExpr,SetType),
    get_texpr_pos(TExpr,Pos),
    unify_types(set(Type),SetType,Pos,TExpr,TRin,TRout).

sort_record_fields(Var,F,R) :- var(Var),!,
    add_internal_error('Illegal variable argument:',sort_record_fields(Var,F,R)).
sort_record_fields([],_,[]).
sort_record_fields([field(Id,_Type)|Trest],Fields,[Field|Frest]) :-
    Field = field(Id,_Value),
    member(Field,Fields),!,
    sort_record_fields(Trest,Fields,Frest).

lookup_eventb_operator(Id,Args1,Args2,Pos,Env,ExPr,Result,Type,Info,TRin,TRout) :-
    ( env_lookup_infos(operator(Id),Env,[Module:Callback]) ->
        append(Args1,Args2,Args), % do not distinguish between expressions and predicates
        btype_l(Args,Env,_Types,TArgs,TRin,TR1),
        call(Module:Callback,Id,TArgs,Pos,Env,ExPr,TExpr,TR1,TRout),
        create_texpr(Result,Type,Info,TExpr)
    ; otherwise ->
        ajoin(['Unknown operator identifier ',Id],Msg),
        store_typecheck_error(Msg,Pos,TRin,TRout),
        failure_syntax_element(ExPr,Id,Result,Type)).
% TODO[DP,2013-01-31]: copied from bmachine_eventb -> refactor!
failure_syntax_element(expr,Id,identifier(Id),_).
failure_syntax_element(pred,_Id,falsity,pred).

get_operation_infos(Info,OpInfos) :-
    operation_infos(OpInfos),
    get_operation_infos2(OpInfos,Info).
get_operation_infos2([],_).
get_operation_infos2([Info|Rest],AllInfos) :-
    memberchk(Info,AllInfos),
    get_operation_infos2(Rest,AllInfos).

% the following information is required for each operation identifier
% it is "filled" by the type-checker (the operation/4 case of btype2)
operation_infos(Infos) :- findall(I,operation_info(I),Infos).
operation_info(modifies(_)).
operation_info(reads(_)).
operation_info(non_det_modifies(_)).


% unify type arguments of two terms
%   N is the number of the argument which should be processed
%   Arity is the arity of the term, so 1 <= N <= Arity
%   Env is the type environment
%   TypeTerm is a term that corresponds to the syntax element
%     e.g. "1<5" ~ less(1,5) has a type term less(integer,integer)
%     in the arguments of the TypeTerm, special constructs may be used:
%     "[Type]" declares that there is a list of expressions with the same Type
%     "ids" declares that there is a list of identifiers wich build a
%       sub-environment which will be used in the following arguments
%       (strict left-to-right)
%     "ids(T)" is like "ids" but the types of the identifiers are merged
%
%   TExpr is the typed expression of the complete expression
%   GI as usual: Introduced identifier and errors
type_and_unify_args(N,Arity,Pos,Env,Expr,TypeTerm,TExpr,TRin,TRout) :-
  %print(chck(N,Arity,Expr,TypeTerm,TExpr)),nl,
    ( N =< Arity ->
        arg(N,Expr,Arg), % Expr is the raw expression of the Nth argument
        arg(N,TypeTerm,ArgType), % ArgType is the corresponding type
        arg(N,TExpr,ArgTExpr), % TExpr is the corresponding typed expression
        type_and_unify_args_aux(ArgType,Arg,Pos,Expr,Env,ArgTExpr,NextEnv,TRin,TR2),
        N2 is N+1,
        type_and_unify_args(N2,Arity,Pos,NextEnv,Expr,TypeTerm,TExpr,TR2,TRout)
    ; otherwise ->
        TRin=TRout
    ).
type_and_unify_args_aux(ArgType,Arg,Pos,Expr,Env,ArgTExpr,NextEnv,TRin,TRout) :-
    % new identifiers are introduced, they are
    % available in the environment of the following arguments
    nonvar(ArgType),id_introduced(ArgType,CoupledType,Duplicates,Infos),!,
    functor(Expr,ExprFunctor,_), % Stored inside the information of newly created variables
    flatten_illegal_couples(Arg,FixedArg),
    ( Duplicates = unique  -> check_for_duplicates(FixedArg,Env,all,TRin,TR1)
    ; Duplicates = allowed -> check_for_duplicates(FixedArg,Env,[definition],TRin,TR1)), % just check that there are no clashes with definitions
    add_introduced_by_info(ExprFunctor,Infos,NewInfos),
    add_ext_variables_with_info(FixedArg,Env,NewInfos,ArgTExpr,NextEnv,TR1,TR2),
    % convert the list of types to a "couplised" type,
    %   (e.g. [a] -> a, or [a,b] -> couple(a,b) )
    % in many cases (if ArgType==ids), the "couplised" type is just ignored
    idlist_to_type(ArgTExpr,CoupledType1),
    unify_types(CoupledType,CoupledType1,Pos,ArgTExpr,TR2,TRout).
type_and_unify_args_aux(ArgType,Arg,_Pos,_Expr,Env,ArgTExpr,NextEnv,TRin,TRout) :-
    % if the argument is of the form [...] we expect a list of
    % expressions with the same type
    nonvar(ArgType),ArgType = [CommonType],!,
    btype_same(Arg,Env,CommonType,ArgTExpr,TRin,TRout),
    NextEnv = Env.
type_and_unify_args_aux(ArgType,Arg,Pos,Expr,Env,ArgTExpr,NextEnv,TRin,TRout) :-
    % standard case: just type-check the argument
    btype(type_and_unify_args(Pos,Expr,Arg,ArgType),Arg,Env,ArgType,ArgTExpr,TRin,TRout),
    NextEnv = Env.

% Note: Nothing seems to use introduced_by(_); except for exists in b_intepreter
add_introduced_by_info(exists,Infos,Res) :- !, Res=[introduced_by(exists)|Infos].
add_introduced_by_info(_,Infos,Infos).

% id_introduced(+Pattern,-CoupledType,-DuplicateIdentifiers,-Infos):
%  Pattern is a pattern in the type checking database (::/2).
%  id_introduced/4 extracts the CoupledType (may be a free variable),
%    (see the declaration of "comprehension_set" for an example use)
%  and returns a flag if an introduction of an already existing variable
%  is allowed. (E.g. LET does not allow to declare an existing identifier).
%  Additional infos that are added to the type environment are supported.
id_introduced(ids,_,allowed,[]).
id_introduced(ids(CoupledType),CoupledType,allowed,[]).
id_introduced(new_ids,_,unique,[]).
id_introduced(new_readonly_ids,_,unique,[readonly]).
id_introduced(new_readonly_ids(CoupledType),CoupledType,unique,[readonly]).
% Atelier-B actually does allow ANY/LET/VAR to override existing ids !
% TO DO: maybe only generate a warning in check_for_duplicates!

% check_for_duplicates(IdList,Env,TRin,TRout):
%  IdList is a list of (not typechecked) identifiers
%  If an identifier already exists in the environment Env, an error
%  is added (if the type matches the InvalidTypes argument)
%check_for_duplicates(IdList,Env,TRin,TRout) :- check_for_duplicates(IdList,Env,all,TRin,TRout).
check_for_duplicates(IdList,Env,InvalidTypes,TRin,TRout) :-
    foldl(check_for_duplicates_aux(Env,InvalidTypes),IdList,TRin,TRout).
check_for_duplicates_aux(Env,InvalidTypes,identifier(Pos,Id),TRin,TRout) :-
    ( (env_lookup_type(Id,Env,IdType), check_invalid_type(InvalidTypes,IdType)) ->
        (env_lookup_position_string(Id,Env,PosStr) -> PosInfo = [' at (line:column) =',PosStr] ; PosInfo = []),
        (get_idtype_desc(IdType,Desc)
          -> ajoin(['Identifier ',Id,' has already been declared as ', Desc|PosInfo],Msg)
           ; ajoin(['Identifier ',Id,' has already been declared'|PosInfo],Msg)),
        store_typecheck_warning(Msg,Pos,TRin,TRout)  % we now use store_typecheck_warning instead of store_typecheck_error
    ; otherwise ->
        TRin=TRout
    ).

get_idtype_desc(IdType,_) :- var(IdType),!,fail.
get_idtype_desc(global(S),Desc) :- !, ajoin(['element of SET ',S],Desc).
% TO DO: add more descriptions
get_idtype_desc(IdType,Desc) :- nonvar(IdType), functor(IdType,Desc,_).

% check if the type is one we do not allow in this context
check_invalid_type(all,_).
check_invalid_type([H|T],IdType) :- nonvar(IdType), functor(IdType,F,_), member(F,[H|T]).

% remove illegal couple terms introduced by Prolog BParser: TO DO: adapt Prolog BParser for UNION/INTER/...
flatten_illegal_couples([],Res) :- !, Res=[].
flatten_illegal_couples([couple(_Pos,Ids)],Res) :- !, Res = Ids.
flatten_illegal_couples([H|T],Res) :- !, Res = [H|T].
flatten_illegal_couples(identifier(Pos,ID),Res) :-
   add_warning(btypechecker,'Fixing error in AST, missing list:',ID,Pos),
   Res = [identifier(Pos,ID)].

% environment access
lookup_type(Id,Env,Pos,Type,TRin,TRout) :-
    ( env_lookup_type(Id,Env,Type) ->
        TRin=TRout
    ; otherwise ->
        get_current_context(Ctxt),
        (Id = op(Op)
          -> UMsg = 'Unknown operation "', Id0 = Op
          ;  UMsg = 'Unknown identifier "', Id0 = Id
        ),
        ( Id \= op(_), env_lookup_type(op(Id),Env,_OpType) ->
            ajoin(['Illegal use of operation identifier "',Id,'"',Ctxt],Msg)
        ; Id \= op(_), is_primed_of(Id,UnprimedId), env_lookup_type(UnprimedId,Env,_UpType) ->
            ajoin(['Primed version "', Id, '" of identifier ', UnprimedId, ' not available',Ctxt],Msg)
        ; fuzzy_find_possible_ids(Id0,Env,AlternativeIds,_Len) ->
            (AlternativeIds=[AlternativeId] ->
                ajoin([UMsg,Id0,'"',Ctxt,', did you mean "',AlternativeId,'"?'],Msg)
              ; wrap_ids_in_quotes(AlternativeIds,WIds),
                ajoin([UMsg,Id0,'"',Ctxt,', did you mean one of: '|WIds],Msg)
            )
        ; find_possible_completion_ids(Id0,Env,AlternativeIds,Len) ->
            (AlternativeIds=[AlternativeId] ->
               ajoin([UMsg,Id0,'"',Ctxt,', the possible completion is "',AlternativeId,'"'],Msg)
             ; Len=<3 ->
               wrap_ids_in_quotes(AlternativeIds,WIds),
               ajoin([UMsg,Id0,'"',Ctxt,', the ', Len, ' possible completions are '|WIds],Msg)
             ; prefix_length(AlternativeIds,AIds3,3), wrap_ids_in_quotes(AIds3,WIds),
               ajoin([UMsg,Id0,'"',Ctxt,', 3 possible completions (out of ',Len,') are '|WIds],Msg)
             )
        ; otherwise ->
            ajoin([UMsg,Id0,'"',Ctxt],Msg) % (lookup type)
            % TO DO: maybe look if the identifier exists in the machine project (b_get_machine_constants,b_get_machine_variables,...), but is not visible; we need bmachine to register the ids
            %, env_get_visible_ids(Env,Ids), format(user_error,'Known identifiers are: ~w~n',[Ids])
        ),
         % print(err(Id)),nl,print(Env),nl,trace,
        store_typecheck_error(Msg,Pos,TRin,TRout)).
lookup_infos(Id,Env,Infos) :-
    ( env_lookup_infos(Id,Env,Infos) -> !
    ; otherwise -> Infos=[]).

is_primed_of(Id0,Id) :- atom(Id0), atom_concat(Id,'$0',Id0).

:- use_module(tools_matching,[fuzzy_match_codes/2, codes_to_lower_case/2,get_current_keywords/1]).

fuzzy_find_possible_ids(Id,Env,AlternativeIds,Len) :-
  findall(AId,fuzzy_find_possible_id(Id,Env,AId),AlternativeIds),
  length(AlternativeIds,Len),
  Len>0.

fuzzy_find_possible_id(Id,Env,AlternativeId) :- atom_codes(Id,Codes1),
   codes_to_lower_case(Codes1,LCodes1),
   env_get_visible_ids_and_keywords(Env,Ids),
   member(AIdOrOp,Ids), id_or_op_codes(AIdOrOp,AlternativeId,Codes2),
   codes_to_lower_case(Codes2,LCodes2),
   fuzzy_match_codes(LCodes1,LCodes2).

env_get_visible_ids_and_keywords(Env,AllIds) :-
  env_get_visible_ids(Env,Ids),
  get_current_keywords(Keywords),
  ord_union(Ids,Keywords,AllIds).

wrap_ids_in_quotes([],[]).
wrap_ids_in_quotes([ID],['"',ID,'"']) :- !.
wrap_ids_in_quotes([ID|T],['"',ID,'", '|WT]) :- wrap_ids_in_quotes(T,WT).


find_possible_completion_ids(Id,Env,AlternativeIds,Len) :-
  findall(AId,find_possible_completion_id(Id,Env,AId),AlternativeIds),
  length(AlternativeIds,Len),
  Len>0.

% find possible completion of an (unknown) identifier
find_possible_completion_id(Id,Env,AlternativeId) :- atom_codes(Id,Codes1),
   env_get_visible_ids_and_keywords(Env,Ids),
   member(AIdOrOp,Ids), id_or_op_codes(AIdOrOp,AlternativeId,Codes2),
   prefix(Codes2,Codes1).

id_or_op_codes(op(ID),RealID,Codes) :- !, RealID=ID, atom_codes(ID,Codes).
id_or_op_codes(ID,ID,Codes) :- !, atom_codes(ID,Codes).


% check if the left hand side of an assignment is writeable
check_assign_lhs(Env,TExpr,TRin,TRout) :-
    ( id_or_function(TExpr,Id) ->
        get_texpr_type(TExpr,Type),
        lookup_infos(Id,Env,Infos),
        ( member(readonly,Infos) ->
            get_texpr_pos(TExpr,Pos),
            store_typecheck_error('read-only expression on left hand side of assignment ',Pos,TRin,TRout)
        ; nonvar(Type), Type = op(_,_) ->
            get_texpr_pos(TExpr,Pos),
            store_typecheck_error('operation on left hand side of assignment ',Pos,TRin,TRout)
        ; otherwise ->
            TRin=TRout
        )
    ; otherwise ->
        get_texpr_pos(TExpr,Pos),
        store_typecheck_error('unsupported expression on left hand side of assignment ',Pos,TRin,TRout)).


id_or_function(TIdFunc,Id) :-
    get_texpr_expr(TIdFunc,IdFunc),
    id_or_function2(IdFunc,Id).
id_or_function2(identifier(Id),Id).
id_or_function2(function(Func,_Arg),Id) :-
    id_or_function(Func,Id).

% extract information for identifiers from the environment
% currently, use all information but the position
extract_id_information(InfosIn,InfosOut) :-
    exclude(is_information_to_remove,InfosIn,InfosOut).
is_information_to_remove(nodeid(_)).
is_information_to_remove(remove(_)).


% some helper predicates
:- assert_must_succeed(btypechecker:couplise_list([a],a)).
:- assert_must_succeed(btypechecker:couplise_list([a,b],couple(a,b))).
:- assert_must_succeed(btypechecker:couplise_list([a,b,c],couple(couple(a,b),c))).
:- assert_must_succeed(btypechecker:couplise_list([a,b,c,d],couple(couple(couple(a,b),c),d))).
couplise_list([A],R) :- !,R=A.
couplise_list([A,B],R) :- !,R=couple(A,B).
couplise_list([A,B|Rest],Result) :-
    couplise_list([couple(A,B)|Rest],Result).

couplise_list_pos([A],_,A) :- !.
couplise_list_pos([A,B],Pos,couple(Pos,A,B)) :- !.
couplise_list_pos([A,B|Rest],Pos,Result) :-
    couplise_list_pos([couple(Pos,A,B)|Rest],Pos,Result).

idlist_to_type(Exprs,Type) :-
    get_texpr_types(Exprs,Types),
    couplise_list(Types,Type).

% remove choice_or constructs in a list
% there has to be at least one occurrence of choice_or
remove_choice_ors(Options,Substs) :-
    memberchk(choice_or(_,_),Options), % TODO: Is it ok that this could fail?
    maplist(remove_choice_or2,Options,Substs).
remove_choice_or2(choice_or(_,Subst),Subst) :- !.
remove_choice_or2(Subst,Subst).

add_identifiers_to_environment(Ids,In,Out) :-
    foldl(add_identifier_to_environment,Ids,In,Out).
add_identifier_to_environment(TExpr,In,Out) :-
    def_get_texpr_id(TExpr,Id),
    get_texpr_type(TExpr,Type),
    get_texpr_info(TExpr,Info),
    env_store(Id,Type,Info,In,Out).

% can nows also be used to unprime ids:
% Event-B style priming: x' is state of x after
prime_identifiers(TExprs,Primed) :-
    maplist(prime_identifier,TExprs,Primed).
prime_identifier(b(identifier(Id),Type,Info),b(identifier(Primed),Type,Info)) :-
    atom_concat(Id,'\'',Primed).
% Clasccial B semantics: x$0 is state of x before
prime_identifiers0(TExprs,Primed) :-
    maplist(prime_identifier0,TExprs,Primed).
prime_identifier0(b(identifier(Id),Type,Info),b(identifier(Primed),Type,Info)) :-
    atom_concat(Id,'$0',Primed).
prime_atom0(Id,Primed) :- atom_concat(Id,'$0',Primed).


% allow access to abstract variables by removing all "error/1"
% information from identifiers in the environment which also
% carry the "abstraction" information
allow_access_to_abstract_vars(OldEnv,NewEnv) :-
    env_transform_infos(OldEnv,_,OldInfos,NewInfos,NewEnv),
    maplist(allow_access_to_abstract_var,OldInfos,NewInfos).
% Iterate over the information list
allow_access_to_abstract_var(OldInfo,NewInfo) :-
    % abstract variable where no access is permitted
    member(abstraction,OldInfo),
    % remove the error flag
    select(error(_),OldInfo,NewInfo),!.
allow_access_to_abstract_var(Info,Info).

% for each local concrete_variable, add a variable x$0 of the same
% type to the scope. (Needed in while loop invariants)
% TODO: This is not completely correct. The variable should only be added
% if there is an abstract variable of the same name.
add_primed_old_value_variables(InvEnv,PInvEnv) :-
    env_transform_infos(InvEnv,Ids,Infos,_,_),
    foldl(add_primed_old_value_variable,Ids,Infos,InvEnv,PInvEnv).
add_primed_old_value_variable(Id,Info,In,Out) :-
    % We add a primed version if it is a concrete variable declared in this machine
    memberchk(loc(local,_,Sec),Info),memberchk(Sec,[concrete_variables,abstract_variables]),!,
    atom_concat(Id,'$0',Primed),
    env_lookup_type(Id,In,Type),
    env_store(Primed,Type,[refers_to_old_state(Id)],In,Out).
add_primed_old_value_variable(_Id,_Info,Env,Env).

% check if a reference to the original value of a variable is made.
% This is only possible in the loop invariant of a while statement
% The reference to an original value is an identifier marked with an refers_to_old_state(Id) tag,
% see add_primed_old_value_variables/2 above.
check_for_old_state_references(TExpr,Info) :-
    check_for_old_state_references2(TExpr,Refs,[]),
    sort(Refs,SortedRefs),
    ( SortedRefs = []    -> Info = []
    ; SortedRefs = [_|_] -> Info = [refers_to_old_state(SortedRefs)]).
check_for_old_state_references2(TExpr) -->
    { syntaxtraversion(TExpr,Expr,Type,Infos,Subs,_) },
    check_for_old_state_references3(Expr,Type,Infos),
    foldl(check_for_old_state_references2,Subs).
check_for_old_state_references3(identifier(Id),Type,Infos) -->
    {memberchk(refers_to_old_state(OrigId),Infos),!}, [oref(Id,OrigId,Type)].
check_for_old_state_references3(_Expr,_Type,_Infos) --> !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% error handling
store_typecheck_error(Msg,Pos,TRin,TRout) :-
    (typecheck_result_add_error(error(Msg,Pos),TRin,TRout) -> true
      ; add_internal_error('Could not store typecheck error: ', Msg),
        TRin=TRout).
store_typecheck_warning(Msg,Pos,TRin,TRout) :-
    (typecheck_result_add_error(warning(Msg,Pos),TRin,TRout) -> true
      ; add_internal_error('Could not store typecheck warning: ', Msg),
        TRin=TRout).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type unification

unify_types_werrors_l(A,B,Pos,Ein,Eout) :-
    typecheck_result_init(Ein,[],[],TRStart),
    unify_types_l(A,B,Pos,TRStart,TR),
    typecheck_result_get_errors(TR,Eout),
    typecheck_result_get_delayed_rules(TR,[]),
    typecheck_result_get_identifiers(TR,[]).
unify_types_werrors(A,B,Pos,Ein,Eout) :-
    unify_types_werrors_l([A],[B],[Pos],Ein,Eout).

unify_types_l([],[],[],TR,TR).
unify_types_l([A|Arest],[B|Brest],[Pos|Prest],TRin,TRout) :-
    unify_types(A,B,Pos,no_expression,TRin,TR1),
    unify_types_l(Arest,Brest,Prest,TR1,TRout).



unify_types(A,B,_,_,TR,TR) :-
    unify_types_strict(A,B),!.
unify_types(A,B,Pos,Expr,TRin,TRout) :-
    unify_failed(A,B,Pos,Expr,TRin,TRout).

unify_failed(A,B,Pos,Expr,TRin,TRout) :-
    %print(unify_failed(A,B,Pos,Expr,TRin,TRout)),nl,trace,
    ( nonvar(Expr),Expr\=no_expression,
      translate_bexpression(Expr,ExprString) % calling predicate with unwrapped AST
       ->
        TailMsg = [' in \'', ExprString, '\'']
    ; otherwise -> TailMsg = []),
    unify_failed_msg(A,B,MsgList,TailMsg),
    ajoin(MsgList,Msg),
    store_typecheck_error(Msg,Pos,TRin,TRout).

unify_failed_msg(A,B,MsgList,Rest) :-
    match_instance(A,record([field(Field,_)|_])),
    match_instance(B,record(Fields)),
    get_field_names(Fields,Names),
    nonmember(Field,Names),
    !,
    ajoin_with_sep(Names,',',NL),
    MsgList = ['Type mismatch: Expected record with field ', Field, ', but had only fields {',NL,'}'|Rest].
unify_failed_msg(A,B,['Type mismatch: Expected ', PA, ', but was ', PB|Rest],Rest) :- % TO DO: provide better user-feedback
    pretty_type(A,PA), pretty_type(B,PB).

get_field_names(V,_) :- var(V),!,fail.
get_field_names([],[]).
get_field_names([H|T],[FH|FT]) :- get_field_name(H,FH), get_field_names(T,FT).
get_field_name(F,Field) :- nonvar(F),F=field(Field,_).
match_instance(V,Pat) :- var(Pat),!,Pat=V.
match_instance(V,_Pat) :- var(V),!,fail.
match_instance(F,Pat) :- nonvar(F), functor(F,N,A), functor(Pat,N,A),
    (A>0 -> F=..FL, Pat =.. PL, maplist(match_instance,FL,PL) ; true).

%unify_failed(record([field(b,_54661)|_54843]),record([field(a,integer)]),pos(11,-1,1,18,1,20),b(identifier(r),record([field(a,integer)]),[nodeid(pos(12,-1,1,18,1,18))]),tr(_51997,_53271,_53273,[]),_54857)

:- assert_must_succeed(( btypechecker:unify_types_strict(set(couple(boolean,string)), set(couple(boolean,string))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(seq(integer), set(couple(integer,integer))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(seq(set(couple(integer,integer))),
                                                         set(couple(integer,seq(integer)))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(seq(integer), any) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(any, set(couple(integer,integer))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(_X, set(couple(integer,integer))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(seq(any), set(couple(integer,boolean))) )).

unify_types_strict(A,B) :-
    ( var(A) ->
        ( var(B)    -> A=B
        ; otherwise -> unify_types1(B,A))
    ; otherwise ->
        unify_types1(A,B)
    ),!.

unify_types1(A,B) :- var(B),contains_var(B,A),!,fail.
unify_types1(A,B) :- unify_types2(A,B).

unify_types2(any,B) :- !,
   (var(B) -> B=any
            ; true). % TO DO: set inner variables to any ?
unify_types2(_,B) :- B==any, !. % Maybe these two cases should be removed and type errors generated
                                % where the type cannot be inferred
unify_types2(set(A),set(B)) :- !, unify_types_strict(A,B).
unify_types2(seq(A),seq(B)) :- !, unify_types_strict(A,B).
unify_types2(set(A),seq(B)) :- !, unify_types_strict(seq(B),set(A)).
unify_types2(seq(A),set(B)) :- !, unify_types_strict(couple(integer,A),B).
unify_types2(couple(A,B),couple(X,Y)) :- !,unify_types_strict(A,X),unify_types_strict(B,Y).
unify_types2(string,string).
unify_types2(integer,integer).
unify_types2(boolean,boolean).
unify_types2(global(G),global(G)).
unify_types2(pred,pred).
unify_types2(subst,subst).
unify_types2(record(A),record(B)) :- !, unify_fields(B,A).
unify_types2(op(ParamsA,ResultsA),op(ParamsB,ResultsB)) :-
    unify_types_l(ParamsA,ParamsB),
    unify_types_l(ResultsA,ResultsB).
unify_types2(freetype(Name),freetype(Name)).
unify_types2(freetype(Name,TypeParams1),freetype(Name,TypeParams2)) :-
    unify_types_l(TypeParams1,TypeParams2).
unify_types2(witness,witness).
unify_types2(status,status).


unify_types_l(A,B) :- var(A),!,A=B.
unify_types_l([],[]) :- !.
unify_types_l([T1|Rest1],[T2|Rest2]) :-
    unify_types_strict(T1,T2),unify_types_l(Rest1,Rest2).

unify_fields(Sub,Super) :-
    nonvar(Sub),nonvar(Super),!,
    ( Sub = [field(Name,Type)|SubRest] ->
        extract_field(Super,Name,Type,SuperRest),
        unify_fields(SubRest,SuperRest)
    ; Sub = [] ->
        Super = []
    ; add_internal_error('Illegal type: ',unify_fields(Sub,Super)),fail
    ).
unify_fields(Fields,Fields).

extract_field(L,Name,Type,Rest) :-
    extract_field2(L,Name,Type,Rest).
extract_field2(L,Name,Type,Rest) :- nonvar(L),!,
    L = [field(FName,FType)|LRest],
    ( Name = FName ->
        unify_types_strict(Type,FType),
        Rest=LRest
    ; otherwise ->
        Rest=[field(FName,FType)|NRest],
        extract_field2(LRest,Name,Type,NRest)).
extract_field2(L,Name,Type,LRest) :-
    L = [field(Name,Type)|LRest].

% a special rule that checks whether a given record with name Id matches an expected type in a Field List
% used for record_field accesses
check_field_type(Fields,Id,Type,Pos,Trin,TRout) :-
    %when(nonvar(Fields), (Fields==[] -> print(empty_rec(Id,Pos)),nl,trace ; true)),
    typecheck_result_add_delayed_rule(Trigger,Errors,Trin,TRout),
    delayed_check_field_type(Fields,Id,Type,Pos,Trigger,Errors).

:- block delayed_check_field_type(-,?,?,?,-,?).
delayed_check_field_type(Fields,Id,_Type,Pos,Trigger,Errors) :- var(Fields),!,
   ajoin(['Could not determine type for record field access ',Id],Msg),
   delay_add_errors(Trigger,Errors,[error(Msg,Pos)]).
delayed_check_field_type([],Id,_Type,Pos,Trigger,Errors) :-
   ajoin(['Missing field for record field access ',Id],Msg),
   delay_add_errors(Trigger,Errors,[error(Msg,Pos)]).
delayed_check_field_type([field(Id,FType)|_],Id,Type,Pos,Trigger,Errors) :- !,
   (unify_types_strict(FType,Type) -> Errors=[]
    ; unify_failed_msg(Type,FType,MsgList,['for record field access ',Id]),
      ajoin(MsgList,Msg),
      delay_add_errors(Trigger,Errors,[error(Msg,Pos)])).
delayed_check_field_type([_|TF],Id,Type,Pos,Trigger,Errors) :-
    delayed_check_field_type(TF,Id,Type,Pos,Trigger,Errors).

% add errors when the delayed rule trigger is activated
:- block delay_add_errors(-,?,?).
delay_add_errors(_,Errors,Errors).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check if all types of the variables are ground
check_ground_types_dl(TExprs,Exceptions,Ein,Eout) :-  %print(check_ground(TExprs)),nl,
    foldl(check_ground_type_dl(Exceptions),TExprs,Ein,Eout).
check_ground_type_dl(Exceptions,TExpr,Ein,Eout) :-
    get_texpr_type(TExpr,Type),
    ( is_ground_type(Exceptions,Type) ->
        Ein=Eout
     % TO DO allow polymorphic typing, without any instantiation
    ; otherwise ->
        % (ground(Type) -> true ; format('Not ground, exceptions = ~w~n~w~n~n',[Exceptions,Type])),
        get_texpr_expr(TExpr,identifier(Id)),
        ( \+ get_preference(allow_untyped_identifiers,false),
             instantiate_non_ground_type(Exceptions,Type)
          -> translate:pretty_type(Type,TS),
             debug_format(20,'% Instantiated type of ~w to ~w~n',[Id,TS]),
             Ein=Eout
          ;  get_texpr_pos(TExpr,Pos),
             ajoin(['Could not infer type of ',Id], Msg),
             Ein=[error(Msg,Pos)|Eout]
        )
    ).
is_ground_type(_Exceptions,Type) :- ground(Type),!.
is_ground_type(Exceptions,Type) :- var(Type),!,exact_member(Type,Exceptions).
is_ground_type(Exceptions,Type) :-
    Type =.. [_|TArgs],
    maplist(is_ground_type(Exceptions),TArgs).

:- use_module(b_global_sets,[generate_fresh_supplementary_global_set/1]).
instantiate_non_ground_type(_Exceptions,Type) :- ground(Type),!.
instantiate_non_ground_type(Exceptions,Type) :- var(Type),!,
   (exact_member(Type,Exceptions) -> true
      ; get_preference(allow_untyped_identifiers,true_with_string_type) ->
        Type = string
      ; % if preference set to true we allow those untyped vars and use new deferred sets for them
        generate_fresh_supplementary_global_set(GS),
        Type = global(GS)
    ).
instantiate_non_ground_type(Exceptions,Type) :-
    Type =.. [_|TArgs],
    maplist(instantiate_non_ground_type(Exceptions),TArgs).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% rewrite rules to clean up the syntax tree

% substitute definitions
btype_rewrite(RawExpr,Env,Pos,Result,NewPos) :-
    is_definition_call(RawExpr,Env),!,
    rewrite_definition(RawExpr,Pos,Env,Result,NewPos).
% transform operation calls inside expressions
btype_rewrite(function(Fun,Args),Env,Pos,Result,Pos) :- %print(fun(Fun)),nl, portray_env(Env),
    is_operation_call(Fun,Env),!,
    debug_println(19,allowing_op_call_in_expression(Fun)),
    Result = operation_call_in_expr(Fun,Args).
% other (not definition-related) rules
btype_rewrite(Expr,_Env,Pos,New,Pos) :- btype_rewrite2(Expr,Pos,New).

is_operation_call(identifier(_,Name),Env) :-  lookup_operation_name(Name,Env,_,_).
% lookup OPERATION with Name in Environment Env
lookup_operation_name(Name,Env,ArgTypes1,ResTypes1) :-
    env_lookup_type(op(Name),Env,Type), nonvar(Type), %print(Type),nl,
    Type = op(ArgTypes1,ResTypes1).

% a shorter version of the
is_definition_call(identifier(Name),Env) :-
    lookup_definition(Name,Env,_,_Params,_Body).
is_definition_call(function(DefIdP,_Args),Env) :-
    remove_pos(DefIdP,identifier(Name)),
    % the definition should have parameters, if not, it could be
    % that a function is defined and normal function application is
    % needed
    % Note: There is a rewrite rule for functions that converts a list
    %   of arguments into couples. That rewrite rule must not be called
    %   before this rule.
    lookup_definition(Name,Env,_,[_|_],_Body).
is_definition_call(definition(_Name,_Args),_Env).

rewrite_definition(identifier(Name),Pos,Env,Result,NewPos) :-
    rewrite_definition2(Name,[],Pos,Env,Result,NewPos).
rewrite_definition(function(DefIdP,Args),Pos,Env,Result,NewPos) :-
    % A function application where the function is an identifier
    % may be a definition with arguments
    remove_pos(DefIdP,identifier(Name)),
    rewrite_definition2(Name,Args,Pos,Env,Result,NewPos).
rewrite_definition(definition(Name,Args),Pos,Env,Result,NewPos) :-
    % the standard way of using a definition
    rewrite_definition2(Name,Args,Pos,Env,Result,NewPos).

rewrite_definition2(Name,Args,CallPos,Env,Result,NewPos) :-
    ( lookup_definition(Name,Env,DefType,Params,Body) ->
        (btype_rewrite_definition(DefType,Params,Name,Env,Args,Body,NewBody)
         -> % add context_pos infos
            add_raw_ast_pos_info_context(definition_call(Name),CallPos,NewBody,NewBody1),
            remove_pos(NewBody1,Result,NewPos) % remove position info
            %,print(def_rewrote(Name,CallPos,Args)),nl,print(Result),nl
         ;  add_internal_error('Rewriting definition failed: ',Name),fail
         )
    ; otherwise ->
        ajoin(['Definition ',Name,' not found'],ErrMsg),
        throw(rewrite_exception(ErrMsg,Name))).

btype_rewrite_definition(_DefType,Params,Name,Env,Args,Body,NewBody) :-
    ( same_length(Params,Args) ->
        maplist(definition_expansion_on_raw_def_paras_with_pos(Env),Args,NewArgs),
        maplist(create_def_replace(Name),Params,NewArgs,Replaces),
        %% replacing DEFINITION parameters by actual arguments
        %% THIS CAN LEAD TO DUPLICATION of computation !
        %% TO DO: try and introduce let_expression,... (or use CSE later); problem: we are not yet in a typed context here; we may not always know whether this is a predicate, expression or substitution !?
        %% print(old_body(Name,Body,replaces(Replaces))),nl,
        raw_replace(Body,Replaces,NewBody)  %%  ,print(replaced_def(Name,NewBody)),nl %%
        %%TODO: keep track of original position as well !
    ; otherwise ->
        length(Params,PN),length(Args,AN),
        (AN=0
         -> ajoin(['Expected ',PN,' arguments for definition ',Name],ErrMsg) % this happens when we provide more than PN arguments; the parser seems to translate this into a definition call with 0 args and then a function call
          ; ajoin(['Expected ',PN,' instead of ',AN,' arguments for definition ',Name],ErrMsg)
        ),
        throw(rewrite_exception(ErrMsg,Name))).
create_def_replace(_,identifier(_,Old),New,Res) :- !, Res=replace(Old,New).
create_def_replace(Name,Term,_,_) :- (Term=definition(_Pos,Old,_) -> true ; Old=''),
        ajoin(['Cannot apply definition ',Name,', formal parameter ',Old, ' is not an identifier'],ErrMsg),
        throw(rewrite_exception(ErrMsg,Name)).
% lookup DEFINITION with Name in Environment Env
lookup_definition(Name,Env,DefType,Params,Body) :-
    env_lookup_type(Name,Env,Type), nonvar(Type),
    Type = definition(DefType,Params,Body).

% expressions
btype_rewrite2(couple([A,B]),_,couple(A,B)) :- !.
btype_rewrite2(couple(Elements),Pos,couple(X,Y)) :- couplise_list_pos(Elements,Pos,couple(_,X,Y)).
btype_rewrite2(function(Fun,Args),Pos,function(Fun,Arg)) :-
    couplise_list_pos(Args,Pos,Arg).

% substitutions and operations
btype_rewrite2(if(Pred,Subst,Elsifs,Else),Pos,if(IfList)) :-
    (Else=skip(_) -> IfList = [if_elsif(Pos,Pred,Subst)|Elsifs]
     ; append([if_elsif(Pos,Pred,Subst)|Elsifs],[if_elsif(none,truth(none),Else)],IfList)).
btype_rewrite2(case(Expr,Either,EitherSubst,Ors,Else),Pos,
                  case(Expr,[case_or(Pos,Either,EitherSubst)|Ors],Else)).
btype_rewrite2(select(Pred,Then,Whens,Else),Pos,select([select_when(Pos,Pred,Then)|Whens],Else)).
btype_rewrite2(select(Pred,Then,Whens),Pos,select([select_when(Pos,Pred,Then)|Whens])).
btype_rewrite2(choice([]),_,choice([])).
btype_rewrite2(choice(Options),_,choice(Substs)) :-
    remove_choice_ors(Options,Substs).

% integer sets
btype_rewrite2(integer_set,_,integer_set('INTEGER')).
btype_rewrite2(natural_set,_,integer_set('NATURAL')).
btype_rewrite2(natural1_set,_,integer_set('NATURAL1')).
btype_rewrite2(nat_set,_,integer_set('NAT')).
btype_rewrite2(nat1_set,_,integer_set('NAT1')).
btype_rewrite2(int_set,_,integer_set('INT')).

% applying rewriting of definitions on the raw (un-typed input) syntax tree
% this is used for parameters of definitions (which do not have their own position info)
definition_expansion_on_raw_def_paras_with_pos(Env,RawExprPos,OutAst) :-
    compound(RawExprPos),ext2int(RawExprPos,RawExpr,Pos,_,_,_,_),!,
    definition_expansion_on_compound_para(Pos,Env,RawExpr,NewPos,OutAst1),
    OutAst1 =.. [Functor|Args], OutAst =.. [Functor,NewPos|Args].
definition_expansion_on_raw_def_paras_with_pos(_Env,Ast,Ast).
definition_expansion_on_compound_para(Pos,Env,RawExpr,NewPos2,OutAst) :-
    is_definition_call(RawExpr,Env),!, % definition all within parameters of a definition
    rewrite_definition(RawExpr,Pos,Env,InterAst,NewPos),
    definition_expansion_on_compound_para(NewPos,Env,InterAst,NewPos2,OutAst).
definition_expansion_on_compound_para(Pos,Env,InAst,Pos,OutAst) :-
    compound(InAst),
    InAst =.. [Functor|Args],!,
    same_functor(InAst,OutAst),
    OutAst =.. [Functor|NewArgs],
    maplist(definition_expansion_on_raw_def_paras_with_pos(Env),Args,NewArgs).
definition_expansion_on_compound_para(Pos,_Env,Ast,Pos,Ast).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% extract syntax tree positions

remove_pos(Ext,Expr,Pos) :-
    ext2int(removepos,Ext,Expr,Pos,_,_,_,_).
remove_pos(Ext,Expr) :-
    ext2int(removepos,Ext,Expr,_,_,_,_,_).

% create a enhanced syntax element from a raw input
% ext2int(+Ext,-Expr,-Pos,+Type,+TExpr,+RestInfos,-Result):
%  Ext:  raw information with position info (from parser)
%  Expr: the underlying expression without position info (e.g. identifier(x))
%  Pos:  the position extracted from the raw expression
%  Type: the type used for the typed expression
%  TExpr: the new expression in the enhanced syntax tree format
%        (but without the information for the root element)
%  RestInfos: Additional list of informations that will be stored inside
%        the element aside from the position information
%  Result: the new completed element (with type and additional information)
%     Example: "1+x"
%          Ext:    add(pos1,integer(pos2,1),identifier(pos3,x)
%          Expr:   add(integer(pos2,1),identifier(pos3,x))
%          Result: b(TExpr,Type,[nodeid(pos1)|RestInfos])
ext2int(Ext,Expr,Pos,Type,TExpr,RestInfos,Res) :-
    ext2int(unknown_src,Ext,Expr,Pos,Type,TExpr,RestInfos,Res).
ext2int(Src,Ext,Expr,_Pos,_Type,_TExpr,_RestInfos,_) :-
    var(Ext),!,add_error(ext2int,'Variable Ext: ',Src:Expr,Expr),fail.
ext2int(_Src,Ext,Expr,Pos,Type,TExpr,RestInfos,b(TExpr,Type,[nodeid(Pos)|Infos])) :-
    % the first argument contains the position information
    safe_functor(ext2int_ext,Ext,Functor,Arity),
    (Arity=0 -> add_internal_error('Not raw AST:',Ext),fail ; true),
    arg(1,Ext,ExtPos),
    extract_optional_external_info(ExtPos,Pos,Infos,RestInfos),
    % now construct a copy without the first argument
    NArity is Arity-1,
    safe_functor(ext2int_expr,Expr,Functor,NArity),
    copy_ext_term(Arity,Ext,Expr).
% It is possible to inject information into the AST from outside
% by using info(Pos,Info) or info(Pos,[Info1,...,InfoN]) as position
% Those terms are copied directly into the AST.
% The result is returned in form of a difference list I1-I2.
extract_optional_external_info(info(Pos,ExtInfo),Pos,I1,I2) :- !,
    ( ExtInfo = []    -> I1=I2
    ; ExtInfo = [_|_] -> append(ExtInfo,I2,I1)
    ; otherwise       -> I1=[ExtInfo|I2]).
extract_optional_external_info(Pos,Pos,Infos,Infos).

copy_ext_term(1,_,_) :- !.
copy_ext_term(N1,Ext,Int) :-
    N2 is N1-1, arg(N1,Ext,Arg), arg(N2,Int,Arg),
    copy_ext_term(N2,Ext,Int).


check_for_duplicate_raw_ids([],_,_,_,[],TR,TR).
check_for_duplicate_raw_ids([identifier(Pos,IdName)|T],List,Construct,Name,Res,TRin,TRout) :- !,
    %print(check(IdName,Name,Construct,List)),nl,
    (member(IdName,List)
     -> ajoin(['Duplicate identifier ',IdName,' for ',Construct,Name],Msg),
        store_typecheck_error(Msg,Pos,TRin,TR2),
        check_for_duplicate_raw_ids(T,List,Construct,Name,Res,TR2,TRout)
      ; Res = [identifier(Pos,IdName)|ResT],
        check_for_duplicate_raw_ids(T,[IdName|List],Construct,Name,ResT,TRin,TRout)).
check_for_duplicate_raw_ids([H|T],List,Construct,Name,[H|RT],TRin,TRout) :-
    check_for_duplicate_raw_ids(T,List,Construct,Name,RT,TRin,TRout).

% add variables to the environment
% - the variables are given as raw parser terms
%   and returned in typed format
% - additionally the identifier is stored in the list of newly introduced
% identifiers
add_ext_variables(Identifiers, Env, TExprs, NewEnv, TRin, TRout) :-
    add_ext_variables_with_info(Identifiers, Env, [], TExprs, NewEnv, TRin, TRout).

add_ext_variables_with_info(Identifiers, Env, AddInfo, TExprs, NewEnv, TRin, TRout) :-
    foldl(add_ext_variable_with_info(AddInfo), Identifiers, TExprs, TRin/Env, TRout/NewEnv).
add_ext_variable_with_info(AddInfo, Identifier, TExpr, TRin/Env, TRout/NewEnv) :-
    % convert the raw format to the typed format
    Expr = identifier(Name),
    % Krings: in the following line, AddInfo was set to []
    % thus, additional information was only added to the type environment,
    % not to the identifier itself. introduced_by was only store inside of
    % expressions, not at the identifier declarations. this might break something?
    ext2int(add_ext_variables_with_info,Identifier,Expr,_,Type,Expr,AddInfo,TExpr),!,
    % add the identifier's type and info to the type environment
    get_texpr_info(TExpr,Infos),
    % append not needed anymore, as additional information is stored in
    % the identifier. see comment above.
    %append(AddInfo,Infos,AllInfos),
    env_store(Name,Type,Infos,Env,NewEnv),
    % store the identifier in the list of newly introduced identifiers
    typecheck_result_add_identifier(TExpr,TRin,TRout).
add_ext_variable_with_info(_AddInfo, Expr, TExpr, TRin/Env, TRout/Env) :-
    % We expected an identifier but there was something different
    ext2int(add_ext_variables_with_info,Expr,_,Pos,_,_,_,_),
    ( translate_raw_expr_functor(Expr,TS)
      -> ajoin(['Expected identifier, but got operator: ',TS],Msg),
         ajoin(['not an identifier:',TS],ID)
      ;  functor(Expr,F,_), ajoin(['not an identifier:',F],ID),
         Msg='Expected identifier'
    ),
    store_typecheck_error(Msg,Pos,TRin,TRout),
    % just to make sure that there is no variable as AST element:
    create_texpr(identifier(ID),_,[found(Expr)],TExpr).

translate_raw_expr_functor(Expr,Translation) :-
    % note: we cannot call translate_bexpression yet; not yet wrapped and typed !
    nonvar(Expr), functor(Expr,F,_),
    translate_prolog_constructor(F,Translation).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type rules

% predicates
truth :: pred.
falsity :: pred.
conjunct(pred,pred) :: pred.
negation(pred) :: pred.
disjunct(pred,pred) :: pred.
implication(pred,pred) :: pred.
equivalence(pred,pred) :: pred.
equal(A,A) :: pred.
not_equal(A,A) :: pred.
member(A,set(A)) :: pred.
not_member(A,set(A)) :: pred.
subset(set(A),set(A)) :: pred.
subset_strict(set(A),set(A)) :: pred.
not_subset(set(A),set(A)) :: pred.
not_subset_strict(set(A),set(A)) :: pred.
less_equal(integer,integer) :: pred.
less(integer,integer) :: pred.
greater_equal(integer,integer) :: pred.
greater(integer,integer) :: pred.
finite(set(_)) :: pred.
exists(ids,pred) :: pred.
partition(set(T),[set(T)]) :: pred.

% expressions
boolean_true :: boolean.
boolean_false :: boolean.
max_int :: integer.
min_int :: integer.
empty_set :: set(_).
bool_set :: set(boolean).
string_set :: set(string).
convert_bool(pred) :: boolean.
add(integer,integer) :: integer.
minus(integer,integer) :: integer.
unary_minus(integer) :: integer.
multiplication(integer,integer) :: integer.
cartesian_product(set(A),set(B)) :: set(couple(A,B)).
div(integer,integer) :: integer.
floored_div(integer,integer) :: integer.
modulo(integer,integer) :: integer.
power_of(integer,integer) :: integer.
successor :: set(couple(integer,integer)).
predecessor :: set(couple(integer,integer)).
max(set(integer)) :: integer.
min(set(integer)) :: integer.
card(set(_)) :: integer.
couple(A,B) :: couple(A,B).
pow_subset(set(A)) :: set(set(A)).
pow1_subset(set(A)) :: set(set(A)).
fin_subset(set(A)) :: set(set(A)).
fin1_subset(set(A)) :: set(set(A)).
interval(integer,integer) :: set(integer).
union(set(A),set(A)) :: set(A).
intersection(set(A),set(A)) :: set(A).
set_subtraction(set(A),set(A)) :: set(A). /* ?? */
general_union(set(set(A))) :: set(A).
general_intersection(set(set(A))) :: set(A).
relations(set(A),set(B)) :: set(set(couple(A,B))).
identity(set(A)) :: set(couple(A,A)).
event_b_identity :: set(couple(A,A)). % for Rodin 1.0
reverse(set(couple(A,B))) :: set(couple(B,A)). % function inverse ~
first_projection(set(A),set(B)) :: set(couple(couple(A,B),A)).
event_b_first_projection(set(couple(A,B))) :: set(couple(couple(A,B),A)). % should no longer appear after Rodin 1.0
event_b_first_projection_v2 :: set(couple(couple(A,_B),A)). % for Rodin 1.0
second_projection(set(A),set(B)) :: set(couple(couple(A,B),B)).
event_b_second_projection(set(couple(A,B))) :: set(couple(couple(A,B),B)). % should no longer appear after Rodin 1.0
event_b_second_projection_v2 :: set(couple(couple(_A,B),B)). % for Rodin 1.0
composition(set(couple(A,B)),set(couple(B,C))) :: set(couple(A,C)).
ring(set(couple(B,C)),set(couple(A,B))) :: set(couple(A,C)).
direct_product(set(couple(E,F)),set(couple(E,G))) :: set(couple(E,couple(F,G))).
parallel_product(set(couple(E,F)),set(couple(G,H))) :: set(couple(couple(E,G),couple(F,H))).
trans_function(set(couple(A,B))) :: set(couple(A,set(B))).
trans_relation(set(couple(A,set(B)))) :: set(couple(A,B)).
iteration(set(couple(A,A)),integer) :: set(couple(A,A)).
reflexive_closure(set(couple(A,A))) :: set(couple(A,A)).
closure(set(couple(A,A))) :: set(couple(A,A)).
domain(set(couple(A,_))) :: set(A).
range(set(couple(_,B))) :: set(B).
image(set(couple(A,B)),set(A)) :: set(B).
domain_restriction(set(A),set(couple(A,B))) :: set(couple(A,B)).
domain_subtraction(set(A),set(couple(A,B))) :: set(couple(A,B)).
range_restriction(set(couple(A,B)),set(B)) :: set(couple(A,B)).
range_subtraction(set(couple(A,B)),set(B)) :: set(couple(A,B)).
overwrite(set(couple(A,B)),set(couple(A,B))) :: set(couple(A,B)).
partial_function(A,B)::T    :- relations(A,B)::T.
total_function(A,B)::T      :- relations(A,B)::T.
partial_injection(A,B)::T   :- relations(A,B)::T.
total_injection(A,B)::T     :- relations(A,B)::T.
partial_surjection(A,B)::T  :- relations(A,B)::T.
total_surjection(A,B)::T    :- relations(A,B)::T.
total_bijection(A,B)::T     :- relations(A,B)::T.
partial_bijection(A,B)::T     :- relations(A,B)::T.
total_relation(A,B)::T      :- relations(A,B)::T.
surjection_relation(A,B)::T :- relations(A,B)::T.
total_surjection_relation(A,B)::T :- relations(A,B)::T.
seq(set(A)) :: set(seq(A)).
seq1(X)::T  :- seq(X)::T.
iseq(X)::T  :- seq(X)::T.
iseq1(X)::T :- seq(X)::T.
perm(set(A)) :: set(seq(A)).
empty_sequence :: seq(_).
size(seq(_)) :: integer.
first(seq(A)) :: A.
last(seq(A)) :: A.
front(seq(A)) :: seq(A).
tail(seq(A)) :: seq(A).
rev(seq(A)) :: seq(A).
concat(seq(A),seq(A)) :: seq(A).
insert_front(A,seq(A)) :: seq(A).
insert_tail(seq(A),A) :: seq(A).
restrict_front(seq(A),integer) :: seq(A).
restrict_tail(seq(A),integer) :: seq(A).
general_concat(seq(seq(A))) :: seq(A).
function(set(couple(A,B)),A) :: B.
set_extension([T]) :: set(T).
sequence_extension([T]) :: seq(T).
comprehension_set(ids(T),pred) :: set(T).
event_b_comprehension_set(ids,T,pred) :: set(T).
lambda(ids(A),pred,R) :: set(couple(A,R)).
general_sum(ids,pred,integer) :: integer.
general_product(ids,pred,integer) :: integer.
quantified_union(ids,pred,set(T)) :: set(T).
quantified_intersection(ids,pred,set(T)) :: set(T).
mu(set(T)) :: T.
if_then_else(pred,T,T) :: T.
tree(set(A)) :: set(set(couple(seq(integer),A))). % Section 5.20 of Atelier-B manrefb.pdf
btree(A) :: T :-  tree(A) :: T.
tree_over(A) :: set(couple(seq(integer),A)).  % dummy function to ease writing of rules below
const(A,seq(TreeA)) :: TreeA :- tree_over(A) :: TreeA.
top(TreeA) :: A :- tree_over(A) :: TreeA.
prefix(TreeA) :: seq(A) :- tree_over(A) :: TreeA.
postfix(TreeA) :: seq(A) :- tree_over(A) :: TreeA.
sizet(TreeA) :: integer :- tree_over(_) :: TreeA.
mirror(TreeA) :: TreeA :- tree_over(_) :: TreeA.
rank(TreeA,seq(integer)) :: integer :- tree_over(_) :: TreeA.  % seq(integer) is the type of paths
father(TreeA,seq(integer)) :: seq(integer) :- tree_over(_) :: TreeA.
son(TreeA,seq(integer),integer) :: seq(integer) :- tree_over(_) :: TreeA.
subtree(TreeA,seq(integer)) :: TreeA :- tree_over(_) :: TreeA.
arity(TreeA,seq(integer)) :: integer :- tree_over(_) :: TreeA.
sons(TreeA) :: seq(TreeA) :- tree_over(_) :: TreeA.
bin(A) :: TreeA :- tree_over(A) :: TreeA.
bin(TreeA,A,TreeA) :: TreeA :- tree_over(A) :: TreeA.
infix(TreeA) :: seq(A) :- tree_over(A) :: TreeA.
left(TreeA) :: TreeA :- tree_over(_) :: TreeA.
right(TreeA) :: TreeA :- tree_over(_) :: TreeA.

% substitutions
skip :: subst.
block(subst) :: subst.
precondition(pred,subst) :: subst.
sequence([subst]) :: subst.
any(new_readonly_ids,pred,subst) :: subst.
if([subst]) :: subst.
if_elsif(pred,subst) :: subst.
assertion(pred,subst) :: subst.
let(new_readonly_ids,pred,subst) :: subst.
var(new_ids,subst) :: subst. % was ids instead of new_ids; see test 1658
select([subst],subst) :: subst.
select([subst]) :: subst.
select_when(pred,subst) :: subst.
choice([subst]) :: subst.

% other stuff
witness(_,pred) :: witness.
ordinary :: status.
anticipated(_) :: status.
convergent(_) :: status.


% error_rewrites: common mistakes users make
% the message will be wrapped in 'Did you ' and ' ?'
:- use_module(specfile,[eventb_mode/0]).
error_rewrite(insert_front(A,B),insert_tail(A,B),'use -> instead of <- (sequence append)').
error_rewrite(insert_front(A,B),couple(A,B),'use -> instead of |-> (maplet) or total function (-->)').
error_rewrite(insert_front(A,B),total_function(A,B),'use -> instead of --> (total function)').
error_rewrite(insert_tail(A,B),insert_front(A,B),'use <- instead of -> (sequence prepend)').
error_rewrite(image(A,B),image(A,set_extension(unknown_src,[B])),'forget the curly braces {.} inside the relational image [.]'). % :- get_texpr_type(B,TB). % f[1] instead of f[{1}]
%error_rewrite(image(A,B),function(A,B),'use relational image .[.] instead of function application .(.)'). % does not seem to trigger
error_rewrite(concat(A,B),power_of(A,B),'use ^ (concat) instead of ** (exponentation)').
error_rewrite(power_of(A,B),cartesian_product(A,B),'use ** (exponentation) instead of * (cartesian product)') :- \+ eventb_mode.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%+
% type environment
env_empty(env(Avl)) :- empty_avl(Avl).

env_lookup_type(Id,Env,Type) :-
    env_lookup2(Env,Id,Type,_Infos).
env_lookup_infos(Id,Env,Infos) :-
    env_lookup2(Env,Id,_Type,Infos).
env_lookup_position_string(Id,Env,PosStr) :-
    env_lookup2(Env,Id,_Type,Infos),
    get_info_pos(Infos,Pos), Pos \= none,
    translate:translate_span(Pos,PosStr).
env_lookup2(env(Avl),Id,Type,Infos) :-
    avl_fetch(Id,Avl,entry(Type,Infos)).
env_lookup2(openenv(Open,Env),Id,Type,Infos) :-
    ( env_lookup2(Env,Id,Type,Infos) ->
        true
    ; otherwise ->
        openenv_lookup(Open,Id,Type,Infos)).

env_get_visible_ids(env(Avl),Ids) :- avl_domain(Avl,Ids).

env_store(Id,Type,Infos,In,Out) :-
    env_store2(In,Id,Type,Infos,Out).
env_store2(env(In),Id,Type,Infos,env(Out)) :-
    avl_store(Id,In,entry(Type,Infos),Out).
env_store2(openenv(Open,In),Id,Type,Infos,openenv(Open,Out)) :-
    env_store2(In,Id,Type,Infos,Out).

env_store_definition(definition_decl(Name,DefType,_Pos,Params,Body,_Deps), In, Out) :-
    env_store(Name, definition(DefType,Params,Body), [], In, Out).

% store a callback predicate in the environment
env_store_operator(Id,Module:Callback,In,Out) :-
    % the type is ignored, we just use the info field
    env_store(operator(Id),boolean,[Module:Callback],In,Out).

% the open environment acts like the normal type environment with the
% exception that a lookup of an identifier always works, in case the
% identifier was not in the environment, it will be added to it.
openenv(Env,openenv(_,Env)).
openenv_lookup([Head|Tail],Id,Type,Infos) :-
    Infos=[],
    ( Head = entry(Id,_,_) ->
        Head = entry(Id,Type,Infos)
    ; otherwise ->
        openenv_lookup(Tail,Id,Type,Infos)).
openenv_identifiers(openenv(List,_),Ids) :-
    openenv_identifiers2(List,Ids).
openenv_identifiers2(List,[]) :- var(List),!.
openenv_identifiers2([],[]).
openenv_identifiers2([entry(Id,Type,Infos)|Rest],[TId|IdRest]) :-
    create_texpr(identifier(Id),Type,Infos,TId),
    openenv_identifiers2(Rest,IdRest).

env_identifiers(env(Avl),Ids) :-
    avl_domain(Avl,Ids).
env_identifiers(openenv(Open,Env),Ids) :-
    env_identifiers(Env,Ids1),
    open_env_identifiers(Open,Ids2),
    append(Ids2,Ids1,UIds),sort(UIds,Ids).
open_env_identifiers(L,Ids) :- var(L),!,Ids=[].
open_env_identifiers([],[]).
open_env_identifiers([entry(Id,_,_)|Orest],[Id|Irest]) :-
    open_env_identifiers(Orest,Irest).


% gives a list of all known identifiers and their information in a type
% environment. A second list of the same size is used to build up a new
% environment whith the same identifiers and types but with new information
env_transform_infos(env(Avl),Ids,OldInfos,NewInfos,env(NewAvl)) :-
    % transform a standard environment by converting it to a list,
    % iterate over the list and convert the result back
    avl_to_list(Avl,OldEntries),
    maplist(env_transinfo2,OldEntries,Id_OldInfo_NewInfo,NewEntries),
    maplist(split_slash,Id_OldInfo_NewInfo,Ids,Old_New_Info),
    maplist(split_slash,Old_New_Info,OldInfos,NewInfos),
    ord_list_to_avl(NewEntries,NewAvl).
env_transform_infos(openenv(Open,Env),Ids,OldInfos,NewInfos,openenv(NewOpen,NewEnv)) :-
    % an open environment iterates first over the (open) list,
    % calls then the code for the standard environment and concatenates
    % the generated lists.
    env_transinfo_open(Open,Ids1,OldInfos1,NewInfos1,NewOpen),
    env_transform_infos(Env,Ids2,OldInfos2,NewInfos2,NewEnv),
    append(Ids1,Ids2,Ids),
    append(OldInfos1,OldInfos2,OldInfos),
    append(NewInfos1,NewInfos2,NewInfos).
% iterate over standard environments
env_transinfo2(Id-entry(Type,OldInfo),Id/(OldInfo/NewInfo),Id-entry(Type,NewInfo)).
% iterate over the open list (last element is not [] but a variable)
env_transinfo_open(OpenEnv,[],[],NewInfo,NewEnv) :- var(OpenEnv),!,OpenEnv=NewEnv,NewInfo=[].
env_transinfo_open([],[],[],[],[]).
env_transinfo_open([entry(Id,Type,OldInfo)|ORest],[Id|IdRest],[OldInfo|OIRest],[NewInfo|NIRest],[entry(Id,Type,NewInfo)|NRest]) :-
    env_transinfo_open(ORest,IdRest,OIRest,NIRest,NRest).
split_slash(A/B,A,B).


% utility to portray environment
:- public portray_env/1.
portray_env(env(In)) :- !, format('ENV:~n',[]),portray_avl_env(In).
portray_env(openenv(Open,In)) :- !, format('OPENENV (~w): ',[Open]),portray_avl_env(In).
portray_env(E) :- !, format('*** UNKNOWN ENV: ~w~n',[E]).
portray_avl_env(A) :- avl_member(Id,A,entry(Type,_Infos)), format('  ~w : ~w~n',[Id,Type]),fail.
portray_avl_env(_) :- nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% fast manual generation of syntax trees
%   this code is only for testing, not for use in the regular type checking
%   of user-given expressions

:- if(current_prolog_flag(dialect,sicstus)).
:- if((current_prolog_flag(version_data,sicstus(4,X,_,_,_)),X<3)).
:- use_module(library(terms),[term_variables/2]). % is built-in in SICSTUS 4.3
:- endif.
:- endif.

fasttype(Expr,TExpr) :-
    term_variables(Expr,OldVars),
    ( Expr= +_  -> fasttype2(Expr,TExpr,unknown)
    ; otherwise -> fasttype2(Expr,TExpr,_Type)),
    term_variables(TExpr,NewVars),
    ( fasttype_check_new_vars(NewVars,OldVars) -> true
    ; add_error(btypechecker, 'fasttype/2 introduced new variables'),fail).
fasttype2(Type<<Expr,b(TExpr,Type,[]),Type) :- !,
    safe_functor(fasttype2,Expr,F,Arity),
    safe_functor(fasttype2,TExpr,F,Arity),
    safe_functor(fasttype2,Pattern,F,Arity),
    ( Pattern :: Type ->
        fasttype_args(Arity,Expr,TExpr,Pattern)
    ; otherwise ->
        fasttype_all_unknown(Arity,Pattern)),
    fasttype_args(Arity,Expr,TExpr,Pattern).
fasttype2(+Expr,TExpr,Type) :- !,
    ( Type==unknown ->
        safe_functor(fasttype2,Expr,F,Arity),
        safe_functor(fasttype2,Pattern,F,Arity),
        ( fasttype_lookup(Pattern,NType) ->
            fasttype2(NType<<Expr,TExpr,NType)
        ; otherwise ->
            add_error(btypechecker,'Not able to simply infer type:',Expr),fail)
    ; otherwise ->
        fasttype2(Type<<Expr,TExpr,Type)).
fasttype2([H|T],[TH|TT],_) :- !,fasttype2(H,TH,unknown),fasttype2(T,TT,unknown).
fasttype2(Expr,Expr,_).

fasttype_all_unknown(0,_) :- !.
fasttype_all_unknown(N,Pattern) :-
    arg(N,Pattern,unknown),
    N1 is N-1, fasttype_all_unknown(N1,Pattern).

fasttype_args(0,_,_,_) :- !.
fasttype_args(N,Expr,NExpr,Pattern) :-
    arg(N,Expr,Arg), arg(N,NExpr,NArg),
    arg(N,Pattern,ArgType),
    fasttype2(Arg,NArg,ArgType),
    N1 is N-1,
    fasttype_args(N1,Expr,NExpr,Pattern).

fasttype_lookup(integer(_),integer).
fasttype_lookup(Pattern,Type) :- Pattern :: Type.

fasttype_check_new_vars(Old,New) :-
    sort(Old,SOld), sort(New, SNew),
    fasttype_check_new_vars2(SOld,SNew).
fasttype_check_new_vars2([],_).
fasttype_check_new_vars2([N|NRest],[O|ORest]) :-
    ( N==O      -> fasttype_check_new_vars(NRest,ORest)
    ; otherwise -> fasttype_check_new_vars([N|NRest],ORest)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type expressions without declaring the identifiers

% opentype type-checks an expression without having to declare every used
% identifier in advance.
%   Exprs        - a list of raw expressions to be type-checked
%   InitEnv      - an initial environment
%   Identifiers  - used identifiers that weren't in the initial environment
%   DeferredSets - new introduced deferred sets for variables whose type couldn't be infered
%   CleanExprs   - the type-checked (and cleaned-up) list of expressions
opentype(Exprs,InitEnv,Identifiers,DeferredSets,CleanExprs) :-
    openenv(InitEnv,OpenEnv),
    do_typecheck(Exprs,OpenEnv,[],TypedExprs,Errors),
    ( Errors = [] ->
        true
    ; Errors = [_|_] ->
        throw(type_errors(Errors))),
    % all identifier I whose type have the form set(_),
    % introduce implicitly a new type global(I)
    foldl(declare_deferred_sets,TypedExprs,DS1,[]),
    % ensure that all bounded identifiers have ground types
    ground_name_types_by_defsets(TypedExprs,DS2),
    % ensure that all free identifiers have ground types
    openenv_identifiers(OpenEnv,AllIdentifiers),
    ground_free_types_by_defsets(AllIdentifiers,DS3),
    % join all deferred sets
    append([DS1,DS2,DS3],DeferredSets),
    % the returned identifiers should not contain the global sets
    exclude(tid_in_idlist(DeferredSets),AllIdentifiers,Identifiers),
    % clean the expression up
    clean_up_l_wo_optimizations(TypedExprs,[],CleanExprs,opentype).

% used by opentype/4
tid_in_idlist(IdList,TId) :-
    def_get_texpr_id(TId,Id),
    memberchk(Id,IdList).

% used by opentype/4
declare_deferred_sets(TExpr) -->
    {syntaxtraversion(TExpr,Expr,Type,_,Subs,_)},
    declare_deferred_sets2(Expr,Type),
    foldl(declare_deferred_sets,Subs).
declare_deferred_sets2(identifier(Id),Type) -->
    {nonvar(Type),Type=set(SType),var(SType),!,SType=global(Id)},[Id].
declare_deferred_sets2(_,_) --> [].

% used by opentype/4, get all types of introduced identifiers
% (e.g. in quantifiers) and ground them by introducing deferred sets
ground_name_types_by_defsets(TExprs,DefSets) :-
    get_all_name_types_l(TExprs,Types),
    ground_by_defsets(Types,'_NSet',DefSets).
% used by opentype/4, get all types of the given identifiers
% and ground them by introducing deferred sets
ground_free_types_by_defsets(Ids,DefSets) :-
    get_texpr_types(Ids,Types),
    ground_by_defsets(Types,'_FSet',DefSets).
% takes a list of types and ground them by introducing
% deferred sets, each set's name starting with Prefix
ground_by_defsets(Types,Prefix,DefSets) :-
    term_variables(Types,VarTypes),
    foldl(ground_by_defsets2(Prefix),VarTypes,DefSets,1,_).
ground_by_defsets2(Prefix,global(DS),DS,N,N2) :-
    number_chars(N,CN),
    safe_atom_chars(Prefix,CS,ground_by_defsets2_1),
    append(CS,CN,Chars),
    safe_atom_chars(DS,Chars,ground_by_defsets2_2),
    N2 is N+1.

% get the types of all introduced identifiers
% (e.g. in quantifiers)
get_all_name_types(TExpr,Types) :-
    syntaxtraversion(TExpr,_,_,_,Subs,Names),
    get_texpr_types(Names,NTypes),
    append(NTypes,RTypes,Types),
    get_all_name_types_l(Subs,RTypes).
get_all_name_types_l(Exprs,Types) :-
    maplist(get_all_name_types,Exprs,LTypes),
    append(LTypes,Types).

%check_nonvar(Src,V) :- var(V) -> add_error(check_nonvar,'Is a variable: ',Src:V) ; true.

% type-checking of external functions/predicates/substitutions
btype_external_call(FunName,rewrite_protected(TypeParams),FunParams,rewrite_protected(Decl),
                    DeclType,InputType,Pos,Env,TFunParams,TRin,TRout) :-
    add_ext_variables(TypeParams,Env,_TTypeParams,DeclSubenv,TRin,TR1),
    btype(extcall_decl,Decl,DeclSubenv,DeclType,_TypedDecl,TR1,TR2),
    % TODO: check against originally (for Atelier-B) defined function
    % first identify the types of the parameters
    btype_l(FunParams,Env,ParamTypes,TFunParams,TR2,TR3),
    % now check if the function parameters match the input type
    (ParamTypes = []
     ->  /* the external call has no parameters */
         (InputType = empty_type -> TR3=TRout
          ; add_error(btype_external_call,'External Definition has no arguments:',FunName),
            fail)
     ; couplise_list(ParamTypes,JoinedParamType),
       unify_types(InputType,JoinedParamType,Pos,no_expression,TR3,TRout)).

% ----------------------------


%   add CallPos (if /= unknown) as additional position information for sub-expressions

add_raw_ast_pos_info_context(Ctxt,CPos,Raw,NewRaw) :-
      % for definition calls we only store the actual position of the call,
      %  not in which scope this call itself has occurred (to avoid overly complex messages)
      simplify_call_pos(CPos,SimpleCPos),
      add_raw_ast_pos_info_context2(Ctxt,SimpleCPos,Raw,NewRaw).

simplify_call_pos(pos_context(P1,_,_),Res) :- !,
   simplify_call_pos(P1,Res).
simplify_call_pos(P,P).

add_raw_ast_pos_info_context2(_,_,Raw,NewRaw) :-
      simple_raw(Raw),!,
      NewRaw=Raw.
add_raw_ast_pos_info_context2(Ctxt,CPos,Raw,NewRaw) :-
      list_operator(Raw,Pos,L,NewRaw,NewPos,NewL),!,
      combine_pos(Pos,Ctxt,CPos,NewPos),
      maplist(add_raw_ast_pos_info_context2(Ctxt,CPos),L,NewL).
add_raw_ast_pos_info_context2(Ctxt,CPos,Raw,NewRaw) :-  % print(Raw),nl,
      Raw =.. [Op,Pos1|T], raw_operator(Op),
      !,
      maplist(add_raw_ast_pos_info_context2(Ctxt,CPos),T,T2),
      combine_pos(Pos1,Ctxt,CPos,NewPos),
      NewRaw =.. [Op,NewPos|T2].
add_raw_ast_pos_info_context2(_,_,R,R). % :- print(uncovered_raw(R)),nl,nl.

%combine_pos(none,P,R) :- !, R=P.
combine_pos(P1,Ctxt,P2,pos_context(P1,Ctxt,P2)).

simple_raw(F) :- functor(F,_,1).
simple_raw(identifier(_,_)).
simple_raw(integer(_,_)).
simple_raw(string(_,_)).

list_operator(comprehension_set(P,Args,L),P,[L],comprehension_set(P2,Args,L2),P2,[L2]).
list_operator(couple(P,L),P,L,couple(P2,L2),P2,L2).
list_operator(definition(P,F,L),P,L,definition(P2,F,L2),P2,L2).
list_operator(external_function_call(P,N,L,F,T1,T2),P,L,
              external_function_call(P2,N,L2,F,T1,T2),P2,L2).
list_operator(function(P,F,L),P,[F|L],function(P2,F2,L2),P2,[F2|L2]).
list_operator(general_sum(P,Args,B,E),P,[B,E|Args],
              general_sum(P2,Args2,B2,E2),P2,[B2,E2|Args2]).
list_operator(lambda(P,Args,B,E),P,[B,E],lambda(P2,Args,B2,E2),P2,[B2,E2]).
list_operator(sequence_extension(P,L),P,L,sequence_extension(P2,L2),P2,L2).
list_operator(set_extension(P,L),P,L,set_extension(P2,L2),P2,L2).

raw_operator(add).
raw_operator(card).
raw_operator(cartesian_product).
raw_operator(conjunct).
raw_operator(convert_bool).
%raw_operator(couple).
%raw_operator(definition).
raw_operator(disjunct).
raw_operator(div).
raw_operator(domain).
raw_operator(equal).
raw_operator(equivalence).
raw_operator(first).
raw_operator(first_projection).
raw_operator(front).
raw_operator(greater).
raw_operator(greater_equal).
raw_operator(identifier).
raw_operator(image).
raw_operator(implication).
raw_operator(interval).
raw_operator(intersection).
raw_operator(last).
raw_operator(less).
raw_operator(less_equal).
raw_operator(max).
raw_operator(member).
raw_operator(minus_or_set_subtract).
raw_operator(minus).
raw_operator(min).
raw_operator(multiplication).
raw_operator(mult_or_cart).
raw_operator(negation).
raw_operator(not_equal).
raw_operator(not_member).
raw_operator(overwrite).
raw_operator(power_of).
raw_operator(range).
raw_operator(reverse).
raw_operator(seq).
raw_operator(size).
raw_operator(subset).
raw_operator(tail).
raw_operator(unary_minus).
raw_operator(union).
%raw_operator(Op) :- print(uncovered_op(Op)),nl,fail.
% TO DO: add more operators
