% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)
:- module(bvisual,
[
    create_tree_node/4, get_tree_node_info/4,
    get_tree_from_expr/4,
    get_tree_from_expr2/4,
    write_dot_graph_to_file/2,
    %print_dot_graph/1,
    get_any/5,
    %get_any_most_satisfier/5,

    get_operation_pre/3,
    get_properties/2
]
).

%:- meta_predicate visit_tree(3,2,-,-,-).
:- meta_predicate visit_tree(:,2,-,-,-).

:- use_module(module_information,[module_info/2]).
:- module_info(group,dot).
:- module_info(description,'Generate a dot visualization of B predicates and expressions.').

:- use_module(library(lists), [maplist/3, scanlist/4]).
%:- use_module(library(terms)).

:- use_module(bsyntaxtree).
:- use_module(translate, [translate_bvalue_with_limit/3,
                          translate_bexpression_with_limit/3]).
:- use_module(kernel_waitflags, [
    init_wait_flags/2,
    ground_wait_flags/1]).
:- use_module(b_interpreter, [
    b_test_boolean_expression_wf/3,
    b_not_test_boolean_expression_wf/3,
    b_compute_expression_nowf/4]).

:- use_module(bvisual_any_maxsolver).
:- use_module(error_manager).


%:- use_module(self_check).



% PARAMETERS

max_chars_per_line(40).
max_lines(10).
max_chars_total(400). % :- max_chars_per_line(X), max_lines(ML), T is X*ML.

%%
% Tree format:
% Trees used by write_dot_graph_to_file/2 are modelled by the bind/3
% predicate:
%   bind(Children, NewLocalState, Metas).
% Children is a list of subexpressions, whose numbers may be greater than two.
% NewLocalState is a new local state for the subexpressions.
% Metas are informations about the expression, like value, label etc.
%
% Important: Avoid access bind directly since it's structure may
% be changed. Use accessors defined below.
% They provide a convenient way to access the meta informations.
%
% Use get_tree_from_expr(Tree, Expr, LocalStale, State) to gain
% a bind based tree from a predicate or expression evaluated in LocalState
% and State.
%
% Use get_tree_from_expr2(State, LocalState, TExpr, Tree) to inspect the
% predicate or expression only without walking down the children recursively.
%

%%
% accessors & shortcuts
%

create_long_tree_node(bind(Children, Label, Type, Value, LongDesc),
                           Children, Label, Type, Value, LongDesc).
get_tree_node_info(NODE, Children, NewLocalState, Metas) :-
   (create_tree_node(NODE, C,N,M) -> Children=C, NewLocalState=N, Metas=M
     ; add_error(bvisual,'Illegal node: ',NODE)).
create_tree_node(bind(Children, NewLocalState, Metas),
                      Children, NewLocalState, Metas).


% bind([],cv_o,nonboolean,137,b(identifier(cv_o),integer,[nodeid(pos(139677,56,192,18,192,21))]))

has_no_children(Node) :- child_nodes(Node,[]).

child_nodes(bind(ChildNodes,_,_,_,_), ChildNodes).
child_nodes(bind(ChildNodes,_,_), ChildNodes).

label(bind(_,Label,_,_,_), Label).
label(bind(_,_, Metas), Val) :- member(label(Val), Metas),!.

value(bind(_,_,_,Val,_), Val).
value(bind(_,_, Metas), Val) :- member(value(Val), Metas),!.

long_desc(bind(_,_,_,_,Val), Val).
long_desc(bind(_,_, Metas), Val):- member(long_desc(Val), Metas),!.

is_boolean(bind(_,_,boolean,_,_)).
is_boolean(bind(_,_, Metas)):- member(type(boolean), Metas),!.

%type(bind(_,_, Metas), Val):- member(type(Val), Metas),!.

is_explicit_value_node(Node) :- long_desc(Node,Val), is_explicit_value(Val).

%%
% Evaluate an expression and it's subexpressions and
% gain their values as a tree.
%
:- use_module(tools,[catch_call/1]).
:- use_module(b_global_sets,[add_prob_deferred_set_elements_to_store/3]).
get_tree_from_expr(Tree, Expr, LocalState, State) :-
  (get_preference(dot_use_unicode,true) -> translate:set_unicode_mode ; true),
  add_prob_deferred_set_elements_to_store(State,State1,visible),
	catch_call(get_tree_from_expr3(State1, LocalState, Expr, Tree)),
	(get_preference(dot_use_unicode,true) -> translate:unset_unicode_mode ; true).
	% catch: as this can sometimes generate resource_error(memory) error exceptions


% whole tree at once
get_tree_from_expr3(State, LocalState, Expr, Tree):-
	  get_tree_from_expr2(State, LocalState, Expr, NewTree),
	  get_tree_node_info(NewTree, SubExprs, NewLocalState, _),
   label(NewTree, Label),
   value(NewTree, Value),
   create_long_tree_node(Tree, Children, Label, Type, Value, Expr),
   % print(value(Value)),nl,
    ( is_boolean(NewTree) -> Type = boolean ; Type = nonboolean ),
      maplist(get_tree_from_expr3(State, NewLocalState), SubExprs, Children).




% only one layer
get_tree_from_expr2(State, LocalState, TExpr, Tree):-
    get_texpr_expr(TExpr, Expr),
    functor(Expr, Op, _),
	x_is_quantifier_like(Op),!,
	x_get_metas(State, LocalState, TExpr, Metas),
	member(value(Value), Metas),
    (create_new_local_state(TExpr, Value, State, LocalState, NewLocalState) ->
        x_get_sub_expressions(TExpr, State, LocalState, Children),
        create_tree_node(Tree, Children, NewLocalState, Metas)
        ;
        create_tree_node(Tree, [], LocalState, Metas)
    ),!.
get_tree_from_expr2(State, LocalState, TExpr, Tree):-
	x_get_metas(State, LocalState, TExpr, Metas),
	x_get_sub_expressions(TExpr, State, LocalState, Children),
	create_tree_node(Tree, Children, LocalState, Metas),!.



%%
% Get all meta information for an expression/predicate
%
x_get_metas(State, LocalState, TExpr, Metas):-
	findall(MetaValue, x_get_meta(State, LocalState, TExpr, MetaValue), Metas).


%%
% Helpers, each single meta info.
% There are value/1, type/1, label/1 and long_desc/1
%
% compute the various attributes of a node:
x_get_meta(State, LocalState, Expr, value(Value)):-
	x_get_value_of_expr(State, LocalState, Expr, Value).
x_get_meta(_State, _LocalState, Expr, type(Type)):-
	( check_if_typed_predicate(Expr) ->
		Type = boolean
		;
		Type = nonboolean
	).
x_get_meta(_State, _LocalState, Expr, label(Label)):-
	once(x_get_label(Expr, Label)).
x_get_meta(_State, _LocalState, TExpr, long_desc(Expr)):-
    get_texpr_expr(TExpr, Expr).
x_get_meta(_State, _LocalState, TExpr, rodin_label(Label)):-
    once(get_texpr_label(TExpr,Label)).



%%
% Generate a new local state for subexpressions if necessary, like a forall
% predicate. called for x_is_quantifier/1 predicate.
%

create_new_local_state(TExpr, Value, State, LocalState, NewLocalState):-
    get_texpr_expr(TExpr, Expr),
    (create_new_local_state2(Expr, Value, State, LocalState, NewLocalState)
      -> true
      ; add_error(bvisual,'Uncovered Expr in create_new_local_state: ',Expr:Value),
        NewLocalState=LocalState).

create_new_local_state2(forall(IDs, LHS, _), true, State, LocalState, NewLocalState) :-  !,
	(get_any(IDs, LHS, State, LocalState, NewLocalState)
	  -> true /* there is a solution for LHS */
	  ;  /* Forall true because no solution for LHS */
	      get_any(IDs, b(truth,pred,[]), State, LocalState, NewLocalState)
	).
create_new_local_state2(forall(IDs, LHS, RHS), false,State, LocalState,  NewLocalState) :- !,
    safe_create_texpr(negation(RHS), pred, NotRHS),
    conjunct_predicates([LHS, NotRHS], Condition),
	get_any(IDs, Condition, State, LocalState, NewLocalState).
create_new_local_state2(let_predicate(Ids,AssignmentExprs,Pred), Value, State, LocalState, NewLocalState) :-
    %print(let(Ids,AssignmentExprs,Pred)),nl,
    construct_pred(Ids,AssignmentExprs,Pred,Preds),
    conjunct_predicates(Preds,NewPred),
    EXISTS = exists(Ids,NewPred), % print(exists),nl,translate:print_bexpr(EXISTS),nl,
    create_new_local_state2(EXISTS, Value, State, LocalState, NewLocalState).
create_new_local_state2(exists(IDs, LHS), true, State, LocalState, NewLocalState) :- !,
	get_any(IDs, LHS, State, LocalState, NewLocalState).
create_new_local_state2(exists(IDs, LHS), Status, State, LocalState, NewLocalState) :-
    (Status=false ; Status=undefined),!, % for Status=undefined get_any can fail but get_any_most_satisfier should always succeed
	get_any_most_satisfier(IDs, LHS, State, LocalState, NewLocalState).
create_new_local_state2(X, undefined, State, LocalState, NewLocalState) :- !,
    create_new_local_state2(X, true, State, LocalState, NewLocalState).
create_new_local_state2(X, x_both_true_and_not_true, State, LocalState, NewLocalState) :-
    %print(x_both_true_and_not_true(X)),nl, % probably means a PROB BUG !
    (X = forall(_,_,_) -> VAL=false ; VAL=true),
    create_new_local_state2(X, VAL, State, LocalState, NewLocalState).
create_new_local_state2(comprehension_set(IDs, LHS), Val, State, LocalState, NewLocalState) :- !,
    % format(user_output,'Comprehension set ~w, Val=~w~n~n',[IDs,Val]),
	  (is_empty_set_value(Val) ->
	     get_any_most_satisfier(IDs, LHS, State, LocalState, NewLocalState)
	    ; get_any(IDs, LHS, State, LocalState, NewLocalState)).

% a bit of a hack; it would be better to look at the Prolog version of the value
is_empty_set_value('{}').
is_empty_set_value('[]').
is_empty_set_value(X) :- translate:unicode_translation(empty_set,X).


construct_pred([],[],Pred,[Pred]).
construct_pred([ID|T],[EXPR|ET],Pred,[b(equal(ID,EXPR),pred,[])|TR]) :-
  construct_pred(T,ET,Pred,TR).


%%
% Get subexpressions, merge sub expressions with associative operators, like
% 'and'. For example the predicate and(A,and(B,C)) has three sub expressions:
% A, B and C. This behaviour is controlled by x_associative_operator/1.
%

x_get_sub_expressions(TExpr, Children) :-
  x_get_sub_expressions(TExpr, [], [], Children).

x_get_sub_expressions(TExpr, _, _, Children) :- nonvar(TExpr),
  do_not_expand(TExpr), !,
  Children=[].
x_get_sub_expressions(TExpr, State, LocalState, Children) :-
  syntaxtraversion(TExpr,Expr,_,_,Children0,_),
  % add custom explanation rules for certain nodes:
  x_get_custom_explanation_subexpressions(Expr,State, LocalState, Children0,Children).

:- use_module(preferences,[get_preference/2]).
do_not_expand(b(equal(A,B),pred,_)) :- get_preference(pp_propositional_logic_mode,true),
   A = b(identifier(_),boolean,_), % do not expand such nodes
   is_explicit_value(B).

:- use_module(library(lists)).
% is a value which does not need to be further decomposed
is_explicit_value(b(Val,_,_)) :- is_explicit_value2(Val).
is_explicit_value2(integer(_N)).
is_explicit_value2(string(_S)).
is_explicit_value2(boolean_true).
is_explicit_value2(boolean_false).
is_explicit_value2(identifier(N)) :- b_global_sets:lookup_global_constant(N,_).
is_explicit_value2(couple(A,B)) :- is_explicit_value(A), is_explicit_value(B).
is_explicit_value2(value(_)).
% to do: add further cases


b_compute_expression_no_wd(Fun, LocalState, State, FunValue) :-
	  temporary_set_preference(disprover_mode,true,CHNG), % avoid WD Errors, we detect WD issues by failure below
	  temporary_set_preference(find_abort_values,false,CHNG2),
    (on_exception(E,
       b_compute_expression_nowf(Fun, LocalState, State, FunValue),
       (add_warning(bvisual,'Exception in bvisual while evaluating expression: ',E),fail)
      )
      -> reset_temporary_preference(disprover_mode,CHNG),
         reset_temporary_preference(find_abort_values,CHNG2)
      ;  reset_temporary_preference(disprover_mode,CHNG),
         reset_temporary_preference(find_abort_values,CHNG2), fail).

x_get_custom_explanation_subexpressions(function(Fun,Value),State,LocalState,_C0,Children) :-
    % a custom rule for function application of closures
    % print(function(Fun)),nl, print(value(Value)),nl,print(C0),nl,
    simple_value(Fun),
    b_compute_expression_no_wd(Fun, LocalState, State, FunValue),
    FunValue = closure(P,T,Pred),
    P = [X,_Res], % TO DO: extend to functions with multiple arguments
    closures:is_lambda_closure(P,T,Pred, _OtherIDs,_OtherTypes, DomainPred, Expr),!,
    bsyntaxtree:replace_id_by_expr(Expr,X,Value,NewExpr),
    bsyntaxtree:replace_id_by_expr(DomainPred,X,Value,NewDomainPred),
    Children = [Fun,Value, NewDomainPred,NewExpr].
x_get_custom_explanation_subexpressions(Expr,_State,_LocalState, Children0,Children) :-
    x_get_custom_explanation_subexpressions(Expr, Children0,Children).

simple_value(b(V,_,_)) :- simple_value_aux(V).
simple_value_aux(identifier(_)).
simple_value_aux(value(_)).

% TO DO: add explanation rules, e.g., f:A --> B ==>  f:TA +-> B & dom(f)=A ,...
x_get_custom_explanation_subexpressions(Expr,_Children0, Children):-
    functor(Expr, Op, _),
    x_operator_with_local_variables(Op),!,Children=[].
x_get_custom_explanation_subexpressions(Expr,Children0, Children):-
    functor(Expr, Op, _), Op = set_extension,
    length(Children0,Len), (Len>50 ; maplist(is_explicit_value,Children0)),
    % can be a very large explicit node; no sense in expanding and showing every element
    % print(no_child(Expr)),nl,
    % TO DO: check if simple value
    !,Children=[].
x_get_custom_explanation_subexpressions(Expr,Children0, Children):-
    functor(Expr, Op, _),
    x_associative_operator(Op),!,
    maplist(x_collect_exprs_with_op(Op), Children0, Children1),
    scanlist(x_append_reverse, Children1, [], Children).
x_get_custom_explanation_subexpressions(member(X,b(partial_function(Dom,Range),T,Info)),_,Children) :-
    T=set(set(couple(DomT,RanT))),
    %print(pf(Dom,Range)),nl,
    (\+ bsyntaxtree:is_just_type(Dom) ; \+ bsyntaxtree:is_just_type(Range)),
    nonmember(already_expanded,Info),
    % otherwise we may have an infinite recursion; maybe better to add other operator
    %print(not_type),nl,fail,
    !,
    Children = [A|TB],
    translate:set_type_to_maximal_texpr(DomT,DomType),
    translate:set_type_to_maximal_texpr(RanT,RanType),
    A = b(member(X,b(partial_function(DomType,RanType),T,[already_expanded|Info])),pred,[]),
    add_subset_check(b(domain(X),set(DomT),[]), Dom, TB,TC),
    add_subset_check(b(range(X),set(RanT),[]), Range, TC,[]),
    TB \= [],!.
x_get_custom_explanation_subexpressions(member(X,b(total_function(Dom,Range),T,Info)),_,Children) :-
    T=set(set(couple(DomT,RanT))), !,
    Children = [A,B|TC],
    %print(total_fun_check(DomT,RanT)),nl,
    translate:set_type_to_maximal_texpr(DomT,DomType),
    translate:set_type_to_maximal_texpr(RanT,RanType),
    A = b(member(X,b(partial_function(DomType,RanType),T,Info)),pred,[]),
    B = b(equal(b(domain(X),set(DomT),[]), Dom),pred,[]),
    add_subset_check(b(range(X),set(RanT),[]), Range, TC,[]).
x_get_custom_explanation_subexpressions(member(X,b(total_surjection(Dom,Range),T,Info)),_,Children) :-
    T=set(set(couple(DomT,RanT))), !,
    Children = [A,B,C],
    translate:set_type_to_maximal_texpr(DomT,DomType),
    translate:set_type_to_maximal_texpr(RanT,RanType),
    A = b(member(X,b(partial_function(DomType,RanType),T,Info)),pred,[]),
    B = b(equal(b(domain(X),set(DomT),[]), Dom),pred,[]),
    C = b(equal(b(range(X),set(RanT),[]), Range),pred,[]).
x_get_custom_explanation_subexpressions(member(X,b(partial_surjection(Dom,Range),T,Info)),_,Children) :-
    T=set(set(couple(DomT,RanT))), !,
    Children = [A|TB],
    translate:set_type_to_maximal_texpr(DomT,DomType),
    translate:set_type_to_maximal_texpr(RanT,RanType),
    A = b(member(X,b(partial_function(DomType,RanType),T,Info)),pred,[]),
    add_subset_check(b(domain(X),set(DomT),[]), Dom, TB,[C]),
    C = b(equal(b(range(X),set(RanT),[]), Range),pred,[]).
x_get_custom_explanation_subexpressions(_Expr,Children, Children). % :- filter_children(Children,Res).


%filter_children([],[]).
%filter_children([H|T],Res) :-  (not_interesting(H) -> Res=R ; Res=[H|R]), filter_children(T,R).
%not_interesting(b(integer(_),integer,_)).


add_subset_check(LHS,RHS,Result,EndOfList) :-
% add subset check, unless "obvious"
  (bsyntaxtree:is_just_type(RHS) -> Result=EndOfList
    ; Result = [b(subset(LHS, RHS),pred,[])|EndOfList]
  ).

x_collect_exprs_with_op(Op, TExpr, Children):-
    syntaxtraversion(TExpr,Expr,_,_,_,_),
    ( functor(Expr, Op, _) ->
        x_get_sub_expressions(TExpr, Children);
        Children = [TExpr]
    ).

x_append_reverse(X,Y,L):- append(Y,X,L).



%%
% Get label for an expression
%

x_get_label(TExpr, Label):-  get_texpr_expr(TExpr,Expr),
   x_get_label_aux(Expr,Label).
x_get_label_aux(value(V), Label) :- !, get_value_label(V,Label).
x_get_label_aux(Expr, Label) :-
    %Expr =.. [Op, Label],
    functor(Expr,Op,1), arg(1,Expr,Label),
    x_literal_value_operator(Op).
x_get_label_aux(Expr, Label):-
    functor(Expr, Fkt, _),
    x_functor_to_label(Fkt, Label).

get_value_label(V,Label) :- var(V),!, Label='VAR-VALUE'.
%get_value_label(global_set(GS),Label) :- !, Label=GS. % should we pretty print the value
%get_value_label(V,stored_value(F)) :- functor(V,F,_N).
get_value_label(V,Label) :- translate_bvalue_with_limit(V,40,Label).
%%
% Helpers
%

%%
% Maps functor of prolog term to correct label
%
:- use_module(translate,[translate_prolog_constructor_in_mode/2]).
x_functor_to_label(F,R) :- translate_prolog_constructor_in_mode(F,PPVersion), !, R=PPVersion.
%x_functor_to_label(conjunct, conjunction):-!.
%x_functor_to_label(disjunct, disjunction):-!.
x_functor_to_label(function, R):-!, R='func.applic.'.
x_functor_to_label(X, X).

%%
% List of operators which may generate a new local state,
% like a forall predicate.
%
x_is_quantifier_like(comprehension_set).
x_is_quantifier_like(X) :- x_is_quantifier(X).

x_is_quantifier(forall).
x_is_quantifier(exists).
x_is_quantifier(let_predicate).

%%
% List of associative operators
%

x_associative_operator(conjunct).
x_associative_operator(disjunct).
x_associative_operator(add).
x_associative_operator(mul).


%%
% List of operators with local variables
%

% x_operator_with_local_variables(comprehension_set). is now supported in create_new_local_state2
x_operator_with_local_variables(lambda).
x_operator_with_local_variables(general_sum).
x_operator_with_local_variables(general_product).
x_operator_with_local_variables(quantified_union).
x_operator_with_local_variables(quantified_intersection).
x_operator_with_local_variables(lazy_let_expr).
x_operator_with_local_variables(lazy_let_pred).
x_operator_with_local_variables(lazy_let_subst).


%%
% List of literal values
%
x_literal_value_operator(identifier).
x_literal_value_operator(value).
x_literal_value_operator(integer).
x_literal_value_operator(string).



%%
% Evaluate predicate or expression ... at least it tries XD
% Expressions return value is translated by pretty printer.
% Value is unknown if Expr is neither a predicate or expression.
%

%%
% Evaluate the predicate Expr in LocaleState and State.
% It does check for consistency by checking Expr == !!Expr.
% Thus it yields in addition to true and false also
% x_both_true_and_not_true and undefined
% as flag signaling internal errors.

:- use_module(preferences,[temporary_set_preference/3,reset_temporary_preference/2]).

x_get_value_of_expr(State, LocalState, Expr, Value):-
	check_if_typed_predicate(Expr),!,
	temporary_set_preference(disprover_mode,false,CHNG), % avoid WD Errors, we detect WD issues by failure below
	temporary_set_preference(find_abort_values,false,CHNG2),
	% print('Test: '),translate:print_bexpr(Expr),nl,
    get_bool_expr_value(Expr,LocalState,State,Value),
    reset_temporary_preference(disprover_mode,CHNG),
    reset_temporary_preference(find_abort_values,CHNG2),
    clear_wd_errors. % ,print('Value: '), print(Value),nl.
x_get_value_of_expr(State, LocalState, Expr, Value):-
    check_if_typed_expression(Expr),!,
	% print('Eval: '),translate:print_bexpr(Expr),nl,
    (b_compute_expression_no_wd(Expr, LocalState, State, Value0)
      -> max_chars_total(MaxT),
         translate_bvalue_with_limit(Value0, MaxT, Value)  % we show maximally 10 lines a 40 chars
      ;  Value = unknown).
x_get_value_of_expr(_State, _LocalState, _Expr, unknown):-!.

% (error(representation_error(max_clpfd_integer)

get_bool_expr_value(Expr,LocalState,State,Value) :-
    on_exception(E,call(get_bool_expr_val2(Expr,LocalState,State,Value)),
           (add_warning(bvisual,'Exception in bvisual while evaluating predicate: ',E),Value=undefined)).
get_bool_expr_val2(Expr,LocalState,State,Value) :-
   ( b_test_boolean_expression_wf(Expr,LocalState,State) ->
        ( b_not_test_boolean_expression_wf(Expr,LocalState,State)
          -> Value = x_both_true_and_not_true, % was x_unknown_true_and_not_true
             % we should add an error message here
             % this should never happen
             print('### Error, x_both_true_and_not_true: '), print(Expr),nl
		       ; Value = true
		    )
		 ; b_not_test_boolean_expression_wf(Expr,LocalState,State) ->
            Value = false
		    %, print(value_false),nl
		 ; otherwise ->
		    Value = undefined % was x_undefined_false_and_not_false
		    %, print(value_undefined),nl
    ).



%%
% Try to execute an 'Any' statement on LocalState and State.
% New state with the values is returned by NewLocalState.
% Compare to b_interpreter:b_execute_statement2(any(Parameters,PreCond,Body),...)
%

:- use_module(b_enumerate, [b_tighter_enumerate_values/2]).
:- use_module(b_interpreter, [set_up_typed_localstate/6]).


get_any(IDs, Condition, State, LocalState, NewLocalState) :-
    on_exception(E,call(get_any2(IDs, Condition, State, LocalState, NewLocalState)),
           (add_warning(bvisual,'Exception in bvisual while treating quantfier predicate: ',E),fail)).
get_any2(IDs, Condition, State, LocalState, NewLocalState) :-
    set_up_typed_localstate(IDs, _FreshVars, TypedVals, LocalState, NewLocalState,positive),
    init_wait_flags(WF,[get_any2]),
    b_interpreter:b_test_boolean_expression(Condition,NewLocalState, State, WF), /* check WHERE */
    b_tighter_enumerate_values(TypedVals,WF),
    ground_wait_flags(WF).


get_any_most_satisfier(IDs, Condition, State, LocalState, NewLocalState):-
    ( any_maxsolver(IDs, Condition, State, LocalState, NewLocalState) ->
        true;
        conjunct_predicates([], Truth),
        get_any(IDs, Truth, State, LocalState, NewLocalState)
    ).






%%
% Main predicate for writing dot graph
% Use it to start.
%
% FileSpec is used by tell/1 to open a stream.
% The graph is printed by print_dot_graph_for_tree/1.
%

:- use_module(tools,[catch_call/1]).

write_dot_graph_to_file(Tree, FileSpec) :-
    catch_call(bvisual:write_dot_graph_to_file_aux(Tree, FileSpec)).
write_dot_graph_to_file_aux(Tree, FileSpec) :-
   (get_preference(dot_use_unicode,true) -> translate:set_unicode_mode ; true),
    tell(FileSpec),
    call_cleanup(print_dot_graph(Tree),
      (told,(get_preference(dot_use_unicode,true) -> translate:unset_unicode_mode ; true))).

print_dot_graph(Tree):-
    print('digraph g {'), nl,
    print('rankdir=RL;'), nl,
    visit_tree(x_print_dot_graph_visitor, child_nodes, Tree, [],_),
    print('}'),
    nl.


%%
% Helpers
%
:- public x_print_dot_graph_visitor/5.
x_print_dot_graph_visitor(enter, Node, _Children, [], [1,root]) :-
    x_print_dot_graph_node_only(Node, root),!.
x_print_dot_graph_visitor(enter, Node, _Children, In, [NewID|In]):-
    x_print_dot_graph_visitor(Node, In, [NewID|_]).
x_print_dot_graph_visitor(leave, _Node, _Children, [NewID,_ID|Out], [NewID|Out]).

/* for 3 args : */
x_print_dot_graph_visitor(Node, [], []):-
    x_print_dot_graph_node_only(Node, root),
    !.

x_print_dot_graph_visitor(Node, [ID|Rest], [NewID|Rest]):-
    Rest = [PID|_],
    NewID is ID + 1,
    x_print_dot_graph_node_only(Node, ID),
    format('~8|Node~p -> Node~p;~N', [ID, PID]),
    !.




shorten_atom(MaxLengthPerLine, Atom, NewAtom) :-
    max_lines(MaxLines),
    shorten_atom(MaxLengthPerLine, Atom, NewAtom,MaxLines).
shorten_atom(MaxLengthPerLine, Atom, NewAtom,MaxLines) :-
    ( atom(Atom) ->
        atom_chars(Atom, AtomChars),
        shorten_atom_list(MaxLengthPerLine, AtomChars, NewAtom,MaxLines)
        ;
        NewAtom = Atom
    ).

shorten_atom_list(MaxLengthPerLine, Chars, Atom,MaxLines):-
    %length(Chars, Length),
    split_chars(Chars,MaxLengthPerLine, Prefix,Suffix),
    ( Suffix=[] ->
        atom_chars(Atom, Chars)
        ;
        %lists:append_length(Prefix, Suffix, Chars, MaxLengthPerLine),
        %print( split_chars(Chars,MaxLengthPerLine, Prefix,Suffix) ),nl,
        (MaxLines>1 -> ML1 is MaxLines-1,
           shorten_atom_list(MaxLengthPerLine, Suffix, SuffixAtom,ML1)
          ; SuffixAtom = '...'),
        atom_chars(PrefixAtom, Prefix),
        atom_concat(PrefixAtom, '\\n', PrefixAtomNewLine),
        atom_concat(PrefixAtomNewLine, SuffixAtom, Atom)
    ).

% bvisual:split_chars([a,b,d,e,f],2,P,S), P==[a,b], S==[d,e,f]

% a predicate to split List if it is longer than MaxLen, but only outside of single and double quotes
% List is suppoed to be already encoded using string_escape
split_chars(List,MaxLen,Prefix,Suffix) :-
  split_chars(List,MaxLen,Prefix,Suffix,q(outside,outside)).

split_chars(List,MaxL,R,Suffix,Q) :- MaxL < 1, Q=q(outside,outside), !,
   % only add newline if we are outside of quotes
   R=[], Suffix=List.
split_chars([],_MaxL,[],[],_).
split_chars(L,MaxL,LR,LeftOver,Q) :-
   is_a_unit(L,T,LR,TR),!,
   M1 is MaxL-1, split_chars(T,M1,TR,LeftOver,Q).
split_chars([Slash,Quote|T],MaxL,[Slash,Quote|TR],LeftOver,Q) :- Slash = ('\\'),
   !, %(Quote=34;Quote=39), !, % \" or \'
   toggle_quotes(Quote,Q,TQ),
   M1 is MaxL-1, split_chars(T,M1,TR,LeftOver,TQ).
split_chars([H|T],MaxL,[H|TR],LeftOver,Q) :-
   M1 is MaxL-1, split_chars(T,M1,TR,LeftOver,Q).

toggle_quotes('\'',q(outside,Q2),q(inside,Q2)) :- !.
toggle_quotes('\'',q(inside,Q2),q(outside,Q2)) :- !.
toggle_quotes('\"',q(Q1,outside),q(Q1,inside)) :- !.
toggle_quotes('\"',q(Q1,inside),q(Q1,outside)) :- !.
toggle_quotes(_,Q,Q).

% check for HTML encodings like &agrave; &Euml; &alpha; ...
% these should be counted as a single character and not be split
is_a_unit(['&'|T1],T,['&'|T1R],TR) :- get_lexeme(T1,10,LexUnit,T),
    append(LexUnit,TR,T1R).

get_lexeme([';'|T],_,[';'],TRes) :- !, TRes=T.
get_lexeme([H|T],L,[H|LexT],TRes) :- L>1, L1 is L-1, get_lexeme(T,L1,LexT,TRes).

get_long_label(Node,MaxPerLine,Pretty1h) :-
            long_desc(Node, LongDesc),
            %tools:print_wtime(start_translate(ID)),
            translate_bexpression_with_limit(LongDesc, 250, Pretty0),
            %tools:print_wtime(translated(ID)),
            string_escape(Pretty0, Pretty1),
            shorten_atom(MaxPerLine, Pretty1, Pretty1h). % maximally 10 lines with 40 chars

:- use_module(tools,[string_escape/2]).
x_print_dot_graph_node_only(Node, ID) :- %tools:print_wtime(node(ID,Node)),
    label(Node, Label0),
    max_chars_per_line(MaxPerLine),
    string_escape(Label0, Label1), %tools:print_wtime(escaped(ID)),
        ( ID \= root, is_boolean(Node)
         ->
            Label=Label1,
            get_long_label(Node,MaxPerLine,Pretty1h),
            value(Node, TruthValue),
            string_escape(TruthValue, TVEsc),
            get_predicate_field_seperator(Sep),
            atom_concat(TVEsc, Sep, Pretty4),
            atom_concat(Pretty4, Pretty1h, Pretty)
          ;
            value(Node, Val),
            string_escape(Val, Pretty1), % TO DO: check that we also escape | in case of records!
            (is_boolean(Node) -> get_predicate_field_seperator(Sep)
              ; get_expression_field_seperator(Sep)),
            ((Pretty1=Label1 ; is_explicit_value_node(Node))
             -> Pretty='', Label=Pretty1 % we have a simple value node like integer(1),...
              ; has_no_children(Node)
             -> get_long_label(Node,MaxPerLine,Label),
                shorten_atom(MaxPerLine, Pretty1, Pretty)
              ; Label=Label1,
                shorten_atom(MaxPerLine, Pretty1, Pretty)
            )
        ),
    %tools:print_wtime(node_style(ID)),
    x_print_dot_graph_node_style(Node, Shape, Color,Style),
    %tools:print_wtime(printing(ID)),
    (Pretty=''
     -> format('~4|Node~p [label="~p", shape="~p", fillcolor="~p", style="~p"]~N',  [ID, Label, Shape, Color,Style])
     ;  format('~4|Node~p [label="~p~p~p", shape="~p", fillcolor="~p", style="~p"]~N',  [ID, Label, Sep, Pretty, Shape, Color,Style])). % ,tools:print_wtime(finished(ID)).


x_print_dot_graph_node_style(Node, Shape, 'tomato','filled,rounded') :-
    is_boolean(Node), value(Node, false),!, boolean_shape(Shape).
x_print_dot_graph_node_style(Node, Shape, 'olivedrab2','filled,rounded') :-
    is_boolean(Node), value(Node, true),!, boolean_shape(Shape).
x_print_dot_graph_node_style(Node, Shape, 'orange','filled,rounded') :-  % styles could be dashed,dotted, rounded, ...
    is_boolean(Node),!, boolean_shape(Shape).
x_print_dot_graph_node_style(Node, Shape, 'yellow','filled') :-
    value(Node, unknown),!, get_preference(dot_expression_node_shape,Shape).
x_print_dot_graph_node_style(_, Shape, 'white',filled) :-
   get_preference(dot_expression_node_shape,Shape).


boolean_shape(Shape) :- get_preference(dot_predicate_node_shape,Shape). % ellipse, oval, record TO DO: other form for quantifiers

% if we use a Dot record, we can use | to separate fields
get_predicate_field_seperator('|') :- get_preference(dot_predicate_node_shape,record).
get_predicate_field_seperator('\\n').
get_expression_field_seperator('|') :- get_preference(dot_expression_node_shape,record).
get_expression_field_seperator('\\n').

:- use_module(bmachine, [
    b_get_properties_from_machine/1,
    b_get_machine_constants/1,
    b_machine_has_constants_or_properties/0
]).

:- use_module(predicate_debugger).

x_create_exists(IDs, Condition, Exists):-
    %print(x_create_exists(IDs,Condition)),nl,
    get_texpr_expr(Condition,exists(InnerVars,InnerCond)),!,
    % fuse two exists together
    append(IDs,InnerVars,Vars),
    x_create_exists(Vars,InnerCond,Exists).
x_create_exists(IDs, Condition, Exists):-
    create_exists(IDs,Condition,Exists).



get_properties(BExpr,CreateExists):-
    b_get_properties_from_machine(Condition),
    ( (CreateExists=true,b_machine_has_constants_or_properties) ->
        b_get_machine_constants(IDs),
        x_create_exists(IDs, Condition, BExpr)
        ;
	    BExpr = Condition
	).


get_operation_pre(OpName, BExpr, BecomeSuchVars):-
    predicate_debugger:get_operation_enabling_condition(OpName, IDs, Condition,BecomeSuchVars,_Precise,true,false),
    ( IDs = [] ->
        BExpr = Condition;
        x_create_exists(IDs, Condition, BExpr)
	).




%%
% Hierarchical Visitor Pattern for "trees"
% used to be in gtree module
%
% For this implementation a tree is a (root) node and a functor, which
% maps a node to it's child nodes. For example a Prolog term can be
% understand as a tree by the functor:
%   children_fkt(Term, Subterms):- Term=..[_|Subterms].
%
% Important: Do not use trees with cycles. This implementation will not
% detect them and will not terminate.
%
% The indented calling is:
%   visit_tree(+VisitorFkt, +ChildrenFkt, +Node, +EnvIn, -EnvOut)
%
% If no functor is specified, the former mentioned one is used.
% EnvIn and EnvOut are Variables for intermediate values or states,
% which you might need to maintain between calls to VisitorFkt. You
% might omit them, if you do not need them.
%
% During the visiting process the VisitorFkt functor is called for each
% node starting with the "root". Calls are made in following order:
%   - VisitorFkt(Node, EnvIn, EnvOut), if Node is a leaf,
%     i.e. ChildrenFkt(Node, []).
%   - VisitorFkt(enter, Node, Children, EnvIn, EnvOut), before branching,
%   - VisitorFkt(leave, Node, Children, EnvIn, EnvOut), after branching
%     to Node's child nodes.
%  !- Only if VisitorFkt(enter, ...) succeeds, child nodes will be
%     visited and VisitorFkt(leave, ...) will be called.
%

:- use_module(library(lists), [scanlist/4, maplist/2]).


visit_tree(VisitorFkt, ChildrenFkt, Node, EnvIn, EnvOut):-
    call(ChildrenFkt, Node, Children),
    ( Children == [] ->
        call(VisitorFkt, Node, EnvIn, EnvOut)
        ;
        ( call(VisitorFkt, enter, Node, Children, EnvIn, Env1) ->
            scanlist(visit_tree(VisitorFkt, ChildrenFkt), Children, Env1, Env2),
            call(VisitorFkt, leave, Node, Children, Env2, EnvOut)
            ;
            EnvOut = EnvIn
        )
    ),
    !.



%%
% Default ChildrenFkt
%
%:- public x_default_children_fkt/2.
%x_default_children_fkt(Node, Children):-
%    Node =.. [_|Children].




