% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(compile_time_flags,
    [compile_time_flag/1,
     compile_time_flags/1,
     relevant_prolog_flags/1
    ]).

% compile time flags can be set from the command line for SICStus Prolog
% e.g., -Dprob_safe_mode=true when starting ProB from source or when generating the .sav file in the Makefile

:- use_module(module_information,[module_info/2]).

:- module_info(group,infrastructure).
:- module_info(description,'This module describes the possible compile-time flags of ProB.').

compile_time_flags(list(Flags)) :- % list wrapper for Tcl
   findall(Flag, compile_time_flag(Flag), Flags).


:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_release,true)).
compile_time_flag(prob_release).
% set when building an official build of probcli or ProB Tcl/Tk
% excludes Tcl benchmarks, unit tests
:- endif.
:- if(environ(prob_safe_mode,true)).
compile_time_flag(prob_safe_mode).
% perform additional checks
% check_ast: check AST ground, check used_id infos for quantifiers...)
% check typing in equal_object, variable clashes in comprehension_set, call_residue check in my_findall,...
:- endif.
:- if(environ(plspec_patch_libraries,true)).
compile_time_flag(plspec_patch_libraries).
% use plspec to check calls to ordsets, avl and lists libraries
:- endif.
:- if(environ(prob_data_validation_mode,true)).
compile_time_flag(prob_data_validation_mode).
% prevent certain instantiations of sets on the assumption that we do not need powerful constraint solving
:- endif.
:- if(environ(no_wd_checking,true)).
compile_time_flag(no_wd_checking).
% disable Well-Definedness checking for function applications
:- endif.
:- if(environ(prob_profile,true)).
compile_time_flag(prob_profile).
% store profile information for operations and invariant checking
:- endif.
:- if(environ(prob_logging_mode,true)).
compile_time_flag(prob_logging_mode).
% automatically log probcli (add -ll command-line switch)
:- endif.
:- if(environ(no_terminal_colors,true)).
compile_time_flag(no_terminal_colors).
% disable terminal colors, can also be influence by setting NOCOLOR environment variable
:- endif.
:- if(environ(partially_evaluate,true)).
compile_time_flag(partially_evaluate).
:- endif.
:- if(environ(partially_evaluate_compile,true)).
compile_time_flag(partially_evaluate_compile).
:- endif.
:- if(environ(use_pe_cache,true)).
compile_time_flag(use_pe_cache).
:- endif.
:- if(environ(no_interrupts,true)).
compile_time_flag(no_interrupts).
% do not treat CTRL-C user_interrupts
:- endif.
:- if(environ(prob_noopt_mode,true)).
compile_time_flag(prob_noopt_mode).
% now While Loop optimisation
:- endif.
:- if(environ(prob_debug_flag,true)).
compile_time_flag(prob_debug_flag).
% additional debugging performed/enabled
:- endif.
:- if(environ(prob_src_profile,true)).
compile_time_flag(prob_src_profile).
% perform profiling at B source level in source_profiler.pl
:- endif.
:- if(environ(prob_myheap,false)).
compile_time_flag(prob_myheap_false).
% do not use C++ priority queue
:- endif.
:- if(environ(disable_chr,true)).
compile_time_flag(disable_chr). % do not include CHR library at all
:- endif.
:- if(environ(enable_time_out_for_constraints,true)).
compile_time_flag(enable_time_out_for_constraints). % use time_out for posting constraints
:- endif.
:- if(environ('SP_JIT',disabled)).
compile_time_flag('SP_JIT=disabled').
:- endif.
:- if(environ('SP_JIT',yes)).
compile_time_flag('SP_JIT=yes'). % this is the default
:- endif.
:- if(environ('SP_JIT',no)).
compile_time_flag('SP_JIT=no').
:- endif.
:- if(environ('SP_JIT_COUNTER_LIMIT',_)).
compile_time_flag('SP_JIT_COUNTER_LIMIT_SET'). % default is 0
:- endif.
:- if(environ('SP_JIT_CLAUSE_LIMIT',_)).
compile_time_flag('SP_JIT_CLAUSE_LIMIT_SET'). % default is 1024
:- endif.
:- if(environ('SP_SPTI_PATH',verbose)).
compile_time_flag('SP_SPTI_PATH=verbose'). % can be set to verbose
:- endif.
:- if(environ('SP_TIMEOUT_IMPLEMENTATON',legacy)).
compile_time_flag('SP_TIMEOUT_IMPLEMENTATON=legacy'). % can be set to verbose
:- endif.

compile_time_flag(_) :- fail.


relevant_prolog_flags(Flags) :-
   findall(Flag/Val, (relevant_flag(Flag),current_prolog_flag(Flag,Val)), Flags).

relevant_flag(dialect).
relevant_flag(version_data).
relevant_flag(profiling).
