:- module(dpllt_preprocessing, [preprocess_predicate/3]).

:- use_module(library(lists)).
:- use_module(probsrc(b_interpreter), [b_test_boolean_expression_wf/3]).
:- use_module(probsrc(bsyntaxtree), [find_identifier_uses/3,
                                     remove_all_infos/2,
                                     add_texpr_infos/3,
                                     conjunct_predicates/2,
                                     unique_typed_id/3]).

:- use_module(probsrc(module_information), [module_info/2]).
:- module_info(group, cbc).
:- module_info(description,'This module provides several preprocessing steps for the DPLL(T) based solver.').

%% process_candidate_impls(+Pred, -LiftedPred, -FilteredCandidateImplsConj).
preprocess_predicate(Pred, LiftedPred, FilteredCandidateImplsConj) :-
    empty_candidate_impls_acc(CandidateAcc),
    lift_negations_find_impls(Pred, CandidateAcc, LiftedPred, CandidateImpls),
    % TO DO: process_candidate_impls/2 slow?
    process_candidate_impls(CandidateImpls, FilteredCandidateImplsConj).

%% process_candidate_impls(+CandidateImpls, -FilteredCandidateImplsConj).
% Given a candidate like x>y, get partner candidates like y<x, y=<x and x=y.
% Filter those partner candidates that exists in the given formula (prevent introducing redundant SAT variables)
% and create the implication, e.g., x>y => (not(y<x) & not(y=<x) & not(x=y)).
process_candidate_impls(CandidateImpls, FilteredCandidateImplsConj) :-
    CandidateImpls = candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)),
    process_var_candidate_impls_list(Integer, IntImplsConj),
    % TO DO: also enable this knowledge for the SMT solver? (add smt_formula/1 term to info fields of implications)
    process_ground_candidate_impls_list(IntegerGrnd, IntGrndImplsConj),
    process_var_candidate_impls_list(Set, SetImplsConj),
    process_ground_candidate_impls_list(SetGrnd, SetGrndImplsConj),
    conjunct_predicates([IntImplsConj,IntGrndImplsConj,SetImplsConj,SetGrndImplsConj], FilteredCandidateImplsConj).

get_id_from_binary_ast(b(Node,_,_), Id, IdVarInfo) :-
    arg(1, Node, Arg1),
    arg(2, Node, Arg2),
    (   Arg1 = b(identifier(_),_,_)
    ->  remove_all_infos(Arg1, IdVarInfo),
        Id = Arg1
    ;   remove_all_infos(Arg2, IdVarInfo),
        Id = Arg2
    ).

process_ground_candidate_impls_list(GrndCandidateImpls, ImplsConj) :-
    process_ground_candidate_impls_list(GrndCandidateImpls, [], Impls),
    conjunct_predicates(Impls, ImplsConj).

process_ground_candidate_impls_list([], Acc, Acc).
process_ground_candidate_impls_list([GrndCandidateImpl|T], Acc, ImplsConj) :-
    get_id_from_binary_ast(GrndCandidateImpl, Id, IdVarInfo),
    get_ground_partner_candidates(IdVarInfo, T, Partners, RestT),
    Partners \== [],
    !,
    build_implications_from_ground_candidates(Id, GrndCandidateImpl, Partners, PartialImplsConj),
    process_ground_candidate_impls_list(RestT, [PartialImplsConj|Acc], ImplsConj).
process_ground_candidate_impls_list([_|T], Acc, ImplsConj) :-
    process_ground_candidate_impls_list(T, Acc, ImplsConj).

generate_bindings_from_ast(Ast, BindList) :-
    find_identifier_uses(Ast, [], UsedIdNames),
    generate_bindings_from_id_names(UsedIdNames, BindList).

generate_bindings_from_id_names(UsedIdNames, BindList) :-
    generate_bindings_from_id_names(UsedIdNames, [], BindList).

generate_bindings_from_id_names([], Acc, Acc).
generate_bindings_from_id_names([UsedIdName|T], Acc, BindList) :-
    generate_bindings_from_id_names(T, [bind(UsedIdName,_)|Acc], BindList).

%% build_implications_from_ground_candidates(+Id, +GrndCandidateImpl, +Partners, -PartialImplsConj).
% Given GrndCandidateImpl and some matching partner candidates, evaluate (search for a counter example)
% and collect implications:
% p in Partners, add one of 'GrndCandidateImpl => p', 'p => GrndCandidateImpl' or nothing
build_implications_from_ground_candidates(Id, GrndCandidateImpl, Partners, PartialImplsConj) :-
    build_implications_from_ground_candidates(Id, GrndCandidateImpl, Partners, [], PartialImplsConj).

build_implications_from_ground_candidates(_, _, [], Acc, PartialImplsConj) :-
    conjunct_predicates(Acc, PartialImplsConj).
build_implications_from_ground_candidates(Id, GrndCandidateImpl, [Partner|T], Acc, PartialImplsConj) :-
    % TO DO: we could consider typing constraints bounding the domain of identifiers to prevent false negatives
    Impl1 = b(implication(GrndCandidateImpl,Partner),pred,[]),
    Id = b(identifier(IdName),_,_),
    Impl1Test = b(exists([Id],b(conjunct(GrndCandidateImpl,b(negation(Partner),pred,[])),pred,[])),pred,[used_ids([IdName])]),
    generate_bindings_from_ast(Impl1, BindList),
    (   \+ b_test_boolean_expression_wf(Impl1Test, BindList, _)
    ->  NewAcc = [Impl1|Acc]
    ;   (Impl2 = b(implication(Partner,GrndCandidateImpl),pred,[]),
         Impl2Test = b(exists([Id],b(conjunct(Partner,b(negation(GrndCandidateImpl),pred,[])),pred,[])),pred,[used_ids([IdName])]),
         (  \+ b_test_boolean_expression_wf(Impl2Test, BindList, _)
         -> NewAcc = [Impl2|Acc]
         ;  NewAcc = Acc
         ))
    ),
    build_implications_from_ground_candidates(Id, GrndCandidateImpl, T, NewAcc, PartialImplsConj).

%% binary_ast_contains_id(+Partner, +IdVarInfo).
binary_ast_contains_id(Partner, IdVarInfo) :-
    IdVarInfo = b(identifier(_),_,_),
    Partner = b(PNode,_,_),
    functor(PNode, _, 2),
    arg(1, PNode, Arg1),
    (   Arg1 = IdVarInfo,
        !
    ;   (arg(2, PNode, Arg2),
         Arg2 = IdVarInfo)
    ).

%% get_ground_partner_candidates(+Id, +Rest, -Partners, -RestRest).
% ASTs in Rest have the same type.
% idea: given, e.g., 1<x: search for all comparisons between x and a ground value like 0<x, x<10 etc.
% afterwards, evaluate the constraints and try to build implications like 1<x => 0<x
get_ground_partner_candidates(Id, Rest, Partners, RestRest) :-
    get_ground_partner_candidates(Id, Rest, [], Partners, [], RestRest),!.

get_ground_partner_candidates(_, [], PartnersAcc, PartnersAcc, RestAcc, RestAcc).
get_ground_partner_candidates(Id, Rest, PartnersAcc, Partners, RestAcc, RestRest) :-
    select(Partner, Rest, TRest),
    (   binary_ast_contains_id(Partner, Id)
    ->  NPartnersAcc = [Partner|PartnersAcc],
        NRestAcc = RestAcc
    ;   NPartnersAcc = PartnersAcc,
        NRestAcc = [Partner|RestAcc]
    ),
    !,
    get_ground_partner_candidates(Id, TRest, NPartnersAcc, Partners, NRestAcc, RestRest).

process_var_candidate_impls_list(CandidateImpls, ImplsConj) :-
    process_var_candidate_impls_list(CandidateImpls, [], Impls),
    conjunct_predicates(Impls, ImplsConj).

process_var_candidate_impls_list([], Acc, Acc).
process_var_candidate_impls_list([CandidateImpl|T], Acc, Impls) :-
    get_partner_candidates(CandidateImpl, PartnerCandidates),
    filter_and_negate_partner_candidates_for_existence(PartnerCandidates, T, [], Partners),
    Partners \== [],
    !,
    conjunct_predicates(Partners, PartnersConj),
    Impl = b(implication(CandidateImpl,PartnersConj),pred,[]),
    process_var_candidate_impls_list(T, [Impl|Acc], Impls).
process_var_candidate_impls_list([_|T], Acc, Impls) :-
    process_var_candidate_impls_list(T, Acc, Impls).

filter_and_negate_partner_candidates_for_existence([], _, Acc, Acc).
filter_and_negate_partner_candidates_for_existence([PartnerCandidate|T], Rest, Acc, Partners) :-
    select(PartnerCandidate, Rest, RestRest),
    !,
    filter_and_negate_partner_candidates_for_existence(T,  RestRest, [PartnerCandidate|Acc], Partners).
filter_and_negate_partner_candidates_for_existence([_|T], Rest, Acc, Partners) :-
    filter_and_negate_partner_candidates_for_existence(T, Rest, Acc, Partners).

% more knowledge for the SAT solver, e.g., x<y => (not(y<x) & not(x=y) & not(y=<x)) (note: negations are lifted)
% Note: the constraints in the second argument's list are negated afterwards.
% Note: we could consider transitivity but this is most likely too much overhead.
get_partner_candidates(b(less(Arg1,Arg2),pred,_), [b(less(Arg2,Arg1),pred,_),b(less_equal(Arg2,Arg1),pred,_),b(equal(Arg1,Arg2),pred,_)]).
get_partner_candidates(b(less_equal(Arg1,Arg2),pred,_), [b(less(Arg2,Arg1),pred,_)]).
get_partner_candidates(b(subset(Arg1,Arg2),pred,_), [b(subset(Arg2,Arg1),pred,_),b(subset_strict(Arg1,Arg2),pred,_)]). % TO DO: consider equality
get_partner_candidates(b(subset_strict(Arg1,Arg2),pred,_), [b(subset(Arg2,Arg1),pred,_)]). % TO DO: maybe add card(Arg2) < card(Arg1) etc.

% Assuming that the AST has been normalized.
%candidate_impl_binary_operator(greater, integer).
%candidate_impl_binary_operator(greater_equal, integer).
candidate_impl_binary_operator(equal, integer).
candidate_impl_binary_operator(equal, set).
candidate_impl_binary_operator(less, integer).
candidate_impl_binary_operator(less_equal, integer).
candidate_impl_binary_operator(subset, set).
candidate_impl_binary_operator(subset_strict, set).

ground_b_ast(integer, b(integer(Int),_,_)) :-
    ground(Int).
ground_b_ast(set, b(set_extension(Set),_,_)) :-
    ground(Set).

%% extend_candidate_impls_acc(+Type, +Functor, +Arg1, +Arg2, +Ast, +CandidateAcc, -NewCandidateAcc) :-
% Collect ASTs with functor in candidate_impl_binary_operator/2 and both arguments being
% an identifier or ground value.
extend_candidate_impls_acc(Type, Functor, Arg1, Arg2, Ast, CandidateAcc, NewCandidateAcc) :-
    Type == pred,
    candidate_impl_binary_operator(Functor, FType),
    (   (ground_b_ast(FType, Arg1),
         Arg2 = b(identifier(_),_,_))
    ;
        (ground_b_ast(FType, Arg2),
         Arg1 = b(identifier(_),_,_))),
    remove_all_infos(Ast, AstClean),
    ast_not_in_candidate_acc(ground, FType, AstClean, CandidateAcc),
    !,
    extend_candidate_impls_acc_safe_ground(FType, CandidateAcc, Ast, NewCandidateAcc).
extend_candidate_impls_acc(Type, Functor, Arg1, Arg2, Ast, CandidateAcc, NewCandidateAcc) :-
    Type == pred,
    candidate_impl_binary_operator(Functor, FType),
    Arg1 = b(identifier(_),_,_),
    Arg2 = b(identifier(_),_,_),
    remove_all_infos(Ast, AstClean),
    ast_not_in_candidate_acc(var, FType, AstClean, CandidateAcc),
    !,
    extend_candidate_impls_acc_safe(FType, CandidateAcc, Ast, NewCandidateAcc).
extend_candidate_impls_acc(_, _, _, _, _, CandidateAcc, CandidateAcc).

%% ast_not_in_candidate_acc(+VarOrGround, +Type, +AstClean, +CandidateAcc).
ast_not_in_candidate_acc(var, integer, AstClean, candidate_impls(integer(Integer),integer_ground(_),set(_),set_ground(_))) :-
    \+ member(AstClean, Integer).
ast_not_in_candidate_acc(ground, integer, AstClean, candidate_impls(integer(_),integer_ground(IntegerGrnd),set(_),set_ground(_))) :-
    \+ member(AstClean, IntegerGrnd).
ast_not_in_candidate_acc(var, set, AstClean, candidate_impls(integer(_),integer_ground(_),set(Set),set_ground(_))) :-
    \+ member(AstClean, Set).
ast_not_in_candidate_acc(ground, set, AstClean, candidate_impls(integer(_),integer_ground(_),set(_),set_ground(SetGrnd))) :-
    \+ member(AstClean, SetGrnd).

% TO DO: use get_sorted_equality/3 for equalities
extend_candidate_impls_acc_safe_ground(integer, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)), Ast, candidate_impls(integer(Integer),integer_ground([Ast|IntegerGrnd]),set(Set),set_ground(SetGrnd))).
extend_candidate_impls_acc_safe_ground(set, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)), Ast, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground([Ast|SetGrnd]))).

extend_candidate_impls_acc_safe(integer, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)), Ast, candidate_impls(integer([Ast|Integer]),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd))).
extend_candidate_impls_acc_safe(set, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)), Ast, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set([Ast|Set]),set_ground(SetGrnd))).

empty_candidate_impls_acc(candidate_impls(integer([]),integer_ground([]),set([]),set_ground([]))).

%% lift_negations_find_impls(+BTerm, +CandidateAcc, -LPred, -CandidateImpls).
% Unfold negated operators like not_equal to detect equal ASTs when introducing
% boolean variables for the SAT formula. For instance, it improves performance
% to transform 'x={} & x/={}' to 'x={} & not(x={})' first, since we can then translate to
% 'A & not(A)' in SAT rather than 'A & B'.
% Collect specific operators that might imply each other like 'x>0' and 'x>1' result to 'x>1 => x>0' (see candidate_impl_binary_operator/2).
% Note: only binary operators are checked recursively since we transform SMT to SAT on the level of conjunct, disjunct and implication
lift_negations_find_impls(Pred, CandidateAcc, LPred, CandidateImpls) :-
    Pred = b(Node,Type,Info),
    functor(Node, Functor, 2),
    negated_b_operator(Functor, NFunctor), % assumes that no negated_b_operator is in candidate_impl_binary_operator
    !,
    arg(1, Node, Arg1),
    arg(2, Node, Arg2),
    TrueNode =.. [NFunctor,Arg1,Arg2],
    TrueAst = b(TrueNode,Type,Info),
    extend_candidate_impls_acc(Type, NFunctor, Arg1, Arg2, TrueAst, CandidateAcc, CandidateAcc1),
    lift_negations_top_level_functor(NFunctor, TrueNode, CandidateAcc1, NTrueNode, CandidateImpls),
    TTrueAst = b(NTrueNode,Type,Info),
    add_texpr_infos(TTrueAst, [was(Pred)], NTrueAst),
    LPred = b(negation(NTrueAst),pred,[]).
lift_negations_find_impls(b(Node,Type,Info), CandidateAcc, LPred, CandidateImpls) :-
    functor(Node, Functor, 2),
    !,
    arg(1, Node, Arg1),
    arg(2, Node, Arg2),
    extend_candidate_impls_acc(Type, Functor, Arg1, Arg2, b(Node,Type,Info), CandidateAcc, CandidateAcc1),
    lift_negations_top_level_functor(Functor, Node, CandidateAcc1, NNode, CandidateImpls),
    LPred = b(NNode,Type,Info).
lift_negations_find_impls(BTerm, CandidateAcc, BTerm, CandidateAcc).

lift_negations_top_level_functor(Functor, Node, CandidateAcc, NNode, CandidateImpls) :-
    arg(1, Node, Lhs),
    arg(2, Node, Rhs),
    lift_negations_find_impls(Lhs, CandidateAcc, LLhs, CandidateAcc1),
    lift_negations_find_impls(Rhs, CandidateAcc1, LRhs, CandidateImpls),
    NNode =.. [Functor,LLhs,LRhs].

get_sorted_equality(Lhs, Rhs, Equality) :-
    Lhs @> Rhs,
    !,
    Equality = b(equal(Rhs,Lhs),pred,[]).
get_sorted_equality(Lhs, Rhs, b(equal(Lhs,Rhs),pred,[])).

negated_b_operator(not_equal, equal).
negated_b_operator(not_member, member).
negated_b_operator(not_subset, subset).
negated_b_operator(not_subset_strict, subset_strict).

create_bindings([], Acc, Acc).
create_bindings([b(identifier(VarName),_,_)|T], Acc, Bindings) :-
    !,
    create_bindings(T, [bind(VarName,_)|Acc], Bindings).
create_bindings([VarName|T], Acc, Bindings) :-
    create_bindings(T, [bind(VarName,_)|Acc], Bindings).
