:- module(dpllt_sat_solver, [b_to_cnf/3, sat/2]).

:- use_module(library(lists)).

:- use_module(probsrc(bsyntaxtree),[get_texpr_expr/2,create_texpr/4]).
:- use_module(probsrc(error_manager), [add_error_fail/3]).
:- use_module(probsrc(translate), [translate_bexpression/2]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% A sat solver, utilising delay declaration to implement
% watched literals
%
% Authors: Jacob Howe and Andy King
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

sat(Clauses, Vars) :-
    problem_setup(Clauses),
    elim_var(Vars).

elim_var([]).
elim_var([Var|Vars]) :-
    elim_var(Vars),
    (   Var = pred_true
    ;   Var = pred_false
    ).

problem_setup([]).
problem_setup([Clause|Clauses]) :-
    clause_setup(Clause),
    problem_setup(Clauses).

clause_setup([Pol-Var-Name|Pairs]) :-
    set_watch(Pairs, Var, Pol, Name).

set_watch([], Var, Pol, _) :-
    Var = Pol.
set_watch([Pol2-Var2-Name2|Pairs], Var1, Pol1, Name1):-
    watch(Var1, Pol1, Name1, Var2, Pol2, Name2, Pairs).

:- block watch(-, ?, ?, -, ?, ?, ?).
watch(Var1, Pol1, Name1, Var2, Pol2, Name2, Pairs) :-
    (   nonvar(Var1)
    ->  update_watch(Var1, Pol1, Name1, Var2, Pol2, Name2, Pairs)
    ;   update_watch(Var2, Pol2, Name2, Var1, Pol1, Name1, Pairs)
    ).

update_watch(Var1, Pol1, _Name1, Var2, Pol2, Name2, Pairs) :-
    (   Var1 == Pol1
    ->  true
    ;   set_watch(Pairs, Var2, Pol2, Name2)
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

b_pred_or_bool_true(X) :- X == truth; X == boolean_true; X == value(pred_true).
b_pred_or_bool_false(X) :- X == falsity; X == boolean_false; X == value(pred_false).

b_to_cnf(In,State,Out) :-
    get_texpr_expr(In,Expr),
    b_to_cnf_aux(Expr,State,Out),
    !.
b_to_cnf(In,_,_) :-
    translate_bexpression(In,PPIn),
    !,
    add_error_fail(b_to_cnf,'CNF conversion failed', PPIn).

b_to_cnf_aux(identifier(Name),State,[[pred_true-Res-Name]]) :-
    member(bind(Name,Res),State).
b_to_cnf_aux(X,_,[[pred_true-pred_true-nonvar]]) :-
    b_pred_or_bool_true(X).
b_to_cnf_aux(X,_,[[pred_false-pred_false-nonvar]]) :-
    b_pred_or_bool_false(X).

b_to_cnf_aux(equal(A,B),State,CNF) :-
    ground(B),
    get_texpr_expr(B,Expr),
    b_pred_or_bool_true(Expr),
    b_to_cnf(A,State,CNF).
b_to_cnf_aux(equal(A,B),State,CNF) :-
    ground(B),
    get_texpr_expr(B,Expr),
    b_pred_or_bool_false(Expr),
    create_texpr(value(pred_true),boolean,[],True),
    create_texpr(equal(A,True),pred,[],EqualTrue),
    b_to_cnf_aux(negation(EqualTrue),State,CNF).

b_to_cnf_aux(negation(A),State,[[Res]]) :-
    b_to_cnf(A,State,[[Pol-Var-Name]]),
    (   Pol == pred_true
    ->  Res = pred_false-Var-Name
    ;   Pol == pred_false
    ->  Res = pred_true-Var-Name
    ).

b_to_cnf_aux(implication(A,B),State,CNF) :-
    create_texpr(negation(A),pred,[],NotA),
    b_to_cnf_aux(disjunct(NotA,B),State,CNF).

b_to_cnf_aux(disjunct(D1,D2),State,Res) :-
    get_texpr_expr(D1,conjunct(A,B)),
    !,
    create_texpr(disjunct(A,D2),pred,[],DJ1),
    create_texpr(disjunct(B,D2),pred,[],DJ2),
    b_to_cnf_aux(conjunct(DJ1,DJ2),State,Res).
b_to_cnf_aux(disjunct(D1,D2),State,Res) :-
    get_texpr_expr(D2,conjunct(A,B)),
    !,
    create_texpr(disjunct(D1,A),pred,[],DJ1),
    create_texpr(disjunct(D1,B),pred,[],DJ2),
    b_to_cnf_aux(conjunct(DJ1,DJ2),State,Res).

b_to_cnf_aux(disjunct(D1,D2), State, Res) :-
    b_to_cnf(D1, State, ResD1),
    b_to_cnf(D2, State, ResD2),
    (   ResD1 = [Clause1],
        ResD2 = [Clause2]
    ->  append(Clause1, Clause2, Clause),
        Res = [Clause]
    ;   ResD1 \= [_],
        ResD2 = [Clause2]
    ->  maplist(append(Clause2), ResD1, Res)
    ;   ResD1 = [Clause1],
        ResD2 \= [_]
    ->  maplist(append(Clause1), ResD2, Res)
    ).

b_to_cnf_aux(conjunct(A,B), State, Res) :-
    b_to_cnf(A, State, RA),
    b_to_cnf(B, State, RB),
    append(RA, RB, Res).
