:- module(dpllt_solver, [dpllt_solve_predicate/2]).

:- use_module(library(lists)).
:- use_module(library(timeout)).

:- use_module(probsrc('dpllt_solver/dpllt_preprocessing')).
:- use_module(probsrc('dpllt_solver/dpllt_sat_solver')).
:- use_module(probsrc(b_ast_cleanup), [clean_up_pred/3]).
:- use_module(probsrc(preferences), [get_preference/2,
                                     temporary_set_preference/2,
                                     reset_temporary_preference/1]).
:- use_module(probsrc(unsat_cores), [unsat_core/2]).
:- use_module(probsrc(b_enumerate), [b_tighter_enumerate_values/2]).
:- use_module(probsrc(kernel_waitflags), [init_wait_flags/2,ground_wait_flags/1]).
:- use_module(probsrc(b_interpreter), [set_up_typed_localstate/6,
                                       b_test_boolean_expression/4,
                                       b_test_boolean_expression_wf/3,
                                       b_convert_bool/5]).
:- use_module(probsrc(error_manager), [add_error/3]).
:- use_module(probsrc(bsyntaxtree), [find_typed_identifier_uses/3,
                                     find_typed_identifier_uses/2,
                                     find_identifier_uses/3,
                                     remove_all_infos/2,
                                     add_texpr_infos/3,
                                     get_texpr_info/2,
                                     conjunct_predicates/2,
                                     conjunction_to_list/2,
                                     disjunct_predicates/2,
                                     disjunction_to_list/2,
                                     unique_typed_id/3]).
:- use_module(probsrc('kodkod/kodkod_annotator')).
%:- use_module(probsrc(b_expression_sharing), [cse_optimize_predicate/2,
%                                              expand_cse_lazy_lets/2]).
:- use_module(probsrc(module_information), [module_info/2]).

:- module_info(group, cbc).
:- module_info(description,'This module provides a DPLL(T) based solver for B.').

:- dynamic initial_solution/0, local_result_found/0, local_unsat_core/1, local_sat_bindings/1.
:- volatile initial_solution/0, local_result_found/0, local_unsat_core/1, local_sat_bindings/1.

init :-
    error_manager:reset_errors,
    retractall(initial_solution),
    retractall(local_result_found),
    %temporary_set_preference(use_chr_solver, true),
    temporary_set_preference(use_smt_mode, true),
    temporary_set_preference(use_clpfd_solver, true),
    temporary_set_preference(optimize_ast, false),
    temporary_set_preference(normalize_ast, true). % important

reset_preferences :-
    reset_temporary_preference(use_smt_mode),
    reset_temporary_preference(use_clpfd_solver),
    reset_temporary_preference(optimize_ast),
    reset_temporary_preference(normalize_ast).

%% dpllt_solve_predicate(+Pred, -Result).
% Main interface predicate for DPLL(T) based solver which times out
% with respect to ProB's time_out preference.
dpllt_solve_predicate(Pred, Result) :-
    get_preference(time_out, Timeout),
    time_out(dpllt_solve_predicate_no_timeout(Pred, SolverResult),
             Timeout,
             TimeoutResult),
    (   TimeoutResult == time_out
    ->  Result = time_out
    ;   Result = SolverResult
    ),
    reset_preferences.

%% dpllt_solve_predicate(+Pred, -SolverResult).
% Interface predicate for DPLL(T) based solver without a timeout.
dpllt_solve_predicate_no_timeout(Pred, SolverResult) :-
    init,
    b_ast_cleanup:clean_up_pred(Pred, _, CleanPred),
    %b_expression_sharing:cse_optimize_predicate(CleanPred, TCseOptPred),
    %b_expression_sharing:expand_cse_lazy_lets(TCseOptPred, CseOptPred),
    preprocess_predicate(CleanPred, LiftedPred, FilteredCandidateImplsConj),
    predicate_to_sat(normal, [], [], LiftedPred, Env1, AnnotatedPred, SmtBoolFormula, SatVars1),
    predicate_to_sat(normal, SatVars1, Env1, FilteredCandidateImplsConj, _NewEnv, _, ImplBoolFormula, SatVars),
    !,
    conjunct_predicates([ImplBoolFormula,SmtBoolFormula], _BoolFormula),
    find_typed_identifier_uses(AnnotatedPred, SmtVars), % TO DO: improve
    conflict_driven_clause_learning_online(SmtVars, SatVars, SmtBoolFormula, SolverResult).
dpllt_solve_predicate_no_timeout(Pred, _) :-
    add_error('dpllt_solve_predicate', 'Cannot create SAT formula from B predicate:', [Pred]).

% Given a set of boolean B identifiers. Create a list of bindings containing terms bind/2
% and a list of tuples of B identifier and corresponding Prolog variable used for sat solving.
get_bindings_from_ids(SatVars, SatBindings, IdPrologVarTuples, PrologSatVars) :-
    get_bindings_from_ids_acc(SatVars, [], [], [], SatBindings, IdPrologVarTuples, PrologSatVars).

get_bindings_from_ids_acc([], AccBindings, AccTuples, AccPlVars, AccBindings, AccTuples, AccPlVars).
get_bindings_from_ids_acc([b(identifier(Name),IdType,IdInfo)|T], AccBindings, AccTuples, AccPlVars, SatBindings, IdPrologVarTuples, PrologSatVars) :-
    NewAccBindings = [bind(Name,PrologVar)|AccBindings],
    NewAccTuples = [(b(identifier(Name),IdType,IdInfo),PrologVar)|AccTuples],
    NewAccPlVars = [PrologVar|AccPlVars],
    get_bindings_from_ids_acc(T, NewAccBindings, NewAccTuples, NewAccPlVars, SatBindings, IdPrologVarTuples, PrologSatVars).

%% conflict_driven_clause_learning_online(+SmtVars, +SatVars, +BoolFormula, -SolverResult).
% Conflict-driven clause learning from incomplete assignments of the boolean formula by setting up reification constraints
% connecting the SAT and SMT solver.
conflict_driven_clause_learning_online(SmtVars, SatVars, BoolFormula, SolverResult) :-
    get_bindings_from_ids(SatVars, SatBindings, IdPrologVarTuples, PrologSatVars),
    b_to_cnf(BoolFormula, SatBindings, CnfBoolFormula),
    setup_theory_wf_store(SmtVars, SmtBindings, WfStoreSmt),
    setup_reification(IdPrologVarTuples, SmtBindings, WfStoreSmt),
    sat(CnfBoolFormula, PrologSatVars),
    ground_wait_flags(WfStoreSmt),
    SolverResult = solution(SmtBindings),
    log_solution.
conflict_driven_clause_learning_online(_, _, _, contradiction_found) :-
    \+ initial_solution.

%% setup_reification(+IdPrologVarTuples, +SmtBindings, +WfStoreSmt).
% Set up reification constraints to connect SAT and SMT solver.
% Propagates in both directions (b_convert_bool/5 implements theory propagation).
setup_reification([], _, _).
setup_reification([(SatId,SatPrologVar)|T], SmtBindings, WfStoreSmt) :-
    get_texpr_info(SatId, Info),
    memberchk(smt_formula(SmtFormula), Info),
    b_convert_bool(SmtFormula, SmtBindings, [], WfStoreSmt, SatPrologVar),
    setup_reification(T, SmtBindings, WfStoreSmt).

%% conflict_driven_clause_learning_offline(+Env, +SmtVars, +SatVars, +BoolFormula, -Result).
conflict_driven_clause_learning_offline(Env, SmtVars, SatVars, BoolFormula, Result) :-
    set_up_typed_localstate(SatVars, _, SatTypedVals, [], SatBindings, positive),
    init_wait_flags(WFStoreSAT, [wf_sat_cdcl]),
    b_tighter_enumerate_values(SatTypedVals, WFStoreSAT),
    !,
    conflict_driven_clause_learning_offline(WFStoreSAT, SmtVars, Env, SatBindings, SatVars, BoolFormula, Result).

%% conflict_driven_clause_learning_offline(+WFStoreSAT, +SmtVars, +Env, +SatBindings, +SatVars, +BoolFormula, -Result).
% Conflict-driven clause learning from complete assignments of the boolean formula by computing an unsat core.
conflict_driven_clause_learning_offline(WFStoreSAT, SmtVars, Env, SatBindings, SatVars, BoolFormula, Result) :-
    retractall(local_result_found),
    (   b_test_boolean_expression(BoolFormula, SatBindings, [], WFStoreSAT),
        setup_theory_wf_store(SmtVars, SmtBindings, WFStoreSMT),
        (   ground_wait_flags(WFStoreSAT),
            asserta(local_sat_bindings(SatBindings)), % enables to collect all found sat solutions for exclusion later on if backtracking
            sat_to_predicate_from_solution(SatBindings, SatVars, SmtFormula),
            b_test_boolean_expression(SmtFormula, SmtBindings, [], WFStoreSMT),
            ground_wait_flags(WFStoreSMT),
            Result = solution(SmtBindings),
            log_solution
        ;   \+ local_result_found, % no choicepoint if backtracking from exhaustive search after finding all local solutions
            % TO DO: investigate output of uc for static contradiction
            unsat_core_with_timeout(SmtFormula, UnsatCore),
            findall(LocalSatBindings, local_sat_bindings(LocalSatBindings), LLocalSatBindings),
            retractall(local_sat_bindings(_)),
            maplist(get_exclusion_from_sat_bindings, LLocalSatBindings, LSatExcl),
            conjunct_predicates(LSatExcl, SatExcl),
            % only reuse: assuming that unsat core computation minimizes only w/o introducing new formulae
            predicate_to_sat(only_reuse, SatVars, Env, UnsatCore, _NewEnv, _AnnotatedPred, UnsatCoreSat, _NewSatVars),
            NegUnsatCoreSat = b(negation(UnsatCoreSat),pred,[]),
            asserta(local_unsat_core(b(conjunct(NegUnsatCoreSat,SatExcl),pred,[]))),
            fail % we have to fail here in order to unbind SatBindings prior to the recursive call
        )
    ;   retract(local_unsat_core(LocalUnsatCore)),
        !,
        NewBoolFormula = b(conjunct(BoolFormula,LocalUnsatCore),pred,[]),
        conflict_driven_clause_learning_offline(WFStoreSAT, SmtVars, Env, SatBindings, SatVars, NewBoolFormula, Result)
    ;   (   \+ initial_solution
        ->  Result = contradiction_found
        ;   fail % all solutions have been found
        )
    ).

%% setup_theory_wf_store(+SmtVars, -SmtBindings, -WFStoreSMT).
setup_theory_wf_store(SmtVars, SmtBindings, WFStoreSMT) :-
    set_up_typed_localstate(SmtVars, _, SmtTypedVals, [], SmtBindings, positive),
    init_wait_flags(WFStoreSMT, [wf_smt_cdcl]),
    b_tighter_enumerate_values(SmtTypedVals, WFStoreSMT),!.

get_exclusion_from_sat_bindings([], b(truth,pred,[])).
get_exclusion_from_sat_bindings([Binding|T], SatExclusion) :-
    get_sat_inequality_from_binding(Binding, Inequality),
    !,
    get_exclusion_from_sat_bindings_acc(T, Inequality, SatExclusion).
get_exclusion_from_sat_bindings([_|T], SatExclusion) :-
    % skip: binding might be a variable if instantiation is not required
    get_exclusion_from_sat_bindings(T, SatExclusion).

get_exclusion_from_sat_bindings_acc([], Acc, Acc).
get_exclusion_from_sat_bindings_acc([Binding|T], Acc, SatExclusion) :-
    get_sat_inequality_from_binding(Binding, Inequality),
    !,
    NewAcc = b(disjunct(Acc,Inequality),pred,[]),
    get_exclusion_from_sat_bindings_acc(T, NewAcc, SatExclusion).
get_exclusion_from_sat_bindings_acc([_|T], Acc, SatExclusion) :-
    get_exclusion_from_sat_bindings_acc(T, Acc, SatExclusion).

get_sat_inequality_from_binding(bind(SatVarName,Binding), SatEquality) :-
    ground(Binding),
    !,
    get_negated_ast_from_sat_binding(Binding, NegAstValue),
    SatEquality = b(equal(b(identifier(SatVarName),boolean,[]),NegAstValue),pred,[]).
/*get_sat_inequality_from_binding(Binding, _) :-
    add_error(get_sat_inequality_from_binding, 'SAT binding not ground:', [Binding]),!,
    fail.*/

get_negated_ast_from_sat_binding(pred_true, b(boolean_false,boolean,[])).
get_negated_ast_from_sat_binding(pred_false, b(boolean_true,boolean,[])).

%% predicate_to_sat(+Type, +VarAcc, +Env, +Pred, -NewEnv, -NewPred, -BoolFormula, -NVarAcc).
% Create a boolean formula from a B predicate, e.g., return 'A=TRUE & B=TRUE' for 'x:INT & x>1'.
% Introduces fresh unique identifiers if Type is not equal to 'only_reuse'.
% If Type is 'only_reuse': do not introduce new SAT variables but skip parts that need a new SAT variable
% (used to include a (possibly partial) SMT conflict in the SAT formula).
% Env stores pairs of B ASTs (without info field) and its assigned SAT variables.
% Add corresponding SAT variable names to AST's info fields (returned in NewPred).
% Each SAT variable contains a term smt_formula/1 in its info field providing the corresponding SMT formula.
% Note: Predicate fails for scalar expressions like an integer except of truth and falsity.
% Note: SMT formula is split on the level of conjunction, disjunction, implication and equivalence.
predicate_to_sat(_, VarAcc, Env, b(truth,pred,Info), Env, b(truth,pred,Info), b(truth,pred,Info), VarAcc) :- !.
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(conjunct(_,_),_,_),
    conjunction_to_list(Pred, PList),
    map_predicate_to_sat(Type, VarAcc, Env, PList, NewEnv, NPList, SatList, NVarAcc),
    !,
    conjunct_predicates(SatList, BoolFormula),
    conjunct_predicates(NPList, NewPred).
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(disjunct(_,_),_,_),
    disjunction_to_list(Pred, PList),
    map_predicate_to_sat(Type, VarAcc, Env, PList, NewEnv, NPList, SatList, NVarAcc),
    !,
    disjunct_predicates(SatList, BoolFormula),
    disjunct_predicates(NPList, NewPred).
predicate_to_sat(Type, VarAcc, Env, b(negation(Pred),pred,_), NewEnv, NewPred, BoolFormula, NVarAcc) :-
    reuse_or_introduce_bool_var(Type, VarAcc, Env, Pred, boolean_false, NewEnv, NewPred, BoolFormula, NVarAcc),
    NewPred \== b(truth,pred,_),
    !.
predicate_to_sat(Type, VarAcc, Env, b(negation(Pred),pred,_), NewEnv, NewPred, BoolFormula, NVarAcc) :-
    predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, TNewPred, BoolFormula, NVarAcc),
    (   TNewPred \== b(truth,pred,_)
    ->  NewPred = b(negation(TNewPred),pred,[])
    ;   NewPred = b(truth,pred,[])
    ).
predicate_to_sat(_, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, VarAcc) :-
    Pred = b(Functor,pred,Info),
    (   Functor == value(pred_true)
    ;   Functor == value(pred_false)
    ),
    !,
    BoolFormula = b(Functor,pred,Info),
    NewEnv = Env,
    NewPred = Pred.
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(implication(Lhs,Rhs),_,Info),
    map_predicate_to_sat(Type, VarAcc, Env, [Lhs,Rhs], NewEnv, NSmtArgs, NSatArgs, NVarAcc),
    !,
    NSmtArgs = [NSmtLhs,NSmtRhs],
    NewPred = b(implication(NSmtLhs,NSmtRhs),pred,Info),
    NSatArgs = [NSatLhs,NSatRhs],
    BoolFormula = b(implication(NSatLhs,NSatRhs),pred,[smt_formula(Pred)|Info]).
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(equivalence(Lhs,Rhs),_,Info),
    map_predicate_to_sat(Type, VarAcc, Env, [Lhs,Rhs], NewEnv, NSmtArgs, NSatArgs, NVarAcc),
    !,
    NSmtArgs = [NSmtLhs,NSmtRhs],
    NewPred = b(equivalence(NSmtLhs,NSmtRhs),pred,Info),
    NSatArgs = [NSatLhs,NSatRhs],
    BoolFormula = b(equivalence(NSatLhs,NSatRhs),pred,[smt_formula(Pred)|Info]).
predicate_to_sat(Type, VarAcc, Env, Pred, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    Pred = b(Node,_,_),
    functor(Node, _, Arity),
    (   Arity == 2
    ;   Arity == 3
    ),
    !,
    reuse_or_introduce_bool_var(Type, VarAcc, Env, Pred, boolean_true, NewEnv, NewPred, BoolFormula, NVarAcc).
% TO DO: if going into quantifications, we have to watch out for local identifier scopes
%       (i.e., x<7 might not be the same x<7 within a quantification if x is a local id)
% Note that we currently do not go into quantifications so this is not an issue.

map_predicate_to_sat(_, VarAcc, Env, [], Env, [], [], VarAcc).
map_predicate_to_sat(Type, VarAcc, Env, [Pred|T], NewEnv, [NPred|PNT], [BoolFormula|NT], NVarAcc) :-
    predicate_to_sat(Type, VarAcc, Env, Pred, TempEnv, NPred, BoolFormula, TVarAcc),
    map_predicate_to_sat(Type, TVarAcc, TempEnv, T, NewEnv, PNT, NT, NVarAcc).

% Note: BoolFormula is 'a = TRUE' or 'a = FALSE' for (negated) literal a.
reuse_or_introduce_bool_var(_, VarAcc, Env, Pred, BoolAtom, NewEnv, NewPred, BoolFormula, NVarAcc) :-
    remove_all_infos(Pred, CPred),
    member((CPred,SatVar), Env),
    !,
    SatVar = b(identifier(SatVarName),_,_),
    add_texpr_infos(Pred, [sat_var(SatVarName)], NewPred),
    BoolFormula = b(equal(SatVar,b(BoolAtom,boolean,[])),pred,[]),
    NewEnv = Env,
    NVarAcc = VarAcc.
reuse_or_introduce_bool_var(Type, VarAcc, Env, Pred, BoolAtom, [(CleanPred,UniqueId)|Env], NewPred, BoolFormula, NVarAcc) :-
    Type \== only_reuse,
    !,
    unique_typed_id("sat", boolean, TUniqueId),
    remove_all_infos(Pred, CleanPred),
    TUniqueId = b(identifier(SatVarName),_,_),
    %find_identifier_uses(Pred, [], UsedIds),
    add_texpr_infos(Pred, [sat_var(SatVarName)], NewPred),
    add_texpr_infos(TUniqueId, [smt_formula(Pred)], UniqueId),
    BoolFormula = b(equal(UniqueId,b(BoolAtom,boolean,[])),pred,[]),
    NVarAcc = [UniqueId|VarAcc].
reuse_or_introduce_bool_var(only_reuse, VarAcc, Env, Pred, _, Env, Pred, b(truth,pred,[]), VarAcc).

%% sat_to_predicate_from_solution(SatBindings, SatVars, Conjunction).
% Create conjunction of SAT bindings using their corresponding SMT formulae
% which is stored in the bool variable's information field.
sat_to_predicate_from_solution([], _, b(truth,pred,[])).
sat_to_predicate_from_solution([Binding|T], SatVars, Conjunction) :-
    get_smt_formula_from_binding(Binding, SatVars, SMT, RestSatVars),
    !,
    sat_to_predicate_from_solution(T, SMT, RestSatVars, Conjunction).
sat_to_predicate_from_solution([_|T], SatVars, Conjunction) :-
    % skip: binding might be a variable if instantiation is not required
    sat_to_predicate_from_solution(T, SatVars, Conjunction).

sat_to_predicate_from_solution([], SmtAcc, _, SmtAcc).
sat_to_predicate_from_solution([Binding|T], SmtAcc, SatVars, Conjunction) :-
    get_smt_formula_from_binding(Binding, SatVars, SMT, RestSatVars),
    !,
    NewSmtAcc = b(conjunct(SMT,SmtAcc),pred,[]),
    sat_to_predicate_from_solution(T, NewSmtAcc, RestSatVars, Conjunction).
sat_to_predicate_from_solution([_|T], SmtAcc, SatVars, Conjunction) :-
    sat_to_predicate_from_solution(T, SmtAcc, SatVars, Conjunction).

get_smt_formula_from_binding(bind(SatVarName,Binding), SatVars, SMT, RestSatVars) :-
    ground(Binding),
    !,
    select(b(identifier(SatVarName),_,Info), SatVars, RestSatVars),
    member(smt_formula(TSMT), Info),!,
    (   Binding == pred_true
    ->  SMT = TSMT
    ;   Binding == pred_false,
        SMT = b(negation(TSMT),pred,[])
    ).
/*get_smt_formula_from_binding(Binding, _, _, _) :-
    add_error(get_smt_formula_from_binding, 'SAT binding not ground:', [Binding]),
    fail.*/

unsat_core_with_timeout(Formula, UnsatCore) :-
    temporary_set_preference(time_out, 100),
    time_out(unsat_core(Formula, TUnsatCore), 200, Result),
    (   Result == time_out
    ->  UnsatCore = Formula
    ;   UnsatCore = TUnsatCore),
    reset_temporary_preference(time_out),!.

log_solution :-
    \+ initial_solution,
    !,
    asserta(initial_solution),
    asserta(local_result_found).
log_solution :-
    asserta(local_result_found).
