% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(error_manager,[init_error_manager/0,
           add_warning/2, add_warning/3, add_warning/4,
           reset_errors/0, clear_wd_errors/0,
           add_error/2, add_error/3, add_error/4,
           add_error_and_fail/3, add_error_and_fail/2,
           add_error_fail/3,
           add_failed_call_error/1, add_internal_error/2, add_internal_error/4,
           register_optional_error/1, reset_optional_errors/0,
           assert_true/1,
           add_message/2, add_message/3, add_message/4,
           add_all_perrors/1, add_all_perrors/2, add_all_perrors/3,
           no_real_perror_occurred/1,
           get_all_errors/1, tcltk_get_all_errors/4,
           get_all_errors_and_reset/1, % call if you want to start with a fresh top-level error scope
           get_all_errors_with_span_info_and_reset/1, % ditto + obtain location info
           get_all_errors_and_clear/1,
           get_error/2, get_error_with_span/3,
           get_warning/2, check_error_occured/2, check_warning_occured/2,
           get_error_span_linecol/5, get_error_span_linecol/4, get_error_span_linecol_for_main_file/4,
           register_error_span/2,
           check_error_span_file_linecol/6,
           position_is_not_in_main_file/1,
           error_or_warning_occured/0,
           real_error_occurred/0, reset_real_error_occurred/0, assert_real_error_occurred/1,
           test_error_occurred/1, count_errors_occurred/2,
           get_total_number_of_errors/1,
           reset_error_spans/0,

           backup_errors/0, backed_up_error/4,

           set_error_context/1, clear_error_context/0,
           get_error_context/1, restore_error_context/1,

           extract_span_description/2,
           extract_line_col/5, extract_file_line_col/6,
           extract_all_line_col/5,
           extract_line_col_for_main_file/5,
           extract_subsidiary_tail_file_name/2, extract_tail_file_name/2,
           extract_file_number_and_name/3,
           extract_additional_description/2,
           print_error_span/1, print_message_span/1,

           turn_show_error_source_on/0, turn_show_error_source_off/0,

            enter_new_error_scope/1, enter_new_error_scope/2,
            enter_new_clean_error_scope/1,
            exit_error_scope/2, exit_error_scope/3,
            error_occurred_at_level/1, specific_event_occurred_at_level/2,
            %throw_errors_in_scope/3,
            add_new_event_in_error_scope/1, add_new_event_in_error_scope/2,
            event_occurred_in_error_scope/1, error_occurred_in_error_scope/0,
            clear_events_in_error_scope/2, clear_events_in_current_error_scope/1,
            clear_all_errors_in_error_scope/1,
            call_in_fresh_error_scope_for_one_solution/1,
            enumeration_warning_occured_in_error_scope/0,
            abort_error_occured_in_error_scope/0,
            on_enumeration_warning/2, on_enumeration_warning_with_continue/3,
            throw_enumeration_warnings_in_current_scope/0,
            throw_enumeration_warnings_in_scope/3,
            catch_enumeration_warning_exceptions/2, catch_enumeration_warning_exceptions/3,
            catch_enumeration_warning_exceptions/4,
            call_with_enumeration_warning/1,
            observe_enumeration_warnings/2,
            critical_enumeration_warning_occured_in_error_scope/4, critical_enumeration_warning_occured_in_error_scope/0,
            clear_enumeration_warnings/0,
            translate_events_in_current_scope_to_warnings/2,
            time_out_with_enum_warning_one_solution/3, % use this for finding one solution
            time_out_with_enum_warning_one_solution_no_new_error_scope/3,
            is_time_out_result/1,
            time_out_with_enum_warning_for_findall/3, % use this if you cycle through all solutions
            time_out_with_enum_warning_for_findall_in_current_error_scope/4,
            logged_error/4,
            display_error_statistics/0,
            set_error_manager_silent_mode/1
            ]).

% meta_predicate annotations should appear before loading any code:
:- meta_predicate add_failed_call_error(0).
:- meta_predicate add_internal_error(-,:).
:- meta_predicate add_new_event_in_error_scope(-,2).

:- meta_predicate on_enumeration_warning(0,0).
:- meta_predicate on_enumeration_warning_with_continue(0,0,0).

:- meta_predicate catch_enumeration_warning_exceptions(0,0).
:- meta_predicate catch_enumeration_warning_exceptions(0,0,-).
:- meta_predicate catch_enumeration_warning_exceptions(0,0,-,-).

:- meta_predicate observe_enumeration_warnings(0,0).

:- meta_predicate time_out_with_enum_warning_one_solution(0,-,-).

:- meta_predicate call_in_fresh_error_scope_for_one_solution(0).

:- meta_predicate call_with_enumeration_warning(0).

:- meta_predicate time_out_with_enum_warning_one_solution_no_new_error_scope(0,-,-).

:- meta_predicate time_out_with_enum_warning_for_findall(0,-,-).
:- meta_predicate time_out_with_enum_warning_for_findall2(-,0,-,-).

:- meta_predicate time_out_with_enum_warning_for_findall_in_current_error_scope(-,0,-,-).

:- meta_predicate assert_true(0).
% ------------

:- use_module(library(lists)).

:- use_module(tools,[get_tail_filename/2]).
:- use_module(tools_lists,[count_occurences/2]).
:- use_module(tools_printing,[print_error/1,print_message_on_output/2,format_error/2,format_with_colour/4]).
:- use_module(tools_strings,[ajoin/2, string_concatenate/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'The error_manager controls error messages. Error messages can be added, retrieved, counted,etc. The module also takes care of extracting source spans from error messages.').

%:- use_module(translate).
:- use_module(debug, [debug_mode/1, debug_println/1, debug_println/2]).
:- use_module(preferences,[get_preference/2,reset_temporary_preference/2,
                           temporary_set_preference/3]).


:- use_module(extension('counter/counter')).

% cur_block_level is now stored as a C counter, which needs to be initialised
init_error_manager :- counter_init, new_counter(cur_block_level).


:- dynamic optional_error/1.
register_optional_error(Type) :-
   (optional_error(Type) -> true ; assert(optional_error(Type))).
reset_optional_errors :- retractall(optional_error(_)).

% INTERNAL ERRORS


add_failed_call_error(Call) :- !,
   add_internal_error('Internal Error; call failed: ',Call).

add_internal_error(Module,Msg,Term,Span) :-
   add_error(internal_error(Module),Msg,Term,Span).
add_internal_error(Msg,Module:Call) :- !,
   add_error(internal_error(Module),Msg,Call).
add_internal_error(Msg,Call) :-
   add_error(internal_error(unknown_module),Msg,Call).


% ERROR SCOPING

%:- dynamic cur_block_level/1.
%cur_block_level(0). % no realised as a C counter
:- volatile error_occurred_at_level/1.
:- dynamic error_occurred_at_level/1.
:- volatile throw_error/3.
:- dynamic throw_error/3.



enter_new_error_scope(Level,_PP) :-
   enter_new_error_scope(Level). %, tools_printing:print_term_summary(enter(Level,PP)),nl.
enter_new_error_scope(Level) :-
   enter_new_clean_error_scope(Level),
   L1 is Level - 1,
   copy_throw_errors(L1,Level).
copy_throw_errors(L1,L) :- throw_error(L1,Error,Span), %print(copy(throw(L1,L,Error))),nl,
   assert(throw_error(L,Error,Span)),fail.
copy_throw_errors(_,_).
enter_new_clean_error_scope(Level) :-
  inc_counter(cur_block_level,L1),
  retractall(throw_error(L1,_,_)),
  % print(enter(L1)),nl,trace,
  Level=L1.
% mark that everything unifying with Error should be thrown rather than collected
throw_errors_in_scope(Level,Error,Span) :-
   (\+ throw_error(Level,Error,Span)  % TO DO: check that we are an instance of existing error !
    -> asserta(throw_error(Level,Error,Span))
     ; true).

clear_all_errors_in_error_scope(Level) :- %print(clearing(Level)),nl,
   clear_events_in_error_scope(Level,_),
   error_occurred_at_level(L1), L1>Level, %print(clearing(L1)),nl,
   clear_events_in_error_scope(Level,_),fail.
clear_all_errors_in_error_scope(_).

exit_error_scope(Level,ErrorOcc) :- exit_error_scope(Level,ErrorOcc,unknown).
exit_error_scope(Level,ErrorOcc,PP) :- get_counter(cur_block_level,CurL),
    % print(exit(Level,CurL,PP)),nl,
   (CurL = Level -> NewL is CurL-1
    ; Level < CurL -> NewL is Level-1,  format('*** exiting multiple error blocks (~w -> ~w) for ~w ***~n',[CurL,NewL,PP])
    ; format('*** block ~w already exited (currently at ~w) for ~w ***~n',[Level,CurL,PP]),
      NewL is CurL
   ), set_counter(cur_block_level,NewL),
   copy_error_scope_events(CurL,NewL),
   retractall(throw_error(CurL,_,_)),
   (retract(error_occurred_at_level(CurL))
    -> ErrorOcc=true,
       (error_occurred_at_level(NewL) -> true ; assert(error_occurred_at_level(NewL)))
    ; ErrorOcc=false).


% versions of enter_new_error_scope and exit_error_scope which re_enter blocks and clean up upon backtracking
%enter_new_error_scope_bt(Level,PP) :- enter_new_error_scope(Level,PP), exit_upon_fail(Level,PP).
%exit_upon_fail(_,_).
%exit_upon_fail(Level,PP) :-  print(exit_upon_backtrack(Level,PP)),nl,
%   exit_error_scope(Level,_,PP),fail.
%
%exit_error_scope_bt(Level,ErrorOcc,PP) :- exit_error_scope(Level,ErrorOcc,PP).
%exit_error_scope_bt(Level,_ErrorOcc,PP) :- print(re_enter_block(Level,PP)),nl,
%   enter_new_error_scope(Level2,PP),
%   (Level2=Level -> true ; print(mismatch_when_reentering_block(Level,PP)),nl),
%   fail.

% ---------------------------



assert_real_error_in_current_block(Source) :- get_counter(cur_block_level,L),L>0,
  %print(assert_real_error_in_current_block(L)),nl,
  (specific_event_occurred_at_level(L,Source) -> true
    ; assert(specific_event_occurred_at_level(L,Source))), % ensures that clear_events_in_current_error_scope does not remove error_occurred_at_level if used for enumeration warnings, ...
   \+ error_occurred_at_level(L), assert(error_occurred_at_level(L)),!.
assert_real_error_in_current_block(_).


reset_error_scopes :-
   set_counter(cur_block_level,0),
   retractall(error_occurred_at_level(_)),
   retractall(specific_event_occurred_at_level(_,_)).

% for debugging / portraying error scopes:
:- public print_error_scopes/0.
print_error_scopes :-
   if(get_counter(cur_block_level,L),
     format(' Cur Block Level = ~w~n',[L]),
     format(' *** NO Cur Block Level Information !!',[])
     ),
   fail.
print_error_scopes :-
   findall(L,error_occurred_at_level(L),LL),
   format(' Errors occured at levels = ~w~n',[LL]),
   specific_event_occurred_at_level(Level,Err),
   format(' ~w  -->  ~w~n',[Level,Err]),
   fail.
print_error_scopes.

% call a predicate in fresh error scope (assuming one solution is searched for only; failure is also ok)
% and then discard all errors
call_in_fresh_error_scope_for_one_solution(Call) :-
     enter_new_error_scope(Level,call_in_fresh_error_scope_for_one_solution),
     call_cleanup(Call,
                   (clear_all_errors_in_error_scope(_),
                    exit_error_scope(Level,_,call_in_fresh_error_scope_for_one_solution))).

:- volatile specific_event_occurred_at_level/2.
:- dynamic specific_event_occurred_at_level/2.

% can be used to keep track of whether specific events occurred within an error_scope
add_new_event_in_error_scope(Event) :-
    add_new_event_in_error_scope(Event,do_nothing).
do_nothing(_,_).
add_new_event_in_error_scope(Event,PrintMessageCall) :-
   get_counter(cur_block_level,L),
   (specific_event_occurred_at_level(L,Event) -> fail
     ; (throw_error(L,Event,Span)
         -> (debug_mode(off) -> true ; print_message(informational,throwing(Event))), % trace, % <---
            call(PrintMessageCall,throwing,Span),
            % Note: event is not asserted ! See e.g. tests 1089, 1519, 1522, 1661, 1199, 1291, ...
            throw(Event)
         ;  assert(specific_event_occurred_at_level(L,Event)),
            call(PrintMessageCall,not_throwing,unknown)
       )
   ).

% call if you just want to add the error information; without throwing
assert_event_in_error_scope_without_throw(Event) :- get_counter(cur_block_level,L),
   (specific_event_occurred_at_level(L,Event) -> true
    ; assert(specific_event_occurred_at_level(L,Event))).

error_occurred_in_error_scope :- get_counter(cur_block_level,L),error_occurred_at_level(L).
event_occurred_in_error_scope(Event) :- get_counter(cur_block_level,L),specific_event_occurred_at_level(L,Event).

copy_error_scope_events(FromLevel,ToLevel) :- retract(specific_event_occurred_at_level(FromLevel,Event)),
   \+ specific_event_occurred_at_level(ToLevel,Event),
   % print(copying_event(FromLevel,ToLevel,Event)),nl,
   assert(specific_event_occurred_at_level(ToLevel,Event)), % should we throw if this level demands it ?
   (throw_error(ToLevel,Event,_Span)
    -> debug_println(20,throw_copied_event(Event)), throw(Event)),
   fail.
copy_error_scope_events(_,_).

clear_events_in_current_error_scope(Event) :- get_counter(cur_block_level,L),clear_events_in_error_scope(L,Event).

%clear_events_in_error_scope(L,Event) :-retract(specific_event_occurred_at_level(L,Event)), print(retract(L,Event)),nl,fail.
clear_events_in_error_scope(L,Event) :-
   retractall(specific_event_occurred_at_level(L,Event)),
   (specific_event_occurred_at_level(L,_) -> true
    ; retractall(error_occurred_at_level(L))).

throw_enumeration_warnings_in_current_scope :-
   get_counter(cur_block_level,L), throw_enumeration_warnings_in_scope(L,_,unknown).

throw_enumeration_warnings_in_scope(Level,Critical,Span) :-
   simplify_span(Span,SimplifiedSpan), % relevant for test 1562
   throw_errors_in_scope(Level,enumeration_warning(_,_,_,_,Critical),SimplifiedSpan).

on_enumeration_warning(Call, BackupCall) :-
        on_enumeration_warning_with_continue(Call, BackupCall, true).
% sets up a new error scope and forces enumeration_warnings to be thrown
% if an enumeration_warning happens the BackupCall is executed
on_enumeration_warning_with_continue(Call, BackupCall, TrueContinuation) :-
    enter_new_error_scope(Level,on_enumeration_warning_with_continue),
	throw_enumeration_warnings_in_scope(Level,_,unknown),
	on_exception(enumeration_warning(_,_,_,_,_),
		    if(Call,
		         (exit_error_scope(Level,_,on_enumeration_warning_with_continue_true),call(TrueContinuation)),
		         (exit_error_scope(Level,_,on_enumeration_warning_with_continue_false),fail)
		       ),
		    (exit_error_scope(Level,_,on_enumeration_warning_with_continue_exception),BackupCall)).

catch_enumeration_warning_exceptions(Call, BackupCall) :-
    catch_enumeration_warning_exceptions(Call,BackupCall,false,unknown).
catch_enumeration_warning_exceptions(Call, BackupCall, AddAsEventError) :-
    catch_enumeration_warning_exceptions(Call,BackupCall,AddAsEventError,unknown).



% AddAsEventError = true means that caught exceptions are added as event_errors in scope
% (this can be important if we later want to decide whether an enumeration warning occured)
catch_enumeration_warning_exceptions(Call, BackupCall,AddAsEventError,Source) :-
    enter_new_error_scope(Level,catch_enumeration_warning_exceptions),
	% difference: we do not force to throw_errors_in_scope
	on_exception(enumeration_warning(A,B,C,D,E),
		         call_cleanup(Call, exit_error_scope(Level,_,catch_enumeration_warning_exceptions)),
		         % call_cleanup can be tricky in case of success of Call: the error scope may not be exited immediately; TO DO: maybe rewrite using same style as above (using if + on_exception)
		         ((AddAsEventError=false
		           -> (Source=ignore(Src)
		                -> formatsilentmsg(user_output,'%%% Ignoring unbounded enumeration warning (source:~w).~n',[Src]) ; true)
		                   % inform that enum warning was not serious: there is a fall-back solution
		            ; assert_event_in_error_scope_without_throw(enumeration_warning(A,B,C,D,E))),
		          BackupCall)).

% execute some code when enumeration warning occurs; and then throw exception again
observe_enumeration_warnings(Call,ObserverCallToExecute) :-
	on_exception(enumeration_warning(A,B,C,D,E),Call,
	   call_cleanup(ObserverCallToExecute, throw(enumeration_warning(A,B,C,D,E)))).


% sets up a new error scope and forces enumeration_warnings to be thrown
% the exceptions are not caught !
call_with_enumeration_warning(Call) :-
    enter_new_error_scope(Level,call_with_enumeration_warning),
	throw_enumeration_warnings_in_scope(Level,_,unknown),
	on_exception(E,
		    if(Call,
		         exit_error_scope(Level,_,call_with_enumeration_warning_true),
		         (exit_error_scope(Level,_,call_with_enumeration_warning_false),fail)
		       ),
		    (exit_error_scope(Level,_,call_with_enumeration_warning_exception),throw(E))).


% a version of time_out which treats enumeration_warnings like time-outs
% and also checks enumeration_warnings in case of failure
% assumes we only look for one solution (otherwise error scope cleanup has to be changed !! )
time_out_with_enum_warning_one_solution(Call,TO,TimeOutRes) :-
    enter_new_error_scope(Level,time_out_with_enum_warning_one_solution),
    %% print(time_out_with_enum_warning_one_solution(TO)),nl, %%
    call_cleanup(
        on_exception(enumeration_warning(A,B,C,D,E),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear,perform_cut), % no need to clear: we leave error_scope below
                 %% (print(enum_warning),nl,TimeOutRes=time_out)). %%
                 TimeOutRes=virtual_time_out(thrown_enumeration_warning(A,B,C,D,E))),
        exit_error_scope(Level,_,time_out_with_enum_warning_one_solution)).

enumeration_warning_occured_in_error_scope :-
   event_occurred_in_error_scope(enumeration_warning(_,_,_,_,_)).
abort_error_occured_in_error_scope :-
   event_occurred_in_error_scope(abort_error(_)).

% just like above; but does not set up new error_scope (for performance) and clears enumeration warnings in case of failure and success
time_out_with_enum_warning_one_solution_no_new_error_scope(Call,TO,TimeOutRes) :-
    (enumeration_warning_occured_in_error_scope
     -> %trace,
        add_internal_error('Error block already has enumeration warnings',time_out_with_enum_warning_one_solution_no_new_error_scope(Call,TO,TimeOutRes))
      ; true),
    on_exception(enumeration_warning(A,B,C,D,E),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,clear_errors,perform_cut), % clear errors so we can continue using same error scope
                 %% (print(enum_warning),nl,TimeOutRes=time_out)). %%
                 TimeOutRes=virtual_time_out(thrown_enumeration_warning(A,B,C,D,E))).


% check if a result is a time_out:
is_time_out_result(time_out).
is_time_out_result(virtual_time_out(_)).

% a version which is to be used not for a single call but for backtracking
% generate virtual time_out at the end if enumeration warning occured

time_out_with_enum_warning_for_findall(Call,TO,TimeOutRes) :-
    enter_new_error_scope(Level,time_out_with_enum_warning_for_findall),
    time_out_with_enum_warning_for_findall2(Level,Call,TO,TimeOutRes).

time_out_with_enum_warning_for_findall2(_Level,Call,TO,TimeOutRes) :- %print(enter(_Level)),nl,
    on_exception(enumeration_warning(Cause,_,_,_,_),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear,no_cut),
                 get_virtual_time_out_res(Cause,TimeOutRes)). % virtual time_out
time_out_with_enum_warning_for_findall2(Level,_Call,_TO,_TimeOutRes) :- % print(exit(Level)),nl,
    clear_events_in_error_scope(Level,enumeration_warning(_,_,_,_,_)),
    exit_error_scope(Level,_,time_out_with_enum_warning_for_findall2),fail.

get_virtual_time_out_res(kodkod_timeout,Res) :- !, Res= virtual_time_out(kodkod_timeout).
get_virtual_time_out_res(_,virtual_time_out(exception)).

% a version of time_out_with_enum_warning_for_findall where you manually have to enter_new_error_scope before
% and exit_error_scope afterwards; useful if you repeatedly call this predicate and want to set up the block only once
time_out_with_enum_warning_for_findall_in_current_error_scope(_Level,Call,TO,TimeOutRes) :- %print(enter(_Level)),nl,
    on_exception(enumeration_warning(Cause,_,_,_,_),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear,no_cut),
                 get_virtual_time_out_res(Cause,TimeOutRes)). % virtual time_out
time_out_with_enum_warning_for_findall_in_current_error_scope(Level,_Call,_TO,_TimeOutRes) :-
    % does not exit level
    clear_events_in_error_scope(Level,enumeration_warning(_,_,_,_,_)),fail.

:- use_module(tools_meta,[safe_time_out/3]).
% a version of time_out which upon failure checks if enumeration warnings were generated
time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,ClearErrors,CutAfterFirstSol) :-
    safe_time_out(Call,TO,TimeOutRes),  % a safe version which checks that time-out value TO is not too large
    (TimeOutRes = time_out -> !  %, print(real_timeout),nl
       ; (ClearErrors=clear_errors -> clear_enumeration_warnings ; true),
         (CutAfterFirstSol=perform_cut -> ! ; true)
         %,print(sol),nl
    ).
time_out_check_enum_warning_for_failure(_Call,_TO,TimeOutRes,ClearErrors,_) :- % print(failure),nl,
    event_occurred_in_error_scope(enumeration_warning(A,B,C,D,critical)), % print(virtual),nl,
    TimeOutRes = virtual_time_out(failure_enumeration_warning(A,B,C,D,critical)), % we have not investigated all possibilities; generate virtual time_out
    (ClearErrors=clear_errors -> clear_enumeration_warnings ; true).

critical_enumeration_warning_occured_in_error_scope :-
   critical_enumeration_warning_occured_in_error_scope(_,_,_,_).
critical_enumeration_warning_occured_in_error_scope(A,B,C,D) :-
    event_occurred_in_error_scope(enumeration_warning(A,B,C,D,critical)).

clear_enumeration_warnings :-
    clear_events_in_current_error_scope(enumeration_warning(_,_,_,_,_)).

% retract events and raise real warnings
% Src and Msg used if these are not enumeration warnings
translate_events_in_current_scope_to_warnings(Src,Msg) :-
     get_counter(cur_block_level,L),
     retract(specific_event_occurred_at_level(L,Event)),
     (translate_enumeration_warning(Event,EMsg,Span)
       -> add_warning(enumeration_warning,EMsg,'',Span)
       ;  add_warning(Src,Msg,Event)),
     fail.
translate_events_in_current_scope_to_warnings(_,_).

% see also: print_enum_warning
translate_enumeration_warning(enumeration_warning(Enumerating,Type,Range,RestrictedRange,_Throwing), Msg,Span) :-
     translate_enumeration_info(Enumerating,Info,Span),
     ajoin(['Unbounded enumeration of ',Info,':',Type,', narrowing range: ',Range,' ---> ',RestrictedRange],Msg).
     % Throwing=throwing -> print_pending_abort_info ?

translate_enumeration_info(enumerating(Info),Res,Span) :- !, translate_enumeration_info(Info,Res,Span). % TO DO: call get_trigger_info
translate_enumeration_info(b(identifier(Id),_,Info),Res,Span) :- !, Res=Id,Span=Info.
translate_enumeration_info(I,I,I).

% -----------------------

:- volatile logged_error_hash/5, real_error_occurred/0, logged_error_spans/2.
:- dynamic logged_error_hash/5.
:- dynamic real_error_occurred/0.
:- dynamic logged_error_spans/2.


reset_real_error_occurred :- retractall(real_error_occurred).
reset_errors :- %print(reset_errors),nl, trace,
                reset_errors_but_not_scopes,
                reset_error_limit_reached,
                reset_error_scopes. % call if you want to start again at a fresh top-level, without error scopes
reset_errors_but_not_scopes :-
                reset_real_error_occurred,
                retractall(logged_error_hash(_,_,_,_,_)),retractall(logged_error_spans(_,_)),
                clear_error_context.

reset_error_span_if_required :-
	(logged_error_hash(_,_,_,_,_) -> true
		; retractall(logged_error_spans(_,_))).

reset_error_spans :- retractall(logged_error_spans(_,_)).


clear_wd_errors :-
   clear_errors_with_source(well_definedness_error).
clear_errors_with_source(Source) :-
   retractall(logged_error_spans(Source,_)),
   retractall(logged_error_hash(_,Source,_S1,_TContext1,_Span1)),
   (logged_error_hash(_,_Source2,_S2,_TContext2,_Span2) -> true
    ; reset_real_error_occurred).

print_error_span(Span) :- print_error_span(Span,error).
print_error_span(unknown,_) :- !.
print_error_span(Span,Kind) :-
     extract_span_description(Span,PosMsg), !,
     print_error(Kind,PosMsg).
print_error_span(b(E,T,I),Kind) :- translate:translate_subst_or_bexpr_with_limit(b(E,T,I),1000,S),!,
     print_error(Kind,S).
print_error_span(Span,Kind) :- print_error(Kind,Span). % span not recognized !

% the same, but not using user_error and red and only printing when span recognised
% more detailed than translate:print_span
print_message_span(unknown) :- !.
print_message_span(Span) :-
     extract_span_description(Span,PosMsg), !, print(PosMsg).
print_message_span(_). % span not recognized !

% try and extract a string (aka atom) describing the position
extract_span_description(Span,PosMsg) :-
   extract_line_col(Span,StartLine,StartCol,EndLine,EndCol),!,
   POS = ['Line: ',StartLine,' Column: ',StartCol|POS1],
   POS1 = [' until Line: ',EndLine,' Column: ',EndCol|POS2],
   (extract_file_name(Span,Filename),
    Filename \=null -> % Filename = null occurs in eval_strings !?
      POS2 = [' in file: ',Filename|POS3] ; POS2 = POS3),
   (extract_additional_description(Span,ContextMsg)
     -> POS3 = [' within ',ContextMsg] ; POS3=[]),
   ajoin(POS,PosMsg).
extract_span_description(Span,PosMsg) :-
   extract_symbolic_pos(Span,Symbolic),!,
   % TO DO: we could try and find section for Rodin
   ajoin(['@label: ',Symbolic],PosMsg).

assert_true(Code) :-
  (call(Code) -> true ; add_internal_error('Assertion failed: ',Code)).

add_error_fail(Source,ErrMsg, ErrTerm) :- add_error_and_fail(Source,ErrMsg, ErrTerm).
add_error_and_fail(Source,ErrMsg, ErrTerm) :- add_error(Source,ErrMsg, ErrTerm),fail.
add_error_and_fail(Source,ErrMsg) :- add_error(Source,ErrMsg),fail.

% add_error/2, add_error/3 and add_error/4
add_error(Source,ErrMsg) :- add_error4(Source,ErrMsg, '',unknown).
add_error(Source,ErrMsg, ErrTerm) :- add_error4(Source,ErrMsg, ErrTerm,unknown).
add_error(Source,ErrMsg,ErrTerm,Span) :- add_error4(Source,ErrMsg,ErrTerm,Span).

add_error4(Source,ErrMsg,ErrTerm,Span) :-
    assert_real_error_occurred(Source),
    combine_msg_and_error_term(ErrMsg,ErrTerm,S),
    add_error_or_warning(Source,S,Span).

%:- use_module(translate, [translate_error_term/2, translate_context/2]).
combine_msg_and_error_term(Msg,'',Res) :- !, Res=Msg.
combine_msg_and_error_term(ErrMsg,ErrTerm,S) :-
    string_concatenate(ErrMsg,' ',ST),
	  safe_translate_error_term(ErrTerm,RString),
    string_concatenate(ST,RString,S).


:- dynamic show_error_source/0.
show_error_source. % True if error source and span to be printed
turn_show_error_source_on :- (show_error_source -> true ; assert(show_error_source)).
turn_show_error_source_off :-  retractall(show_error_source).

trace_if_user_wants_it :- (preferences:get_preference(trace_upon_error,true) -> safe_trace ; true).

% try Prolog's trace; if this fails (compiled code), then ask user for input.
safe_trace :- translate:install_b_portray_hook,
  on_exception(error(existence_error(_,_),_), trace, type_to_continue).

:- dynamic do_not_interrupt_user_anymore/0.

type_to_continue :- do_not_interrupt_user_anymore,!.
type_to_continue :- format(user_output,'(type <RETURN> to continue, z<RETURN> to zip) ===> ',[]),
   read_line(user_input,X),
   (zip(X) -> assert(do_not_interrupt_user_anymore) ; true).

zip([122|_]). %starts with z


print_error_source(well_definedness_error) :- !,print_error('A well-definedness error occured !').
print_error_source(warning(Source)) :- show_error_source,!,
    format_error('! A warning occurred (source: ~w) !~n',[Source]).
print_error_source(internal_error(Source)) :- show_error_source,!,
    format_error('! An internal error occurred (source: ~w) !~n',[Source]).
print_error_source(Source) :- show_error_source,!,
    format_error('! An error occurred (source: ~w) !~n',[Source]).
print_error_source(_).

print_error_context(Kind,Context) :- ((Context = '' ; \+show_error_source) -> true ; print_error(Kind,Context)).

print_error(message,Term) :- !, print_message_on_output([blue],Term).
print_error(_,Term) :- print_error(Term).

safe_translate_error_term(ErrTerm,RString) :-
    temporary_set_preference(expand_avl_upto,5,CHNG),
    translate:translate_error_term(ErrTerm,RString),
    reset_temporary_preference(expand_avl_upto,CHNG).

safe_translate_context(Result) :-
% avoid infinite loop when error occurs inside translate_context
    current_context(Context),
    !,
    set_error_context(translate_context),
    temporary_set_preference(expand_avl_upto,4,CHNG),
    (translate:translate_context(Context,TContext)
      -> true ; TContext=Context),
    reset_temporary_preference(expand_avl_upto,CHNG),
    restore_error_context(Context),
    Result=TContext.
safe_translate_context(''). % no context


add_error_or_warning(Source,ErrMsg,Span) :-
    safe_translate_context(Context),
    reset_error_span_if_required,
    simplify_span(Span,SSpan),  % for better hashing, assert, retract
    assertz_logged_error_if_new(Source,ErrMsg,Context,SSpan,New),
    log_error(Source,ErrMsg,SSpan),
    (New=false -> true
      ; optional_error(Source) ->
        (extract_span_description(SSpan,PosMsg) -> true ; PosMsg=''),
        formatsilentmsg(user_output,'% Optional error ~w occured: ~w ~w~n',[Source,ErrMsg,PosMsg]),
        register_error_span(Source,SSpan,_)
      ; register_error_span(Source,SSpan,NewSrcSpan),
        (NewSrcSpan=false, error_limit_reached
          -> true % this is not a new Source,SSpan combination and we have already several errors
          ;  print_error_source(Source),
             print_error(ErrMsg),
             print_error_context(error,Context),
             print_error_span(SSpan)
        ),
        trace_if_user_wants_it
    ).

% :- use_module(tools,[b_string_escape_codes/2]).
% optionally log errors in to a ndsjon logfile
log_error(Source,ErrMsg,Span) :- get_preference(error_log_file,LogFile),
      LogFile \= '',
      extract_position_from_span(Span,Filename,StartLine,StartCol,EndLine,EndCol),
      decompose_error_source(Source,Type,Src),
      % TO DO: escape Filename and ErrMsg
      atom_codes(ErrMsg,EMC),tools:b_string_escape_codes(EMC,EscErrMsgC),
      atom_codes(Filename,EFC),tools:b_string_escape_codes(EFC,EscFileC),
      on_exception(E,intelligent_open(LogFile,append,S),
                     (format(user_error,'*** Cannot open error log file: ~w~n*** Exception: ~w~n',[LogFile,E]),fail)),
      !, %ndjson format
      format(S,'{"~w": {"message": "~s", type: "~w", "file": "~s", "start": {"line": ~w, "col": ~w}, "end": {"line": ~w, "col": ~w} }}~n',
         [Type,EscErrMsgC,Src,EscFileC,StartLine,StartCol,EndLine,EndCol]),
      close(S).
log_error(_,_,_).

intelligent_open(user_error,Mode,Stream) :- (Mode=append ; Mode=write), !, Stream=user_error.
intelligent_open(stderr,Mode,Stream) :- (Mode=append ; Mode=write), !, Stream=user_error.
intelligent_open(user_output,Mode,Stream) :- (Mode=append ; Mode=write), !, Stream=user_output.
intelligent_open(stdout,Mode,Stream) :- (Mode=append ; Mode=write), !, Stream=user_output.
intelligent_open(File,Mode,Stream) :- open(File,Mode,Stream).

decompose_error_source(warning(Source),warning,Res) :- !, Res=Source.
decompose_error_source(internal_error(Source),internal_error,Res) :- !, Res=Source.
decompose_error_source(Msg,error,Msg).
:- dynamic total_number_of_errors/1.
total_number_of_errors(0).

% the total number of errors since start; no way to reset this; useful for validation reports
% TO DO: maybe use counter extension; but tricky to get dependency and initialisation right
get_total_number_of_errors(X) :-
  (total_number_of_errors(X) -> true
   ; format(user_error,'No total_number_of_errors/1 fact!~n',[]), X=0).

:- dynamic error_limit_reached_printed/0.
:- dynamic total_number_of_errors_at_last_reset/1.
total_number_of_errors_at_last_reset(0).

reset_error_limit_reached :-
                retractall(error_limit_reached_printed),
                retractall(total_number_of_errors_at_last_reset(_)),
                get_total_number_of_errors(T),
                assert(total_number_of_errors_at_last_reset(T)).

error_limit_reached :- error_limit_reached_printed,!.
error_limit_reached :-  get_total_number_of_errors(Nr), Nr >= 100,
   total_number_of_errors_at_last_reset(Last), Nr-Last >= 100,
   print_error('Error limit reached; only printing new error source and locations!'),
   assert(error_limit_reached_printed).

assert_warning_occured :-
    (preferences:get_preference(strict_raise_warnings,true)
      -> assert_real_error_occurred(warning) ; true).
assert_real_error_occurred(Source) :- %print(assert_real_error_occurred(Source)),nl, trace,
    (real_error_occurred -> true ; assert(real_error_occurred)),
    (retract(total_number_of_errors(X))
     -> X1 is X+1, assert(total_number_of_errors(X1))
      ; assert(total_number_of_errors(1))),
    assert_real_error_in_current_block(Source).

:- use_module(library(terms),[term_hash/3]).
assertz_logged_error_if_new(Source,S,TContext,Span,New) :-
    % limit depth mainly for TContext/Span; avoiding generating many similar error messages with same msg
    % however, for WD errors often the msg varies (e.g. argument for function application); so this will not prevent a lot of errors
    term_hash(logged_error_hash(Source,S,TContext,Span),[if_var(ignore),depth(4)],Hash),
    (logged_error_hash(Hash,Source,S,TContext,Span)
     -> debug_println(repeated_error(Hash)),New=false
      ; %print(new(Hash,Source,S,TContext,Span)),nl,
        assertz(logged_error_hash(Hash,Source,S,TContext,Span)),
        New=true).

logged_error(Source,S,TContext,Span) :- logged_error_hash(_Hash,Source,S,TContext,Span).

add_warning(Source,ErrMsg, ErrTerm) :-
   add_warning(Source,ErrMsg, ErrTerm,unknown).
add_warning(Source,ErrMsg, ErrTerm,Span) :-
    safe_translate_error_term(ErrTerm,RString),
    string_concatenate(ErrMsg,RString,S),
    assert_warning_occured,
    add_error_or_warning(warning(Source),S,Span).
add_warning(Source,Message) :-
    assert_warning_occured,
    add_error_or_warning(warning(Source),Message,unknown).

add_message(Source,ErrMsg, ErrTerm,Span) :- /* same as add_error but does not add err msg */
    safe_translate_error_term(ErrTerm,RString),
    string_concatenate(ErrMsg,RString,S),
    add_message(Source,S),
    (Span=unknown -> true
          ; safe_translate_context(Context),
            print_error_context(message,Context),
            print_error_span(Span,message)).

add_message(Source,ErrMsg, ErrTerm) :- /* same as add_error but does not add err msg */
    add_message(Source,ErrMsg, ErrTerm,unknown).

add_message(Source,ErrMsg) :-
   (show_error_source -> format_with_colour(user_output,[blue],'! Message (source: ~w):~n',[Source]) ; true),
   print_error(message,ErrMsg).
   %assertz(logged_error(Source,ErrMsg,_,_)).

% adds a list of errors with position information
% [error(Msg,Pos)] or [warning(Msg,Pos)...]
no_real_perror_occurred([]).
no_real_perror_occurred([warning(_,_)|T]) :- no_real_perror_occurred(T).

add_all_perrors(Errors) :- add_all_perrors(Errors,[],bmachine).
add_all_perrors(Errors,Files) :- add_all_perrors(Errors,Files,bmachine).
add_all_perrors(Errors,Files,Context) :- add_all_perrors(Errors,Files,Context,0,100).
add_all_perrors([],_,_,_,_).
add_all_perrors([H|Rest],_Filenames,ContextOfError,NrOfErrors,MaxErrors) :-
    NrOfErrors >= MaxErrors,!,
    length([H|Rest],Len),
    add_error(ContextOfError,'Reached maximium number of errors; number of errors not reported: ',Len).
add_all_perrors([Error|Rest],Filenames,ContextOfError,NrOfErrors,MaxErrors) :-
    get_perror(Error,ErrOrWarn,Msg,Pos),
    extract_perror_pos(Pos,Filenames,NewPos),
    (ErrOrWarn == warning -> add_warning(ContextOfError,Msg,'',NewPos)
      ; ErrOrWarn = internal_error -> add_internal_error(ContextOfError,Msg,'',NewPos)
      ; add_error(ContextOfError,Msg,'',NewPos)),
    N1 is NrOfErrors+1,
    add_all_perrors(Rest,Filenames,ContextOfError,N1,MaxErrors).

extract_perror_pos(pos(_,Filenumber,Srow,Scol,Erow,Ecol),Filenames,POS) :-
    get_filename(Filenumber,Filenames,Filename),!,
    %Ecol1 is Ecol+1,
    POS = src_span(Filenumber,Filename,Srow,Scol,Erow,Ecol,[],[]).
extract_perror_pos(pos(_,_Filenumber,Srow,Scol,Erow,Ecol),_,POS) :- !,
    %Ecol1 is Ecol+1,
    POS = src_span(Srow,Scol,Erow,Ecol,[],[]).
extract_perror_pos(pos(Srow,Scol,Filename),_,POS) :- !, % deprecated ?
    S1 is Scol+0, % zero width span
    POS = src_position_with_filename_and_ec(Srow,Scol,Srow,S1,Filename).
extract_perror_pos(pos(Srow,Scol,Erow,Ecol,Filename),_,POS) :- !,
    %Ecol1 is Ecol+1,
    POS = src_position_with_filename_and_ec(Srow,Scol,Erow,Ecol,Filename).
extract_perror_pos(pos(Srow,Scol),_,POS) :- !, print(deprecated_pos(Srow,Scol)),nl,
    POS = src_position(Srow,Scol,none,1).
extract_perror_pos(Pos,_,Pos).


get_perror(error(Msg,Pos),error,Msg,Pos) :- !.
get_perror(internal_error(Msg,Pos),internal_error,Msg,Pos) :- !.
get_perror(warning(Msg,Pos),warning,Msg,Pos) :- !.
get_perror(M,_,M,unknown) :- format(user_error,'!!! unknown perror: ~w~n',[M]).

get_filename(Filenumber,Filenames,Filename) :-
	number(Filenumber),!,
	nth1(Filenumber,Filenames,Filename).
get_filename(N,F,R) :-
   add_internal_error('Cannot get filename: ',get_filename(N,F,R)),
   N=1,F=[R|_]. % return first file if possible

% check if error/warning occured and retract matching error/warning
get_error_with_span(Source,ErrMsg,Span) :-
   retract(logged_error_hash(_Hash,Source,ErrMsg,_Context,Span)).
get_error(Source,ErrMsg) :-
   retract(logged_error_hash(_,Source,ErrMsg,_,_)).
get_warning(Source,ErrMsg) :-
   retract(logged_error_hash(_,warning(Source),ErrMsg,_,_)).

% now the same without actually retracting the error/warning:
check_error_occured(Source,ErrMsg) :- logged_error(Source,ErrMsg,_,_).
check_warning_occured(Source,ErrMsg) :- logged_error(warning(Source),ErrMsg,_,_).

get_error(Msg) :-
    retract(logged_error_hash(_Hash,Source,ErrMsg,Context,Span)),
    (extract_line_col(Span,StartLine,StartCol,EndLine,EndCol)
      -> POS = ['\n ### Line: ',StartLine,', Column: ',StartCol|POS1],
         ((StartLine,StartCol) \= (EndLine,EndCol)
          -> (StartLine=EndLine -> POS1 = [' until ',EndCol|POS2]
               ; POS1 = ['\n ### until Line: ',EndLine,', Column: ',EndCol|POS2]
             )
          ;  POS1 = POS2
         ),
         (extract_additional_description(Span,CMsg)
           -> POS2 = ['\n ### within: ',CMsg,'\n\n']
           ;  POS2 = ['\n\n']
         ),
         (extract_file_name(Span,Filename)
              -> PosMsg = ['\n ### File: ',Filename | POS]
               ; PosMsg = POS
         ),
         tools:ajoin(PosMsg,EndStr)
      ;  EndStr = '\n\n'
     ),
    ((Context='' ; Context = '???')
      ->  string_concatenate(ErrMsg,EndStr,S0)
       ;  safe_translate_error_term(Context,CString),
          string_concatenate(CString,EndStr,CString2),
          string_concatenate(':  Context = ',CString2,CString3),
          string_concatenate(ErrMsg,CString3,S0)
     ),
    (debug_mode(off) -> Msg = S0
      ;  decompose_error_source(Source,_,SrcAtom),
         string_concatenate(SrcAtom,': ',S1),
         string_concatenate(S1,S0,Msg)
     ).



get_error_with_span_info(ErrMsg,Type,ErrLocations) :-
    retract(logged_error_hash(_Hash,Source,ErrMsg,_Context,Span)), % should we integrate source,context ?
    (Source = warning(_) -> Type=warning
     ; Source = internal_error(_) -> Type=internal_error
     ; Type=error),
    findall( error_span(Filename,StartLine,StartCol,EndLine,EndCol),
       extract_position_from_span(Span,Filename,StartLine,StartCol,EndLine,EndCol),
       ErrLocations).

extract_position_from_span(Span,Filename,StartLine,StartCol,EndLine,EndCol) :-
    extract_line_col(Span,StartLine,StartCol,EndLine,EndCol),
    (extract_file_name(Span,FName) -> Filename=FName ; Filename='').

% Tcl/Tk interface of SICStus cannot deal with certain strange Unicode symbols
% see prob_examples/public_examples/B/Puzzles/LogicQuestion.mch where code 55349 appears
tk_escape(Atom,EAtom) :- atom(Atom),!,
    atom_codes(Atom,C),
    maplist(tk_escape_code,C,EC),
    atom_codes(EAtom,EC).
tk_escape(list(L),list(EL)) :- !, maplist(tk_escape,L,EL).
tk_escape(X,X).

tk_escape_code(StrangeUnicode,Res) :-
    StrangeUnicode > 55000,!, % not sure what the limit should be here
    Res = 9633. % white box
tk_escape_code(Code,Code).

tcltk_get_all_errors(Limit,list(ERes),NrErrorsShown,NrErrors) :-
    get_all_errors(All),
    length(All,NrErrors),
    (NrErrors =< Limit -> Res = All, NrErrorsShown = NrErrors
     ; prefix_length(All,Res,Limit), % truncate
       NrErrorsShown = Limit
    ),
    tk_escape(list(Res),list(ERes)).


get_all_errors(All) :-
    findall(Err,get_error(Err),All),
    All \= [], clear_error_context.

get_all_errors_and_clear(All) :- findall(Err,get_error(Err),All),reset_errors_but_not_scopes,All \= [].
get_all_errors_and_reset(All) :- findall(Err,get_error(Err),All),reset_errors,All \= [].
get_all_errors_with_span_info_and_reset(All) :-
    findall(error(ErrMsg,ErrType,ErrLocations),get_error_with_span_info(ErrMsg,ErrType,ErrLocations),All),
    reset_errors.

error_or_warning_occured :- logged_error(_Source,_,_,_),!.
test_error_occurred(Source) :- logged_error(Source,_,_,_),real_error_occurred,!.

%:- use_module(library(lists),[length/2]). length is builtin

count_errors_occurred(Source,NrOfErrors) :-
  findall(1,logged_error(Source,_,_,_),Ls),
  length(Ls,NrOfErrors).

display_error_statistics :- findall(Source,logged_error(Source,_,_,_),L),
  count_occurences(L,Occ),
  format('Error manager summary: ~w~n',[Occ]).



:- volatile current_context/1.
:- dynamic current_context/1.

/* add context information; will be added to any error messages */
/* will be removed upon backtracking */
%push_error_context(_Ctx) :- !.
set_error_context(Ctx) :- % print(ctx(Ctx)),nl,
  retractall(current_context(_)),
  assert(current_context(Ctx)).
%push_error_context(_Ctx) :-
%  (retract(current_context(_Ctx)) -> true ; print_error(could_not_retract_context(_Ctx))).

clear_error_context :- %print(clearing_ctx),nl,
   retractall(current_context(_)).

%pop_error_context :-
%  (retract(current_context(_Ctx)) -> true ; print_error(could_not_pop_context(_Ctx))).

get_error_context(C) :- (current_context(C) -> true ; C='???').

% restore an error context obtained by get_error_context
restore_error_context('???') :- !, clear_error_context.
restore_error_context(C) :- set_error_context(C).

get_error_span(Source,Span) :-
   retract(logged_error_spans(Source,Span)).

% a way for external modules to register error locations, e.g., for display by Tcl/Tk HighlightSyntaxErrors
register_error_span(Source,Span) :- register_error_span(Source,Span,_).
register_error_span(_,unknown,New) :- !,New=false.
register_error_span(Source,Span,New) :- logged_error_spans(Source,Span),!,New=false.
register_error_span(Source,Span,New) :- assert(logged_error_spans(Source,Span)),New=true.

get_error_span_linecol(Line,Col,EndLine,EndCol) :- get_error_span_linecol(_Source,Line,Col,EndLine,EndCol).
get_error_span_linecol(Source,Line,Col,EndLine,EndCol) :-
    get_error_span(Source,Span),  % looks at logged_error_spans and retracts
    extract_line_col(Span,Line,Col,EndLine,EndCol).

get_error_span_linecol_for_main_file(Line,Col,EndLine,EndCol) :-
    get_error_span(_Source,Span),  % looks at logged_error_spans and retracts even if not in main file !
    extract_line_col_for_main_file(Span,Line,Col,EndLine,EndCol).

% just like get_error_span_linecol but does not retract and gets Filename
check_error_span_file_linecol(Source, FullFilename,Line,Col,EndLine,EndCol) :-
    logged_error_spans(Source,Span),
    extract_file_line_col(Span,FullFilename,Line,Col,EndLine,EndCol).

extract_file_line_col(Span,FullFilename,Line,Col,EndLine,EndCol) :-
    (extract_file_number_and_name(Span,_,FullFilename) -> true ; FullFilename = unknown),
    extract_line_col(Span,Line,Col,EndLine,EndCol).

% try and extract symbolic (Rodin label) information
extract_symbolic_pos([],_) :- !,fail.
extract_symbolic_pos([H|T],SymbolicPos) :- !,
   posinfo_member(N,[H|T]),
   extract_symbolic_pos(N,SymbolicPos),!.
extract_symbolic_pos(rodinpos(Label,_),Label).  % see also get_texpr_labels in bsyntaxtree
extract_symbolic_pos(rodinpos(Model,Name,_),Label) :- ajoin([Model,':',Name],Label).
extract_symbolic_pos(label(Label),Label).

% quickly try and simplify a span, e.g., before asserting it in a fact (as it may become big)
simplify_span(b(E,_,Infos),Res) :- !,
   (definite_posinfo_member(N,Infos) -> Res=N
   ; first_sub_expr(E,A) -> simplify_span(A,Res)
   ; otherwise -> Res=unknown).
simplify_span(span_predicate(P,_,_),Res) :- !, simplify_span(P,Res).
simplify_span(S,S).

% choose first sub-expression in case top-expression has no position info (it would be better if extract_info/3 in bsyntaxtree merges position info)
first_sub_expr(negation(A,_),A).
first_sub_expr(conjunct(A,_),A).
first_sub_expr(disjunct(A,_),A).
first_sub_expr(implication(A,_),A).
first_sub_expr(equivalence(A,_),A).
first_sub_expr(exists([A|_],_),A).
first_sub_expr(forall([A|_],_),A).
first_sub_expr(let_predicate([A|_],_,_),A).
first_sub_expr(let_expression([A|_],_,_),A).
first_sub_expr(let_expression_global([A|_],_,_),A).
first_sub_expr(sequence([A|_]),A).
first_sub_expr(parallel([A|_]),A).
first_sub_expr(assign([A|_],_),A).
first_sub_expr(var([A|_],_),A).

%extract_line_col(S,_,_,_,_) :- print(extract(S)),nl,fail.
extract_line_col([],_,_,_,_) :- !,fail.
extract_line_col([H|T],Srow,Scol,Erow,Ecol) :- !,
   posinfo_member(N,[H|T]),
   extract_line_col(N,Srow,Scol,Erow,Ecol),!.
extract_line_col(closure(_,_,B),Srow,Scol,Erow,Ecol) :- !,
   extract_line_col(B,Srow,Scol,Erow,Ecol).
extract_line_col(b(E,_,Infos),Srow,Scol,Erow,Ecol) :- !,
   (extract_line_col(Infos,Srow,Scol,Erow,Ecol) -> true
     ; first_sub_expr(E,A) -> extract_line_col(A,Srow,Scol,Erow,Ecol)).
extract_line_col(span_predicate(P,_,_),Srow,Scol,Erow,Ecol) :- !,
  extract_line_col(P,Srow,Scol,Erow,Ecol).
extract_line_col(pos(_,_Filenumber,Srow,Scol,Erow,Ecol),Srow,S1,Erow,E1) :-  number(Srow),
  % _Filenumber is an index into list provided by b_get_all_used_filenames
  number(Scol), number(Erow), number(Ecol),!,
  (Scol>0 -> S1 is Scol-1 ; S1 = Scol),  % the BParser starts column numbering at 1
  (Ecol>0 -> E1 is Ecol-1 ; E1 = Ecol).
  %E1=Ecol. %Somehow the Ecol info is ok for Tcl
extract_line_col(unknown,_,_,_,_) :- !,fail.
extract_line_col(pos_context(A,_,_),StartLine,StartCol,EndLine,EndCol) :- !,
   extract_line_col(A,StartLine,StartCol,EndLine,EndCol).
extract_line_col(src_span(_Filenumber,_Filename,StartLine,SC,EndLine,EC,Offset,LenBytes),StartLine,StartCol,EndLine,EndCol)
 :- !, extract_line_col(src_span(StartLine,SC,EndLine,EC,Offset,LenBytes),StartLine,StartCol,EndLine,EndCol).
extract_line_col(src_span(StartLine,SC,EndLine,EC,_Off,[]),StartLine,StartCol,EndLine,EndCol)
 :-  !, StartCol is SC-1, EndCol is EC-1.
extract_line_col(src_span(SL,SC,EL,EC,_Off,_LenBytes),StartLine,StartCol,EndLine,EndCol) :- !,
   StartLine=SL,StartCol is SC-1,EndLine = EL, EndCol is EC-1.
 %:- !, StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine = StartLine. %% fix because new library generates erroneous EndLine/EndCol info
% EndCol is EC-1.
extract_line_col(span_info(_Info,Span),SL,SC,EL,EC) :- !, extract_line_col(Span,SL,SC,EL,EC).
extract_line_col(src_span_operator(SP1,SP2),SL,SC,EL,EC) :- !,
 (extract_line_col(SP1,SL,SC,EL,EC) ; extract_line_col(SP2,SL,SC,EL,EC)).
extract_line_col(multi_span(SL,SC,EL,EC,_Off,_Len,TAIL),StartLine,StartCol,EndLine,EndCol) :- !,
  % TO DO: replace by multi_span/2
  (  StartLine = SL, StartCol is SC-1,
     EndLine = EL, EndCol is EC-1
   ; extract_line_col(TAIL,StartLine,StartCol,EndLine,EndCol)).
%extract_line_col(multi_span(A,B),StartLine,StartCol,EndLine,EndCol) :- !,
%    (extract_line_col(A,StartLine,StartCol,EndLine,EndCol) ;
%     extract_line_col(B,StartLine,StartCol,EndLine,EndCol)).
extract_line_col(src_position(StartLine,SC,_Off,LenBytes),StartLine,StartCol,EndLine,EndCol) :- !,
   StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine is StartLine.
extract_line_col(src_position_with_filename(StartLine,SC,LenBytes,_Filename),StartLine,StartCol,EndLine,EndCol) :-
   !,
   StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine is StartLine.
extract_line_col(src_position_with_filename_and_ec(SL,SC,EL,EC,_Fname),StartLine,StartCol,EndLine,EndCol) :-
   !,
   StartLine=SL,StartCol is SC-1,EndLine = EL, EndCol is EC-1.
extract_line_col(lineCol(StartLine,StartCol),StartLine,StartCol,EndLine,EndCol) :- !,
   EndLine=StartLine, EndCol=StartCol.
extract_line_col(Struct,StartLine,StartCol,EndLine,EndCol) :-
    is_argument_of_compound_term(Struct,Arg),
      /* recursively look for error spans inside */
    extract_line_col(Arg,StartLine,StartCol,EndLine,EndCol),!.
%extract_line_col(S,_,_,_,_) :- print('COULD NOT EXTRACT LINE/COL INFO: '), print(S),
%    functor(S,F,N), print(' '),print(F/N),nl,nl,fail.

:- use_module(library(between),[between/3]).
is_argument_of_compound_term(Term,Arg) :- functor(Term,_,Arity),
   between(1,Arity,ArgNr), arg(ArgNr,Term,Arg).

/* For TCL TK: extract all line_col information in one go and in 4 lists */
extract_all_line_col(Span, StartLineList,StartColList,EndLineList,EndColList) :-
    findall(lc(SL,SC,EL,EC), extract_line_col(Span,SL,SC,EL,EC),All),
    %print(all(All,Span)),nl,
    split_lc(All,StartLineList,StartColList,EndLineList,EndColList).

split_lc([], [],[],[],[]).
split_lc([lc(SL,SC,EL,EC)|T],[SL|T1],[SC|T2],[EL|T3],[EC|T4]) :-
   split_lc(T,T1,T2,T3,T4).

% TO DO: merge with line_col and in Tcl/Tk use info to highlight proper file !

extract_file_name(Pos,Filename) :- extract_valid_file_number_and_name(Pos,_,Filename).
extract_tail_file_name(Pos,Filename) :-
   extract_valid_file_number_and_name(Pos,_,FullFilename),
   tools:get_tail_filename(FullFilename,Filename).
extract_subsidiary_tail_file_name(Pos,Filename) :-
   extract_valid_file_number_and_name(Pos,Nr,FullFilename),
   not_main_file(Nr,FullFilename),
   tools:get_tail_filename(FullFilename,Filename).

extract_valid_file_number_and_name(Pos,Nr,Filename) :-
   extract_file_number_and_name(Pos,Nr,Filename),
   valid_filename(Filename).
valid_filename(F) :- atom(F). % not unknown(-1) which can come from REPL

% a version of extract_line_col which fails if the error is not in the main file
extract_line_col_for_main_file(Span,Line,Col,EndLine,EndCol) :-
   extract_line_col(Span,Line,Col,EndLine,EndCol),
   \+((extract_file_number_and_name(Span,Nr,FullFilename),
       not_main_file(Nr,FullFilename))).


extract_file_number_and_name(V,_,_) :- var(V),!,fail.
extract_file_number_and_name(value(_),_,_) :- !,fail.
extract_file_number_and_name(pos_context(A,_,_), Filenumber,Filename) :- !,
   extract_file_number_and_name(A, Filenumber,Filename).
extract_file_number_and_name(pos(_,Filenumber,_,_,_,_), ResFileNr,Res) :-
   bmachine:b_get_all_used_filenames(FN),
   get_filename(Filenumber,FN,Filename),!,
   ResFileNr=Filenumber,Res=Filename.
extract_file_number_and_name(pos(_,Filenumber,_,_,_,_), ResFileNr,Res) :- !, ResFileNr=Filenumber,Res=unknown(Filenumber). % unknown(-1) which can come from REPL
extract_file_number_and_name(src_span(_,Filename,_,_,_,_,_,_), ResFileNr, Res) :- !,ResFileNr=unknown, Res=Filename.
extract_file_number_and_name(src_position_with_filename(_,_,_,Filename), ResFileNr, Res) :- !,ResFileNr=unknown,Res=Filename.
extract_file_number_and_name(src_position_with_filename_and_ec(_,_,_,_,Filename), ResFileNr, Res) :- !,ResFileNr=unknown,Res=Filename.
extract_file_number_and_name(section(RodinFilename), ResFileNr, Res) :- !,ResFileNr=unknown,Res=RodinFilename.
extract_file_number_and_name([H|T],Filenumber,Filename) :- !,
    posinfo_member(Arg,[H|T]),
    extract_file_number_and_name(Arg,Filenumber,Filename),!.
extract_file_number_and_name(closure(_,_,B),Filenumber,Filename) :- !,
   extract_file_number_and_name(B,Filenumber,Filename).
extract_file_number_and_name(b(E,_,Infos),Filenumber,Filename) :- !,
   (extract_file_number_and_name(Infos,Filenumber,Filename) -> true
     ; first_sub_expr(E,A) -> extract_file_number_and_name(A,Filenumber,Filename)).
extract_file_number_and_name(Struct,Filenumber,Filename) :-
      /* recursively look for error spans inside */
    is_argument_of_compound_term(Struct,Arg),
    extract_file_number_and_name(Arg,Filenumber,Filename),!.

definite_posinfo_member(POS,Args) :- member(nodeid(POS),Args).
posinfo_member(Arg,Args) :-  % first look for a nodeid(.) member
  (member(nodeid(POS),Args) -> Arg=POS ; member(Arg,Args)).

%:- use_module(bmachine,[b_get_all_used_filenames/1,b_get_main_filenumber/1]).
position_is_not_in_main_file(pos(_,Filenumber,_,_,_,_)) :- !, % B position info
   not_main_filenumber(Filenumber).
position_is_not_in_main_file(src_span(_,Filename,_,_,_,_,_,_)) :- !,
   specfile:currently_opened_file_status(Main,_), Main \= Filename.
position_is_not_in_main_file(src_position_with_filename(_,_,_,Filename)) :- !,
   specfile:currently_opened_file_status(Main,_), Main \= Filename.
position_is_not_in_main_file(src_position_with_filename_ec(_,_,_,_,Filename)) :- !,
   specfile:currently_opened_file_status(Main,_), Main \= Filename.
position_is_not_in_main_file(Struct) :- \+ atomic(Struct), Struct =.. [_|Args],
      /* recursively look for error spans inside */
    member(Arg,Args), position_is_not_in_main_file(Arg),!.


not_main_file(Filenumber,_) :- number(Filenumber),!, % B Position info
    not_main_filenumber(Filenumber).
not_main_file(_,Filename) :- % probably CSP position info; or unknown
    specfile:currently_opened_file_status(Main,_), \+ same_file_name(Main,Filename). %Main \= Filename.

:- use_module(tools,[get_tail_filename/2]).
same_file_name(F1,F2) :-
   (F1=F2 -> true
  ; get_tail_filename(F1,T1), get_tail_filename(F2,T2), T1=T2).

not_main_filenumber(Filenumber) :-
  bmachine:b_get_main_filenumber(MainFN),
  MainFN \= Filenumber.



% extract additional information about error span, e.g., Definition call context, other spans,...
extract_additional_description(b(E,_,Infos),Desc) :- !,
  (extract_additional_description(Infos,Desc) -> true
    ; first_sub_expr(E,A),extract_additional_description(A,Desc)).
extract_additional_description(span_predicate(P,_,_),Desc) :- !,
  extract_additional_description(P,Desc).
extract_additional_description([H|T],Desc) :- !,
   posinfo_member(N,[H|T]),
   extract_additional_description(N,Desc).
extract_additional_description(closure(_,_,B),Desc) :- !,
   extract_additional_description(B,Desc).
extract_additional_description(pos_context(MainPos,Context,CtxtPos),Desc) :- !,
   %nl,print(pos_context(MainPos,Context,CtxtPos)),nl,
   (extract_inner_context_description(MainPos,InnerCMsg)
    -> extract_context_msg(Context,OuterCMsg),
       ajoin([InnerCMsg,' within ',OuterCMsg],CMsg)
    ; extract_context_msg(Context,CMsg)
   ),
   (extract_span_description(CtxtPos,PosMsg)
    -> ajoin([CMsg,' at ',PosMsg],Desc)
    ;  Desc=CMsg). % no position info available

extract_context_msg(definition_call(Name),Msg) :- !, ajoin(['DEFINITION call of ',Name],Msg).
extract_context_msg(C,Msg) :- ajoin(['CONTEXT: ',C],Msg).

extract_inner_context_description(pos_context(MainPos,Context,CtxtPos),Desc) :- !,
     extract_additional_description(pos_context(MainPos,Context,CtxtPos),Desc).
% ---------------------------------
:- dynamic backed_up_error/4.

% copy current error facts into backed_up_error facts
backup_errors :-
   retractall(backed_up_error(_,_,_,_)),
   logged_error(Source,S,TContext,Span),
   assert(backed_up_error(Source,S,TContext,Span)),
   fail.
backup_errors.

% ---------------------------------

:- dynamic error_manager_silent_mode/1.
% separate from debug:silent_mode to avoid dependency to debug module
error_manager_silent_mode(off).

set_error_manager_silent_mode(X) :- retractall(error_manager_silent_mode(_)),
   assert(error_manager_silent_mode(X)).


formatsilentmsg(Stream,FS,Args) :- error_manager_silent_mode(off) -> format(Stream,FS,Args) ; true.
