% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_records,[is_a_record_wf/3, not_is_a_record_wf/3,
    construct_record_wf/3,
    access_record_wf/4,
    normalise_record_type/2, record_has_multiple_field_names/2,
    check_field_name_compatibility/3
    ]).


:- use_module(self_check).
:- use_module(debug).
:- use_module(tools).
:- use_module(typechecker).
:- use_module(error_manager).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module contains predicates that operate on the record type.').

:- assert_must_succeed(( kernel_records:normalise_record_type(
 record([field(name,global('ID')),field(balance,integer)]),X),
 X= record([field(balance,integer),field(name,global('ID'))])  )).
:- assert_must_succeed(( kernel_records:normalise_record_type(
  record([field(balance,integer),field(name,global('ID'))]),X),
 X= record([field(balance,integer),field(name,global('ID'))]) )).
:- assert_must_succeed(( kernel_records:normalise_record_type(
  record([field(b,d),field(a,c)]),X),
 X= record([field(a,c),field(b,d)]) )).

normalise_record_type(record(Fields),record(SortedFields)) :-
   check_known_field_names(Fields),
   sort(Fields,SortedFields).


:- use_module(kernel_waitflags).
:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
       kernel_records:is_a_record_wf(rec([field(balance,int(1)),field(name,fd(2,'Name'))]),
              rec([field(balance,global_set('NAT')),field(name,global_set('Name'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).
:- assert_must_fail(( kernel_waitflags:init_wait_flags(WF),
       kernel_records:is_a_record_wf(rec([field(a,int(0)),field(b,fd(2,'Name'))]),
              rec([field(a,global_set('NAT1')),field(b,global_set('Name'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).
:- assert_must_fail(( kernel_waitflags:init_wait_flags(WF),
      kernel_records:is_a_record_wf(rec([field(a,int(1)),field(b,fd(2,'Name')),field(c,int(0))]),
              rec([field(a,global_set('NAT1')),field(b,global_set('Name')),field(c,global_set('NAT1'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).


is_a_record_wf(rec(Fields),rec(FieldSets),WF) :-  %print_message(check_is_a_record(Record,FieldSets)),
   check_field_values(Fields,FieldSets,WF).


:-use_module(kernel_objects,[check_element_of_wf/3, membership_test_wf/4]).

:- block check_field_values(-,-,?).
check_field_values([],[],_WF).
check_field_values([field(Name1,V)|ValueRest],[field(Name2,VT)|TypeRest],WF) :-
   check_field_name_compatibility(Name1,Name2,check_field_values),
   check_element_of_wf(V,VT,WF),
   check_field_values(ValueRest,TypeRest,WF).


 :- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
      kernel_records:not_is_a_record_wf(rec([field(balance,int(0)),field(name,fd(2,'Name'))]),
              rec([field(balance,global_set('NAT1')),field(name,global_set('Name'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).
 :- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
      kernel_records:not_is_a_record_wf(rec([field(a,int(1)),field(b,fd(2,'Name')),field(c,int(0))]),
              rec([field(a,global_set('NAT1')),field(b,global_set('Name')),field(c,global_set('NAT1'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).
 :- assert_must_fail(( kernel_waitflags:init_wait_flags(WF),
      kernel_records:not_is_a_record_wf(rec([field(a,int(1)),field(b,fd(2,'Name')),field(c,int(1))]),
              rec([field(a,global_set('NAT1')),field(b,global_set('Name')),field(c,global_set('NAT1'))]),WF),
        kernel_waitflags:ground_wait_flags(WF))).

not_is_a_record_wf(rec(Fields),rec(FieldSets),WF) :-
   not_check_field_values(Fields,FieldSets,WF).
   %print_message(not_is_a_record(Record,FieldSets,WF)).


:- block not_check_field_values(-,-,?).
not_check_field_values([field(Name1,V)|ValueRest],[field(Name2,VT)|TypeRest],WF) :-
   check_field_name_compatibility(Name1,Name2,not_check_field_values),
   membership_test_wf(VT,V,MemRes,WF),
   not_check_field_values2(MemRes,ValueRest,TypeRest,WF).

:- block not_check_field_values2(-,?,?,?).
not_check_field_values2(pred_false,_,_,_).
not_check_field_values2(pred_true,ValueRest,TypeRest,WF) :-
      not_check_field_values(ValueRest,TypeRest,WF).





:- assert_must_fail(( X = [field(b,int(33)),field(a,int(2))],  X2 = [field(a,int(33)),field(b,int(2))], kernel_records:construct_record_wf(X,rec(X2),_WF) )).
:- assert_must_succeed(( kernel_records:construct_record_wf(X,rec([field(a,int(2)),field(b,int(3))]),_WF),
    X = [field(b,int(3)),field(a,int(2))] )).
:- assert_must_succeed(( kernel_records:construct_record_wf(X,rec([field(a,int(2)),field(b,int(3))]),_WF),
    X = [field(a,int(2)),field(b,int(3))] )).
:- assert_must_succeed(( kernel_records:construct_record_wf(X,Y,_WF),
    X = [field(b,int(4)),field(a,int(2))], Y = rec([field(a,int(2)),field(b,int(4))]) )).
:- assert_must_succeed(( kernel_records:construct_record_wf(X,rec([field(a,[int(2),int(3)]),field(b,int(3))]),_WF),  X = [field(a,[int(3),int(2)]),field(b,int(3))]  )).

:- block construct_record_wf(-,?,?).
construct_record_wf(Fields,Res,WF) :-
   check_known_field_names(Fields),
   sort(Fields,SortedFields),
   kernel_objects:equal_object_wf(Res,rec(SortedFields),WF).

check_known_field_names(Var) :- var(Var), !,
   add_internal_error('Illegal var fields: ',check_known_field_names(Var)).
check_known_field_names([]).
check_known_field_names([field(Name,_)|T]) :-
   (var(Name) -> add_internal_error('Illegal var fields: ',check_known_field_names([field(Name,_)|T]))
     ; true),
   check_known_field_names(T).



%instantiate output argument when all field names known + indicate whether sorted or not
%:- block known_field_names(-,?).
%known_field_names([],sorted).
%known_field_names([field(Name,_)|T],Known) :- known_field_names4(T,Name,sorted,Known).
%
%:- block known_field_names4(-,?,?,?),known_field_names4(?,-,?,?).
%known_field_names4([],_,K,K).
%known_field_names4([field(Name,_)|T],PrevName,SortedSoFar,Known) :-
%    (PrevName @< Name -> S2=SortedSoFar ; S2=not_sorted),
%    known_field_names4(T,Name,S2,Known).



% check if a sorted records has multiple field names
record_has_multiple_field_names([field(N,_)|T],Res) :- multiple_fields_aux(T,N,Res).
multiple_fields_aux([field(N,_)|T],N1,Res) :-
  (N1=N -> Res=N ; multiple_fields_aux(T,N,Res)).

:- use_module(library(lists)).

%%:- assert_must_fail(( kernel_records:access_record(rec([]),a,_) )). %% actually: must generate error; not just fail
:- assert_must_fail(( kernel_records:access_record(rec([field(a,int(2))]),a,int(3)) )).
:- assert_must_succeed(( kernel_records:access_record(rec(X),a,int(2)), X=[field(b,_),field(a,int(2))|_] )).
:- assert_must_succeed(( kernel_records:access_record(rec(X),b,int(2)), X=[field(b,_),field(a,int(3))|_] )).

access_record(Rec,Field,V) :- access_record_wf(Rec,Field,V,no_wf_available).

%access_record(X,Y,Z) :- print_message(access_record(X,Y,Z)),fail.
:- block access_record_wf(?,-,?,?). % when we add access_record_wf(-,?,?,?) we do not instantiate to rec(_), but PROB-356 runs slower.
access_record_wf(rec(Fields),FieldName,Value,WF) :-
%   access_record2(Fields,FieldName,Value).
%:- block access_record2(?,-,?).
%access_record2(Fields,FieldName,Value) :-
   get_field(Fields,FieldName,Value,Fields,WF).

:- block get_field(-,?,?,?,?). % note: FieldName is already known
get_field([],FieldName,_Value,OrigFields,_WF) :-
   add_internal_error('Could not get field: ', FieldName:OrigFields),fail.
get_field([field(SFieldName,SValue)|T], FieldName,Value,OrigFields,WF) :-
   get_field2(SFieldName,SValue,T, FieldName,Value,OrigFields,WF).

:- block get_field2(-,?,?, ?,?,?,?).
get_field2(SFieldName,SValue,T, FieldName,Value,OrigFields,WF) :-
     (SFieldName=FieldName
       -> kernel_objects:equal_object_wf(Value,SValue,get_field2,WF)
       ;  get_field(T,FieldName,Value,OrigFields,WF)
     ).



% try and unify field names and throw error if unification fails
% indicates an incorrectly sorted field
:- assert_must_succeed(( kernel_records:check_field_name_compatibility(a,a,test))).
check_field_name_compatibility(Name1,Name2,Origin) :-
   nonvar(Name1), nonvar(Name2), Name1 \= Name2, !,
   add_internal_error('Incompatible fields: ',check_field_name_compatibility(Name1,Name2,Origin)),
   fail.
check_field_name_compatibility(Name,Name,_).
