% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(parsercall, [load_b_machine_as_term/3,
                       load_b_machine_probfile_as_term/2, % a way to directly load a .prob file
                       load_b_machine_list_of_facts_as_term/2,
                       get_parser_version/1,
                       get_java_command_path/1,
                       check_java_version/2,
                       get_java_fullversion/1, get_java_fullversion/3, get_java_version/1,
                       ensure_console_parser_launched/0, % just make sure the console parser is up and running
                       release_console_parser/0,
                       call_ltl_parser/3,
                       call_tla2b_parser/1, tla2b_filename/2,
                       %call_promela_parser/1, promela_prolog_filename/2,
                       call_alloy2pl_parser/2,
                       tla2prob_filename/2,
                       parse/3,
                       parse_formula/2, parse_predicate/2, parse_expression/2,
                       parse_substitution/2,
                       call_fuzz_parser/2
                       ]).


:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module takes care of calling the Java B Parser if necessary.').

:- use_module(library(lists)).
:- use_module(library(process)).
:- use_module(library(file_systems)).
:- use_module(library(codesio)).
:- use_module(library(system)).
:- use_module(library(sockets)).

:- use_module(error_manager,[add_error/2,add_error/3,add_error/4,add_error_fail/3,add_failed_call_error/1, add_internal_error/2, add_warning/3, real_error_occurred/0]).
:- use_module(self_check).
:- use_module(preferences).
:- use_module(tools,[host_platform/1, split_filename/3, get_PROBPATH/1, safe_atom_chars/3, platform_is_64_bit/0]).
:- use_module(tools_strings,[ajoin_with_sep/3, ajoin/2]).
:- use_module(tools_printing,[print_error/1]).
:- use_module(debug).
:- use_module(bmachine,[b_get_definition/5]).
:- use_module(specfile, [b_mode/0]).

:- dynamic java_parser_version/1.
:- dynamic parser_location/1.

% stores the java process of the parser, consisting of ProcessIdentifier, Socket Stream, STDErrStream
:- dynamic java_parser_process/3.
:- volatile java_parser_process/3.


send_definition(Stream,def(Name,Type,Arity)) :-
    write(Stream,'definition'), nl(Stream),
    format(Stream,'~w\n~w\n~w\n',[Name,Type,Arity]).

send_definitions(ToParser) :-
    b_mode, !,
    findall(def(A,B,Ar),(b_get_definition(A,B,C,_,_),length(C,Ar)),L),
    %format('~nSending definitions ~w~n~n',[L]),
    maplist(send_definition(ToParser),L).
send_definitions(_).


crlf(10).
crlf(13).

% remove newlines in a B formula so that we can pass the formula as a single line to the parser
cleanup_newlines([],[]).
cleanup_newlines([H|T],CleanCodes) :- crlf(H),!, cleanup_newlines(T,CleanCodes).
cleanup_newlines([47,47|T],CleanCodes) :- % a B Comment: skip until end of line
   skip_until_crlf(T,T2), !,
   cleanup_newlines(T2,CleanCodes).
cleanup_newlines([H|T],[H|R]) :- cleanup_newlines(T,R).

skip_until_crlf([],[]).
skip_until_crlf([H|T],T) :- crlf(H),!.
skip_until_crlf([_|T],R) :- skip_until_crlf(T,R).

send_extended_kind(Extended,Kind,In) :-
    (Extended == true
     -> ajoin([extended,Kind],ToParser)
     ;  Kind = ToParser),
    write(In,ToParser), nl(In),
    flush_output(In).

:- mode parse_x(+Kind,+Extended,+Codes,-Tree).
parse_x(Kind,Extended,CodesWithNewlines,Tree) :-
    % Java parser expects only one line of input -> remove newlines
    cleanup_newlines(CodesWithNewlines,Codes),
    % statistics(walltime,[Tot,Delta]), format('~s~n~w~n0     ~w (Delta ~w) ms  ~n',[Codes,Kind,Tot,Delta]),
    get_console_parser(Stream,Err),
    send_definitions(Stream),
    send_extended_kind(Extended,Kind,Stream),
    format(Stream,'~s',[Codes]),
    nl(Stream),
    flush_output(Stream),
    read_line(Stream,CodesIn),
    %format('read: ~s~n',[CodesIn]),
    catch(
         (my_read_from_codes(CodesIn,Tree,Stream,Err),handle_parser_exception(Tree)),
          error(_,_),
          (append("Illegal parser result exception: ",CodesIn,ErrMsgCodes),
           atom_codes(ErrMsg,ErrMsgCodes),
           throw(parse_errors([internal_error(ErrMsg,none)])))).
    %statistics(walltime,[Tot4,Delta4]), format('4     ~w (Delta ~w) ms ~n ~n',[Tot4,Delta4]).


my_read_from_codes(end_of_file,_,_,Err) :- !,
    read_line_if_ready(Err,Err1Codes),  % we just read one line
    safe_name(ErrMsg,Err1Codes),
    add_error(parsercall,'Abnormal termination of parser: ',ErrMsg),
    missing_parser_diagnostics,
    ajoin(['Parser not available (',ErrMsg,')'],FullErrMsg),
    read_lines_and_add_as_error(Err),  % add other info on error stream as single error
    throw(parse_errors([error(FullErrMsg,none)])).
my_read_from_codes(ErrorCodes,_Tree,Out,_) :-
    append("Error",_,ErrorCodes),!,
    atom_codes(ErrMsg,ErrorCodes),
    add_error(parsercall,'Error running/starting parser: ',ErrMsg),
    read_lines_and_add_as_error(Out),
    fail.
my_read_from_codes(CodesIn,Tree,_,_) :- read_from_codes(CodesIn,Tree).

% a version of my_read_from_codes which has no access to error stream
my_read_from_codes(end_of_file,_) :- !,
    missing_parser_diagnostics,
    throw(parse_errors([error('Parser not available (end_of_file on input stream)',none)])).
my_read_from_codes(ErrorCodes,_Tree) :-
    append("Error",_,ErrorCodes),!,
    atom_codes(ErrMsg,ErrorCodes),
    add_error(parsercall,'Error running/starting parser: ',ErrMsg),
    fail.
my_read_from_codes(CodesIn,Tree) :- read_from_codes(CodesIn,Tree).

missing_parser_diagnostics :-
    parser_location(Classpath), debug_println(9,classpath(Classpath)),fail.
missing_parser_diagnostics :- runtime_application_path(P),
   atom_concat(P,'/lib/probcliparser.jar',ConParser),
   (file_exists(ConParser)
      -> add_error(parsercall,'The Java B parser (probcliparser.jar) cannot be launched: ',ConParser),
         add_error(parsercall,'Please check that Java and B Parser are available, e.g., using "probcli -check_java_version" command.'),
         (get_preference(jvm_parser_additional_args,ExtraArgStr), ExtraArgStr\=[]
           -> add_error(parsercall,'Also check your JVM_PARSER_JVM_ARGS preference value:',ExtraArgStr)
           ; true)
      ;  add_internal_error('Java B parser (probcliparser.jar) is missing from lib directory in: ',P)
    ).

% the position reported by the parser is slightly off,
% because the formula has to be submitted as
% "#FORMULA ...", "#EXPRESSION ..." or "#PREDICATE ..."
%handle_parser_exception(E) :- nl,print(E),nl,fail.
handle_parser_exception(compound_exception(Exs)) :- !,
    maplist(handle_parser_exception_aux,Exs,ExAuxs),
    throw(parse_errors(ExAuxs)).
handle_parser_exception(Ex) :-
    handle_parser_exception_aux(Ex,ExAux), !,
    throw(parse_errors([ExAux])).
handle_parser_exception(_).

handle_parser_exception_aux(parse_exception(Pos,Msg),error(SanitizedMsg,Pos)) :-
    remove_msg_posinfo_known(Msg,SanitizedMsg).
handle_parser_exception_aux(io_exception(Pos,Msg),error(Msg,Pos)).
handle_parser_exception_aux(io_exception(Msg),error(Msg,unknown)).
handle_parser_exception_aux(exception(Msg),error(Msg,unknown)). % TO DO: get Pos from Java parser !

:- mode parse_formula(+Codes,-Tree).
parse_formula(Codes,Tree) :-
    get_preference(bool_expression_as_predicate,Extended),
    parse_x(formula,Extended,Codes,Tree).

:- mode parse_expression(+Codes,-Tree).
parse_expression(Codes,Tree) :-
    get_preference(bool_expression_as_predicate,Extended),
    parse_x(expression,Extended,Codes,Tree).

:- mode parse_predicate(+Codes,-Tree).
parse_predicate(Codes,Tree) :-
    get_preference(bool_expression_as_predicate,Extended),
    parse_x(predicate,Extended,Codes,Tree).

:- mode parse_substitution(+Codes,-Tree).
parse_substitution(Codes,Tree) :-
    parse_x(substitution,false,Codes,Tree).

:- mode parse(+Kind,+Codes,-Tree).
parse(formula,Codes,Tree) :- parse_formula(Codes,Tree).
parse(expression,Codes,Tree) :- parse_expression(Codes,Tree).
parse(predicate,Codes,Tree) :- parse_predicate(Codes,Tree).
parse(substitution,Codes,Tree) :- parse_substitution(Codes,Tree).

:- mode parse_temporal_formula(+Kind,+LanguageExtension,+Formula,-Tree).
parse_temporal_formula(Kind,LanguageExtension,Formula,Tree) :-
    write_to_codes(Formula,CodesWithNewlines),
    % Java parser expects only one line of input -> remove newlines
    cleanup_newlines(CodesWithNewlines,Codes),
    get_console_parser(Stream,Err),
    send_definitions(Stream),
    format(Stream,'~s', [Kind]), nl(Stream),
    format(Stream,'~s', [LanguageExtension]), nl(Stream),
    format(Stream,'~s',[Codes]), nl(Stream),
    flush_output(Stream),
    read_line(Stream,CodesIn), % TODO: Improve feedback from java!
    catch(my_read_from_codes(CodesIn,Tree,Stream,Err),_,throw(parse_errors([error(exception,none)]))).

% throws parse_errors(Errors) in case of syntax errors
load_b_machine_as_term(Filename, Machine,Options) :- %print(load_b_machine(Filename)),nl,
    prob_filename(Filename, ProBFile),
    debug_println(9,prob_filename(Filename,ProBFile)),
    ( dont_need_compilation(Filename,ProBFile,Machine,LoadResult,Options) ->
        % no compilation is needed, .prob file was loaded
        release_parser_if_requested(Options),
        (LoadResult=loaded -> debug_println(20,'.prob file up-to-date')
        ; print_error('*** Loading failed *** '),nl,fail)
    ; otherwise ->
        debug_println(20,'Calling Parser'),
        (debug:debug_mode(on) ->
            time(call_console_parser(Filename,ProBFile))
        ;  call_console_parser(Filename,ProBFile)
        ),
        release_parser_if_requested(Options),
        load_b_machine_probfile_as_term(ProBFile, Machine)
    ).
release_parser_if_requested(Options) :- (member(release_java_parser,Options) -> release_console_parser ; true).

% converts the filename of the machine into a filename for the parsed machine
prob_filename(Filename, ProB) :-
    split_filename(Filename,Basename,_Extension),
    safe_atom_chars(Basename,BasenameC,prob_filename1),
    append(BasenameC,['.','p','r','o','b'],ProBC),
    safe_atom_chars(ProB,ProBC,prob_filename2),!.
prob_filename(Filename, ProB) :-
    add_failed_call_error(prob_filename(Filename,ProB)),fail.

check_version_and_read_machine(ProBFile, ProbTime, Version, Machine,Options) :-
    (Version='$UNKNOWN'
      -> add_warning(parsercall,'*** Unknown parser version number. Trying to load parsed file: ',ProBFile)
        /* keep CheckVersionNr as variable */
      ; CheckVersionNr=Version),
    read_machine_term_from_file(ProBFile,parser_version(CheckVersionNr),
                                [check_if_up_to_date(ProbTime)|Options],Machine).

read_machine_term_from_file(ProBFile,VersionTerm,Options,CompleteMachine) :-
   (debug:debug_mode(on)
     ->  time(read_machine_term_from_file_aux(ProBFile,VersionTerm,Options,CompleteMachine))
     ;   read_machine_term_from_file_aux(ProBFile,VersionTerm,Options,CompleteMachine)).

read_machine_term_from_file_aux(ProBFile,VersionTerm,Options,complete_machine(MainName,Machines,Files)) :-
    open(ProBFile, read, S),
    call_cleanup(( read(S,VersionTerm),
                   %read(S,classical_b(MainName,Files)), read_machines(S,Machines)),
                   read_machines(S,Options,Machines,MainName,Files),
                   (var(MainName) -> add_internal_error('Missing classical_b fact',ProBFile), MainName=unknown,Files=[]
                     ; true)
                   ),
                 close(S)).

load_b_machine_list_of_facts_as_term(Facts,Machine) :-
    read_machine_term_from_list_of_facts(Facts,_,Machine).
read_machine_term_from_list_of_facts(Facts,VersionTerm,complete_machine(MainName,Machines,FileList)) :-
    (select(parser_version(VERS),Facts,F2) -> VersionTerm = parser_version(VERS)
      ; add_internal_error('No parser_version available',read_machine_from_list_of_facts),
        F2=Facts, VersionTerm = unknown),
    (select(classical_b(MainName,FileList),F2,F3) -> true
      ; add_internal_error('No classical_b file information available',read_machine_from_list_of_facts),
        fail),
     debug_println(9,loading_classical_b(VersionTerm,MainName,FileList)),
     include(is_machine_term,F3,F4),
     maplist(get_machine_term,F4,Machines).

is_machine_term(machine(_M)).
get_machine_term(machine(M),M).

load_b_machine_probfile_as_term(ProBFile, Machine) :-
    debug_println(9,consulting(ProBFile)),
    ( read_machine_term_from_file(ProBFile,_VersionTerm,[],Machine) -> true
    ; otherwise ->
        add_error(parsercall,'Failed to read parser output: wrong format?',ProBFile),fail).

:- use_module(tools_fastread,[read_term_from_stream/2]).
read_machines(S,Options,Machines,MainName,Files) :-
    (member(use_fastread,Options)
     -> read_term_from_stream(S,Term) , debug_println(9,use_fastread)
      ; read(S,Term)
    ),
    %,  write_term(Term,[max_depth(5),numbervars(true)]),nl,
    ( Term == end_of_file ->
        Machines = []
    ; Term = classical_b(M,F) ->
        ((M,F)=(MainName,Files) -> true ; add_internal_error('Inconsistent classical_b fact: ',classical_b(M,F))),
        (member(check_if_up_to_date(ProbTime),Options)
           -> all_older(Files, ProbTime) % avoid reading rest if not all_older !
          ; true),
        read_machines(S,Options,Machines,MainName,Files)
    ; Term = machine(M) ->
        Machines = [M|Mrest],
        read_machines(S,Options,Mrest,MainName,Files)
    ).

% checks if all files that the prob-file depends on
% are older than the prob-file
dont_need_compilation(Filename,ProB,Machine,LoadResult,Options) :-
    file_exists(Filename),
    catch( (   get_parser_version(Version),
               file_property(Filename, modify_timestamp, BTime),
               debug_format(19,'Parser Version: ~w~n',[Version]),
               debug_format(9, 'Filename Timestamp: ~w : ~w~n',[Filename,BTime]),
               file_exists(ProB), % the .prob file exists
               file_property(ProB, modify_timestamp, ProbTime),
               debug_format(9, '.prob Timestamp: ~w : ~w~n',[ProB,ProbTime]),
               % print(time_stamps(ProB,ProbTime,Filename,BTime)),nl,
               BTime < ProbTime, % .mch file is older than .prob file
               check_version_and_read_machine(ProB, ProbTime, Version, Machine,Options),
               %Machine = complete_machine(_,_,_),
               LoadResult = loaded),
           error(A,B),
           (debug_println(9,error(A,B)),
            (A=resource_error(memory)
              -> add_error(load_b_machine,'Insufficient memory to load file.\nTry starting up ProB with more memory\n (e.g., setting GLOBALSTKSIZE=500M before starting ProB).'),
                 LoadResult=A
              ; fail))).

% check that all included files are older than the .prob file (ProbTime):
all_older([],_).
all_older([File|Rest],ProbTime) :- % print(checking(File)),nl,trace,
    (file_exists(File)
     -> file_property(File, modify_timestamp, BTime),
        debug_format(9,' Subsidiary File Timestamp: ~w:~w~n',[File,BTime]),
        ( BTime < ProbTime -> true
        ; format('File has changed : ~w : ~w~n',[File,BTime]),
          fail )
     ; virtual_dsl_file(File) -> debug_format(9,'Virtual file generated by DSL translator: ~w~n',[File])
     ; debug_format(20,'File does not exist anymore: ~w; calling parser.~n',[File]),fail),
    all_older(Rest,ProbTime).

virtual_dsl_file(File) :- atom_codes(File,Codes),
   Codes = [95|_], % file name starts with _ underscore
   nonmember(47,Codes), % no slash
   nonmember(92,Codes). % no backslash

:- use_module(system_call).

% use the former LARGE_JVM / use_large_jvm_for_parser by default
% if we are running on a 64-bit system
get_jvm_options([HeapOpt|T]) :- get_preference(jvm_parser_heap_size_mb,Heap), Heap>0, !,
  ajoin(['-Xmx',Heap,'m'],HeapOpt),
  get_jvm_options2(T).
%get_jvm_options(['-Xmx20m','-Xss1m']) :- !. % comment in to simulate low-memory situations
  % use -Xmx221500000m instead of -Xss5m or Xmx to force exception and test ProB's response
get_jvm_options(O) :- get_jvm_options2(O).

get_jvm_options2(ExtraArgs) :- get_preference(jvm_parser_additional_args,ExtraArgsStr),
    ExtraArgsStr \= '',
    atom_codes(ExtraArgsStr,C),
    split_chars(C," ",CA),
    maplist(atom_codes,ExtraArgs,CA),!,
    debug_format(19,'Using additional arguments for JVM (JVM_PARSER_JVM_ARGS): ~w~n',[ExtraArgs]).
get_jvm_options2(['-Xss5m']) :- platform_is_64_bit, !.
get_jvm_options2([]).


% we could do something like this: or allow user to provide list of JVM options
%get_gc_options([]) :- !,get_preference(jvm_parser_aggressive_gc,true).
%get_gc_options(['-XX:GCTimeRatio=19', '-XX:MinHeapFreeRatio=20', '-XX:MaxHeapFreeRatio=30']).
% see https://stackoverflow.com/questions/30458195/does-gc-release-back-memory-to-os

% ensure that the console Java process is running
ensure_console_parser_launched :-
   on_exception(E,parse_formula("1",_),
      (add_internal_error('Cannot launch console parser (probcliparser.jar)',E),fail)).


get_console_parser(Stream,Err) :-
    java_parser_process(_PID,Stream,Err),
    % calling is_process seams to be unreliable (if the process has crashed)
    % hence, we call process_wait with a timeout of 0
    % if this does not return an exit value of 'timeout'
    % the process has either been killed or died on its own
    % -- turns out that using process_wait does not work either
    % -- process_wait(PID,timeout,[timeout(0)]), !.
    % now we just check if in and output streams are still available
    \+ at_end_of_stream(Stream), !,
    (at_end_of_stream(Err)
     -> add_error(parsercall,'Java process not running'),
        read_lines_and_add_as_error(Stream)
     ;  true).
get_console_parser(Stream,Err) :-
    clear_old_console_parsers,
    get_java_command(JavaCmd,Classpath),
    % maybe move to a predicate similar to get_jvm_options for Properties
    get_PROBPATH(PROBPATH),
    atom_concat('-Dprob.stdlib=', PROBPATH, Stdlibpref),
    %
    get_jvm_options(XOpts),
    append(XOpts, [Stdlibpref, '-cp',Classpath,'de.prob.cliparser.CliBParser',
                   %% '-v', % comment in for verbose
                   %% -time for timing info
                   %% -fastprolog instead of Prolog
                   '-prolog','-lineno','-prepl'], FullOpts),
    debug_println(9,launching_java_console_parser(JavaCmd,FullOpts,probpath(PROBPATH))),
    system_call_keep_open(JavaCmd,FullOpts,PID,_In,Out,Err,[]), % note: if probcliparser.jar does not exist we will not yet get an error here !
    debug_println(9,java_console_parser_launched(PID)),
    % read first line to get port number
    safe_read_line(Out,Err,1,[],Term),
    socket_client_open('':Term, Stream, [type(text),encoding('UTF-8')]),
    assert(java_parser_process(PID,Stream,Err)), !.
get_console_parser(_,_) :- missing_parser_diagnostics, fail.

clear_old_console_parsers :-
    retract(java_parser_process(PID,_,_)),
    process_release(PID),
    fail.
clear_old_console_parsers.

% useful to try and free up memory
% alternatively we could add some JVM options: -XX:GCTimeRatio=19 -XX:MinHeapFreeRatio=20 -XX:MaxHeapFreeRatio=30
% see https://stackoverflow.com/questions/30458195/does-gc-release-back-memory-to-os
release_console_parser :-
   (java_parser_process(PID,Stream,_)
    -> format('Releasing Java parser id=~w (note: access to DEFINITIONS is lost)~n',[PID]),
        write(Stream,'halt'), nl(Stream), flush_output(Stream), % handled in Java in runPRepl in CliBParser.java;
        retract(java_parser_process(PID,_,_)),
        process_release(PID)   % or should we call process_kill ?
        %process_kill(PID)
       % Note: parser maybe launched again, but then not have access to DEFINITIONS
     ; true).

:- use_module(runtime_profiler,[profile_single_call/3]).
call_console_parser(Filename, ProB) :-
    profile_single_call('$parsing',unknown,parsercall:call_console_parser2(Filename, ProB)).
call_console_parser2(Filename, ProB) :-
    get_console_parser(Stream,Err),

    % path to the .prob file
    replace_windows_path_backslashes(ProB,ProBW),
    % the input file
    replace_windows_path_backslashes(Filename,FilenameW),
    format(Stream,'machine~n~w~n~w~n',[FilenameW,ProBW]), % tell parser to write output to ProBW .prob file
    flush_output(Stream),

    safe_read_line(Stream,Err,1,[],Term),
    %print(result(Term)),nl,
    handle_console_parser_result(Term,Err).

safe_read_line(Out,Err,LineNr,SoFar,Term) :-
    safe_read_line_aux(read_line,Out,Err,LineNr,SoFar,Term).
safe_read_line_if_ready(Out,Err,LineNr,SoFar,Term) :-
    safe_read_line_aux(read_line_if_ready,Out,Err,LineNr,SoFar,Term).

safe_read_line_aux(Pred,Out,Err,LineNr,SoFar,Term) :-
    catch(call(Pred,Out,Codes),Exception,    % read another line
          (add_error(parsercall,'Exception while reading next line: ',Exception),throw(Exception))),
    Codes \= end_of_file, !, % happens if parser crashed and stream is closed
    append(SoFar,Codes,NewCodes),
    catch(my_read_from_codes(NewCodes,Term,Out,Err),Exception,
           safe_read_exception(Out,Err,LineNr,Exception,NewCodes,Term)).
safe_read_line_aux(_Pred,_Out,Err,_LineNr,CodesSoFar,_) :-
    safe_name(T,CodesSoFar),
    add_error(parsercall,'Unexpected error while parsing machine: ',T),
    read_lines_and_add_as_error(Err),fail.


safe_read_exception(Out,_,1,_Exception,CodesSoFar,_Term) :- append("Error",_,CodesSoFar),!,
    % probably some Java error, we do not obtain a correct Prolog term, but probably something like:
    % "Error occurred during initialization of VM"
    safe_name(T,CodesSoFar),
    add_error(parsercall,'Java error while parsing machine: ',T),
    read_lines_and_add_as_error(Out),fail.
safe_read_exception(Out,Err,LineNr,error(syntax_error(M),_),CodesSoFar,Term) :- LineNr<10, % avoid infinite loop in case unexpected errors occur which cannot be solved by reading more input lines
    !,
    %(debug:debug_mode(off) -> true ; format('Syntax error in parser result (line ~w): ~w~nTrying to read another line.~n',[LineNr,M])),
    add_warning(parsercall,'Syntax error in parser result; trying to read another line. Line:Error = ',LineNr:M),
    L1 is LineNr+1,
    safe_read_line_if_ready(Out,Err,L1,CodesSoFar,Term). % try and read another line; we assume the syntax error is because the TERM is not complete yet and transmitted on multiple lines (Windows issue)
safe_read_exception(_,_,_LineNr,parse_errors(Errors),_CodesSoFar,_Term) :- !,
    debug_println(4,read_parse_errors(Errors)),
    throw(parse_errors(Errors)).
safe_read_exception(_,_,_LineNr,Exception,_CodesSoFar,_Term) :-
    add_internal_error('Unexpected exception occurred during parsing: ',Exception),fail.
    %safe_name(T,CodesSoFar), add_error_fail(parsercall,'Unexpected error while parsing machine: ',T).

safe_name(Name,Codes) :- var(Name),Codes == end_of_file,!,
   %add_internal_error('Parser does not seem to be available',''),
   Name=end_of_file.
safe_name(Name,Codes) :- number(Name),!,
   safe_number_codes(Name,Codes).
safe_name(Name,Codes) :-
  on_exception(E,atom_codes(Name,Codes),
               (print('Exception during atom_codes: '),print(E),nl,nl,throw(E))).

safe_number_codes(Name,Codes) :-
  on_exception(E,number_codes(Name,Codes),
               (print('Exception during number_codes: '),print(E),nl,nl,throw(E))).

% see also handle_parser_exception
%handle_console_parser_result(E,_) :- nl,print(E),nl,fail.
handle_console_parser_result(exit(0),_) :- !.
handle_console_parser_result(io_exception(Msg),_) :- !,
    add_error_fail(parsercall,'Error while opening the B file: ',Msg).
handle_console_parser_result(preparse_exception(Tokens,Msg),_) :- !,
    remove_msg_posinfo(Msg,SanitizedMsg,MsgPos),
    findall(error(SanitizedMsg,Pos),member(Pos,Tokens),Errors1),
    (Errors1 = []
     -> /* if there are no error tokens: use pos. info from text Msg */
        Errors = [error(SanitizedMsg,MsgPos)]
     ;  Errors = Errors1),
    debug_println(4,preparse_exception_parse_errors(Tokens,Msg,Errors)),
    throw(parse_errors(Errors)).
handle_console_parser_result(parse_exception(Pos,Msg),_) :- !,
    remove_msg_posinfo_known(Msg,SanitizedMsg),  %print(sanitized(SanitizedMsg)),nl,
    debug_println(4,parse_exception(Pos,Msg,SanitizedMsg)),
    throw(parse_errors([error(SanitizedMsg,Pos)])).
handle_console_parser_result(exception(Msg),_) :- !,
    %add_error(bmachine,'Error in the classical B parser: ', Msg),
    remove_msg_posinfo(Msg,SanitizedMsg,Pos),
    debug_println(4,exception_in_parser(Msg,SanitizedMsg,Pos)),
    throw(parse_errors([error(SanitizedMsg,Pos)])).
handle_console_parser_result(end_of_file,Err) :- !,
     % probably NullPointerException or similar in parser; no result on StdOut
     debug_println(9,end_of_file_on_parser_output_stream),
     format(user_error,'Detected abnormal termination of B Parser~n',[]), % print in case we block
     read_line_if_ready(Err,Err1),
     (Err1=end_of_file -> Msg = 'Abnormal termination of B Parser: EOF on streams'
      ; append("Abnormal termination of B Parser: ",Err1,ErrMsgCodes),
        safe_name(Msg,ErrMsgCodes),
        read_lines_and_add_as_error(Err) % also show additional error infos
     ),
     throw(parse_errors([error(Msg,none)])).
handle_console_parser_result(compound_exception(Exs),_) :- !,
    maplist(handle_parser_exception_aux,Exs,ExAuxs),
    debug_println(4,compound_exception(Exs,ExAuxs)),
    throw(parse_errors(ExAuxs)).
handle_console_parser_result(Term,_) :- !,
    write_term_to_codes(Term,Text,[]),
    safe_name(Msg,Text),
    %add_error(bmachine,'Error in the classical B parser: ', Msg),
    remove_msg_posinfo(Msg,SanitizedMsg,Pos),
    debug_println(4,unknown_error_term(Term,SanitizedMsg,Pos)),
    throw(parse_errors([error(SanitizedMsg,Pos)])).

% now replaced by remove_msg_posinfo:
%extract_position_info(Text,pos(Row,Col,Filename)) :-
%    atom_codes(Text,Codes),
%    epi(Row,Col,Filename,Codes,_),!.
%extract_position_info(_Text,none).

:- assert_must_succeed((parsercall:epi(R,C,F,"[2,3] xxx in file: /usr/lib.c",[]), R==2,C==3,F=='/usr/lib.c')).
:- assert_must_succeed((parsercall:epi(R,C,F,"[22,33] xxx  yyy ",_), R==22,C==33,F=='unknown')).

% we expect error messages to start with [Line,Col] information
epi(Row,Col,Filename) --> " ",!,epi(Row,Col,Filename).
epi(Row,Col,Filename) --> "[", epi_number(Row),",",epi_number(Col),"]",
   (epi_filename(Filename) -> [] ; {Filename=unknown}).
epi_number(N) --> epi_numbers(Chars),{safe_number_codes(N,Chars)}.
epi_numbers([C|Rest]) --> [C],{is_number(C)},epi_numbers2(Rest).
epi_numbers2([C|Rest]) --> [C],{is_number(C),!},epi_numbers2(Rest).
epi_numbers2("") --> !.
is_number(C) :- member(C,"0123456789"),!.
epi_filename(F) --> " in file: ",!,epi_path2(Chars),{safe_name(F,Chars)}.
epi_filename(F) --> [_], epi_filename(F).
epi_path2([C|Rest]) --> [C],!,epi_path2(Rest). % assume everything until the end is a filename
epi_path2([]) --> [].

:- assert_must_succeed((parsercall:remove_rowcol(R,C,"[22,33] xxx  yyy ",_), R==22,C==33)).
% remove parser location info from message to avoid user clutter
remove_rowcol(Row,Col) --> " ",!, remove_rowcol(Row,Col).
remove_rowcol(Row,Col) --> "[", epi_number(Row),",",epi_number(Col),"]".

:- assert_must_succeed((parsercall:remove_filename(F,C,"xxx  yyy in file: a.out",R), F == 'a.out', C == "xxx  yyy", R==[])).
remove_filename(F,[]) --> " in file: ",!,epi_path2(Chars),{safe_name(F,Chars)}.
remove_filename(F,[C|T]) --> [C], remove_filename(F,T).

:- assert_must_succeed((parsercall:remove_filename_from_codes("xxx  yyy in file: a.out",F,C), F == 'a.out', C == "xxx  yyy")).
:- assert_must_succeed((parsercall:remove_filename_from_codes("xxx  yyy zzz",F,C), F == 'unknown', C == "xxx  yyy zzz")).
remove_filename_from_codes(Codes,F,NewCodes) :-
   (remove_filename(F,C1,Codes,C2) -> append(C1,C2,NewCodes) ; F=unknown, NewCodes=Codes).

% obtain position information from a message atom and remove the parts from the message at the same time
remove_msg_posinfo_known(Msg,NewMsg) :- remove_msg_posinfo(Msg,NewMsg,_,pos_already_known).
remove_msg_posinfo(Msg,NewMsg,Pos) :- remove_msg_posinfo(Msg,NewMsg,Pos,not_known).
remove_msg_posinfo(Msg,NewMsg,Pos,AlreadyKnown) :-  %print(msg(Msg,AlreadyKnown)),nl,
   atom_codes(Msg,Codes),
   (remove_rowcol(Row,Col,Codes,RestCodes)
    -> Pos=pos(Row,Col,Filename),
       remove_filename_from_codes(RestCodes,Filename,NewCodes)
     ; AlreadyKnown==pos_already_known ->
       Pos = none,
       remove_filename_from_codes(Codes,_Filename2,NewCodes)
     ; otherwise ->
       NewCodes=Codes, Pos=none
       % do not remove filename: we have not found a position and cannot return the filename
       % see, e.g., public_examples/B/Tickets/Hansen27_NestedMchErr/M1.mch
    ),
   !,
   atom_codes(NewMsg,NewCodes).
remove_msg_posinfo(M,M,none,_).


% ----------------------

get_parser_version(Version) :- java_parser_version(Version),!.
get_parser_version(Version) :-
    get_version_from_parser(Version),
    assert(java_parser_version(Version)),
    !.
get_parser_version(_Vers) :-
    (real_error_occurred -> true % already produced error messages in get_version_from_parser
     ; add_error(get_parser_version,'Could not get parser version.')),
    %missing_parser_diagnostics, % already called if necessary by get_version_from_parser
    fail.


get_version_from_parser(Version) :-
    get_console_parser(Stream,Err),
    write(Stream,'version'), nl(Stream),
    flush_output(Stream),
    read_line(Stream,Text),  % if parser cannot be started we receive end_of_file; length will fail
    (Text = end_of_file -> add_error(parsercall,'Could not get parser version: '),
                           read_lines_and_add_as_error(Err),fail
     ; length(Text,Length),
        ( append("Error",_,Text)
                       -> atom_codes(ErrMsg,Text),
                          add_error(parsercall,'Error getting parser version: ',ErrMsg),
                          read_lines_and_add_as_error(Stream),fail
        ; Length < 100 -> atom_codes(Version,Text)
        ; otherwise    -> prefix_length(Text,Prefix,100),
                          append(Prefix,"...",Full),
                          safe_name(Msg,Full),
                          add_error_fail(parsercall, 'B parser returned unexpected version string: ', Msg))
    ).

:- use_module(library(sockets),[socket_select/7]).
% check if a stream is ready for reading
stream_ready_to_read(Out) :- var(Out),!,
    add_internal_error('Illegal stream: ',stream_ready_to_read(Out)),fail.
stream_ready_to_read(Out) :-
    socket_select([],_, [Out],RReady, [],_, 1), % wait 1 sec at most %print(ready(Out,RReady)),nl,
    RReady == [Out].


read_lines_and_add_as_error(Out) :- read_lines_until_eof(Out,false,ErrMsgCodes,[]),
    (ErrMsgCodes=[] -> true
      ; ErrMsgCodes=[10] -> true % just a single newline
      ; ErrMsgCodes=[13] -> true % ditto
      ; safe_name(ErrMsg,ErrMsgCodes),
        add_error(parsercall,'Additional information: ',ErrMsg)).

read_line_if_ready(Stream,Line) :- stream_ready_to_read(Stream), !, read_line(Stream,Line).
read_line_if_ready(_,end_of_file). % pretend we are at the eof

read_lines_until_eof(Out,NewlineRequired) --> {read_line_if_ready(Out,Text)},
    !,
    ({Text = end_of_file} -> []
      ; ({NewlineRequired==true} -> "\n" ; []),
         Text, %{format("read: ~s~n",[Text])},
         read_lines_until_eof(Out,true)).
read_lines_until_eof(_Out,_) --> [].

get_java_command(JavaCmdW,ClasspathW) :-
    FileList = ['probcliparser.jar'],
    (parser_location(Classpath) -> true
        ; otherwise -> libraries_as_classpath(FileList,Classpath)),
    get_java_command_path(JavaCmdW),
    replace_windows_path_backslashes(Classpath,ClasspathW).


% a predicate to check whether we have correct java version number and whether java seems to work ok
% ResultStatus=compatible if everything is ok
check_java_version(VersionCheckResult,ResultStatus) :- get_java_fullversion(Version),!,
   check_java_version_aux(Version,VersionCheckResult,ResultStatus).
check_java_version(VersionCheckResult,error) :- get_java_command(JavaCmdW,ClasspathW),!,
   ajoin(['*** Unable to launch java command: ',JavaCmdW,' (classpath: ',ClasspathW,')!'],VersionCheckResult).
check_java_version('*** Unable to launch Java or get path to java command!',error). % should not happen

check_java_version_aux(VersionCodes,VersionCheckResult,ResultStatus) :-
     get_java_fullversion_numbers(VersionCodes,V1,V2,V3Atom),!,
     check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,ResultStatus).
check_java_version_aux(VersionCodes,VersionCheckResult,error) :-
     atom_codes(VersionA,VersionCodes),
     ajoin(['*** Unable to identify Java version number: ',VersionA, ' !'],VersionCheckResult).

check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,compatible) :- java_version_ok_for_parser(V1,V2,V3Atom),
     get_java_version(_),!, % if get_java_version fails, this is an indication that Java not fully installed with admin rights
     ajoin(['Java is correctly installed and version ',V1,'.',V2,'.',V3Atom,
            ' is compatible with ProB requirements (>= 1.7).'],VersionCheckResult).
check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,error) :- java_version_ok_for_parser(V1,V2,V3Atom),!,
     ajoin(['*** Java version ',V1,'.',V2,'.',V3Atom,
            ' is compatible with ProB requirements (>= 1.7) ',
            'but does not seem to be correctly installed: reinstall Java with admin rights!'],VersionCheckResult).
check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,incompatible) :-
     get_java_version(_),!,
     ajoin(['*** Java is correctly installed but version ',V1,'.',V2,'.',V3Atom,
            ' is *not* compatible with ProB requirements (>= 1.7)!'],VersionCheckResult).
check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,error) :-
     ajoin(['*** Java is not correctly installed and version ',V1,'.',V2,'.',V3Atom,
            ' is *not* compatible with ProB requirements (>= 1.7)!'],VersionCheckResult).


% we need Java 1.7 or higher
java_version_ok_for_parser(V1,_V2,_) :- V1>1.
java_version_ok_for_parser(1,V2,_) :- V2>=7.

get_java_fullversion(V1,V2,V3Atom) :- get_java_fullversion(VersionCodes),
     get_java_fullversion_numbers(VersionCodes,V1,V2,V3Atom).

:- use_module(tools,[split_chars/3]).
get_java_fullversion_numbers(VersionCodes,V1,V2,V3Atom) :-
    append("java full version """,Tail,VersionCodes),
    get_java_fullversion_numbers_aux(Tail,V1,V2,V3Atom).
get_java_fullversion_numbers(VersionCodes,V1,V2,V3Atom) :-
    append("openjdk full version """,Tail,VersionCodes),
    get_java_fullversion_numbers_aux(Tail,V1,V2,V3Atom).
get_java_fullversion_numbers_aux(Tail,V1,V2,V3Atom) :-
    split_chars(Tail,".",[V1C,V2C|FurtherC]),
    append(FurtherC,V3C),
    number_codes(V1,V1C), number_codes(V2,V2C),
    (append(V3C2,[_,_],V3C) % try remove trailing " and newline
     -> atom_codes(V3Atom,V3C2) ; atom_codes(V3Atom,V3C)).
get_java_fullversion_numbers_aux(Tail,1,V2,V3Atom) :-
    % new version scheme (https://openjdk.java.net/jeps/223), leading 1 dropped, 13+33 stands for 1.13.33
    split_chars(Tail,"+",[V2C,V3C]),
    number_codes(V2,V2C), atom_codes(V3Atom,V3C).

% get java fullversion as code list; format "java full version "...."\n"
get_java_fullversion(Version) :-
   get_java_command_path(JavaCmdW),
   (my_system_call(JavaCmdW, ['-fullversion'],java_version,Text) -> Text=Version
     ; format(user_error,'Could not execute ~w -fullversion~n',[JavaCmdW]),
       fail).

% when java -fullversion works but not java -version it seems that Java was improperly
% installed without admin rights (http://stackoverflow.com/questions/11808829/jre-1-7-returns-java-lang-noclassdeffounderror-java-lang-object)
get_java_version(Version) :-
   get_java_command_path(JavaCmdW),
   (my_system_call(JavaCmdW, ['-version'],java_version,Text) -> Text=Version
     ; format(user_error,'Could not execute ~w -version~n',[JavaCmdW]),
       fail).
% should we check for JAVA_PATH first ? would allow user to override java; but default pref would have to be set to ''
get_java_command_path(JavaCmdW) :-
    %host_platform(darwin)
    get_preference(path_to_java,X), debug_println(8,path_to_java_preference(X)),
    (X=''
     ; preference_default_value(path_to_java,Default), debug_println(8,default(Default)),
       Default=X
    ),
    % only try this when the Java path has not been set explicitly
    % on Mac: /usr/bin/java exists even when Java not installed ! it is a fake java command that launches a dialog
    absolute_file_name(path(java),
                       JavaCmd,
                       [access(exist),extensions(['.exe','']),solutions(all),file_errors(fail)]),
    debug_println(8,obtained_java_cmd_from_path(JavaCmd)),
    replace_windows_path_backslashes(JavaCmd,JavaCmdW),
    !.
get_java_command_path(JavaCmdW) :- get_preference(path_to_java,JavaCmd),
    JavaCmd \= '',
    debug_println(8,using_path_to_java_preference(JavaCmd)),
    (file_exists(JavaCmd) -> true
     ; directory_exists(JavaCmd) -> add_warning(get_java_command_path,'No Java Runtime (7 or newer) found; please correct the JAVA_PATH advanced preference (it has to point to the java tool, *not* the directory enclosing it): ',JavaCmd)
     ; add_warning(get_java_command_path,'No Java Runtime (7 or newer) found; please correct the JAVA_PATH advanced preference: ',JavaCmd)),
     replace_windows_path_backslashes(JavaCmd,JavaCmdW),
     !.
get_java_command_path(_JavaCmd) :-
    add_error(get_java_command_path,'Could not get path to the java tool. Make sure a Java Runtime (7 or newer) is installed',''),
    fail.

libraries_as_classpath(Libs,Classpath) :-
   %findall(PLD,user:file_search_path(prob_lib,PLD),LL2), print(prob_lib_directory(LL2)),nl,
    maplist(library_abs_name,Libs,ALibs),
    ( host_platform(windows) -> Sep = (';')
    ; otherwise              -> Sep = (':')),
    ajoin_with_sep(ALibs,Sep,Classpath).
library_abs_name(Lib,Abs) :- absolute_file_name(prob_lib(Lib),Abs).


call_ltl_parser(Formulas, CtlOrLtl, Result) :-
    get_ltl_lang_spec(LangSpec),
    maplist(parse_temporal_formula(CtlOrLtl,LangSpec),Formulas,Result).

:- use_module(specfile,[b_or_z_mode/0,csp_with_bz_mode/0]).
get_ltl_lang_spec('B,none') :- csp_with_bz_mode,!.
get_ltl_lang_spec('B') :- b_or_z_mode,!.
get_ltl_lang_spec(none).

/* unused code :
generate_formula_codes([F]) -->
    !,generate_formula_codes2(F),"\n".
generate_formula_codes([F|Rest]) -->
    generate_formula_codes2(F), "###\n",
    generate_formula_codes(Rest).
generate_formula_codes2(F) -->
    write_to_codes(F).
*/


% ---------------------------

% call the TLA2B parser to translate a TLA file into a B machine
call_tla2b_parser(TLAFile) :-
   get_java_command_path(JavaCmdW),
   replace_windows_path_backslashes(TLAFile,TLAFileW),
   get_jvm_options(XOpts),
   absolute_file_name(prob_lib('TLA2B.jar'),TLA2BJAR),
   append(XOpts,['-jar',file(TLA2BJAR),file(TLAFileW)],FullOpts),
   debug_println(9,calling_java(FullOpts)),
   (my_system_call(JavaCmdW, FullOpts,tla2b_parser) -> true
     ; print_error('Be sure to download TLA2B.zip from http://nightly.cobra.cs.uni-duesseldorf.de/tla,'),
       print_error('then unzip the archive and put TLA2B.jar into the ProB lib/ directory.'),
       fail).

%call_cspmj_parser(CSPFile,PrologFile) :-
%   get_java_command_path(JavaCmdW),
%   replace_windows_path_backslashes(CSPFile,CSPFileW),
%   get_jvm_options(XOpts),
%   absolute_file_name(prob_lib('cspmj.jar'),CSPMJAR),
%   get_writable_compiled_filename(CSPFile,'.pl',PrologFile),
%   tools:string_concatenate('--prologOut=', PrologFile, PrologFileOutArg),
%   append(XOpts,['-jar',CSPMJAR,'-parse',CSPFileW,PrologFileOutArg],FullOpts),
%   debug_println(9,calling_java(FullOpts)),
%   (my_system_call(JavaCmdW, FullOpts,cspmj_parser) -> true
%     ; print_error('Be sure that cspmj.jar parser is installed.'),
%       fail).


tla2prob_filename(TLAFile,GeneratedProbFile) :-
    split_filename(TLAFile,Basename,_Extension),
    atom_chars(Basename,BasenameC),
    append(BasenameC,['.','p','r','o','b'],TLABC),
    atom_chars(GeneratedProbFile,TLABC),!.
tla2prob_filename(Filename, GeneratedProbFile) :-
    add_failed_call_error(tla2b_filename(Filename,GeneratedProbFile)),fail.

tla2b_filename(TLAFile,BFile) :-
    split_filename(TLAFile,Basename,_Extension),
    atom_chars(Basename,BasenameC),
    append(BasenameC,['_',t,l,a,'.','m','c','h'],TLABC),
    atom_chars(BFile,TLABC),!.
tla2b_filename(Filename, BFile) :-
    add_failed_call_error(tla2b_filename(Filename,BFile)),fail.


my_system_call(Command,Options,Origin) :- my_system_call(Command,Options,Origin,_Text).
my_system_call(Command,Options,Origin,ErrText) :-
   system_call(Command,Options,ErrText,Exit), %nl,print(exit(Exit)),nl,nl,
   treat_exit_code(Exit,Command,Options,ErrText,Origin).

my_system_call5(Command,Options,Origin,OutputText,ErrText) :-
   system_call(Command,Options,OutputText,ErrText,Exit), %nl,print(exit(Exit)),nl,nl,
   treat_exit_code(Exit,Command,Options,ErrText,Origin).

treat_exit_code(exit(0),_,_,_,_) :- !.
treat_exit_code(Exit,Command,Options,ErrText,Origin) :-
   debug_println(9,treat_exit_code(Exit,Command,Options,ErrText,Origin)),
   catch( my_read_from_codes(ErrText,Term),
          _,
          (safe_name(T,ErrText),Term = T)), !,
    %print_error(Term),
    (Term = end_of_file -> ErrMsg = '' ; ErrMsg = Term),
    tools:ajoin(['Exit code ',Exit,' for command ',Command, ', error message: '],Msg),
    (get_error_position_from_term(Origin,Term,Pos)
     -> add_error(Origin,Msg,ErrMsg,Pos)
     ;  add_error(Origin,Msg,ErrMsg)
     ),
    fail.


get_error_position_from_term(alloy2b,Term,Pos) :- atom(Term),
   atom_codes(Term,Codes), get_error_position_from_codes(Codes,Pos).

% Alloy error message: ! Exception in thread "main" Type error in .../Restrictions.als at line 41 column 23:
get_error_position_from_codes([97,116,32,108,105,110,101,32|Tail],lineCol(Line,Col)) :- % "at line "
   epi_number(Line,Tail,RestTail),
   (RestTail = [32,99,111,108,117,109,110,32|Tail2], % "column "
     epi_number(Col,Tail2,_) -> true
     ; Col=0).
get_error_position_from_codes([_|T],Pos) :- get_error_position_from_codes(T,Pos).

% replace backslashes by forward slashes in atoms, only under windows
% TODO(DP,14.8.2008): Check if still needed with SICStus 4.0.4 and if
% everybody uses that version
replace_windows_path_backslashes(Old,New) :-
    ( host_platform(windows) ->
        safe_name(Old,OldStr),
        replace_string_backslashes(OldStr,NewStr),
        safe_name(New,NewStr)
    ; otherwise ->
        Old=New).

replace_string_backslashes([],[]) :- !.
replace_string_backslashes([C|OldRest],[N|NewRest]) :- !,
    ( C=92 /* backslash */ ->
        N=47 /* forward slash */
    ; otherwise ->
        N=C ),
    replace_string_backslashes(OldRest,NewRest).
replace_string_backslashes(X,Y) :-
  add_error(parsercall,'Illegal call: ',replace_string_backslashes(X,Y)),
  fail.

% ---------------------------
:- use_module(pathes,[runtime_application_path/1]).

call_jar_parser(PMLFile,JARFile,Options,ParserName) :-
   get_java_command_path(JavaCmdW),
   replace_windows_path_backslashes(PMLFile,PMLFileW),
   get_jvm_options(XOpts),
   absolute_file_name(prob_lib(JARFile),FullJAR),
   append(XOpts,['-jar',FullJAR,PMLFileW|Options],FullOpts),
   debug_println(9,calling_java(FullOpts)),
   my_system_call(JavaCmdW, FullOpts,ParserName).



call_alloy2pl_parser(AlloyFile,BFile) :-
   alloy_pl_filename(AlloyFile,BFile),
   call_alloy2pl_parser_aux(AlloyFile,'alloy2b-1.0-SNAPSHOT.jar',BFile).

call_alloy2pl_parser_aux(AlloyFile,JARFile,BFile) :-
   compilation_not_needed(AlloyFile,BFile,JARFile),!.
call_alloy2pl_parser_aux(AlloyFile,JARFile,BFile) :-
   (file_exists(BFile) -> delete_file(BFile) ; true), % TODO: only delete generated files
   replace_windows_path_backslashes(BFile,BFileW),
   call_jar_parser(AlloyFile,JARFile,['-toProlog',BFileW],alloy2b).

alloy_pl_filename(AlloyFile,BFile) :- atom_codes(AlloyFile,BasenameC),
    append(BasenameC,".pl",NewC),
    atom_codes(BFile,NewC),!.
alloy_pl_filename(Filename, BFile) :-
    add_failed_call_error(alloy_pl_filename(Filename,BFile)),fail.

compilation_not_needed(BaseFile,DerivedFile,LibraryFile) :-
    file_exists(BaseFile),
    file_property(BaseFile, modify_timestamp, BTime),
    %system:datime(BTime,datime(_Year,Month,Day,Hour,Min,Sec)), format('~w modfied on ~w/~w at ~w:~w (~w sec)~n',[BaseFile,Day,Month,Hour,Min,Sec]),
    file_exists(DerivedFile),
    file_property(DerivedFile, modify_timestamp, DTime),
    DTime > BTime,
    absolute_file_name(prob_lib(LibraryFile),FullLibFile),
    (file_exists(FullLibFile)
     -> file_property(FullLibFile, modify_timestamp, LibTime),
        DTime > LibTime % check that file was generated by this library (hopefully); ideally we should put version numbers into the DerivedFiles
     ;  true % Jar/command not available anyway; use the derived file
     ).

% ---------------------------

call_fuzz_parser(TexFile,FuzzFile) :-
    get_command_in_lib(fuzz,FuzzCmd),
    absolute_file_name(prob_lib('fuzzlib'),FUZZLIB), % TO DO: windows
    fuzz_filename(TexFile,FuzzFile),
    % fuzz options:
    % -d       Allow use before definition
    % -l       Lisp-style echoing of input
    % -p file  Use <file> in place of standard prelude
    debug_format(19,'Fuzz: running ~w -d -l with library ~w on ~w~n',[FuzzCmd,FUZZLIB,TexFile]),
    (file_exists(FUZZLIB)
      -> my_system_call5(FuzzCmd,['-d', '-l', '-p', file(FUZZLIB), file(TexFile)],call_fuzz_parser,Text,_ErrText)
      ;  my_system_call5(FuzzCmd,['-d', '-l', file(TexFile)],call_fuzz_parser,Text,_ErrText)
    ),
    format('Writing fuzz AST to ~s~n',[FuzzFile]),
    open(FuzzFile,write,Stream),
    call_cleanup(format(Stream,'~s~n',[Text]), close(Stream)).

get_command_in_lib(Name,Command) :-
    get_binary_extensions(Extensions),
    absolute_file_name(prob_lib(Name),
                       Command,
                       [access(exist),extensions(Extensions),solutions(all),file_errors(fail)]).

get_binary_extensions(List) :-
   host_platform(windows),!, List=['.exe','']. % will backtrack in get_command_in_lib
get_binary_extensions(['']).

fuzz_filename(TexFile,FuzzFile) :-
    split_filename(TexFile,Basename,_Extension),
    atom_chars(Basename,BasenameC),
    append(BasenameC,['.','f','u','z','z'],CC),
    atom_chars(FuzzFile,CC),!.
fuzz_filename(TexFile, FuzzFile) :-
    add_failed_call_error(fuzz_filename(TexFile,FuzzFile)),fail.
