% (c) 2015-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(pge_algo, [
          %%%% Static Analysis Predicates %%%%
          tcltk_disabling_analysis/1,
          %%%% Model Checker predicates %%%%
          compute_transitions_pge/2,
          %%%% INFO %%%%
          disabled_operations_stats/1,
          do_not_evaluate_guard/0,
          is_pge_opt_on/0,
          is_pge_opt_on/1,
          compute_and_update_disabled_enabled_operations/3,
          compute_operation_relations_for_pge/1,
          compute_operation_relations_for_pge_if_necessary/2
           ]).

/* SICSTUS Libraries */
:- use_module(library(ordsets)).
:- use_module(library(lists)).
%% :- use_module(library(timeout)).

/* B module sources */
:- use_module(probsrc(bmachine),[b_top_level_operation/1,b_machine_has_constants_or_properties/0]).
:- use_module(probsrc(b_state_model_check), [get_guard/2,get_negated_guard/3]).
:- use_module(probsrc(state_space),[clear_context_state/0]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2,create_negation/2]).
:- use_module(probporsrc(static_analysis), [
      get_conj_inv_predicate/3,
      catch_and_ignore_well_definedness_error/2,
      test_path_non_failing/6]).
:- use_module(probsrc(enabling_analysis), [init_or_op/1, print_enable_table/1]).
:- use_module(probsrc(b_read_write_info),[b_get_read_write/3,b_get_operation_read_guard_vars/4]).

/* Importing predicates for handling the states of a respective B/Event-B machine */
:- use_module(probsrc(eventhandling),[register_event_listener/3]).
:- use_module(probsrc(specfile),[specfile_possible_trans_name/2,prepare_state_for_specfile_trans/2,animation_mode/1]).
:- use_module(probsrc(state_space),[ %visited_expression/2,
                                    assert_time_out_for_node/3,
                                    use_no_timeout/1,assert_max_reached_for_node/1]).
:- use_module(probsrc(value_persistance), [save_constants/1,add_new_transitions_to_cache/1]).
:- use_module(probsrc(succeed_max),[reset_max_reached/0,max_reached/0]).
:- use_module(probsrc(preferences),[get_preference/2]).
:- use_module(probsrc(error_manager),[enter_new_error_scope/1, exit_error_scope/3,specific_event_occurred_at_level/2,clear_all_errors_in_error_scope/1,
                                      is_time_out_result/1,set_error_context/1,clear_error_context/0,time_out_with_enum_warning_for_findall/3,
                                      add_warning/3, add_error_fail/3]).
:- use_module(probsrc(tools),[cputime/1,print_message/1]).
:- use_module(probsrc(debug)).
:- use_module(probltlsrc(ltl),[init_states/1]).

/* Modules and Infos for the code coverage analysis */
:- use_module(probsrc(module_information)).
:- module_info(group,model_checker).
:- module_info(description,'This module contains predictes for model checking B and Event-B models using the partial guard evaluation optimisastion.').

%------------------------------------

%%%%%% Dynamic predicates used for: saving the enabling relations, number of skipped guard evaluations,
%%%%%% and sets of enabled and disabled events at the various states. %%%%%

% for the PGE analysis
:- dynamic disables/2, enables/2, keeps/2.
%%% enabling_relations(Op1,Op2,Enable,KeepEnabled,Disable,KeepDisabled)
:- dynamic enabling_relations/6.
:- dynamic pge_predicates_computed/0.

% for the state space exploration by means of PGE
:- dynamic disabled_operations/3.
:- dynamic do_not_evaluate_guard/0.

% for keeping track of the skipped guard evaluations
:- dynamic nr_skipped_guard_evaluations/1.

:- register_event_listener(clear_specification,clear_dynamic_predicates_for_pge,
                            'Partial Guard Evaluation.').

clear_dynamic_predicates_for_pge :-
   retractall(keeps(_,_)),
   retractall(disables(_,_)),
   retractall(disabled_operations(_,_,_)),
   retractall(pge_predicates_computed),
   retractall(do_not_evaluate_guard),
   retractall(nr_skipped_guard_evaluations(_)),
   retractall(enabling_relations(_,_,_,_,_,_)).

disabled_operations_stats(ResPGE) :-
  is_pge_opt_on,
  animation_mode(b),!,
  nr_skipped_guard_evaluations(NrSkippedEvaluations),
  ResPGE = ['SKIPPED_GUARD_EVALUATIONS: ',NrSkippedEvaluations].
disabled_operations_stats(ResPGE) :- % otherwise no information is available
  ResPGE = [].

is_pge_opt_on :-
  get_preference(pge,Pref),
  is_pge_opt_on(Pref).

is_pge_opt_on(off) :- !,fail.
is_pge_opt_on(Pref) :-
  member(Pref,[disabled,disabled2,enabled,enabled2,full,full2]),!.
is_pge_opt_on(Pref) :-
  add_error_fail(pge,'Unknown Partial Guard Evaluation (PGE) preference: ', Pref).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%% PGE Analysis %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

compute_operation_relations_for_pge_if_necessary(off,_FindInvViolations).
compute_operation_relations_for_pge_if_necessary(PGEPref,FindInvViolations) :-
  is_pge_opt_on(PGEPref),!,
  compute_operation_relations_for_pge(FindInvViolations).
compute_operation_relations_for_pge_if_necessary(_PGEPref,_FindInvViolations).

compute_operation_relations_for_pge(FindInvViolations) :-
  (pge_predicates_computed
    -> true
  ; debug_println(19,'********** START PGE ANALYSIS *************'),
    get_preference(timeout_cbc_analysis,Timeout),
    cputime(T1),
    enter_new_error_scope(ScopeID),
    get_preference(pge,Pref),
    (member(Pref,[disabled,enabled,full]) ->
      call_cleanup(compute_impossible_keep_guaranteed(FindInvViolations,Timeout,Pref),
                   my_exit_error_scope(ScopeID))
    ; member(Pref,[disabled2,enabled2,full2]) -> % now used for experimental issues
      call_cleanup(compute_enabling_relations(FindInvViolations,Timeout,Pref),
                   my_exit_error_scope(ScopeID))
    ; otherwise ->
      add_warning(pge,'Unknown PGE preference: ',Pref)
    ),
    assert_disabled_ops_to_init_states_if_necessary,
    cputime(T2),
    debug_println(19,'********** PGE ANALYSIS FINISHED **********'),
    D is T2-T1,
    formatsilent('PGE Analysis Checking Time: ~w ms.~n',[D]),
    assert(pge_predicates_computed),
    assert(nr_skipped_guard_evaluations(0)),
    !
  ).

%%%%%%%% Analysis for the enabling relations: keep, guaranteed, and impossible (in case 'pge' is set to 'enabled', 'disabled', or 'full') %%%%%%%%%%
% failure loop
compute_impossible_keep_guaranteed(FindInvViolations,Timeout,Pref) :-
  init_or_op(OpName1),
  b_top_level_operation(OpName2),
  b_get_read_write(OpName1,_Reads1,Writes1),
  b_get_operation_read_guard_vars(OpName2,true,GReads2,_Precise),
  (OpName1\='INITIALISATION',\+ord_intersect(Writes1,GReads2) ->
    assert(keeps(OpName1,OpName2)),
    debug_println(9,keeps(OpName1, OpName2))
  ; (Pref\=enabled, disables(OpName1,OpName2,FindInvViolations,Timeout)) ->
    assert(disables(OpName1,OpName2)),
    debug_println(9,disables(OpName1,OpName2))
  ; (Pref\=disabled, enables(OpName1,OpName2,FindInvViolations,Timeout)) ->
    assert(enables(OpName1,OpName2)),
    debug_println(9,enables(OpName1,OpName2))
  ; true
  ),fail.
compute_impossible_keep_guaranteed(_FindInvViolations,_Timeout,_Pref).

disables('INITIALISATION',OpName2,FindInvViolations,Timeout) :-
  get_goal_predicate(OpName2,FindInvViolations,no,GoalPred),
  (catch_and_ignore_well_definedness_error(sap:testcase_path_timeout(init,Timeout,[],GoalPred,_,_,_,_,_R),_Result) -> fail; true).
disables(OpName1, OpName2, FindInvViolations, Timeout) :-
  get_start_predicate(OpName1,FindInvViolations,StartPred),
  get_goal_predicate(OpName2,FindInvViolations,no,GoalPred),
  (catch_and_ignore_well_definedness_error(sap:testcase_path_timeout(StartPred,Timeout,[OpName1],GoalPred,_,_,_,_,_R),_Result) ->fail; true).

enables('INITIALISATION',OpName2,FindInvViolations,Timeout) :-
  get_goal_predicate(OpName2,FindInvViolations,yes,GoalPred),
  (catch_and_ignore_well_definedness_error(sap:testcase_path_timeout(init,Timeout,[],GoalPred,_,_,_,_,_R),_Result) ->fail; true).
enables(OpName1, OpName2, FindInvViolations, Timeout) :-
  get_start_predicate(OpName1,FindInvViolations,StartPred),
  get_goal_predicate(OpName2,FindInvViolations,yes,GoalPred),
  (catch_and_ignore_well_definedness_error(sap:testcase_path_timeout(StartPred,Timeout,[OpName1],GoalPred,_,_,_,_,_R),_Result) ->fail; true).

get_start_predicate(OpName,FindInvViolations,StartPred) :-
  get_guard(OpName,PosGuard),
  (FindInvViolations=1 -> StartPred = pred(PosGuard); StartPred=typing(PosGuard)).

get_goal_predicate(OpName,FindInvViolations,Negated,GoalPred) :-
  get_negated_guard(OpName,PosGuard,NegGuard),
  (Negated = yes -> GuardPred = NegGuard; GuardPred = PosGuard),
  get_conj_inv_predicate([GuardPred],FindInvViolations,GoalPred).

%%%%%%%% Analysis for the enabling relations: ER(e1,e2) (in case 'pge' is set to 'enabled2', 'disabled2', or 'full2') %%%%%%%%%%
% failure loop
compute_enabling_relations(FindInvViolations,Timeout,Pref) :-
  init_or_op(OpName1),
  b_top_level_operation(OpName2),
  b_get_read_write(OpName1,_Reads1,Writes1),
  b_get_operation_read_guard_vars(OpName2,true,GReads2,_Precise),
  (OpName1='INITIALISATION' ->
    true % we do not predict the enabling status of the events in the init state, for now...
  ; otherwise ->
    ((\+ord_intersect(Writes1,GReads2)) ->
      assert(enabling_relations(OpName1,OpName2,false,ok,false,ok)),
      debug_println(9,enabling_relations(OpName1,OpName2,false,ok,false,ok))
    ; otherwise ->
      compute_cbc_enabling_relation_and_assert(OpName1,OpName2,FindInvViolations,Timeout,Pref)
    )),fail.
compute_enabling_relations(_FindInvViolations,_Timeout,_Pref).

compute_cbc_enabling_relation_and_assert(OpName1,OpName2,FindInvViolations,Timeout,Pref) :-
  b_get_read_write(OpName1,_Reads1_,Writes1),
  get_negated_guard(OpName2,PosGuard2,NegGuard2),
  enabling_analysis: partition_predicate(PosGuard2,Writes1,FilteredPosGuard2,StaticPosGuard2),
  create_negation(FilteredPosGuard2,FilteredNegGuard2),
  (OpName1==OpName2 ->
    KeepDisabled=false, Enable=false,
    (Pref=enabled2 ->
      test_path_non_failing(PosGuard2,[OpName1],FilteredNegGuard2,FindInvViolations,Timeout,Disable),
      KeepEnabled = unknown
    ;Pref=disabled2 ->
      translate: print_bexpr(PosGuard2),
      translate: print_bexpr(FilteredPosGuard2),
      test_path_non_failing(PosGuard2,[OpName1],FilteredPosGuard2,FindInvViolations,Timeout,KeepEnabled),
      Disable = unknown
    ; otherwise ->
      test_path_non_failing(PosGuard2,[OpName1],FilteredNegGuard2,FindInvViolations,Timeout,Disable),
      test_path_non_failing(PosGuard2,[OpName1],FilteredPosGuard2,FindInvViolations,Timeout,KeepEnabled)
    )
  ; otherwise ->
    conjunct_predicates([StaticPosGuard2,FilteredNegGuard2],NegEnableGuard2),
    (Pref=enabled2 -> % need to known just Disable and KeepDisabled
      Enable = unknown, KeepEnabled = unknown,
      test_path_non_failing(PosGuard2,[OpName1],FilteredNegGuard2,FindInvViolations,Timeout,Disable),
      test_path_non_failing(NegGuard2,[OpName1],NegGuard2,FindInvViolations,Timeout,KeepDisabled)
    ;Pref=disabled2 -> % need to known just Enable and KeepEnabled
      test_path_non_failing(NegEnableGuard2,[OpName1],FilteredPosGuard2,FindInvViolations,Timeout,Enable),
      test_path_non_failing(PosGuard2,[OpName1],FilteredPosGuard2,FindInvViolations,Timeout,KeepEnabled),
      Disable = unknown, KeepDisabled = unknown
    ; otherwise -> % we have full2
      test_path_non_failing(NegEnableGuard2,[OpName1],FilteredPosGuard2,FindInvViolations,Timeout,Enable),
      test_path_non_failing(PosGuard2,[OpName1],FilteredPosGuard2,FindInvViolations,Timeout,KeepEnabled),
      test_path_non_failing(PosGuard2,[OpName1],FilteredNegGuard2,FindInvViolations,Timeout,Disable),
      test_path_non_failing(NegGuard2,[OpName1],NegGuard2,FindInvViolations,Timeout,KeepDisabled)
    )
  ),
  prove_and_assert_enabling_relation(OpName1,OpName2,Enable,KeepEnabled,Disable,KeepDisabled),
  debug_println(9,enabling_relations(OpName1,OpName2,Enable,KeepEnabled,Disable,KeepDisabled)).

prove_and_assert_enabling_relation(OpName1,OpName2,Enable,KeepEnabled,Disable,KeepDisabled) :-
  (maplist(nonvar,[Enable,KeepEnabled,Disable,KeepDisabled]) ->
    assert(enabling_relations(OpName1,OpName2,Enable,KeepEnabled,Disable,KeepDisabled))
  ; add_warning(cbc_enabling_relation,'A variable result occurred for one of the enabling relations result',
                [Enable,KeepEnabled,Disable,KeepDisabled])).


my_exit_error_scope(ScopeID) :-
  (specific_event_occurred_at_level(ScopeID,Event) ->
   format('### ERROR/ENUMERATION WARING OCCURED : ~w !~n',[Event])),fail.
my_exit_error_scope(ScopeID) :-  %error_manager:print_error_scopes,
  clear_all_errors_in_error_scope(ScopeID),
  exit_error_scope(ScopeID,_,my_exit_error_scope).

% TODO: consider if we need this at all
tcltk_disabling_analysis(list([list(['Origin'|Ops])|Result])) :-
  get_preference(timeout_cbc_analysis,Timeout),
  findall(Op, b_top_level_operation(Op), Ops),
  findall(list([OpName1|DisableList]),
     (init_or_op(OpName1),
      findall(Disables,(b_top_level_operation(OpName2),(disables(OpName1,OpName2,1,Timeout) -> Disables=yes; Disables=no)),DisableList)),
     Result),
  print_enable_table([list(['Origin'|Ops])|Result]).

%----------------- Auxiliary predicates for the Partial Guard Evaluation implementation

:- use_module(probsrc(tcltk_interface),[add_transition/4]).
add_transition_timeout(CurID,CurTO,CurState,ActionName,NewId) :-
    set_error_context(operation(ActionName,CurID)),
    (CurTO > 0
        -> time_out_with_enum_warning_for_findall(tcltk_interface:add_transition(CurID,CurState,ActionName,NewId),CurTO,TimeOutRes),
           check_timeout_result(CurID,ActionName,CurTO,TimeOutRes)
        ;  add_transition(CurID,CurState,ActionName,NewId)
    ).

check_timeout_result(CurID,ActionName,CurTO,TimeOutRes) :-
    (is_time_out_result(TimeOutRes)
        ->
        ( TimeOutRes = virtual_time_out(_) ->
            true
        ;   print_message('TIME OUT: '), print_message(CurID), print_message(CurTO),
          print_message(ActionName)
        ),
    (var(ActionName)
         -> assert_time_out_for_node(CurID,'unspecified_operation',TimeOutRes)
         ;  assert_time_out_for_node(CurID,ActionName,TimeOutRes))
         ;  true
  ).

setup_constants_state1(csp_and_b_root) :- !.
setup_constants_state1(CurState) :-
  setup_constants_state(CurState).

setup_constants_state(concrete_constants(_)).
setup_constants_state(const_and_vars(_,_)).

% failure loop
assert_disabled_ops_to_init_states_if_necessary :-
  \+b_machine_has_constants_or_properties, % we do not want to explore states prior to firing the model checker
  findall(Op,disables('INITIALISATION',Op),DisabledOps),
  findall(Op,enables('INITIALISATION',Op),EnabledOps),
  % init states finds exactly the initial states of the checked machine
  init_states(Init),
  member(I,Init),
  assert(disabled_operations(I,DisabledOps,EnabledOps)),
  fail.
assert_disabled_ops_to_init_states_if_necessary.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%% State Exploration by means of PGE %%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(probsrc(tcltk_interface),[catch_clpfd_overflow_call_for_state/3]).
compute_transitions_pge(CurID,CurState) :-
   catch_clpfd_overflow_call_for_state(CurID,
                                            pge_algo:compute_transitions_pge2(CurID,CurState),
                                            clear_context_state).
:- public compute_transitions_pge2/2.
compute_transitions_pge2(CurID,CurState) :-
  reset_max_reached,
  prepare_state_for_specfile_trans(CurState,PreparedCurState),
  (use_no_timeout(CurID) ->
    add_transitions_pge(CurID,PreparedCurState,0)
  ;otherwise ->
    preferences: preference(time_out,CurTO),
    add_transitions_pge(CurID,PreparedCurState,CurTO)
  ),
  (setup_constants_state1(CurState) -> true ; retractall(disabled_operations(CurID,_,_))),
  debug_println(9,state_has_been_explored(CurID,CurState)),
  clear_error_context,
  (max_reached -> assert_max_reached_for_node(CurID),MaxReached=true ; MaxReached=false),
  (CurID==root -> save_constants(MaxReached) ; add_new_transitions_to_cache(CurID)).


add_transitions_pge(CurID,CurState,CurTO) :-
  % PossiblyEnActionNames contains all possible actions that are not recognised to be disabled and need to be tested for enabledness
  get_possible_enabled_trans_names(CurID,CurState,DisabledOperations,EnabledOperations,PossiblyEnActionNames),
  explore_current_state_pge(PossiblyEnActionNames,CurID,CurState,CurTO,EnabledOperations,Transitions,DisabledActionsNew),
  list_to_ord_set(DisabledActionsNew,DisabledActionsNewSet),
  ord_union(DisabledOperations,DisabledActionsNewSet,DisabledActionsAtCurState),
  maplist(get_first_argument,Transitions,EnabledActions),
  list_to_ord_set(EnabledActions,EnabledActionsAtCurState),
  compute_and_update_disabled_enabled_operations(Transitions,DisabledActionsAtCurState,EnabledActionsAtCurState).

get_possible_enabled_trans_names(CurID,CurState,DisabledOperations,EnabledOperations,PossiblyEnActionNames) :-
  (disabled_operations(CurID,DisabledOperations,EnabledOperations) -> true; DisabledOperations = [], EnabledOperations = []),
  get_possible_enabled_trans_names_aux(CurID,CurState,DisabledOperations,PossiblyEnActionNames).

get_possible_enabled_trans_names_aux(CurID,CurState,DisabledOperations,PossiblyEnActionNames) :-
  findall(PossibleEnActionName,
    (specfile_possible_trans_name(CurState,PossibleEnActionName),not_a_member_of_disabled(PossibleEnActionName,DisabledOperations)),
          PossiblyEnActionNames),
  debug_println(9,skipped_guard_evaluations_at_state(CurID,DisabledOperations)).

not_a_member_of_disabled(ActionName,DisabledOperations) :-
  (ord_member(ActionName,DisabledOperations) ->
    retract(nr_skipped_guard_evaluations(NrDisabled)),
    NrDisabled1 is NrDisabled + 1,
    assert(nr_skipped_guard_evaluations(NrDisabled1)),
    fail
  ; true).

explore_current_state_pge(ActionNames,CurID,CurState,CurTO,EnabledOperations,Transitions,DisabledActionsNew) :-
    explore_current_state_pge1(ActionNames,CurID,CurState,CurTO,EnabledOperations,[],Transitions,[],DisabledActionsNew).

explore_current_state_pge1([],_CurID,_CurState,_CurTO,_EnabledOperations,Transitions,Transitions,DisabledActions,DisabledActions).
explore_current_state_pge1([ActionName|R],CurID,CurState,CurTO,EnabledOperations,TransitionsGen,Transitions,DisabledActionsGen,DisabledActions) :-
  (ord_member(ActionName,EnabledOperations) ->
    retract(nr_skipped_guard_evaluations(NrDisabled)),
    NrDisabled1 is NrDisabled +1,
    assert(nr_skipped_guard_evaluations(NrDisabled1)),
    debug_println(9,do_not_evaluate_guard(ActionName)),
    assert(do_not_evaluate_guard)
  ;otherwise ->
    true
  ),
  findall(NewId,add_transition_timeout(CurID,CurTO,CurState,ActionName,NewId),NewIds),
  (NewIds = [] -> % ActionName is disabled at CurState
    TransitionsGen1 = TransitionsGen,
    DisabledActionsGen1 = [ActionName|DisabledActionsGen]
  ; otherwise ->
    TransitionsGen1 = [(ActionName,NewIds)|TransitionsGen],
    DisabledActionsGen1 = DisabledActionsGen
  ),retractall(do_not_evaluate_guard),
  explore_current_state_pge1(R,CurID,CurState,CurTO,EnabledOperations,TransitionsGen1,Transitions,DisabledActionsGen1,DisabledActions).

compute_and_update_disabled_enabled_operations([],_DisabledActionsAtCurState,_EnabledActionsAtCurState).
compute_and_update_disabled_enabled_operations([(ActionName,SuccStates)|R],DisabledActionsAtCurState,EnabledActionsAtCurState) :-
  compute_disabled_operations(ActionName,DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateDisabledActions),
  compute_enabled_operations(ActionName,DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateEnabledActions),
  update_disabled_enabled_operations_l(SuccStates,SuccStateDisabledActions,SuccStateEnabledActions),
  compute_and_update_disabled_enabled_operations(R,DisabledActionsAtCurState,EnabledActionsAtCurState).

compute_disabled_operations(ActionName,DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateDisabledActions) :-
  get_preference(pge,Pref),
  (check_only_enabled(Pref) ->
    SuccStateDisabledActions = []
  ; compute_disabled_operations_after_state(Pref,ActionName,DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateDisabledActions)
  ).

compute_enabled_operations(ActionName,DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateEnabledActions) :-
  get_preference(pge,Pref),
  (check_only_disabled(Pref) ->
    SuccStateEnabledActions = []
  ; compute_enabled_operations_after_state(Pref,ActionName,DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateEnabledActions)
  ).

update_disabled_enabled_operations_l(StateIDs,NewDisabledOperations,NewEnabledOperations) :-
  maplist(update_disabled_enabled_operations(NewDisabledOperations,NewEnabledOperations),StateIDs).

update_disabled_enabled_operations(NewDisabledOperations,NewEnabledOperations,StateID) :-
  (disabled_operations(StateID,DisabledOperations,EnabledOperations) ->
    ord_union(DisabledOperations,NewDisabledOperations,DisabledOperations1),
    ord_union(EnabledOperations,NewEnabledOperations,EnabledOperations1),
    retractall(disabled_operations(StateID,_,_)),
    assert(disabled_operations(StateID,DisabledOperations1,EnabledOperations1))
  ; otherwise ->
    assert(disabled_operations(StateID,NewDisabledOperations,NewEnabledOperations))
  ).

compute_disabled_operations_after_state(Pref,OpName1,DisabledActionsAtCurState,_EnabledActionsAtCurState,SuccStateDisabledActions) :-
  (Pref = disabled; Pref = full),!,
  findall(OpName2,disables(OpName1,OpName2),ImpossibleEnabled),
  findall(OpName2,get_disabled_action_by_keep(OpName1,DisabledActionsAtCurState,OpName2),SuccStateDisabledActions1,ImpossibleEnabled),
  list_to_ord_set(SuccStateDisabledActions1,SuccStateDisabledActions).
compute_disabled_operations_after_state(Pref,OpName1,DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateDisabledActions) :-
  (Pref = disabled2; Pref = full2),!,
  findall(OpName2,(member(OpName2,EnabledActionsAtCurState),enabling_relations(OpName1,OpName2,_,false,_,_)),DisbaledOps1),
  findall(OpName2,(animation_mode(b),member(OpName2,DisabledActionsAtCurState),enabling_relations(OpName1,OpName2,false,_,_,_)),DisabledOps2,DisbaledOps1),
  list_to_ord_set(DisabledOps2,SuccStateDisabledActions).
compute_disabled_operations_after_state(_Pref,_OpName1,_DisabledActionsAtCurState,_EnabledActionsAtCurState,[]).

get_disabled_action_by_keep(OpName1,DisabledActionsAtCurState,OpName2) :-
  animation_mode(b),!, % can be applied only in B mode, for CSP||B is generally not sound
  member(OpName2,DisabledActionsAtCurState),
  keeps(OpName1,OpName2).

compute_enabled_operations_after_state(Pref,OpName1,_DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateEnabledActions) :-
  (Pref = enabled; Pref = full),!,
  findall(OpName2,enables(OpName1,OpName2),GuaranteedEnabled),
  findall(OpName2,get_enabled_action_by_keep(OpName1,EnabledActionsAtCurState,OpName2),SuccStateEnabledActions1,GuaranteedEnabled),
  list_to_ord_set(SuccStateEnabledActions1,SuccStateEnabledActions).
compute_enabled_operations_after_state(Pref,OpName1,DisabledActionsAtCurState,EnabledActionsAtCurState,SuccStateEnabledActions) :-
  (Pref = enabled2; Pref = full2),!,
  findall(OpName2,(member(OpName2,EnabledActionsAtCurState),enabling_relations(OpName1,OpName2,_,_,false,_)),EnabledOps1),
  findall(OpName2,(animation_mode(b),member(OpName2,DisabledActionsAtCurState),enabling_relations(OpName1,OpName2,_,_,_,false)),EnabledOps2,EnabledOps1),
  list_to_ord_set(EnabledOps2,SuccStateEnabledActions).
compute_enabled_operations_after_state(_Pref,_OpName1,_DisabledActionsAtCurState,_EnabledActionsAtCurState,[]).

get_enabled_action_by_keep(OpName1,EnabledActionsAtCurState,OpName2) :-
  member(OpName2,EnabledActionsAtCurState),
  keeps(OpName1,OpName2).

get_first_argument((A,_B),A).

check_only_disabled(disabled).
check_only_disabled(disabled2).

check_only_enabled(enabled).
check_only_enabled(enabled2).

%------------------------------
