% (c) 2014-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(static_analysis, [dependent_actions/5, enable_analysis/6, disable_analysis/5, action_dependent_to_itself/4, tcltk_enabling_analysis/2,
                            compute_dependendency_relation_of_all_events_in_the_model/3,
                            enable_graph/1,dependent_act/4,
                            syntactic_independence/3,
                            get_conj_inv_predicate/3, % predicate used in the enabling analysis module
                            catch_and_ignore_well_definedness_error/2,
                            test_path_non_failing/6,
                            test_path/6,
                            is_timeout/1
                            ]).

:- meta_predicate catch_enumeration_warning(0,0).
:- meta_predicate catch_and_ignore_well_definedness_error(0,*).


/* Modules and Infos for the code coverage analysis */
:- use_module(probsrc(module_information)).
:- module_info(group,model_checker).
:- module_info(description,'This module provides predicates for static analysis of Event-B and B models').

% Standard SICSTUS prolog libraries
%% :- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(ugraphs)).

% Classical B prolog modules
:- use_module(probsrc(bmachine),[b_top_level_operation/1, b_get_invariant_from_machine/1, b_get_properties_from_machine/1]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2]).
:- use_module(probsrc(b_state_model_check), [get_negated_guard/3,get_guard/2]).
:- use_module(probsrc(bmachine),[b_get_machine_operation_for_animation/6]). % TODO: predicate used for determining whether an event is self-dependent to itself (consider whether to remove)

% Event-B prolog modules

% POR modules
:- use_module(probporsrc(enabling_predicates),[compute_enabling_predicate/5]).

% Importing predicates for using CBC
:- use_module(probsrc(sap), [testcase_path_timeout/9]).
:- use_module(probsrc(b_read_write_info), [b_get_read_write/3, b_get_read_write_vars/5, b_get_operation_read_guard_vars/4]).

% Other modules (used mostly for debugging)
%:- use_module(probsrc(dot_graph_generator),[gen_dot_graph/6]).
:- use_module(probsrc(preferences),[get_preference/2]).
:- use_module(probsrc(error_manager),[real_error_occurred/0, logged_error/4, reset_errors/0,
                                      call_in_fresh_error_scope_for_one_solution/1,
                                      add_internal_error/2, add_error_fail/3,
                                      enumeration_warning_occured_in_error_scope/0, clear_enumeration_warnings/0]).
:- use_module(probsrc(debug),[debug_println/2, debug_mode/1, debug_format/3, formatsilent/2]).
:- use_module(probsrc(tools),[cputime/1]).

% Importing unit tests predicates
%% :- use_module(probsrc(self_check)).

is_timeout(timeout).
is_timeout(time_out).
is_timeout(clpfd_overflow).
is_timeout(overflow).
is_timeout(virtual_time_out).

% --------------------------- Static analysis for determining the dependece/independence of events (Begin) -----------------------------%

% check syntactically influence of executing OpName1 on OpName2's enabling and effect
syntactic_independence(OpName1,OpName2,Res) :-
   b_top_level_operation(OpName1),
   b_top_level_operation(OpName2),
   b_get_read_write_vars(OpName1,GReads1,AReads1,Reads1,Writes1),
   b_get_read_write_vars(OpName2,GReads2,AReads2,Reads2,Writes2),
   \+ ord_intersect(Writes1,GReads2),% otherwise Op1 can influence guard of Op2
   (ord_intersect(Writes1,AReads2)
      ->  % guard not modified, but effect of OpName2 could change
          Res = syntactic_keep
       ;  (\+ ord_intersect(Writes1,Writes2),  % no race
           \+ ord_intersect(GReads1,Writes2), % no enabling/disabling of OpName1 by OpName2
           \+ ord_intersect(AReads1,Writes2)) % no change of effect
          -> (\+ ord_intersect(Reads1,Reads2)
               -> Res = syntactic_fully_independent
                ; Res = syntactic_independent
             )
       ;  otherwise -> Res = syntactic_unchanged % guard is kept and operation effect of OpName2 unchanged; but there could be race conditions on writes or effects of OpName2 on OpName1
    ).

dependent_actions1(OpName1,OpName2,FindInvViolations,Timeout,Res) :-
  b_top_level_operation(OpName1),
  b_top_level_operation(OpName2),
  b_get_read_write_vars(OpName1,GReads1,AReads1,Reads1,Writes1),
  ( OpName1=OpName2    -> action_dependent_to_itself(OpName1,GReads1,Writes1,Res) % TODO: Do we really need this kind of analysis???
  ; OpName2 @< OpName1 -> Res = '-'  % our checking is symmetric; check only one pair
  ; otherwise ->
      b_get_read_write_vars(OpName2,GReads2,AReads2,Reads2,Writes2),
      ( ord_intersect(Writes1,Writes2) -> Res = race_dependent
      ; (\+ ord_intersect(Writes1,Reads2),\+ ord_intersect(Writes2,Reads1)) -> Res = syntactic_independent
      ; otherwise ->
          (get_preference(use_cbc_analysis,true) -> % sometimes we don't want to use the CBC
            (\+ord_intersect(AReads1,Writes2),
             \+ord_intersect(AReads2,Writes1) ->
              (get_preference(dependency_enable_predicates,true) -> %trace,
                get_dependency_enabling_predicate(OpName1,GReads1,Writes1,OpName2,GReads2,Writes2,FindInvViolations,Timeout,Res)
              ; otherwise ->
                test_enabledness_condition(OpName1,GReads1,Writes1,OpName2,GReads2,Writes2,FindInvViolations,Timeout,Res)
              )
            ; otherwise -> Res = action_dependent
            )
          ; otherwise -> Res = dependent  % use_cbc_analysis = false
          )
      )
  ).

% Testing whether it is possible that the one of the events can disable the other one
test_enabledness_condition(OpName1,GReads1,Writes1,OpName2,GReads2,Writes2,FindInvViolations,Timeout,Res) :-
  get_negated_guard(OpName1,PosGuard1,NegGuard1),
  get_negated_guard(OpName2,PosGuard2,NegGuard2),
  %% get_conj_inv_predicate([PosGuard1,PosGuard2],FindInvViolations,GuardsConj),
  conjunct_predicates([PosGuard1,PosGuard2],GuardsConj), % both events are enabled
  ( ord_intersect(GReads1,Writes2),
    test_path(GuardsConj,[OpName2],NegGuard1,FindInvViolations,Timeout,_R1)
   -> Res = guard_dependent
  ; ord_intersect(Writes1,GReads2),
    test_path(GuardsConj,[OpName1],NegGuard2,FindInvViolations,Timeout,_R2)
   -> Res = guard_dependent
  ; otherwise -> Res = independent
  ).

get_dependency_enabling_predicate(OpName1,GReads1,Writes1,OpName2,GReads2,Writes2,FindInvViolations,Timeout,Res) :-
  get_negated_guard(OpName1,PosGuard1,NegGuard1),
  get_negated_guard(OpName2,PosGuard2,NegGuard2),
  conjunct_predicates([PosGuard1,PosGuard2],GuardsConj),
  ( (ord_intersect(GReads1,Writes2),test_path(GuardsConj,[OpName2],NegGuard1,FindInvViolations,Timeout,_R1)) ->
      % get enabling predicate after executing OpName2
      compute_enabling_predicate(OpName2,true,PosGuard1,FindInvViolations,predicate(Enable1)),
      ( (ord_intersect(Writes1,GReads2),test_path(GuardsConj,[OpName1],NegGuard2,FindInvViolations,Timeout,_R2)) ->
        % get enabling predicate after executing OpName1
        compute_enabling_predicate(OpName1,true,PosGuard2,FindInvViolations,predicate(Enable2)),
        conjunct_predicates([Enable1,Enable2],EnablingPredicate),
        Res = guard_dependent(EnablingPredicate)
      ; otherwise ->
        Res = guard_dependent(Enable1)
      )
  ; (ord_intersect(Writes1,GReads2),test_path(GuardsConj,[OpName1],NegGuard2,FindInvViolations,Timeout,_R2)) ->
    compute_enabling_predicate(OpName1,true,PosGuard2,FindInvViolations,predicate(Enable2)),
    Res = guard_dependent(Enable2)
  ; otherwise ->
    Res = independent
  ),
  (Res = guard_dependent(Pred), debug_mode(on) ->
    print('####### Enabling Dependency Predicate for '), print(OpName1), print('<->'), print(OpName2), print(' #######'),nl,
    print('Predicate: '),translate:print_bexpr(Pred),nl,
    print('####################################'),nl
  ; otherwise -> true).

action_dependent_to_itself(OpName,Reads,Writes,Res) :-
    b_get_machine_operation_for_animation(OpName,_Res,Params,_,_,true),
    length(Params,Len),
    (Len>0,\+ord_intersect(Reads,Writes) ->
      Res = self_independent
    ; Len = 0 ->
      Res = '='
    ; otherwise ->
      Res = self_dependent
    ).

% --------------------------- Static analysis for determining the dependece/independence of events (End) -----------------------------%


% --------------------------- Enabling anaysis used for the computation of ample sets (Begin) -----------------------------%

enable_analysis(OpName1,OpName2,FindInvViolations,UseEnableGraph,Timeout,Enable) :-
  b_top_level_operation(OpName1),
  b_top_level_operation(OpName2),
  (OpName1 == OpName2 ->
    Enable=impossible % on top level an operation cannot enable itself
  ; otherwise ->
    b_get_read_write(OpName1,_Reads1,Writes1),
    b_get_operation_read_guard_vars(OpName2,true,GReads2,_Precise),
    get_negated_guard(OpName2,PosGuard2,NegGuard2),
    %enabling_analysis: filter_predicate(PosGuard2,Writes1,FilteredPosGuard2),
    (get_preference(use_cbc_analysis,false) ->
      (ord_intersect(Writes1,GReads2) ->
        get_enabling_result(UseEnableGraph,OpName1,PosGuard2,FindInvViolations,_CBCResult,Enable)
      ; otherwise ->
        Enable=possible_keep
      )
    ; otherwise ->
      get_guard(OpName1,Guard1),
      conjunct_predicates([Guard1,NegGuard2],StartPred),
      ( (ord_intersect(Writes1,GReads2),test_path(StartPred,[OpName1],PosGuard2,FindInvViolations,Timeout,R)) ->
        get_enabling_result(UseEnableGraph,OpName1,PosGuard2,FindInvViolations,R,Enable)
      ; otherwise ->
        Enable=possible_keep_or_disable
      )
    )
  ).

get_enabling_result(UseEnableGraph,OpName1,PosGuard2,FindInvViolations,CBCResult,Enable) :-
  (UseEnableGraph=true ->   %(OpName1 = close_door, OpName2 = push_call_button -> trace;true),
    % check further if Enable is inconsistent, if so then we do not need to add an edge to the enable graph
    (compute_enabling_predicate(OpName1,true,PosGuard2,FindInvViolations,Enable) ->
      true
    ; otherwise ->
      add_error_fail(enable_analysis,'Error occurred while computing enabling predicate for ', OpName1/PosGuard2)
    )
  ; otherwise ->
      (is_timeout(CBCResult)-> Enable=CBCResult; Enable=possible_enable)
  ).

catch_enumeration_warning(Call,Handler) :-
  % throw/1 predicate raises instantiation_error
  on_exception(enumeration_warning(enumerating(_),_Type,_,_,critical),Call,call(Handler)).

catch_and_ignore_well_definedness_error(Call,Result) :-
  catch_enumeration_warning(Call, true),
  (real_error_occurred ->
    ((logged_error(Error,_S,_TContext,_Span), debug_println(19,error_during_static_analysis(Error)),
      Error==well_definedness_error) ->
      reset_errors,
      formatsilent('Well definedness error occurred, but it will be ignored here.~n',[]),
      Result=error
    ; otherwise ->
      formatsilent('No well definedness error occurred. Error will not be ignored here.~n',[]),
      Result=unexpected_error_occurred
    )
  ; otherwise ->
    Result=success
  ),
  (enumeration_warning_occured_in_error_scope ->
    formatsilent('Enumeration warning occurred and will be ignored here.~n',[]),
    clear_enumeration_warnings
    ; true).

disable_analysis(OpName1,OpName2,FindInvViolations,Timeout,Result) :-
  b_top_level_operation(OpName1),
  b_top_level_operation(OpName2),
  b_get_read_write(OpName1,_Reads1,Writes1),
  b_get_operation_read_guard_vars(OpName2,true,GReads2,_Precise),
  get_negated_guard(OpName2,PosGuard2,NegGuard2),
  (get_preference(use_cbc_analysis,false) ->
      (ord_intersect(Writes1,GReads2) ->
        Result=possible_disable
      ; otherwise ->
        Result=possible_keep
      )
  ;otherwise ->
      get_guard(OpName1,PosGuard1),
      conjunct_predicates([PosGuard1,PosGuard2],StartPred),
      ( (ord_intersect(Writes1,GReads2),test_path(StartPred,[OpName1],NegGuard2,FindInvViolations,Timeout,CBCResult)) ->
        (is_timeout(CBCResult)-> Result=CBCResult; Result=possible_disable)
      ; otherwise ->
        Result=cannot_disable
      )
    ).

% test_path(+Start,+Path,+Goal,+WithInv,+Timeout,-R)
% Runs the CBC to check whether there is a possible execution of sequence of events Path in the loaded model
% starting at a state s satisfying Start and reaching a state s' satisfying Goal whitin Timeout milliseconds
% + Start: B predicate
% + Path: a list of events/operations
% + Goal: B predicate
% + Timeout: specifying a timeout (in milliseconds)
% - R: result (result could be either 'ok', 'timeout', 'interrupt', or 'unknown')
%%% pred(P) - adds P to invariant
%%% typing(P) - just use typing from invariant and add P

test_path(Start,Path,Goal,WithInv,Timeout,Res) :-
  (Start=init ->
    Start1 = init
  ; otherwise ->
    (WithInv=1 ->
      Start1 = pred(Start),
      get_conj_inv_predicate([Goal],WithInv,Goal_Inv)
    ; Start1=typing(Start),
      Goal_Inv = Goal)
  ),
  catch_and_ignore_well_definedness_error(testcase_path_timeout(Start1,Timeout,Path,Goal_Inv,_,_,_,_,R),ER),
  check_result(R,ER,Res).

test_path_non_failing(Start,Path,Goal,WithInv,Timeout,R) :-
  (test_path(Start,Path,Goal,WithInv,Timeout,R) -> true; R = false).

check_result(R,ER,Res) :-
  (nonvar(R) -> Res=R
   ; ER = success -> Res = 'ok'
   ; Res = 'unknown').

% conjoin a predicate with invariant and properties if specified
get_conj_inv_predicate(Preds,FindInvViolations,StartPred) :-
  (FindInvViolations==1 ->
    b_get_invariant_from_machine(Inv),
    b_get_properties_from_machine(Prop),
    conjunct_predicates([Prop,Inv|Preds],StartPred) % TO DO: project away
  ; otherwise ->
    (FindInvViolations==0 ->
       true
    ; otherwise ->
       add_internal_error('Illegal flag: ',get_conj_inv_predicate(Preds,FindInvViolations,StartPred))
    ),
    conjunct_predicates(Preds,StartPred)
  ).

%reads_writes_intersection(Writes, Reads, ComputedResult, PositiveResult, Res) :-
%  (ord_intersect(Writes,Reads) -> Res=PositiveResult
%  ; otherwise ->  Res=ComputedResult ).

tcltk_enabling_analysis(list([list(['Origin'|Ops])|Result]),POR) :-
  get_preference(timeout_cbc_analysis,Timeout),
  %% compute_dependendency_relation_of_all_events_in_the_model(1,por(ample_sets,0,0,0),EnableGraph),
  %% suitable_for_por(EnableGraph,POR),
  POR = option_disabled,
  findall(Op, b_top_level_operation(Op), Ops),
  findall(list([OpName1|EnableList]),
     (b_top_level_operation(OpName1),
      findall(Enable,enable_analysis(OpName1,_OpName2,1,1,Timeout,Enable),EnableList)),
     Result)
  . %,print_enable_table([list(['Origin'|Ops])|Result]).

%% suitable_for_por(EnGraph,Result) :-
%%   (not_all_reachable(EnGraph) ->
%%      Result=true
%%   ;  Result=false
%%   ).

%% not_all_reachable(EnGraph) :-
%%   ample_sets: vertices(EnGraph,Vertices),
%%   member(V,Vertices),
%%   \+ (ample_sets: reachable(V,EnGraph,Reachable),Reachable=Vertices).
% --------------------------- Used for the ample sets (End) -----------------------------%




/************************** DEPENDENCY RELATION (BEGIN) ******************************/

/* predicate which determines the dependency relations between all actions in the model,
   it should be called only once (prior to the model checking) */

:- dynamic enable_graph/1, dependent_act/4.

compute_dependendency_relation_of_all_events_in_the_model(FindInvViolations,por(_TYPE,UseEnableGraph,Depth,_PGE),EnableGraph) :-
  (enable_graph(EnableGraph) ->
    % In case the dependency relations and the enable graph are already computed.
    % Case possible if the user stopped the model checking for a while (and meanwhile didn't reloaded the model)
    % and in case that he wants to continue the verification of the model the enable graph don't have
    % to be computed again.
    true
  ; otherwise ->
    findall(Action, b_top_level_operation(Action),Actions),
    sort(Actions,AllActions),
    retractall(enable_graph(_)),
    debug_println(19,'********** DETERMINE ACTION DEPENDENCIES AND ENABLE GRAPH *************'),
    get_preference(timeout_cbc_analysis,Timeout),
    cputime(T1),
    call_in_fresh_error_scope_for_one_solution(
        determine_dependency_enabling_relations(AllActions,Timeout,FindInvViolations,UseEnableGraph,Depth,EnableGraph)),
    cputime(T2),
    debug_println(19,'********** FINISHED ANALYSIS *************'),
    D is T2-T1, debug_format(19,'Dependency Analysis Time: ~w ms~n',[D]),
    assert(enable_graph(EnableGraph))
  ),
  debug_println(9,enable_graph(EnableGraph)).

determine_dependency_enabling_relations(AllActions,Timeout,FindInvViolations,UseEnableGraph,Depth,EnableGraph) :-
  determine_dependency_enabling_relations(AllActions,AllActions,Timeout,FindInvViolations,UseEnableGraph,Depth,Edges),
  (UseEnableGraph = true ->
     enable_graph: vertices_edges_predicates_to_egraph(AllActions,Edges,EnableGraph)
  ; otherwise -> % to be removed later (here we just want to compare the graph implementation's performances of enable_graph and ugraph)
     vertices_edges_to_ugraph(AllActions,Edges,EnableGraph)
  ).


%determine_dependency_enabling_relations(AllActions,Timeout,FindInvViolations,Depth,EnableGraph) :-
%  determine_dependency_enabling_relations(AllActions,AllActions,Timeout,FindInvViolations,Depth,Edges),
%  enable_graph: vertices_edges_predicates_to_egraph(AllActions,Edges,EnableGraph).

determine_dependency_enabling_relations([],_AllActions,_Timeout,_FindInvViolations,_UseEnableGraph,_Depth,[]).
determine_dependency_enabling_relations([Act|Actions],AllActions,Timeout,FindInvViolations,UseEnableGraph,Depth,Edges) :-
    determine_dependency_enabling_relations1(AllActions,Act,Timeout,FindInvViolations,UseEnableGraph,Depth,Edges1),
    append(Edges1,Rest,Edges),
    determine_dependency_enabling_relations(Actions,AllActions,Timeout,FindInvViolations,UseEnableGraph,Depth,Rest).

determine_dependency_enabling_relations1([],_Act,_Timeout,_FindInvViolations,_UseEnableGraph,_Depth,[]).
determine_dependency_enabling_relations1([Act2|Acts],Act1,Timeout,FindInvViolations,UseEnableGraph,Depth,Result) :-
    ( (Act1==Act2,\+dependent_act(Act1,Act1,_Status,_)) ->
       assert(dependent_act(Act1,Act1,self,true)),
       debug_println(9,dependent_act(Act1,Act1,self,true))
    ; dependent_actions_symm(Act1,Act2,FindInvViolations,Timeout,Status,Coenabled) ->
       assert(dependent_act(Act1,Act2,Status,Coenabled)), % the dependency relation is symmetric
       assert(dependent_act(Act2,Act1,Status,Coenabled)),
       debug_println(9,dependent_act(Act1,Act2,Status,Coenabled))
    ; otherwise ->
       true % no new dependencies have been discovered
    ),
    (may_enable(Act1,Act2,FindInvViolations,Timeout,UseEnableGraph,Depth,Edge) ->
       Result = [Edge|Res1]
    ; otherwise ->
       Result = Res1
    ),
    %compute_if_coenabled(Act1,Act2,FindInvViolations,Timeout),
    determine_dependency_enabling_relations1(Acts,Act1,Timeout,FindInvViolations,UseEnableGraph,Depth,Res1).

may_enable(Act1,Act2,FindInvViolations,Timeout,UseEnableGraph,_Depth,Edge) :-
  enable_analysis(Act1,Act2,FindInvViolations,UseEnableGraph,Timeout,Enable),
  (memberchk(Enable,[guaranteed,possible,possible_enable]) ->
    (UseEnableGraph = true ->
        Edge = Act1-b(truth,pred,[])-Act2
    ; otherwise ->
        Edge = Act1-Act2
    )
  ; Enable = predicate(Expr) -> Edge = Act1-Expr-Act2
  ; is_timeout(Enable) -> Edge = Act1-Act2 % assume it is possible to enable Act2
  ; otherwise -> fail
  ).

/*
   dependent_actions/2 determine statically or dynamically if two actions are dependent
   two events/actions are syntactically dependent in the following three cases:
  1. If both events modify at least one common variable.
  2. Act1 modifies a variable in the guard of Act2. (Act1 can enable or disable Act2)
  3. Act2 modifies a variable in the guard of Act1. (Act2 can enable or disable Act1)
*/

%%% Act1 and Act2 are enabled in current state State,
%%% we don't have to check if both actions are enabled in
%%% some state
dependent_actions_symm(Act1,Act2,_FindInvViolations,_Timeout,Status,Coenabled) :-
  dependent_act(Act1,Act2,Status,Coenabled),!,fail. % dependent relation is already computed
dependent_actions_symm(Act1,Act2,FindInvViolations,Timeout,Status,Coenabled) :-
  dependent_actions_coenabled(Act1,Act2,FindInvViolations,Timeout,Status,Coenabled).

dependent_actions_coenabled(Act1,Act2,FindInvViolations,Timeout,Status,Coenabled) :-
  dependent_actions1(Act1,Act2,FindInvViolations,Timeout,Res),
  ( Res == '-' -> dependent_actions_coenabled(Act2,Act1,FindInvViolations,Timeout,Status,Coenabled) % Res = '-' : not checked due to symmetry
  ; Res == syntactic_independent -> fail
  ; Res == independent -> fail
  ; Res == self_independent -> fail
  ; otherwise ->
    get_dependency_status(Res,Status),
    check_if_coenabled_nonfailing(Act1,Act2,FindInvViolations,Timeout,Coenabled)
  ).

dependent_actions(Act1,Act2,FindInvViolations,Timeout,Status) :-
  dependent_actions1(Act1,Act2,FindInvViolations,Timeout,Res),
  ( Res == '-' -> dependent_actions(Act2,Act1,FindInvViolations,Timeout,Status) % Res = '-' : not checked due to symmetry
  ; Res == syntactic_independent -> fail
  ; Res == independent -> fail
  ; Res == self_independent -> fail
  ; otherwise ->
    get_dependency_status(Res,Status)).


get_dependency_status(race_dependent,Status) :- !,Status=race.
get_dependency_status(action_dependent,Status) :- !,Status=action.
get_dependency_status(guard_dependent,Status) :- !,Status=guard.
get_dependency_status(guard_dependent(Pred),Status) :- !,Status=predicate(Pred).
get_dependency_status(dependent,Status) :- !,Status=general.
get_dependency_status(self_dependent,Status) :- !,Status=general. % not sure this is correct
get_dependency_status('=',Status) :- !, Status=general.
get_dependency_status(DStatus,_) :- add_error_fail(get_dependency_status, 'Unknown dependency status: ', DStatus).

% currently unused
% computes which events are co-enabled
% co-enabled relations is reflexive and symmetric
%:- dynamic coenabled/2.
%compute_if_coenabled(OpName1,OpName2,FindInvViolations,Timeout) :-
%  (OpName1 == OpName2 ->
%     assert(coenabled(OpName1,OpName1))
%  ;OpName2 @< OpName1 ->
%    true % due symmetry we skip the computation of coenabled events in that case
%  ;check_if_coenabled(OpName1,OpName2,FindInvViolations,Timeout) ->
%    assert(coenabled(OpName1,OpName2)),
%    assert(coenabled(OpName2,OpName1))
%  ; otherwise -> % both events cannot be co-enabled
%    true).

check_if_coenabled_nonfailing(OpName1,OpName2,FindInvViolations,Timeout,Result) :-
  (check_if_coenabled(OpName1,OpName2,FindInvViolations,Timeout) ->
    Result = true
  ; Result = false).

check_if_coenabled(OpName1,OpName2,FindInvViolations,Timeout) :-
  get_guard(OpName1,PosGuard1),
  get_guard(OpName2,PosGuard2),
  conjunct_predicates([PosGuard1,PosGuard2],GuardsConj), % both events are enabled
  (FindInvViolations=1 -> Pred = pred(GuardsConj) ; Pred=typing(GuardsConj)),
  sap:testcase_predicate_timeout(Pred,Timeout,_R).

/************************** DEPENDENCY RELATION (END) ******************************/


/*

The meaning of the following comments here is to give an idea how to validate the results from the Enabling analysis.
We use here the Linear Time Logic (proposed by Amir Pnueli) in order to give LTL-formulas to each possible table entry result that can prove that
the relation results between the operations of machine M produced by the Enabling analysis are correct. For instance,
if the Enabling analysis says that it is impossible operation B to be preceded by operation A in the state space of M (i.e. the result in
cell (A,B) of the table is 'impossible') then this property can be expressed by the LTL-formula "G ([a] => not X e(b))" and checked by
the LTL model checker of ProB.

* impossible:
    The impossibility an action 'a' to enable action 'b' can be expressed by means of LTL-formulas as follows:
            "G ([a] => not X e(b))" (dis)

    If a disables b then (dis) should be satisfied by the model. The result of the enabling analysis
    returns 'impossible' in cell (a,b) of the Enabling analysis table.

* guaranteed:
    Action 'a' guaranteed enables action 'b' if there is an execution in the state space where after 'a' follows 'b' and
    'a' and 'b' are never enabled simultaneously in the same state:
            "G ( ([a] => X e(b)) & (not (e(a) & e(b))) )" (en)

* keep:
    Action 'a' keeps action 'b' enabled:
                "(G ((e(b) & [a]) => X e(b))) or (G ((not e(b) & [a]) => X not e(b)))" (keep)
*/
