% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(sap,[%reset_sap/0,
               explore_and_generate_testcases/7,
               create_testcase_path/5,
               create_testcase_path/3,
               create_testcase_path_nondet/4,
               cbc_gen_test_cases/5, cbc_gen_test_cases_from_string/5,
               cbc_get_uncovered_events/1, cbc_get_nr_uncovered_events/1,
               test_generation_by_xml_description/1,
               tcl_create_local_testsuite/5,
               tcl_get_events_preselected/2, tcl_convert_event_numbers_to_names/2,
               tcl_generate_cb_testcases/5,
               tcl_generate_testcases/6,
               tcl_get_stored_test_cases/1, tcl_execute_stored_test_case/1,
               stored_test_case_op_trace/2,
               cb_timeout_path/2, cb_path/3, cb_path_testcase_only/3, % provide access to cbc-tests tree
               write_all_deadlocking_paths_to_xml/1,
               testcase_path_timeout/9, testcase_predicate_timeout/3,
               cbc_gen_test_cases_task_step/7, cbc_gen_test_cases_finish_tasks/0,
               get_counter_info/4
              ]).

:- use_module(library(lists)).
:- use_module(library(xml)).
:- use_module(library(timeout)).

:- use_module(error_manager).
:- use_module(state_space).
:- use_module(bmachine).
:- use_module(b_enumerate,[b_tighter_enumerate_values/2]).
:- use_module(b_state_model_check,[set_up_transition/7,set_up_initialisation/5]).
:- use_module(preferences,[get_preference/2]).
:- use_module(specfile,[expand_const_and_vars_to_full_store/2,expand_to_constants_and_variables/3]).
:- use_module(translate,[translate_bvalue/2]).
:- use_module(prob2_interface,[set_eclipse_preference/2]).
:- use_module(kernel_objects,[not_equal_object/2]).

:- use_module(bsyntaxtree).
:- use_module(tools).
:- use_module(b_global_sets,[static_symmetry_reduction_for_global_sets/1]).
:- use_module(debug,[debug_mode/1,debug_println/2, printsilent/1,nls/0,silent_mode/1]).

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,test_generation).
:- module_info(description,'This module contains code to generate test cases based on Event-B or Classical B models').

% This is quite a hack, TODO: move the init_states/1 predicate to
% another place
:- use_module(probltlsrc(ltl)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% some code for information in the GUI

% get_events_preselected(-Events,-Preselect):
%   Events: a list of all events
%   Preselect: a list of numbers, each an index (starting with 0) of which event in Events should
%     be preselected because it is already covered in the state space
% Both lists are wrapped by list(...) to use it from Tcl
tcl_get_events_preselected(list(Events),list(Preselect)) :-
    get_all_events(Events),
    select_uncovered_operations(Events,0,Preselect).
select_uncovered_operations([],_,[]).
select_uncovered_operations([Event|Erest],N,Preselect) :-
    ( operation_already_covered_or_infeasible(Event) ->
        Preselect = PSrest
    ; otherwise ->
        Preselect = [N|PSrest]),
    N2 is N+1,
    select_uncovered_operations(Erest,N2,PSrest).

:- use_module(enabling_analysis,[infeasible_operation_cache/1]).
operation_already_covered_or_infeasible(Event) :-
    \+ operation_name_not_yet_covered(Event).
operation_already_covered_or_infeasible(Event) :-
    infeasible_operation_cache(Event). % check if determined to be infeasible; but do not trigger computation

% returns a list of all events (their namese) in the model
get_all_events(Events) :-
    findall(E,b_top_level_operation(E),Events).

tcl_convert_event_numbers_to_names(list(EventNumbers),list(EventNames)) :-
    get_all_events(Events),
    maplist(get_event_name(Events),EventNumbers,EventNames).
get_event_name(List,Number,Res) :-
  ((number(Number),nth0(Number,List,Res)) -> true
    ; add_error(get_event_name,'Illegal event number: ',Number),fail).

% tcl_generate_testcases(+EventSelection,+TargetString,+MaxDepth,+MaxNewNodes,+Testfile,-NumberOfTestcases):
%  EventSelection: a list of numbers containing all indeces of events that should be covered
%  TargetString: the predicate that identifies target nodes as string (atom)
%  MaxDepth: the maximum depth of the search
%  MaxNewNodes: the maximum number of newly explored nodes
%  Testfile: the file where we store the found test cases
%  NumberOfTestcases: The number of test cases that where found
tcl_generate_testcases(EventSelection,TargetString,MaxDepth,MaxNewNodes,Testfile,NumberOfTestcases) :-
    % convert the list of event indeces into a list of events
    get_selected_events(EventSelection,AllEvents,Events),
    b_parse_optional_machine_predicate(TargetString,Target),
    set_max_nr_of_new_impl_trans_nodes(MaxNewNodes),
    safe_get_inits(Inits),
    ( explore_covered_statespace(Inits, Events, Target, MaxDepth, _ExploredDepth) ->
        true
    ; otherwise ->
        show_uncovered_events_as_errors),
    generate_test_cases(Inits,Testcases),
    length(Testcases,NumberOfTestcases),
    testcase_output(Testfile,AllEvents,Events,Testcases).

% get_selected_events(+TkNumbers,-ListOfAllEvents,-SelectedEvents)
%    Input: TkNumbers: numbers starting at 0; comes from Tk Selection Box
%    Output: ListOfAllEvents
%    Output: SelectedEvents: Name of Selected Events
get_selected_events(Selection,AllEvents,Events) :-
    get_all_events(AllEvents),
    (Selection==all -> Events=AllEvents
     ; get_selected_events2(Selection,AllEvents,Events)).
get_selected_events2([],_AllEvents,[]).
get_selected_events2([Selection|Srest],AllEvents,[Event|Erest]) :-
    nth0(Selection,AllEvents,Event),!,
    get_selected_events2(Srest,AllEvents,Erest).

safe_get_inits(Inits) :-
    ltl:init_states(Inits),
    ( Inits = [] -> add_error(sap,'no initial states found'),fail
    ; otherwise -> true).

% this predicate is called if the breadth-first search fails
% it shows an error for every event, for which we did not find
% an appropiate path
show_uncovered_events_as_errors :-
    target_needed(Event),
    (infeasible_operation_cache(Event)
     -> ajoin(['No test path exists for infeasible (!) event ',Event],Msg)
      ; ajoin(['No test path found for event ',Event],Msg)
    ),
    add_error(sap,Msg),
    fail.
show_uncovered_events_as_errors.

get_uncovered_events(Uncovered) :-
    findall(Event, target_needed(Event), Uncovered).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% model checking based test-case generation mcm

explore_and_generate_testcases(InEvents,Target,MaxDepth,MaxNewNodes,
                               Testfile,NumberOfTestcases,UncoveredEvents) :-
    % If no event is specified to be covered, we assume that all events
    % should be covered
    ( InEvents = []    ->
        printsilent('No events to cover specified, assuming all'),nls,
        get_all_events(Events)
    ; InEvents = [_|_] -> InEvents = Events
    ; add_internal_error('Illegal events: ',InEvents),fail
    ),
    statistics(walltime, [Start|_]),
    set_max_nr_of_new_impl_trans_nodes(MaxNewNodes),
    safe_get_inits(Inits),
    ( explore_covered_statespace(Inits, Events, Target, MaxDepth, _ExploredDepth) ->
        UncoveredEvents = []
    ; otherwise ->
        get_uncovered_events(UncoveredEvents)),
    generate_test_cases(Inits, Testcases),
    length(Testcases, NumberOfTestcases),
    get_all_events(AllEvents),
    testcase_output(Testfile,AllEvents,Events,Testcases),
    statistics(walltime, [End|_]),
    Runtime is End - Start,
    format_msg('Runtime for test case generation: ~w ms~n',[Runtime]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate test cases from an explored state-space


% generate_test_cases(+Inits,-TestCases):
%  Inits: List of initial states
%  TestCases: list of test cases, each of the form global_test_case_for_event(Event,Trace),
%    where Trace is a list of events
generate_test_cases(Inits,TestCases) :-
    init_wanted_event_to_cover,
    findall(T, generate_test_case(Inits,T), Ts),
    sort(Ts,TestCases).
    % TO DO: if generate_minimal_nr_of_testcases is true we could still have redundant tests; eliminate them.

:- dynamic wanted_event_to_cover/1.
init_wanted_event_to_cover :- retractall(wanted_event_to_cover(_)),
    wanted_event(E), assert(wanted_event_to_cover(E)),fail.
init_wanted_event_to_cover.
assert_covered(E) :- retractall(wanted_event_to_cover(E)).
% find a single test case by first choosing one event and one initial state
generate_test_case(Inits,global_test_case_for_events(WantedEventsCovered,Trace)) :-
    %fail, % SAP wants all test cases, not just for each event the shortest,
    !,
    % generate a single test_case per event
    wanted_event_to_cover(Event),
    (minimum_path_length(Event,MaxSteps) -> true ; debug_println(20,cannot_cover(Event)),fail),
    % this code could be improved: maybe a previous test-case has already covered this event !
    debug_println(20,generating_testcase(Event,MaxSteps)),
    % now find a path that can serve as a test case
    (find_test_case_with_init(Event,Inits,MaxSteps,Trace)
     -> trace_covers_wanted_events(Trace,WantedEventsCovered),
        debug_println(44,found_trace(WantedEventsCovered,Trace)),
        (get_preference(generate_minimal_nr_of_testcases,true)
          -> maplist(assert_covered,WantedEventsCovered)
          ;  true)
     ).
generate_test_case(Inits,global_test_case_for_events(WantedEventsCovered,Trace)) :-
    % generate all paths up to certain depth; was requested by SAP, but can blow up
    find_global_maximum(MaxSteps), % print(global_max(MaxSteps)),nl,trace,
    member(Init,Inits),
    find_way_to_target(Init,MaxSteps,Trace),
    %print(trace(Trace)),nl,
    trace_covers_wanted_events(Trace,WantedEventsCovered).

trace_covers_wanted_events(Trace,WantedEventsCovered) :-
    findall(TestEvent,contains_an_event_to_test(TestEvent,Trace),Covered),
    Covered\=[],
    sort(Covered,WantedEventsCovered).
contains_an_event_to_test(Event,Trace) :-
    member('$full_event'(Event,_Args),Trace),wanted_event_to_cover(Event).


% find variables or constants which are non-deterministically assigned
% TO DO: check max_reached of predecessor ?!
non_deterministic_values(_,[],NonDetState) :- !, NonDetState=[].
non_deterministic_values(Init,OtherInits,NonDetState) :-
    findall(bind(ID,Value),non_deterministic_value(Init,OtherInits,ID,Value),NonDetState).

non_deterministic_value(Init,OtherInits,ID,Value) :-
    expand_initalised_state(Init,State0),
    maplist(expand_initalised_state,OtherInits,OtherStates),
    member(bind(ID,Value),State0),
    (other_value_exists(ID,Value,OtherStates) -> true).

other_value_exists(ID,Value,OtherStates) :-
    member(OtherState,OtherStates),
    member(bind(ID,OtherValue),OtherState),
    kernel_objects:not_equal_object(Value,OtherValue).

% find the maximum of the minimum mapth length
find_global_maximum(MaxSteps) :-
    findall(M, minimum_path_length(_Event,M), [First|Maxis]),
    find_global_maximum2(Maxis,First,MaxSteps).
find_global_maximum2([],Max,Max).
find_global_maximum2([H|T],A,Max) :- B is max(H,A), find_global_maximum2(T,B,Max).


% find a single test case for a single event and a set of initial states
find_test_case_with_init(Event,Inits,MaxSteps,ResTrace) :-
    select(Init,Inits,OtherInits),
    find_test_case(Event,Init,MaxSteps,TraceWithoutInit),
    non_deterministic_values(Init,OtherInits,NonDetState),
    (NonDetState=[] -> ResTrace = TraceWithoutInit
     ; ResTrace = ['$non_det_init'(NonDetState)|TraceWithoutInit]).

% find a single test case for a single event and an initial state
find_test_case(Event,Init,MaxSteps,Trace) :-
    % first find a path where Event occurs
    find_way_to_event(Init,MaxSteps,Event,Trace,TraceB,State,RestSteps),
    % then find a path from that state to a target state
    find_way_to_target(State,RestSteps,TraceB).

% find_way_to_event(+A,+Max,+Event,-Trace,TraceR,-B,-RestSteps):
%   A: node where we start the search
%   Max: maximum number of steps in the path
%   Event: the event we are searching for
%   Trace: a trace to the state, the last event is Event, the list
%     has the variable TraceR as tail
%   B: the state after Event occurs
%   RestSteps: the number of remaining steps (i.e. RestSteps=Max-length(Trace))
find_way_to_event(A,Max,Event,['$full_event'(Operation,Args)|Trace],TraceR,B,RestSteps) :-
    dec(Max,Max2),
    do_full_event(A,Operation,Args,C), % TO DO: extract operation arguments
    ( Event = Operation ->
        Trace = TraceR,
        B = C,
        RestSteps = Max2
    ; otherwise ->
        find_way_to_event(C,Max2,Event,Trace,TraceR,B,RestSteps)).

find_way_to_target(A,_Max,[]) :-
    target_node(A). % removed the cut means that if target found straight away, no paths were generated : ,!.
find_way_to_target(A,Max,['$full_event'(Operation,Args)|Trace]) :-
    dec(Max,Max2),
    do_full_event(A,Operation,Args,B),
    find_way_to_target(B,Max2,Trace).

% decrease N, but never become negative
dec(N,N2) :- N>0, N2 is N-1.

:- use_module(specfile,[get_operation_arguments/2, get_operation_name/2]).
do_event(A,Event,B) :-
    transition(A,FullOperation,B),
    get_operation_name(FullOperation,Event).
do_full_event(A,Event,Args,B) :-
    transition(A,FullOperation,B),
    get_operation_name(FullOperation,Event),
    get_operation_arguments(FullOperation,Args).

:- use_module(library(ordsets)).
% :- use_module(probsrc(sap)), sap:path_to_deadlock_or_loop(root,[],P), print(P),nl,fail.
% compute all paths until a deadlock or non-interesting node is reached
% assumes no loop !! otherwise an infinite number of solutions will be returned
path_to_deadlock_or_loop(A,History,Path) :- ord_member(A,History),
    !, % loop
    Path = [loop(A)].
path_to_deadlock_or_loop(A,History,Path) :- ord_add_element(History,A,NH),
    if(%do_event(A,Event,B),
       %(transition(A,FullOperation,B), translate:translate_event(FullOperation,Event)),
       my_event(A,Event,B),
       (Path = [Event|P2], path_to_deadlock_or_loop(B,NH,P2)),
       Path = [] %[deadlock(A)]
      ).

my_event(Src,(TransId,Op,Src,Dst),Dst) :-
    transition(Src,Op,TransId,Dst).

% :- use_module(probsrc(sap)), sap:write_all_deadlocking_paths_to_xml('~/Desktop/all.xml').
write_all_deadlocking_paths_to_xml(Filename) :-
    retractall(path_nr(_)), assert(path_nr(0)),
    open_xml_testcase(Filename),
    call_cleanup(write_path_to_deadlock_or_loop, close_xml_testcase),
    get_path_nr(Nr),
    format_msg('Wrote ~w tests to ~w.~n',[Nr,Filename]).

write_path_to_deadlock_or_loop :- path_to_deadlock_or_loop(root,[],Trace),
     get_path_nr(Id),
     (debug_mode(on) -> format_msg('Trace ~w = ~w~n',[Id,Trace]) ; true),
     %write_testcase(global_test_case_for_event(unknown,Trace),Id),
     get_initialisation(Trace,RestTrace,State),
     cb_write_test_case(RestTrace,State),
     (exceeded_limit(Id) -> format_warn('Exceeded limit of number of traces: ~w~n',[Id])
      ; Id>0, Id mod 500 =:= 0, format_msg('Wrote ~w tests ...~n',[Id]),
        fail). % continue backtracking
write_path_to_deadlock_or_loop.

exceeded_limit(100000). % do not write more than 100,000 states to avoid filling up disk
% TO DO: add loop detection and/or user settable upper bound

:- use_module(specfile,[state_corresponds_to_initialised_b_machine/2]).
:- use_module(state_space,[visited_expression/2]).
get_initialisation([(_,_,_,Dst)|T],Trace,InitBState) :-
     expand_initalised_state(Dst,InitBState),!, Trace=T.
get_initialisation([_|T],Trace,State) :- get_initialisation(T,Trace,State).

expand_initalised_state(ID,InitBState) :-
     state_space:visited_expression(ID,State),
     state_corresponds_to_initialised_b_machine(State,InitBState).

:- dynamic path_nr/1.
path_nr(0).
get_path_nr(N) :- retract(path_nr(N)), N1 is N+1, assert(path_nr(N1)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% explore the state-space so that it is usable for test-case generation

% bfs_node(NodeId,Depth): open nodes in the state space, for breadth first search,
%   the order is relevant, Depth is the distance from the initialisation nodes
%   (later we call this the breadt-first FIFO queue)
% predecessor(NodeId,PreNodeId): store predecessors of a node
% target_reachable(NodeId,TargetDist): A target state is reachable from this state
% event_dest(NodeId,Event): One of the events we want to cover leads to this node
% node_covered(NodeId,Depth): This node was encountered before by the search at depth Depth.
% target_needed(Event): We need to find a final from a state that is the destination of the event,
%  in the beginning, each given event is stored here
% wanted_event(Event): Event if one of the events we are interested in
%  in the beginning, each given event is stored here
% target_node(NodeId): The node is a target node

:- volatile bfs_node/2, predecessor/2, target_reachable/2, event_dest/2.
:- volatile node_covered/2, target_needed/1, wanted_event/1.
:- volatile target_node/1, minimum_path_length/2.
:- volatile maximum_depth_reached/1.
:- dynamic bfs_node/2, predecessor/2, target_reachable/2, event_dest/2.
:- dynamic node_covered/2, target_needed/1, wanted_event/1.
:- dynamic target_node/1, minimum_path_length/2.
:- dynamic maximum_depth_reached/1.

% explore_covered_statespace(+Inits,+Events,+Target,-ExploredDepth):
%  Inits: a list of initialisation states
%  Events: the list of events that should be covered
%  Target: the predicate that is used to identify a target state
%  MaxDepth: the maximum depth that will be explored
%  ExploredDepth: the depth until the state-space was explored
% This predicate explores a part of the state-space by breadt-first search
% that contains for each event a path where the event occurs and that ends
% in a target state.
% Additionally, the state-space is explored to a fixed depth.
% (i.e. if we needed a state that is reachable by 4 steps, than all states
%  that are reachable by 4 steps are explored with their transitions and
%  successor states)
explore_covered_statespace(Inits,Events,Target,MaxDepth,ExploredDepth) :-
    % initialise the data structures
    reset_bfs(Events),

    % put initial states into the breadt-first FIFO queue
    assert_bfs_nodes(Inits,0,_,_),
    % check if the initialised states are target states
    mark_target_nodes(Inits,Target),

    % do the breadt-first search
    explore(Target,MaxDepth,-1,ExploredDepth).

% explore(+Target,+PrevExploredDepth,-ExploredDepth):
%  Target: the predicate that is used to identify target nodes
%  MaxDepth: the maximum depth that will be explored
%  PrevExploredDepth: the deepest node that is explored before calling this predicate
%  ExploredDepth: the deepest node that was explored by the search
explore(Target,_MaxDepth,ExploredDepth,ExploredDepth) :-
    % we do not have to find any more targets, we're finished
    \+ target_needed(_),!,
    % we did a breadth-first search until a certain depth, now we
    % complete the exploration of the state space to that depth
    complete_depth(Target,ExploredDepth).
explore(Target,MaxDepth,_,ExploredDepth) :-
    % get a node from the breadth-first FIFO queue
    retract( bfs_node(Node,Depth) ),!,
    % and explore it
    explore_node(Node, Target, Depth),
    !, % just to be safe that no back-tracking occurs
    Depth =< MaxDepth,
    update_depth(Depth,MaxDepth),
    explore(Target,MaxDepth,Depth,ExploredDepth).

:- use_module(state_space,[get_state_space_stats/3]).
update_depth(Depth,MaxDepth) :- maximum_depth_reached(MD), Depth>MD,
    !,
    get_state_space_stats(NrNodes,NrTransitions,ProcessedNodes),
    format_msg('Depth reached: ~w (max.: ~w), States: ~w (~w processed), Transitions: ~w~n',[Depth,MaxDepth,NrNodes,ProcessedNodes,NrTransitions]),
    retractall(maximum_depth_reached(_)),
    assert(maximum_depth_reached(Depth)).
update_depth(_,_).

% explore_node(+Node,+Target,+Depth):
%  Node: the node that should be explored
%  Target: the predicate that is used to identify target nodes
%  Depth: depth of the given node
explore_node(Node,Target,Depth) :-
    % newly encountered nodes are one level deeper
    NewDepth is Depth+1,
    % get all transitions from this node to successor nodes
    impl_trans_term_all(Node,Transitions),
    % check if there are events we are interested in and
    % get the successor nodes
    analyse_transitions(Transitions,Successors),
    % find out wich of the successor nodes are new and
    % store those in the breadt-first FIFO queue
    assert_bfs_nodes(Successors,NewDepth,NewNodes,OldNodes),
    % store that this node is a predecessor node of
    % all the successor nodes encountered
    assert_predecessors(Successors, Node),
    % find out which of the new nodes are target nodes
    % and in such a case propagate the information that a target
    % note is reachable back into the network.
    mark_target_nodes(NewNodes,Target),
    mark_target_nodes_by_old(OldNodes,Node).

mark_target_nodes_by_old([],_Node).
mark_target_nodes_by_old([PotentialTargetNode|Rest],Node) :-
    ( target_reachable(PotentialTargetNode,DistToTarget) ->
        NewDist is DistToTarget + 1,
        update_minimal_path_for_events(Node,PotentialTargetNode,DistToTarget),
        assert_target_reachable(Node,NewDist)
    ; otherwise ->
        true ),
    mark_target_nodes_by_old(Rest,Node).


% mark_target_nodes(+Nodes,+Target).
%  Nodes: a list of node IDs
%  Target: the predicate to check whether the node is a target node
%  PreDepth: the depth of the node that led to this node,
% checks for each node in the given list if it is a target
% node, and if yes then propagate the information that the
% target is reachable to all predecessor nodes.
:- use_module(eclipse_interface,[test_boolean_expression_in_node/2]).
mark_target_nodes([],_Target).
mark_target_nodes([Node|NRest],Target) :-
    ( test_boolean_expression_in_node(Node,Target) ->
        assert( target_node(Node) ),
        assert_target_reachable(Node,1)
    ; otherwise -> true ),
    mark_target_nodes(NRest,Target).

% for all transitions, check if the resp. event is interesting
% and if yes, mark the successor node.
% and return the list of successor nodes.
analyse_transitions([],[]).
analyse_transitions([op(_Id,Operation,Succ)|Orest],[Succ|Srest]) :-
    % the arguments of the event are not interesting to us
    specfile:get_operation_name(Operation,Event),     %functor(Operation,Event,_),
    % if Event is one of the events we monitor, we mark
    % the destination node as a node from where we look for
    % a path to a target node.
    ( wanted_event(Event) ->
        assert_event_dest(Succ,Event)
    ; otherwise -> true),
    analyse_transitions(Orest,Srest).

% store the information, that the Node is reached by Event.
% We need this, because we need such a node fro
assert_event_dest(Node,Event) :-
    ( event_dest(Node,Event) -> true   % information was already present, ignore
    ; otherwise ->
        assert( event_dest(Node,Event) ),
        % if a target node can be reached from Node, we have a path in the explored
        % state-space that includes the Event and ends in a target node.
        ( target_reachable(Node,_) ->
            ( retract( target_needed(Event) ) -> true; true)
        ; otherwise -> true) ).

% reset all data structures needed for the breadth-first search
reset_bfs(Events) :- %print(reset_bfs),nl,
    retractall(bfs_node(_,_)),
    retractall(predecessor(_,_)),
    retractall(target_reachable(_,_)),
    retractall(node_covered(_,_)),
    retractall(event_dest(_,_)),
    retractall(target_needed(_)),
    retractall(wanted_event(_)),
    retractall(target_node(_)),
    retractall(minimum_path_length(_,_)),
    retractall(maximum_depth_reached(_)),
    assert( maximum_depth_reached(0)),
    reset_bfs2(Events).
reset_bfs2([]).
reset_bfs2([Event|Rest]) :-
    (b_top_level_operation(Event) -> true
      ; add_warning(sap,'Unknown event: ',Event)),
    assert( target_needed(Event) ),
    assert( wanted_event(Event) ),
    reset_bfs2(Rest).

% assert_bfs_nodes(+Nodes,+Depth,-NewNodes):
%  Nodes: a list of nodes
%  Depth: the current depth of the search, this is stored for finding
%    out later until which depth the search has been done
%  NewNodes: a sublist of Nodes, that includes all nodes that have not been
%    encountered before
% This predicate checks if the given nodes are already encountered and if
%  not, it stores them into the breadt-first FIFO queue
assert_bfs_nodes([],_,[],[]).
assert_bfs_nodes([Node|Rest],Depth,NewNodes,OldNodes) :-
    ( node_covered(Node,_NodeDepth) -> % we already encountered Node, ignore
        NewNodes = NewRest,
        OldNodes = [Node|OldRest]
    ; otherwise -> % this is a new node
        NewNodes = [Node|NewRest],
        OldNodes = OldRest,
        assert( node_covered(Node,Depth) ), % mark as covered
        assertz( bfs_node(Node,Depth) )),   % put it into the FIFO queue for the breadth-first search
    assert_bfs_nodes(Rest,Depth,NewRest,OldRest).

% assert_predecessors(+Nodes,+Predecessor):
%  Nodes: a list of nodes
%  Predecessor: the predecessor node, a node from where a transition
%    leads to each of the nodes in Nodes
% This predicate stores for each node in Nodes that Predecessor is a
% predecessor of the node.
% This information is needed to be able to propagate information of
% a reachable target state back into the graph.
assert_predecessors([],_).
assert_predecessors([Node|Rest],Pre) :-
    ( predecessor(Node,Pre) -> true % information already there, ignore
    ; otherwise ->
        assert( predecessor(Node,Pre) )),
    assert_predecessors(Rest,Pre).

% store the information that a target node is reachable from a node
% propagate this information to predecessor nodes if necessary
assert_target_reachable(Node,DistToTarget) :-
    ( target_reachable(Node,OldDistToTarget) ->
        ( OldDistToTarget =< DistToTarget ->
            true                % information already there, ignore
        ; otherwise ->
            retract( target_reachable(Node,_) ),
            assert_target_reachable2(Node,DistToTarget)
        )
    ; otherwise ->
        assert_target_reachable2(Node,DistToTarget)
    ).
assert_target_reachable2(Node,DistToTarget) :-
    assert( target_reachable(Node,DistToTarget) ), fail.
assert_target_reachable2(Node,_DistToTarget) :-
    % the node can be reached by Event, so we know
    % that we have a path in the explored state-space that contains
    % Event a ends in a target node.
    event_dest(Node,Event),
    retract( target_needed(Event) ),
    % (this is a fail-loop)
    fail.
assert_target_reachable2(Node,DistToTarget) :-
    % propagate the information to the predecessors
    % (if we can reach a target node from this node, we also can
    %  reach the target node from each predecessor node.)
    predecessor(Node,Pre),
    update_minimal_path_for_events(Pre,Node,DistToTarget),
    NewDistToTarget is DistToTarget + 1,
    assert_target_reachable(Pre,NewDistToTarget),
    % (this is a fail-loop)
    fail.
assert_target_reachable2(_,_).

% Pre: node where the event starts
% Post: node where the events leads to
% DistToTarget: number of steps to reach the target when starting in the
%      pre node.
% For each event that should be covered, we store the minimal length of the
% path needed to cover that event.
% We store and compute that information while going back from the target node
% to set the target_reachable flag for states.
% To do that, we add the distance to the target to the depth of the node. The
% encountered minimum for each event is stored in minimum_path_length/2.
update_minimal_path_for_events(Pre,Post,DistToTarget) :-
    do_event(Pre,Event,Post),
    wanted_event(Event),
    node_covered(Pre,NodeDepth),
    MinLength is DistToTarget + NodeDepth,
    ( minimum_path_length(Event,Old) ->
        Old > MinLength,
        retractall( minimum_path_length(Event,_) ),
        assert( minimum_path_length(Event,MinLength) )
    ; otherwise ->
        format_ok('Covered: ~w (length: ~w)~n',[Event,MinLength]),
        assert( minimum_path_length(Event,MinLength) )),
    fail.
update_minimal_path_for_events(_Pre,_Post,_DistToTarget).

% explore the state-space to the given depth
complete_depth(Target,Depth) :-
    retract( bfs_node(Node,NodeDepth) ),!,
    ( NodeDepth =< Depth ->
        explore_node(Node,Target,Depth),
        complete_depth(Target,Depth)
    ; otherwise ->
        % the next nodes are deeper, just revert the previous retract
        % (this might not be necessary, because the data may not be used
        %  anymore after this, but just to keep it cleaner)
        asserta( bfs_node(Node,NodeDepth) )).
complete_depth(_Target,_Depth).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% XML output

testcase_output(Filename,_AllEvents,_SelectedEvents,Testcases) :-
    open_xml_testcase(Filename),
    % write_selected_events(AllEvents,SelectedEvents),
    call_cleanup((write_testcases(Testcases,1) -> true ; add_internal_error('Generating testcases failed: ',Filename)),
                 close_xml_testcase).

:- use_module(tools_io).
:- dynamic no_xml_file/0.

open_xml_testcase('') :- assert(no_xml_file),!.
open_xml_testcase(Filename) :-
    safe_open_file(Filename,write,_Stream,[alias(xml_testcase),encoding('UTF-8')]),
    absolute_file_name(Filename,AF), format_msg('% Writing testcases to ~w~n',[AF]),
    xml_write('<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<extended_test_suite>\n').

close_xml_testcase :- retract(no_xml_file),!.
close_xml_testcase :-
    xml_write('</extended_test_suite>\n'),
    close(xml_testcase).

xml_write(Atom) :- no_xml_file,!, (debug_mode(on) -> write(Atom) ; true).
xml_write(Atom) :- write(xml_testcase,Atom).

:- use_module(tools_files,[put_codes/2]).
put_codes_xml(Codes) :- no_xml_file,!, (debug_mode(on) -> put_codes(Codes,user_output) ; true).
put_codes_xml(Codes) :- put_codes(Codes,xml_testcase).

flush_output_xml :- no_xml_file,!.
flush_output_xml :- flush_output(xml_testcase).

/*
write_selected_events(AllEvents,Events) :-
    xml_write('  <global_events>\n'),
    write_selected_events2(AllEvents,Events),
    xml_write('  </global_events>\n').
write_selected_events2([],_Selected).
write_selected_events2([Event|Rest],Selected) :-
    ( member(Event,Selected) -> Sel = true
    ; otherwise -> Sel = false),
    xml_write('    <event name=\"'),xml_write(Event),
    xml_write('\" covered=\"'),xml_write(Sel),
    xml_write('\"/>\n'),
    write_selected_events2(Rest,Selected).
    */


write_testcases([],_).
write_testcases([Testcase|TCrest],N) :-
    write_testcase(Testcase,N),
    N2 is N+1,
    write_testcases(TCrest,N2).

write_testcase(global_test_case_for_events(Events,Trace),N) :-
    number_codes(N,Id),
    write_testcase2(Trace,[],Id,Events).
write_testcase(global_test_case(Id,Trace),_N) :-
    write_testcase2(Trace,[],Id,unknown).
write_testcase(local_test_case(Id,Global,Local),_N) :-
    write_testcase2(Global,[Local],Id,unknown).

write_testcase2(GlobalTrace,LocalTraces,Id,Events) :-
    xml_write('  <test_case id=\"'),put_codes_xml(Id),
    (Events=unknown -> true
      ; xml_write('\" targets=\"'), xml_write(Events)),
    xml_write('\">\n'),
    write_trace(global,GlobalTrace),
    write_traces(LocalTraces,local),
    xml_write('  </test_case>\n').

write_traces([],_Type).
write_traces([Trace|Trest],Type) :-
    write_trace(Type,Trace),
    write_traces(Trest,Type).
write_trace(Type,Trace) :-
    xml_write('    <'),xml_write(Type),xml_write('>\n'),
    write_trace2(Trace,1),
    xml_write('    </'),xml_write(Type),xml_write('>\n').
write_trace2([],_).
write_trace2([Step|Trace],N) :-
    write_step(Step,N),
    N2 is N+1,
    write_trace2(Trace,N2).

write_step('$non_det_init'(NonDetState),N) :- !,
    write_step_tag(N,'INITIALISATION'),
    maplist(xml_write_binding(value),NonDetState),
    xml_write('    </step>\n').
write_step('$full_event'(Event,[]),N) :- !,write_step(Event,N).
write_step('$full_event'(Event,Args),N) :-
    write_step_tag(N,Event),
    b_get_machine_operation(Event,_,Parameters,_),
    xml_write_parameters(Event,Parameters,Args),
    xml_write('    </step>\n').
write_step(Event,N) :-
    %xml_write('      <step id=\"'), xml_write(N), xml_write('\">'), xml_write(Step), xml_write('    </step>\n').
    xml_write('      <step id=\"'),
    xml_write(N),
    xml_write('\" name=\"'),
    xml_write(Event),
    xml_write('\" />\n').

write_step_tag(N,Event) :-
    xml_write('      <step id=\"'),
    xml_write(N),
    xml_write('\" name=\"'),
    xml_write(Event),
    xml_write('\">\n').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% local test case generation

:- volatile is_local_event/1, is_global_event/2.
:- dynamic is_local_event/1, is_global_event/2.

tcl_create_local_testsuite(GlobalFile,LocalFile,MaxNewNodes,NumSat,NumUnsat) :-
    read_xml_testcases(GlobalFile,Global),
    set_max_nr_of_new_impl_trans_nodes(MaxNewNodes),
    process_testsuite(Global,Local,Unsatisfied),
    length(Local,NumSat),length(Unsatisfied,NumUnsat),
    append(Local,Unsatisfied,Testcases),
    testcase_output(LocalFile,[],[],Testcases).

process_testsuite(GlobalTestsuite,LocalTestsuite,Unsatisfied) :-
    reset_local_search,
    safe_get_inits(Inits),
    process_testsuite2(GlobalTestsuite,Inits,LocalTestsuite,Unsatisfied).
process_testsuite2([],_Inits,[],[]).
process_testsuite2([Global|Grest],Inits,LocalTestsuite,Unsatisfied) :-
    get_testcase_id(Global,Id),
    ( find_local_testcase(Global,Inits,Local) ->
        write('++ found local testcase: '),print(Id),nl,
        LocalTestsuite = [Local|Lnext],
        Unsatisfied = Unext
    ; otherwise ->
        write('--    no local testcase: '),print(Id),nl,
        LocalTestsuite = Lnext,
        Unsatisfied = [Global|Unext]),
    process_testsuite2(Grest,Inits,Lnext,Unext).

get_testcase_id(global_test_case(Id,_),I) :- atom_codes(I,Id).

% find_local_testcase(+Global,+Inits,-Local):
%   Global: a global test case
%   Inits: list of initial states
%   Local: a local test case
% search for a local trace of a global test case in the current model.
% A local trace consists of refined events of events of the global model,
% interspered by local events (i.e. events that refine skip)
find_local_testcase(global_test_case(Id,Global),Inits,local_test_case(Id,Global,Local)) :-
    member(Init,Inits),
    find_local_trace(Global,Init,Local).

find_local_trace([],_State,[]).
find_local_trace([GlobalEvent|Grest],State,LocalTrace) :-
    % do a number of local steps (without loop) until the global event
    find_local_path_to_global_event(State,GlobalEvent,[State],LocalTrace,RTrace,Dest),
    % continue with the next global event
    find_local_trace(Grest,Dest,RTrace).

% find_local_path_to_global_event(+State,+GlobalEvent,+History,-Trace,?RTrace,-End):
%   State: the state where the search starts
%   GlobalEvent: the global event that we are searching for, until reaching it, only
%     local events are taken
%   History: the states that we encountered before in this search, to prevent loops
%   Trace: the resulting trace, the last element is the global event, the tail of the
%     list is uninstantiated
%   RTrace: the tail of Trace
%   End: the last state in the search, the global event leads into this state
:- use_module(state_space,[compute_transitions_if_necessary/1]).
find_local_path_to_global_event(State,GlobalEvent,History,[Event|Trace],RTrace,End) :-
    % make sure this node is explored
    compute_transitions_if_necessary(State),
    % do an (non-deterministic) step
    do_event(State,Event,Dest),
    ( is_local_event(Event) ->
        % local events can be taken as long we do not loop
        \+ member(Dest,History),
        % continue the search with the new state
        find_local_path_to_global_event(Dest,GlobalEvent,[Dest|History],Trace,RTrace,End)
    ; is_global_event(Event,GlobalEvent) ->
        % if we encounter the global event we search for, we can stop
        Trace = RTrace,
        Dest = End).

% initial the data structures for the search
reset_local_search :-
    get_all_events(Events),
    retractall(is_local_event(_)),
    retractall(is_global_event(_,_)),
    assert_local_events(Events).

% for each event, we store whether it's local or global,
% and if it's global, we store the refined event
assert_local_events([]).
assert_local_events([Event|Erest]) :-
    get_operation_info(Event,Info),
    ( member(refines(Ref),Info) ->
        assert_abstract_events(Event,Ref)
    ; otherwise -> assert( is_local_event(Event) )),
    assert_local_events(Erest).
assert_abstract_events(Event,Refines) :-
    member(Abstract,Refines), assert( is_global_event(Event,Abstract) ),
    % (fail-loop)
    fail.
assert_abstract_events(_,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% read XML testcases
read_xml_testcases(Filename,TestCases) :-
    read_string_from_file(Filename,Codes),
    xml_parse(Codes,xml(_Atts,Content)),
    member(element(extended_test_suite,_,XmlTestSuite),Content),
    extract_testcases(XmlTestSuite,TestCases).
extract_testcases(Content,TestCases) :-
    findall(TC, extract_testcase(Content,TC), TestCases).
extract_testcase(Content,global_test_case(Id,GlobalTrace)) :-
    member(element(test_case,Atts,EContent),Content),
    member(id=Id, Atts),
    member(element(global,_,GContent),EContent),
    findall(Step,
            ( member(element(step,_,[pcdata(P)]),GContent),atom_codes(Step,P) ),
            GlobalTrace).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dot debugger

:- use_module(library(process),[process_create/3,process_wait/2]).
:- public sapdebug/0.
sapdebug :-
    process_create(path(dot),
                   ['-T','pdf','-o','sap_debug.pdf'],
                   [stdin(pipe(S)),process(P)]),
    write(S,'digraph sap_debug {\n'),
    sapwrite_nodes(S),
    write(S,'\n'),
    sapwrite_edges(S),
    write(S,'}\n'),
    close(S),
    process_wait(P,_).

sapwrite_nodes(S) :-
    node_covered(Node,_),
    ( bfs_node(Node,_) -> Shape = triangle; Shape = box ),
    ( target_reachable(Node,_) -> Style = ',filled'; Style = ''),
    write_l(S,['  ',Node,' [shape=',Shape,', style=\"rounded',Style,'\"];\n']),
    fail.
sapwrite_nodes(_).

sapwrite_edges(S) :-
    predecessor(Post,Pre),
    write_l(S,['  ',Pre,' -> ',Post,';']),
    fail.
sapwrite_edges(_).

write_l(_S,[]) :- !.
write_l(S,[H|T]) :-
    write(S,H),write_l(S,T).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% test case generation for a given list of event by
% constraint solving

:- use_module(b_interpreter).
%:- use_module(b_interpreter_eventb).
:- use_module(kernel_waitflags).
:- use_module(store).
:- use_module(state_space).

create_testcase_path(Events,Timeout,Result) :-
   create_testcase_path(init,Events,b(truth,pred,[]),Timeout,Result).
%create_testcase_path(INIT,Events,Timeout,Result) :- % version with arity 4; not used
%   create_testcase_path(INIT,Events,b(truth,pred,[]),Timeout,Result).
create_testcase_path(INIT,Events,EndPredicate,Timeout,Result) :-
    % allow user interrupts during the search, just fail if that happens
    testcase_path_timeout(INIT,Timeout,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,R),
    ( R==ok ->
        add_constants_to_state_space(ConstantsState,TestStates,ConstTestStates,SetupSequence,StartStateId),
        add_operations_to_state_space(Operations,StartStateId,ConstTestStates,TransInfos,OpSequence),
        append(SetupSequence,OpSequence,Result)
    ; otherwise -> Result = R).

% same as create_testcase_path: does not react to interrupts, no time-out, but can be backtracked !
create_testcase_path_nondet(INIT,Events,EndPredicate,Result) :-
    % allow user interrupts during the search, just fail if that happens
    testcase_path(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos),
    add_constants_to_state_space(ConstantsState,TestStates,ConstTestStates,SetupSequence,StartStateId),
    add_operations_to_state_space(Operations,StartStateId,ConstTestStates,TransInfos,OpSequence),
    append(SetupSequence,OpSequence,Result).

testcase_path_timeout(INIT,0,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,R) :-
    !,testcase_path_interrupt(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,R).
testcase_path_timeout(INIT,TimeoutMs,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,Result) :-
    time_out(testcase_path_interrupt(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,R),
             TimeoutMs,
             T),
    ( T==success -> Result=R
    ; otherwise  -> Result=timeout).
:- use_module(clpfd_interface,[catch_clpfd_overflow_call2/2]).
testcase_path_interrupt(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,Result) :-
    catch_clpfd_overflow_call2(
         testcase_path_interrupt2(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,Result),
         Result=clpfd_overflow).
testcase_path_interrupt2(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,Result) :-
    user_interruptable_call_det(testcase_path(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos),
                                Interrupt),
    ( Interrupt==ok -> Result=ok
    ; otherwise     -> Result=interrupt).

add_constants_to_state_space(ConstantsState,TestStates,TestStates,[],root) :-
    empty_state(ConstantsState),!.
add_constants_to_state_space(ConstantsState,TestStates,ConstTestStates,[Tuple],ConstantsID) :-
    Tuple = (SetupID,Setup,root,ConstantsID),
    specfile:create_setup_constants_operation(ConstantsState,Setup),
    tcltk_interface:tcltk_add_new_transition_transid(root,Setup,ConstantsID,concrete_constants(ConstantsState),[],SetupID),
    remove_constants(TestStates,ConstantsState,ConstantsID,ConstTestStates).
remove_constants([],_,_,[]).
remove_constants([InState|Irest],Constants,ConstId,[const_and_vars(ConstId,OutState)|Orest]) :-
    append(OutState,Constants,InState),!,
    remove_constants(Irest,Constants,ConstId,Orest).


add_operations_to_state_space([],_,[],[],[]).
add_operations_to_state_space([Operation|Orest],PrevId,[State|Srest],[TransInfo|Trest],[Tuple|Irest]) :-
    Tuple = (OpId,Operation,PrevId,NextId),
    tcltk_interface:tcltk_add_new_transition_transid(PrevId,Operation,NextId,State,TransInfo,OpId),
    add_operations_to_state_space(Orest,NextId,Srest,Trest,Irest).

testcase_path(INITIALISE,Events,EndPredicate,NormalisedConstantsState,OPS,StateSequence,TINFOS) :-
    init_wait_flags(WF,[testcase_path]),
    (INITIALISE=init
     -> OPS =  [OpInit|Operations], TINFOS = [ITransInfo|OpTransInfos],
        tc_initialise(ConstantsState,InitialState,OpInit,ITransInfo,WF)
     ;  OPS = Operations, TINFOS = OpTransInfos,
        (just_typing_opt(INITIALISE) -> set_up_a_typed_state(ConstantsState,InitialState,WF) % only set up types
          ; set_up_a_valid_state(ConstantsState,InitialState,WF)),
        % TO DO: project away constants and variables not needed for OPS and Predicates
        (has_predicate(INITIALISE,InitialPredicate)
         -> eval_predicate(InitialPredicate,InitialState,WF)
         ; true)
    ),
    set_up_events(Events,InitialState,ConstantsState,Operations,States,FinalState,OpTransInfos,WF),
    eval_predicate(EndPredicate,FinalState,WF),
    ground_wait_flags(WF),
    normalise_states([InitialState|States],StateSequence),
    normalise_state(ConstantsState,NormalisedConstantsState).

just_typing_opt(typing).
just_typing_opt(typing(_)).
has_predicate(pred(P),P). % add P to invariant
has_predicate(typing(P),P). % just use typing from invariant and add P
% other valid value for INITIALISE: init, invariant

testcase_predicate_timeout(Pred,Timeout,Result) :-
  on_exception(enumeration_warning(_,_,_,_,_),
          (time_out(testcase_predicate(Pred),
                    Timeout,
                    T),
           ( T==success -> Result=ok
           ; otherwise  -> Result=timeout)
          ),
          Result=timeout).

testcase_predicate(Predicate) :-
  init_wait_flags(WF,[testcase_predicate]),
  (just_typing_opt(Predicate) ->
    set_up_a_typed_state(_ConstantsState,InitialState,WF) % only set up types
  ;otherwise ->
    set_up_a_valid_state(_ConstantsState,InitialState,WF)),
  (has_predicate(Predicate,Pred) ->
    eval_predicate(Pred,InitialState,WF)
  ; true),
  ground_wait_flags(WF).

eval_predicate(Predicate,State,WF) :-
    empty_state(EmptyState),
    b_test_boolean_expression(Predicate,EmptyState,State,WF).

set_up_events([],State,_ConstState,[],[],State,[],_WF).
set_up_events([Event|Erest],InState,ConstantState,[Operation|Orest],
              [InterState|Srest],OutState,[TransInfo|Trest],WF) :-
    set_up_transition(Event,Operation,ConstantState,InState,InterState,TransInfo,WF),
    set_up_events(Erest,InterState,ConstantState,Orest,Srest,OutState,Trest,WF).

tc_initialise(ConstantState,InitialState,Initialisation,TransInfo,WF) :-
    set_up_constants_state(ConstantState,WF),
    set_up_initialisation(ConstantState,InitialState,Initialisation,TransInfo,WF). % TO DO: here I would like to be able to set up the invariant instead, if requested by another parameter

set_up_constants_state(ConstantsState,_WF) :-
    \+ not_all_transitions_added(root),
    \+ state_space:max_reached_or_timeout_for_node(root),
    % We should also check enumeration warnings : [DONE with max_reached_or_timeout_for_node]
    (state_space:transition(root,_,N) -> true),
    \+ ( state_space:transition(root,_,N2), N2\=N ),
    state_space:visited_expression(N,concrete_constants(ConstantsState)),
    %print(constants(ConstantsState)),nl,
    !. % there is only one possible valuation of the constants
set_up_constants_state(ConstantState,WF) :-
    b_get_properties_from_machine(Properties),
    b_get_machine_constants(Constants),
    empty_state(EmptyState),
    set_up_typed_localstate(Constants,_FreshVars,TypedVals,EmptyState,ConstantState,positive),
    static_symmetry_reduction_for_global_sets(ConstantState), % from b_global_sets
    b_tighter_enumerate_values(TypedVals,WF),
    b_test_boolean_expression(Properties,EmptyState,ConstantState,WF).

% set up a valid state satisfying the invariant
set_up_a_valid_state(ConstantState,ValidState,WF) :-
    set_up_constants_state(ConstantState,WF),
    b_get_machine_variables(Variables),
    % create_target_state(Variables,Values,ConstantsState,ValidState,WF),
    set_up_typed_localstate(Variables,_Values,TypedVals,ConstantState,ValidState,positive),
    b_tighter_enumerate_values(TypedVals,WF),
    b_get_invariant_from_machine(Invariant),
    empty_state(EmptyState),
    b_test_boolean_expression(Invariant,EmptyState,ValidState,WF).



set_up_constants_typed_state(ConstantState,WF) :-
    b_get_machine_constants(Constants),
    empty_state(EmptyState),
    set_up_typed_localstate(Constants,_FreshVars,TypedVals,EmptyState,ConstantState,positive),
    b_tighter_enumerate_values(TypedVals,WF).

% set up a valid state satisfying the typing conditions of the invariant; but not necessarily the invariant
set_up_a_typed_state(ConstantState,ValidState,WF) :-
    set_up_constants_typed_state(ConstantState,WF),
    b_get_machine_variables(Variables),
    % create_target_state(Variables,Values,ConstantsState,ValidState,WF),
    set_up_typed_localstate(Variables,_Values,TypedVals,ConstantState,ValidState,positive),
    b_tighter_enumerate_values(TypedVals,WF).


normalise_states([],[]).
normalise_states([I|Irest],[O|Orest]) :-
    normalise_state(I,O),
    normalise_states(Irest,Orest).
normalise_state(I,O) :-
    extract_vars(I,VI,VO,O),
    normalise_store(VI,VO).
extract_vars(const_and_vars(ConstID,IV),IV,OV,const_and_vars(ConstID,OV)) :- !.
extract_vars(IV,IV,OV,OV).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% constraint based test case generation

tcl_generate_cb_testcases(EventSelection,TargetString,MaxDepth,Filename,list(Uncovered)) :-
    % first translate Tk EventSelection Number list into Events
    get_selected_events(EventSelection,_AllEvents,Events),
    cbc_gen_test_cases_from_string(Events,TargetString,MaxDepth,Filename,Uncovered).

% TargetString can be #not_invariant for BMC - bounded model checking
cbc_gen_test_cases_from_string(Events,TargetString,MaxDepth,Filename,Uncovered) :-
    % parse target predicate for test-case sequences
    parse_cbc_target_predicate(TargetString,Target),
    cbc_gen_test_cases(Events,Target,MaxDepth,Filename,Uncovered).

% cb_uncovered_event/1: event is not covered in any test-case; but could by covered in a path
:- volatile cb_path/3, cb_path_testcase_only/3, cb_uncovered_event/1, cb_target_event/1, cb_final_event/1.
:- volatile cb_try_state_caching/0.
:- dynamic cb_path/3, cb_path_testcase_only/3, cb_uncovered_event/1, cb_target_event/1, cb_final_event/1.
:- dynamic cb_try_state_caching/0.

print_cb_uncovered_events :- %print_prefs
    findall(E,cb_uncovered_event_in_path(E),L1),
    findall(E,(cb_uncovered_event(E), \+cb_uncovered_event_in_path(E)),L2),
    (debug_mode(on)
     -> format_silent_msg('Completely uncovered events: ~w~n',[L1]),
        format_silent_msg('Covered but no test-case yet: ~w~n',[L2])
      ; length(L1,Len1), length(L2,Len2),
        format_silent_msg('Completely uncovered events: ~w, Covered but no test-case: ~w~n',[Len1,Len2])
     ).

%print_prefs :-
%    get_preference(use_smt_mode,SMT), get_preference(use_chr_solver,CHR),
%    get_preference(use_clpfd_solver,CLPFD), get_preference(use_po,PO),
%    format('Preferences: SMT = ~w, CHR=~w, CLPFD=~w, PROOF_INFO=~w~n',[SMT,CHR,CLPFD,PO]).
% constraint-based test case generation: one step predicate
:- use_module(solver_interface, [call_with_smt_mode_enabled/1]).
cbc_gen_test_cases(EventsToCover,TargetPredicate,MaxLength,Filename,UncoveredEvents) :-
    cbc_reset_gen_test_cases(EventsToCover),
    open_xml_testcase(Filename),
    get_preference(time_out,Timeout),
    call_with_smt_mode_enabled(
        call_cleanup(cb_timed_testcases(TargetPredicate, MaxLength, Timeout),
                     close_xml_testcase)),
    cbc_get_uncovered_events(UncoveredEvents).

cbc_reset_gen_test_cases(EventsToCover) :-
    reset_test_cases,
    cb_test_cases_init_search(EventsToCover).

cbc_get_uncovered_events(UncoveredEvents) :-
    findall(E,cb_uncovered_event(E),UncoveredEvents).
cbc_get_nr_uncovered_events(Len) :-
    cbc_get_uncovered_events(UncoveredEvents),
    length(UncoveredEvents,Len).

% constraint-based test case generation: multi step predicate; for use by prologTasks module
:- use_module(library(between),[between/3]).
cbc_gen_test_cases_task_step(EventSelection,TargetString,MaxLength,Filename, Final, Len, Task) :-
    % first translate Tk EventSelection Number list into Events
    get_selected_events(EventSelection,_AllEvents,EventsToCover),
    % parse target predicate for test-case sequences
    parse_cbc_target_predicate(TargetString,TargetPredicate),

    (Final=1 -> cbc_reset_gen_test_cases(final(EventsToCover))
     ; cbc_reset_gen_test_cases(EventsToCover)),
    get_preference(time_out,Timeout),

    open_xml_testcase(Filename),
    Task = sap: cb_test_cases_step(Len,TargetPredicate,MaxLength,Timeout,_),
    between(1,MaxLength,Len).

cbc_gen_test_cases_finish_tasks :-
    close_xml_testcase.

cb_timed_testcases(TargetPred,MaxLengthOfTestCase,Timeout) :-
    statistics(walltime, [Start|_]),
    InitialLength = 1,
    cb_test_cases_loop(InitialLength,TargetPred,MaxLengthOfTestCase,Timeout),
    statistics(walltime, [End|_]),
    Runtime is End - Start,
    write_results(Runtime).

cb_test_cases_loop(CurLength,TargetPred,MaxLengthOfTestCase,Timeout) :-
   cb_test_cases_step(CurLength,TargetPred,MaxLengthOfTestCase,Timeout,Result),
   (Result=paths_found
     -> L1 is CurLength+1,
        cb_test_cases_loop(L1,TargetPred,MaxLengthOfTestCase,Timeout)
     ;  Result=limit_reached -> check_progress_possible
     ;  Result=no_test_found -> check_progress_possible
     ; true
    ).

:- use_module(solver_interface, [call_with_smt_mode_enabled/1]).
:- use_module(tools_printing,[format_with_colour/4]).

% try and find test-cases of a particular depth: CurLengthOfTestCase
cb_test_cases_step(1,TargetPred,_MaxLengthOfTestCase,Timeout,Res) :- bounded_model_checking,
    % check if INITIALISATION can violate INVARIANT
    Events = [], Length=0,
    count_path(Length),
    call_with_smt_mode_enabled(create_testcase_path(init,Events,TargetPred,Timeout,PathResult)),
    debug_println(19,init_bmc_check(PathResult)),
    PathResult = [_],  % Result contains e.g., (0,$initialise_machine(int(-1)),root,0)
    !,
    Res =paths_found,
    format_err('~nBMC counterexample found, Length=~w~n',[Length]),
    assert(bounded_model_checking_complete), count_found(0,testcase),
    store_test_case(['$initialise_machine'],PathResult).
cb_test_cases_step(CurLengthOfTestCase,_TargetPred,_MaxLengthOfTestCase,_Timeout,success) :-
    cb_test_case_algorithm_finished,
    !,
    L1 is CurLengthOfTestCase-1,
    format_msg('~nCBC Algorithm finished (~w)~n',[L1]),
    (debug_mode(on), specfile:currently_opened_file(F) -> format_msg('  File: ~w~n',[F]) ; true).
cb_test_cases_step(CurLengthOfTestCase,_TargetPred,MaxLengthOfTestCase,_Timeout,limit_reached) :-
    CurLengthOfTestCase > MaxLengthOfTestCase,!,
    format_warn('~nLimit Reached: ~w~n',[MaxLengthOfTestCase]).
cb_test_cases_step(CurLengthOfTestCase,TargetPred,MaxLengthOfTestCase,Timeout,paths_found) :-
    % now try to find a path of length CurLengthOfTestCase
    call_with_smt_mode_enabled(cb_find_paths(CurLengthOfTestCase,TargetPred,MaxLengthOfTestCase,Timeout)),
    !,
    (bounded_model_checking_complete -> format_err('~nBMC counterexample found, Length=~w~n',[CurLengthOfTestCase])
     ; \+ cb_has_uncovered_events -> format_ok('~nEverything covered, Length=~w~n',[CurLengthOfTestCase])
     ; \+ cb_has_uncovered_feasible_events -> format_ok('~nEverything feasible covered, Length=~w~n',[CurLengthOfTestCase])
     ; true).
cb_test_cases_step(CurLengthOfTestCase,_TargetPred,MaxLengthOfTestCase,_Timeout,Res) :-
    % either everything covered or MaxLength reached or no way to extend current paths
    (CurLengthOfTestCase = MaxLengthOfTestCase
     -> format_warn('~nCould not generate a test case of length ~w~n',[CurLengthOfTestCase]),
        Res = no_test_found
      ; format_warn('~nCould not generate a path of length ~w~n',[CurLengthOfTestCase]),
        Res = no_path_found
    ).

% utilities to print colored messages
format_msg(FormatStr,Args) :- format_with_colour(user_output,[blue],FormatStr,Args).
format_err(FormatStr,Args) :- format_with_colour(user_output,[red,bold],FormatStr,Args).
format_warn(FormatStr,Args) :- format_with_colour(user_output,[cyan,bold],FormatStr,Args).
format_ok(FormatStr,Args) :- format_with_colour(user_output,[green,bold],FormatStr,Args).

format_silent_msg(_,_) :- silent_mode(on),!.
format_silent_msg(FormatStr,Args) :- format_msg(FormatStr,Args).

format_silent_ok(_,_) :- silent_mode(on),!.
format_silent_ok(FormatStr,Args) :- format_with_colour(user_output,[green],FormatStr,Args).


cb_test_case_algorithm_finished :-
   (cb_has_uncovered_feasible_events ->  bounded_model_checking_complete ; true).

:- dynamic cb_uncovered_event_in_path/1. % uncovered in any path (and thus also test-case)
% reset datastructures for CBC test case generation
cb_test_cases_init_search(Events) :-
    reset_counters,
    retractall(cb_path(_,_,_)), % retract all found paths
    retractall(cb_path_testcase_only(_,_,_)),
    retractall(cb_timeout_path(_,_)),
    assert(cb_path(0,[],no_state)), % generate an intial path; Note: we assume the initialisation is feasible
    % set up all coverage targets:
    retractall(cb_uncovered_event(_)),
    retractall(cb_uncovered_event_in_path(_)),
    retractall(cb_target_event(_)),
    retractall(cb_final_event(_)),
    cb_test_cases_init_uncovered_in_path,
    add_target_events(Events).
cb_test_cases_init_uncovered_in_path :-
    b_top_level_operation(Event),
    assert(cb_uncovered_event_in_path(Event)),
    fail.
cb_test_cases_init_uncovered_in_path.

add_target_events(Desc) :- add_events_aux(Desc),fail.
add_target_events(_).
add_events_aux(all) :- b_top_level_operation(Event), assert_cb_uncovered_event(Event).
add_events_aux([H|T]) :- member(Event,[H|T]),
    match_event(Event,MatchedEvent),
    assert_cb_uncovered_event(MatchedEvent).
add_events_aux(final(List)) :- member(Event,List),
   match_event(Event,MatchedEvent),
   assert_cb_uncovered_event(MatchedEvent), assert(cb_final_event(MatchedEvent)).

match_event(match_event(E),Match) :- !, % try and find operation/event with E occuring inside string
   atom_codes(E,EC), append(EC,_,EC2),
   if((b_top_level_operation(Match),
       atom_codes(Match,MC),
       append(_,EC2,MC)),
      debug_println(19,match_event(E,Match)),
      (add_error(cbc_cover_match,'No matching event for: ',E),fail)).
match_event(E,E).

assert_cb_uncovered_event(E) :- \+ b_top_level_operation(E),!,
   add_error(cbc_cover,'Unknown event: ',E).
assert_cb_uncovered_event(E) :- assert(cb_uncovered_event(E)), assert(cb_target_event(E)).

get_feasible_event_name(Event) :-
    b_is_operation_name(Event),
    \+ infeasible_operation_cache(Event).
cb_feasible_uncovered_event(E) :- cb_uncovered_event(E), \+ infeasible_operation_cache(E).

useless_prefix(PrefixPath) :-
    last(PrefixPath,LastPrefix),
    \+ cb_uncovered_event(LastPrefix), % we still need to try and navigate into a valid end state
    bmachine:b_operation_cannot_modify_state(LastPrefix). % no use in trying such prefixes


:- dynamic stat_last_print_time/1.
print_cb_stats(Level,StartWTime) :-
    statistics(walltime,[CurWTime,_]),
    ToTime is CurWTime-StartWTime,
    (stat_last_print_time(Last)
     -> (CurWTime - Last > 60000
          -> retract(stat_last_print_time(_)),
             assert(stat_last_print_time(CurWTime)),
             format_msg('~nTotal runtime for level ~w so far: ~w ms~n',[Level,ToTime]),
             findall(Op,cb_feasible_uncovered_event(Op),Uncov),
             length(Uncov,LU),
             (LU>160 -> format_msg('Uncovered events: ~w~n',[LU])
               ; format_msg('Uncovered events: ~w : ~w~n',[LU,Uncov])),
             write_results,nls
          ;  true)
     ; assert(stat_last_print_time(CurWTime))).

cb_find_paths(Length,TargetPredicate,MaxLength,Timeout) :-
    (all_uncovered_events_are_final
      -> (CoveredStatus = uncovered ; CoveredStatus = covered), % force using uncovered operations first in b_ordered_get_operation
         % this only works if once an uncovered operation becomes covered it no longer needs to be enumerated; otherwise we will miss paths
         format_silent_msg('~nCBC test search: starting length ~w (~w events)~n',[Length,CoveredStatus])
       ; format_silent_msg('~nCBC test search: starting length ~w~n',[Length])),
    print_cb_uncovered_events,
    statistics(walltime,[StartWTime,_]),
    % set up template for Path = PrefixPath ^ [Event]
    cb_prefix_trace(Length,PreLength,PrefixPath,Event,Path),
    % choose any existing prefix of length PreLength for which a CBC solution was found:
    cb_path(PreLength,PrefixPath,LastStateId),
     % TO DO: maybe also try cb_timeout_path(PreLength,PrefixPath), LastStateId = no_state
    \+ useless_prefix(PrefixPath), % avoid using operations that do not change the state
    statistics(walltime,[WT,_]), Delta is WT-StartWTime,
    (Delta>1000 -> nl ; true), % too long lines on terminal slow things down
    (debug_mode(on) -> format_msg('~nTrying to extend path of length ~w: ~w~n',[Length,PrefixPath])
     ; printsilent('*'),printsilent(Length)),flush_output,
    % print(prefix(PrefixPath)),nl,
    % check if full coverage not yet achieved:
    \+ cb_test_case_algorithm_finished,
    % now choose any event; uncovered ones first:
    b_ordered_get_operation(Event,CoveredStatus),
    \+ bounded_model_checking_complete, % check again ; we may have found a solution with the previous event
    possible_trace(PrefixPath,Event,Length), % check if trace is in principle possible
    printsilent('.'),flush_output,
    %user:do_one_gui_event,
    % statistics information: store that we attempt another path of length Length:
    count_path(Length),
    (Length>=MaxLength -> Result=testcase, % no need to find partial solutions now
        cb_feasible_uncovered_event(Event) % only set up paths for feasible Events we want to cover
      ; true
      ),
    print_cb_stats(Length,StartWTime),
    % and try to set up that path PrefixPath ^ [Event]
    debug_println(9,setting_up_path(Path,Event)),
    cb_set_up_path(Path,Event,TargetPredicate,LastStateId,Timeout,Result),
    % now store info that we have found a solution for length Length with Result
    debug_println(9,found_solution(Length,Result,Path)),
    count_found(Length,Result),
    % fail-loop
    fail.
cb_find_paths(Length,_TargetPredicate,_MaxLengthOfTestCase,_Timeout) :-
    % success if a path or testcase of this length was generated, fail otherwise
    (cb_path(Length,_,_) -> true ; counter(testcase_count(Length),TC), TC>0).

% enumerate events with uncovered first:
b_ordered_get_operation(Event,CoveredStatus) :-
   findall(Event,get_feasible_event_name(Event),AllEvents),
   include(cb_uncovered_event,AllEvents,Uncovered),
   exclude(cb_uncovered_event,AllEvents,Covered),
   % Note: we do a findall and exclude/include first to avoid side-effects from modifying cb_uncovered_event
   %  and possibly forgetting an event
   %print(uncovered(Uncovered)),nl,
   ( CoveredStatus=uncovered,
     member(Event,Uncovered)
    ;
     CoveredStatus=covered,
     member(Event,Covered),
     \+ bmachine:b_operation_cannot_modify_state(Event),
     \+ cb_final_event(Event) % it is already covered; no need to cover it again
   ). % no use in using a query/skip operation at the end

all_uncovered_events_are_final :- cb_uncovered_event(Ev), \+ cb_final_event(Ev),!,fail.
all_uncovered_events_are_final.

cb_prefix_trace(Length,PreLength,PrefixPath,Event,Path) :-
    PreLength is Length-1,
    length(PrefixPath,PreLength),append(PrefixPath,[Event],Path).

cb_has_uncovered_events :-
    cb_uncovered_event(_),!. %%
cb_has_uncovered_feasible_events :-
    cb_feasible_uncovered_event(_),!. %%

cb_set_up_events(Trace,InitialState,ConstantsState,FinalState,DetailedTrace,WF) :-
    DetailedTrace = dt1(States,Operations,OpTransInfos),
    set_up_events(Trace,InitialState,ConstantsState,Operations,States,FinalState,OpTransInfos,WF).

combine_detail_infos(ConstantsState,InitialState,Initialisation,InitInfo,DT1,DT) :-
    DT1 = dt1(States,Operations,OpTransInfos),
    DT = dt(ConstantsState,
            [InitialState|States],
            [Initialisation|Operations],
            [InitInfo|OpTransInfos]).



% check if trace : Trace ^ [Event] is in principle possible using static/cbc enabling analysis results
possible_trace(Trace,Event,Length) :- get_preference(use_po,true),
   last(Trace,PreviousEvent),
   !,
   %format('Try "~w" after "~w"~n',[Event,PreviousEvent]),nl,
   check_operation_sequence_possible(PreviousEvent,Event,Length),
      % TO DO: we could delay computing this until after cb_set_up_path has failed; if we managed to find a solution from the intialisation no need to do this check which starts from Invariant
   true.
possible_trace(_,_,_).


:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(specification_initialised,reset_sap,
                           'Initialise module sap.').

:- use_module(enabling_analysis,[operation_sequence_possible/3]).
:- dynamic operation_sequence_possible_cache/3.
:- dynamic operation_can_be_enabled_by_cache/3.
reset_sap :-
   retractall(operation_sequence_possible_cache(_,_,_)),
   retractall(operation_can_be_enabled_by_cache(_,_,_)).

check_operation_sequence_possible(PreviousEvent,Event,Length) :-
   (operation_sequence_possible_cache(PreviousEvent,Event,R)
       -> chk_possible(R,PreviousEvent,Event,Length)
     ; operation_sequence_possible(PreviousEvent,Event,R) ->
        assert(operation_sequence_possible_cache(PreviousEvent,Event,R)),
        chk_possible(R,PreviousEvent,Event,Length)
     ; assert(operation_sequence_possible_cache(PreviousEvent,Event,failure))
       % we assume it is possible to be on the safe side
   ).

%chk_possible(R,PreviousEvent,Event) :- print(chk(R,PreviousEvent,Event)),nl,fail.
chk_possible(impossible,_,Event,_) :- !,
  (debug_mode(on) -> format_warn('Impossible event: ~w~n',[Event]) ; printsilent('-')), fail.
chk_possible(activation_independent,Prev,Event,_) :- cb_uncovered_event_in_path(Event), !,
   % PreviousEvent cannot influence the variables of the guards of Event;
   %   as Event was previously not covered it cannot possible be enabled after PreviousEvent
   (debug_mode(on) -> format_msg('Activation independent: ~w (previous ~w)~n',[Event,Prev]) ; printsilent('||')),
   fail.
% this is probably not very often useful ; to do: enable only if we repeatedly fail to satisfy this combination?
% or do this only at deeper levels ?
chk_possible(_,PreviousEvent,Event,Length) :-
   Length>3, % only start doing this check later, when hopefully many events are already covered
   cb_uncovered_event_in_path(Event),!,
   (check_operation_can_be_enabled_by(PreviousEvent,Event) -> true
      ; (debug_mode(on) -> format_msg('Event ~w cannot be enabled by ~w~n',[Event,PreviousEvent]) ; printsilent('|^|')),fail).
chk_possible(_,_,_,_).

:- use_module(enabling_analysis,[operation_can_be_enabled_by/3]).

check_operation_can_be_enabled_by(PreviousEvent,Event) :-
   (operation_can_be_enabled_by_cache(PreviousEvent,Event,R) -> R \= failure
     ; operation_can_be_enabled_by(PreviousEvent,Event,R) ->
        assert(operation_can_be_enabled_by_cache(PreviousEvent,Event,R)),
        R \= failure
     ; assert(operation_can_be_enabled_by_cache(PreviousEvent,Event,failure)),
       fail
   ).

% TO DO: use this check at regular intervals ?!? and use it to abort cbc_tests if no progress possible and or prune the events that can be reached
% check if there exists a covered event that can enable one of the uncovered ones:
check_progress_possible :-
% to do: check only valid if target pred = truth
    (progress_possible(PreviousEvent,Event)
       -> format_msg('Progress possible, e.g., covered ~w could enable uncovered ~w~nTry increasing depth bound for cbc_tests.~n',[PreviousEvent,Event])
       ;  cbc_get_uncovered_events(Uncov),
          format_warn('Impossible to activate any of the uncovered events ~w.~nNo further progress possible!~n',[Uncov])).

progress_possible(PreviousEvent,Event) :- cb_feasible_uncovered_event(Event),
    cb_covered_event(PreviousEvent),
    check_operation_can_be_enabled_by(PreviousEvent,Event).

cb_covered_event(Event) :- % TO DO: this may give false positives; if an event was not marked to be covered initially and has not yet occured
  b_top_level_operation(Event), \+ cb_uncovered_event(Event).


%cb_set_up_path([Event],Event,TargetPredicate,no_state,Timeout,Result) :-
%   cb_find_constants_and_initial_vars_in_statespace(State,InitID),
%   do_event(InitID,Event,SuccId),
%   visited_expression(SuccId,OutState), expand ...
%   cb_finalize_path_with_timeout(TargetPredicate,OutState,Trace,Event,WF,LResult,Timeout,TO_Success).
%cb_set_up_path(Trace,Event,TargetPredicate,StateId,Timeout,Result) :-
%   print(cb_set_up_path(Trace,Event,TargetPredicate,StateId,Timeout,Result)),nl,fail.
cb_set_up_path(Trace,Event,TargetPredicate,StateId,Timeout,Result) :-
   % StateID is the last state of the previously found path Trace (without Event)
    cb_try_state_caching, % if true: try to use model checking paths found; currently disabled
    StateId \== no_state,
    lookup_state(StateId,ConstantState,InState),
    init_wait_flags(WF,[cb_set_up_path]),
    set_up_transition(Event,Operation,ConstantState,InState,OutState,TransInfo,WF),
    cb_finalize_path_with_timeout(TargetPredicate,OutState,Trace,Event,WF,LResult,Timeout,TO_Success),
    ( TO_Success = time_out ->
        cb_write_timeout(Timeout,Trace),fail
    ; otherwise ->
        cb_store_single_step_in_statespace(ConstantState,OutState,StateId,Operation,TransInfo,LastStateId,Sequence),
        cb_save_new_step(Trace,Event,LResult,LastStateId,Sequence),
        % if we did not find a test case, just a normal step, we need to fall back
        % to constraint solving on the full path
        !,
        ( LResult==testcase -> Result=testcase
        ; cb_set_up_path(Trace,Event,TargetPredicate,no_state,Timeout,testcase) -> Result=testcase
        ; otherwise -> Result = path)).
%cb_set_up_path(Trace,Event,TargetPredicate,StateId,Timeout,Result) :- %print(try(Trace,Event,TargetPredicate,StateId)),nl, % see if an outgoing transition event exists for StateId already explored by MC
%    is_truth(TargetPredicate), do_event(StateId,Event,LastStateId),!, print('MC'),Sequence=[], % TO DO: compute sequence
%     Result=testcase,cb_save_new_step(Trace,Event,Result,LastStateId,Sequence).
cb_set_up_path(Trace,Event,TargetPredicate,_StateId,Timeout,Result) :-
    init_wait_flags(WF,[cb_set_up_path]),
    % set up constants and initialisation:
    tc_initialise(ConstantsState,InitialState,Initialisation,TransInfo,WF),
    combine_detail_infos(ConstantsState,InitialState,Initialisation,TransInfo,DT1,DT),
    cb_set_up_events(Trace,InitialState,ConstantsState,FinalState,DT1,WF),
    % either find a real testcase where TargetPredicate holds at end
    % or find a path where TargetPredicate does not hold at end:
    cb_finalize_path_with_timeout(TargetPredicate,FinalState,Trace,Event,WF,Result,Timeout,TO_Success),
    !,
    ( TO_Success = time_out ->
        cb_write_timeout(Timeout,Trace),
        fail
    ; (debug_mode(on) -> format_ok('Found ~w path for ~w (~w)~n',[Result,Event,Trace])
       ; Result=testcase -> printsilent('!') %,print(Trace),nl
       ; otherwise -> printsilent('+')
      ),
        cb_store_path_in_statespace(DT,LastStateId,Sequence), % only do if path required or new testcase generated
        cb_save_new_step(Trace,Event,Result,LastStateId,Sequence)).
cb_save_new_step(Trace,Event,Result,LastStateId,Sequence) :-
    (retract(cb_uncovered_event_in_path(Event))
     -> (debug_mode(on) -> format_ok('Covered event ~w~n',[Event]) ; printsilent('#')) ; true),
    ( Result == testcase
     -> (cb_final_event(Event) -> cb_save_path_testcase_only(Trace,LastStateId) ; cb_save_path(Trace,LastStateId)),
        cb_save_found_test_case(Trace,Sequence)
      ; cb_save_path(Trace,LastStateId)).

lookup_state(StateId,ConstStore,VarStore) :-
    visited_expression(StateId,Store),
    lookup_state2(Store,ConstStore,VarStore).
lookup_state2(const_and_vars(ConstId,VarStore),ConstStore,VarStore) :-
    !,visited_expression(ConstId,concrete_constants(ConstStore)).
lookup_state2(Store,Empty,Store) :- empty_state(Empty).

% finalize the path by checking whether TargetPredicate holds in final state; if so: we have a testcase otherwise just a path
cb_finalize_path_with_timeout(TargetPredicate,_FinalState,Trace,Event,WF,Result,Timeout,TO_Success) :-
    % if the target predicate is just true, we can store every found path
    % directly as a possible test trace, too.
    is_truth(TargetPredicate),!, % avoid ground_wait_flags twice if path infeasible
    % now check if the path makes sense as either testcase or path; if not: no use in enumerating
    (is_potential_new_test_case(Trace) -> Result = testcase % replace by is_potential_test_case if you want more tests in the test xml files
     ; is_potential_path_to_be_stored(Trace,Event) -> Result = path), % paths are not written to xml file
    ground_wait_flags_with_timeout(WF,Timeout,TO_Success),
    !.
cb_finalize_path_with_timeout(TargetPredicate,FinalState,Trace,Event,WF,RESULT,Timeout,TO_Success) :-
    is_potential_new_test_case(Trace),
    % Note: if the target predicate is fulfilled in the last state, we
    % can store the solution as a test trace.
    (bounded_model_checking,
     b_specialized_invariant_for_op(Event,SpecINV) % only computed if get_preference(use_po,true)
     -> bsyntaxtree:create_negation(SpecINV,NegSpecINV),
        eval_predicate(NegSpecINV,FinalState,WF),
        (debug_mode(on) -> print('Specialized Invariant Target: '),translate:print_bexpr(SpecINV),nl ; true)
     ;  eval_predicate(TargetPredicate,FinalState,WF),
        (debug_mode(on) -> print('Target: '),translate:print_bexpr(TargetPredicate),nl ; true)
     ),
    % print(grounding_with_predicate(Event,Trace)),nl,
    ground_wait_flags_with_timeout(WF,Timeout,TO_Success),
    (RESULT == testcase -> true % we are only looking for testcases not paths where we do not reach the target predicate
     ; TO_Success = time_out -> printsilent('@Timeout@'),fail % try and find a solution without target predicate below
     ; true),
    !,
    RESULT=testcase.
cb_finalize_path_with_timeout(_TargetPredicate,_FinalState,Trace,Event,WF,path,Timeout,TO_Success) :-
    % if the path is not feasible with the target predicate applied to the
    % last state, it still might be feasible as a starting point for other traces.
    is_potential_path_to_be_stored(Trace,Event),
    % print(grounding_without_predicate(_Trace)),nl,
    ground_wait_flags_with_timeout(WF,Timeout,TO_Success),!.

% check if a sequence of events is a potential  new test case
is_potential_new_test_case(_Trace) :- bounded_model_checking,
    !. % any trace can potentially lead to an invariant violation % TO DO: check if last event is not a proven event
is_potential_new_test_case(Trace) :- ((member(Ev,Trace), cb_uncovered_event(Ev)) -> true).  %print(uncov(Ev)),nl ;  fail)
% we could add a requirement that only certain events are admissible as last events (a bit like cb_final_event + requiring that if there is one cb_final_event then one must be used)

:- public is_potential_test_case/1.
% check if a sequence of events is a potential test case
is_potential_test_case(_Trace) :- bounded_model_checking,
    !. % any trace can potentially lead to an invariant violation % TO DO: check if last event is not a proven event
is_potential_test_case(Trace) :- ((member(Ev,Trace), cb_target_event(Ev)) -> true).  %print(uncov(Ev)),nl ;  fail)

% check if it is worth storing this trace, i.e., could it later be extended to a test-case
is_potential_path_to_be_stored(_Trace,Event) :-
   (cb_final_event(Event) -> printsilent(f),fail ; true). % this event is only allowed at the end: we cannot extend the trace


ground_wait_flags_with_timeout(WF,Timeout,TO_Success) :-
    time_out(ground_wait_flags(WF),Timeout,TO_Success).

% cb_save_found_test_case(TraceOfOperationNames,Sequence of 4-tuples (Id,Op,Src,Dest))
cb_save_found_test_case(Trace,Sequence) :- % print(found(Trace)),nl,
    % mark all events in Trace as covered
    cb_mark_all_covered(Trace,NewCover),
    (bounded_model_checking -> assert(bounded_model_checking_complete) ; true),
    cb_get_initialised_state(Sequence,InitState),
    % Sequence is a list of (_Id,Op,Src,Dst)
    (NewCover=[] -> true
     ; store_test_case(NewCover,Sequence)
    ),
    cb_write_test_case(Sequence,InitState).

% store main test cases for later replaying in GUI / Tcl/Tk
:- dynamic stored_test_case/3, test_case_nr/1.
test_case_nr(0).

reset_test_cases :- retractall(stored_test_case(_,_,_)), retract(test_case_nr(_)),
    assert(test_case_nr(0)).
store_test_case(NewCover,Sequence) :- retract(test_case_nr(Nr)), N1 is Nr+1,
    assert(test_case_nr(N1)),
    %print(stored_test_case(N1,NewCover,Sequence)),nl,
    assert(stored_test_case(N1,NewCover,Sequence)),
    (bounded_model_checking
     -> write('Found counter example'),nl,
        tcl_execute_stored_test_case(N1) ; true).

% a function to get a textual description of a stored test case
stored_test_case_description(N1,Descr) :-
   stored_test_case(N1,NewCover,Sequence), get_op(Sequence,Ops),
   add_sep(NewCover,NCS),ajoin(NCS,NCA),
   ajoin(['Test ', N1, '. Covering ',NCA,'. Trace = '|Ops],Descr).
% function to get the list of operation names of a stored test case
stored_test_case_op_trace(N1,OpTrace) :-
   stored_test_case(N1,_NewCover,Sequence),
   maplist(get_op_name,Sequence,OpTrace).

add_sep([],[]).
add_sep([A],R) :- !, R=[A].
add_sep([A|T],[A,' '|AT]) :- add_sep(T,AT).
get_op([],[]).
get_op([Tuple|Rest],[Name,' '|T]) :-
    get_op_name(Tuple,Name),
    get_op(Rest,T).
get_op_name((_Id,Op,_Src,_Dst),Name) :-
   specfile:get_operation_name(Op,Name). % deal with '-->'

% a function to execute a stored test case in the animator; should probably move to tcltk_interface.pl
tcl_execute_stored_test_case(Nr) :-
   stored_test_case(Nr,NewCover,Sequence),
   debug_println(19,executing_stored_test_case(Nr,NewCover,Sequence)),
   Sequence = [(_,_,Src,_)|_],
   tcltk_interface:tcltk_execute_trace_to_node(Src), % first find a trace to the starting initialisation
   last(Sequence,(_,_,_,Dest)),
   maplist(get_op_id,Sequence,OpIds),
   maplist(get_dest,Sequence,Ids),
   execute_id_trace_from_current(Dest,OpIds,Ids).

get_op_id((OpID,_,_,_),OpID).
get_dest((_,_,_,Dest),Dest).

tcl_get_stored_test_cases(list(L)) :-
    findall(Descr,stored_test_case_description(_,Descr),L).

cb_get_initialised_state([(_Id,_Op,Src,_Dst)|_],InitState) :-
    visited_expression(Src,Store),
    expand_const_and_vars_to_full_store(Store,InitState).

cb_mark_all_covered([],[]).
cb_mark_all_covered([Event|Rest],NewCover) :-
    ( retract(cb_uncovered_event(Event)) -> NewCover=[Event|RestCover],
        format_silent_ok('~nCBC test search: covered event ~w~n',[Event])
    ; otherwise -> NewCover=RestCover),
    cb_mark_all_covered(Rest,RestCover).

cb_write_test_case(Sequence,InitState) :-
    xml_write('  <test_case>\n'),
    if(cb_xml_initialisation(InitState),true,add_internal_error('Call failed: ',cb_xml_initialisation(InitState))),
    if(cb_xml_sequence(Sequence),true,add_internal_error('Call failed: ',cb_xml_sequence(Sequence))),
    xml_write('  </test_case>\n'),
    flush_output_xml.
cb_xml_initialisation(InitState) :-
    xml_write('    <initialisation>\n'),
    b_get_machine_constants(Constants),cb_xml_init2(InitState,Constants,constant),
    b_get_machine_variables(Variables),cb_xml_init2(InitState,Variables,variable),
    xml_write('    </initialisation>\n').
cb_xml_init2(Store,Identifiers,Type) :-
    get_texpr_id(TId,Id),
    member(TId,Identifiers),
    lookup_value(Id,Store,BValue),
    cb_xml_value(Id,Type,BValue),
    fail.
cb_xml_init2(_Store,_Identifiers,_Type).
cb_xml_value(Id,Type,BValue) :-
    translate_bvalue(BValue,SValue),
    xml_write('      <value type=\"'),
    xml_write(Type),
    xml_write('\" name=\"'),xml_write(Id),xml_write('\">'),
    xml_write(SValue),xml_write('</value>\n').
cb_xml_sequence([]).
cb_xml_sequence([(_Id,Op,Src,Dst)|Rest]) :- !,
    specfile:get_operation_name(Op,Name),
    cb_xml_step(Name,Op,Src,Dst), % TO DO: also provide operation parameters
    cb_xml_sequence(Rest).
cb_xml_sequence([loop(_ID)]).
cb_xml_step(Event,Op,SrcId,DstId) :-
    b_get_machine_operation(Event,_,Parameters,_),
    Parameters = [_|_],!,
    xml_write('    <step name=\"'),xml_write(Event),xml_write('\">\n'),
    specfile:get_operation_arguments(Op,Args),  % TO DO: maybe also write out return arguments ?!
    xml_write_parameters(Event,Parameters,Args),
    expand_state(SrcId,Src),
    expand_state(DstId,Dst),
    maplist(cb_xml_modified(Src),Dst),
    xml_write('    </step>\n').
cb_xml_step(Event,_,_Src,_Dst) :-
    xml_write('    <step name=\"'),xml_write(Event),xml_write('\" />\n').

xml_write_parameters(Event,Parameters,Args) :-
   get_preference(prefix_internal_operation_arguments,Prefix),
   maplist(cb_xml_parameter(Event,Prefix),Parameters,Args).

% write operation parameters
cb_xml_parameter(_Event,Prefix,Parameter,BValue) :-
    get_texpr_id(Parameter,Id),!,
    (is_internal_parameter(Prefix,Id) -> true % do not print it
      ; cb_xml_value_element(value,Id,BValue)).
cb_xml_parameter(Event,Prefix,Parameter,BValue) :-
    add_internal_error('Call failed: ',cb_xml_parameter(Event,Prefix,Parameter,BValue)).

is_internal_parameter(Prefix,Parameter) :- atom_concat(Prefix,_,Parameter).

% write modified variable values
cb_xml_modified(Src,bind(Name,NewValue)) :-
    cb_var_is_modified(Name,NewValue,Src),!,
    cb_xml_value_element(modified,Name,NewValue).
cb_xml_modified(_,_).

cb_var_is_modified(Name,NewValue,Src) :-
    memberchk(bind(Name,OldValue),Src),!,
    not_equal_object(OldValue,NewValue).
cb_var_is_modified(Name,_NewValue,Src) :-
    nonmember(bind(Name,_OldValue),Src).

xml_write_binding(Element,bind(Name,Value)) :- cb_xml_value_element(Element,Name,Value).

cb_xml_value_element(Element,Name,Value) :-
    xml_write('      <'),xml_write(Element),
    xml_write(' name=\"'),xml_write(Name),xml_write('\">'),
    xml_write_b_value(Value),
    !,
    xml_write('</'),xml_write(Element),xml_write('>\n').
cb_xml_value_element(Element,Name,Value) :-
    add_internal_error('Call failed: ',cb_xml_value_element(Element,Name,Value)).

xml_write_b_value((Fst,Snd)) :- !,
    xml_write('<pair><fst>'),
    xml_write_b_value(Fst),
    xml_write('</fst><snd>'),
    xml_write_b_value(Snd),
    xml_write('</snd></pair> ').
xml_write_b_value(avl_set(A)) :- !, custom_explicit_sets:expand_custom_set_to_list(avl_set(A),Elements),
    xml_write('<set>'),
    maplist(xml_write_b_value,Elements),
    xml_write('</set> ').
xml_write_b_value([H|T]) :- !,
    xml_write('<set>'),
    maplist(xml_write_b_value,[H|T]),
    xml_write('</set> ').
xml_write_b_value(rec(Fields)) :- !,
    xml_write('<record>'),
    maplist(xml_write_b_field_value,Fields),
    xml_write('</record> ').
xml_write_b_value(string(S)) :- !,
    xml_write('<string>'),
    xml_write_b_value(S), % TO DO: encode S !
    xml_write('</string> ').
xml_write_b_value(Value) :- translate_bvalue(Value,SValue), xml_write(SValue).
% TO DO: support other types: other set representations, booleans pred_false, pred_true
% [] -> empty_set ?
% interval closures ?  freeval(ID,Case,Value)

xml_write_b_field_value(field(Name,Val)) :-
   xml_write('<field name=\"'), xml_write(Name), xml_write('\">'),
   xml_write_b_value(Val), xml_write('</field>').

expand_state(Id,VariableBindings) :-
    visited_expression(Id,State),
    expand_to_constants_and_variables(State,_ConstantBindings,VariableBindings),
    !.
expand_state(Id,VariableBindings) :-
    add_internal_error('Call failed: ',expand_state(Id,VariableBindings)),
    VariableBindings=[].

:- volatile cb_timeout_path/2.
:- dynamic cb_timeout_path/2.
cb_write_timeout(Timeout,Events) :-
    length(Events,Length),
    count_timeout(Length),
    printsilent(timeout),flush_output, %nl,write(Events),nl,
    assert(cb_timeout_path(Length,Events)),
    xml_write('  <!-- timeout ('),xml_write(Timeout),xml_write(') ms) for events '),
    cb_write_timeouts2(Events),
    xml_write(' --!>\n'),flush_output_xml.
cb_write_timeouts2([Event]) :-
    !,xml_write(Event).
cb_write_timeouts2([Event|Erest]) :-
    xml_write(Event),xml_write(','),cb_write_timeouts2(Erest).

cb_save_path(Trace,LastStateId) :-
    length(Trace,CurDepth),
    assert(cb_path(CurDepth,Trace,LastStateId)).
cb_save_path_testcase_only(Trace,LastStateId) :- % save mainly for Tcl/Tk tree viewer
    length(Trace,CurDepth),
    assert(cb_path_testcase_only(CurDepth,Trace,LastStateId)).

cb_store_path_in_statespace(dt(ConstantsState1,States1,Operations,TransInfos),LastStateId,RealOperations) :-
    normalise_states([ConstantsState1|States1],[ConstantsState|States]),
    add_constants_to_state_space(ConstantsState,States,ConstStates,_SetupSequence,StartStateId),
    add_operations_to_state_space(Operations,StartStateId,ConstStates,TransInfos,OpSequence),
    OpSequence = [_InitOp|RealOperations],
    last(OpSequence,(_,_,_,LastStateId)).

cb_store_single_step_in_statespace(ConstantsState,State1,PrevStateId,Operation,TransInfo,NewStateId,[Tuple]) :-
    normalise_state(State1,State2),
    ( cb_find_constants_state_in_statespace(ConstantsState,ConstId) ->
        remove_constants([State2],ConstantsState,ConstId,[State])
    ; otherwise ->
        State = State2),
    add_operations_to_state_space([Operation],PrevStateId,[State],[TransInfo],[Tuple]),
    Tuple = (_,_,_,NewStateId).

% find a state where the constants have been valued
cb_find_constants_state_in_statespace(ConstantsState,_ConstId) :-
    empty_state(ConstantsState),!,fail.
cb_find_constants_state_in_statespace(ConstantsState,ConstId) :-
    transition(root,_,ConstId),
    visited_expression(ConstId,concrete_constants(ConstantsState)),!.

% %try and find an initial state in the state space
%cb_find_constants_and_initial_vars_in_statespace(InitialState,InitId) :-
%    transition(root,_,Id1),
%    (visited_expression(Id1,concrete_constants(_))
%      ->  transition(Id1,_,InitId)
%      ;   InitId = Id1),
%    visited_expression(InitId,State),
%    expand_const_and_vars_to_full_store(State,InitialState).

test_generation_by_xml_description(Filename) :-
    read_test_generation_description(Filename,Description),
    execute_test_generation(Description).
execute_test_generation(cb_test_description(Events,TargetString,MaxDepth,Preferences,Filename,StateCaching)) :-
    parse_cbc_target_predicate(TargetString,Target),
    maplist(set_ec_preference,Preferences),
    retractall(cb_try_state_caching),
    ( StateCaching = true
      -> assert(cb_try_state_caching), printsilent('State caching (states from model checker) enabled'),nls
      ; otherwise -> true),
    cbc_gen_test_cases(Events,Target,MaxDepth,Filename,_Uncovered).
set_ec_preference(pref(Name,Value)) :-
    set_eclipse_preference(Name,Value).

read_test_generation_description(Filename,Description) :-
    read_string_from_file(Filename,Codes),
    xml_parse(Codes,xml(_Atts,Content)),
    ( read_test_generation_description2(Content,Description) ->
        true
    ; otherwise ->
        add_error(sap,'error while reading test description file'),fail).
read_test_generation_description2(Xml,cb_test_description(Events,Target,MaxDepth,Preferences,File,StateCaching)) :-
    expect_xml_element('test-generation-description',Xml,Content),!,
    expect_xml_element('output-file',Content,[pcdata(FileCodes)]),atom_codes(File,FileCodes),
    get_xml_events(Content,Events),
    expect_xml_element(parameters,Content,XmlParameters),
    % maximum depth in parameters section
    expect_xml_element('maximum-depth',XmlParameters,[pcdata(MaxDepthCodes)]),number_codes(MaxDepth,MaxDepthCodes),
    % optional target predicate
    ( memberchk(element(target,_,[pcdata(Tcodes)]),Content) -> atom_codes(Target,Tcodes)
    ; otherwise -> Target = ''),
    % set preference in parameters section
    findall(PL,member(element(preference,PL,_),XmlParameters),PrefAttLists),
    maplist(convert_xml_preference,PrefAttLists,Preferences),
    % check if state caching should be used
    ( memberchk(element('state-caching',_,_),Content) -> StateCaching = true
    ; otherwise -> StateCaching = false).
convert_xml_preference(PAL,pref(Name,Value)) :-
    ( memberchk(name=CName,PAL) -> atom_codes(Name,CName)
    ; otherwise -> add_error(sap,'Missing preference name'),fail),
    ( memberchk(value=CValue,PAL) -> atom_codes(Value,CValue)
    ; otherwise -> add_error(sap,'Missing preference value'),fail).
get_xml_events(Content,Events) :-
    memberchk(element('event-coverage',_,EventElements),Content),!,
    findall(E,(member(element(event,_,[pcdata(Ecodes)]),EventElements),atom_codes(E,Ecodes)),Events).
get_xml_events(_Content,Events) :-
    get_all_events(Events).


expect_xml_element(ElemName,Xml,Elements) :-
    ( memberchk(element(ElemName,_,Es),Xml) -> Elements=Es
    ; otherwise ->
        atom_concat('did not find xml element ',ElemName,Msg),
        add_error(sap,Msg),fail).

:- volatile counter/2.
:- dynamic counter/2.

reset_counters :-
    retractall(counter(_,_)).

count(Counter) :-
    ( retract(counter(Counter,Count)) -> true
    ; otherwise                       -> Count = 0),
    NCount is Count+1,
    assert(counter(Counter,NCount)).

count_timeout(Length) :- count(timeouts(Length)).
count_path(Length) :- count(path_count(Length)).
count_found(Length,Result) :-
    ( Result==testcase -> count(testcase_count(Length))
    ; otherwise        -> true),
    count(solution_count(Length)).

write_result_item(Desc,Db) :-
    format('~w: ',[Desc]),
    C =..[Db,Length],
    findall( count(Length,Num), counter(C,Num), UResults),
    sort(UResults,Results),
    maplist(write_result_item2,Results),
    maplist(extract_num,Results,Nums),sumlist(Nums,Total),
    format('total: ~w~n',[Total]).
write_result_item2(count(Length,Num)) :-
    format('~w:~w, ',[Length,Num]).
extract_num(count(_,Num),Num).

%total_cbc_testcases(T) :- get_totals(testcase_count,T).
%get_totals(CounterName,Total) :-
%    C =..[CounterName,_Len],
%    findall( Num, counter(C,Num), Ns),
%    sumlist(Ns,Total).

write_results(_) :- silent_mode(on),!.
write_results(Runtime) :-
    write_results,
    format('Runtime (wall): ~w ms~n',[Runtime]).
write_results :-
    write_result_item('Timeouts',timeouts),
    write_result_item('Checked paths (Depth:Paths)',path_count),
    write_result_item('Found paths',solution_count),
    write_result_item('Found test cases', testcase_count).

get_counter_info(CBC_Depth,Paths,Tests,Timeouts) :-
    get_counter(solution_count(CBC_Depth),Paths),
    get_counter(testcase_count(CBC_Depth),Tests),
    get_counter(timeouts(CBC_Depth),Timeouts).

get_counter(C,Num) :- counter(C,R) -> Num=R ; Num=0.

:- dynamic bounded_model_checking/0, bounded_model_checking_complete/0.
parse_cbc_target_predicate(_,_) :-
   retractall(bounded_model_checking),
   retractall(bounded_model_checking_complete),
   fail.
parse_cbc_target_predicate('#invariant',Invariant) :- !,
   b_get_invariant_from_machine(Invariant).
parse_cbc_target_predicate('#not_invariant',TargetPredicate) :- !,
   debug_println(19,bounded_model_checking),
   b_get_invariant_from_machine(Invariant),
   assert(bounded_model_checking), % for bounded model check we will try to use proven invariant for op !
   bsyntaxtree:create_negation(Invariant,TargetPredicate).
parse_cbc_target_predicate(TargetString,TargetPredicate) :- debug_println(15,parse_target_string(TargetString)),
   b_parse_optional_machine_predicate(TargetString,TargetPredicate).
