% (c) 2015-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(model_translation, [translate_smt_model_to_b_values/4]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smt_solvers).
:- module_info(description,'This module translates the models found by an SMT-solvers to a ProB state.').

:- use_module(library(lists), [maplist/3,maplist/4,
                               %convlist/3,delete/3,
                               is_list/1]).
:- use_module(library(avl)).

:- use_module(probsrc(translate), [translate_bvalue/2]).
:- use_module(probsrc(tools), [arg_is_number/2,
                               split_atom/3,
                               ajoin/2]).
:- use_module(probsrc(tools_strings), [convert_to_number/2]).
:- use_module(probsrc(error_manager), [add_error_fail/3]).
:- use_module(probsrc(bsyntaxtree), [safe_create_texpr/3,
                                     unique_typed_id/3,
                                     %conjunct_predicates/2,
                                     get_texpr_type/2,
                                     get_texpr_id/2,
                                     %create_implication/3,
                                     %create_negation/2,
                                     replace_id_by_expr/4]).
:- use_module(probsrc(custom_explicit_sets), [try_expand_custom_set_to_list/4]).
:- use_module(probsrc(b_global_sets), [b_type2_set/2]).

%:- use_module(probsrc(self_check), [assert_must_succeed/1]).

:- use_module(solver_dispatcher).

translate_smt_model_to_b_values(Solver,State,Identifiers,Bindings) :-
    maplist(translate_smt_model_to_b_value(Solver,State),Identifiers,Bindings).

translate_smt_model_to_b_value(Solver,State,b(identifier(Id),Type,_Infos),binding(Id,Value,PPValue)) :-
    member(id(Id,Expr),State),
    smt_solver_interface_call(Solver,get_string(Expr,String)), !,
    atom_codes(String,Codes),
    % format('Result of get_string: ~s~n',[Codes]),
    % trace,
    lisp_style_expression(E,Codes,[]), !,
    (Solver == cvc4 -> translate_cvc4_tree(E,Type,Value) ;
     Solver == z3   -> translate_z3_tree(Id,E,Type,Value) ;
     otherwise      -> fail),

    (translate_bvalue(Value,PPValue) -> true ; PPValue='**pretty-print failed**').

translate_z3_tree(Id,Id,Type,Result) :- !,
    % variable is not assigned (hence z3 returns its name)
    % just ground it somehow
    z3_ground_result(Type,Result).
translate_z3_tree(Id,EscapedId,Type,Result) :-
    % the id might have been escaped in z3 (e.g. due to ticks)
    ajoin(['|',Id,'|'],EscapedId), !,
    % variable is not assigned (hence z3 returns its name)
    % just ground it somehow
    z3_ground_result(Type,Result).
translate_z3_tree(_,E,Type,Result) :-
    translate_z3_tree(E,Type,Result).
translate_z3_tree(['-',Number],integer,int(X)) :- !,
    X is -Number.
translate_z3_tree(Number,integer,int(Number)).
translate_z3_tree(Number,global(Set),fd(ResNumber,Set)) :-
    integer(Number), !,
    ResNumber is Number + 1.
translate_z3_tree(Number,global(Set),fd(ResNumber,Set)) :-
    split_atom(Number,['!'],[Set,val,InternalNumber]),
    convert_to_number(InternalNumber,Nr),
    ResNumber is Nr + 1.
translate_z3_tree(A,boolean,Res) :-
    (A == true  -> Res = pred_true ;
     A == false -> Res = pred_false ;
     otherwise  -> fail).
translate_z3_tree([couple,A,B],couple(X,Y),(AT,BT)) :-
    translate_z3_tree(A,X,AT),
    translate_z3_tree(B,Y,BT).
translate_z3_tree(['_','as-array',AuxFunName],set(X),Res) :-
    % set is represented by aux function
    % get and parse the full model
    smt_solver_interface_call(z3,get_full_model_string(FullModel)),
    % format('z3 full model:~n~w~n',[FullModel]),
    z3_model_to_avl(FullModel,AVL), !,
    z3_set_to_b_set(AuxFunName,X,AVL,Res).
translate_z3_tree([store,Sub,Entry,true],set(X),ListOut) :-
    translate_z3_tree(Sub,set(X),SubEntries),
    translate_z3_tree(Entry,X,BEntry),
    sort([BEntry|SubEntries],ListOut).
translate_z3_tree([[as,const,['Array',_,'Bool']],false],set(_),[]).
translate_z3_tree([[as,const,['Array',_,'Bool']],true],set(Type),Res) :-
    b_type2_set(Type,Res).
translate_z3_tree(['record'|Z3Fields],record(Fields),rec(FieldsOut)) :-
    maplist(translate_z3_record,Fields,Z3Fields,FieldsOut).
translate_z3_tree([let,[[Id,Val]],Function],Type,Out) :-
    z3_replace_let_id_by_value(Id,Val,Function,NewFunction),
    translate_z3_tree(NewFunction,Type,Out).

translate_z3_record(field(Name,BType),CVCVal,field(Name,BVal)) :-
    translate_z3_tree(CVCVal,BType,BVal).

z3_ground_result(integer,int(0)).
z3_ground_result(boolean,pred_true).
z3_ground_result(couple(X,Y),(XG,YG)) :-
    z3_ground_result(X,XG), z3_ground_result(Y,YG).
z3_ground_result(set(_),[]).
z3_ground_result(record(Fields),rec(FieldsOut)) :-
    maplist(z3_ground_record_fields,Fields,FieldsOut).

z3_ground_record_fields(field(Name,BType),field(Name,BVal)) :-
    z3_ground_result(BType,BVal).

:- use_module(probsrc(b_interpreter),[b_compute_expression_nowf/4]).
z3_set_to_b_set(FunName,ResType,AVL,ResultOut) :-
    avl_fetch(FunName,AVL,f([[ParamName,_ParamType]],_ReturnType,Tree)),

    safe_create_texpr(identifier(ParamName),ResType,IdA),
    unique_typed_id("tmp",boolean,IdB),
    z3_function_to_b_pred(Tree,Pred,IdA,IdB,AVL),
    unique_typed_id("tmp",boolean,IDPredType),
    safe_create_texpr(equal(IDPredType,Pred),pred,IsEqual),
    safe_create_texpr(comprehension_set([IdA,IDPredType],IsEqual),set(couple(ResType,boolean)),TheFunction),

    safe_create_texpr(function(TheFunction,IdA),boolean,FCall),
    safe_create_texpr(equal(b(boolean_true,boolean,[]),FCall),pred,NewNewPred),
    safe_create_texpr(comprehension_set([IdA],NewNewPred),set(ResType),CompSet),
    % translate:set_unicode_mode,translate:print_bexpr(CompSet),
    b_compute_expression_nowf(CompSet,[],[],Res),
    (try_expand_custom_set_to_list(Res,ResultOut,true,z3_set_to_b_set)
     -> true
     ;  ResultOut = Res).
z3_set_to_b_set(A,B,_,_) :-
    add_error_fail(model_translation, 'z3_set_to_b_set failed on fun / result pair', [A,B]).

%z3_type_to_b_type('Int',integer).
%z3_type_to_b_type('Bool',boolean).

z3_function_to_b_pred([ite,[Op,_ParamName,Val],Result|Ifs],Pred,IdParam,IdReturn,AVL) :- !,
    z3_function_to_b_pred(Ifs,SubPred,IdParam,IdReturn,AVL),
    get_texpr_type(IdReturn,ReturnType), get_texpr_type(IdParam,ParamType),

    %translate_z3_tree(Result,ReturnType,Value),
    %safe_create_texpr(equal(IdReturn,b(value(Value),ReturnType,[])),pred,Then),
    z3_function_to_b_pred(Result,Then,IdParam,IdReturn,AVL),

    translate_z3_tree(Val,ParamType,Value2),
    (Op = '=' -> safe_create_texpr(equal(IdParam,b(value(Value2),ParamType,[])),pred,If) ;
     Op = '<=' -> safe_create_texpr(less_equal(IdParam,b(value(Value2),ParamType,[])),pred,If) ;
     Op = '>=' -> safe_create_texpr(greater_equal(IdParam,b(value(Value2),ParamType,[])),pred,If) ;
     otherwise -> fail),

    safe_create_texpr(if_then_else(If,Then,SubPred),ReturnType,Pred).
z3_function_to_b_pred([let,[[Id,Val]],Function],Pred,IdParam,IdReturn,AVL) :- !,
    z3_replace_let_id_by_value(Id,Val,Function,NewFunction),
    z3_function_to_b_pred(NewFunction,Pred,IdParam,IdReturn,AVL).
% unwraps lower levels
z3_function_to_b_pred([X],Pred,IdParam,IdReturn,AVL) :- !,
    z3_function_to_b_pred(X,Pred,IdParam,IdReturn,AVL).
% function composition
z3_function_to_b_pred([X,ParamName],Pred,IdParam,IdReturn,AVL) :- % TODO Id return
    get_texpr_id(IdParam,ParamName), !,
    avl_fetch(X,AVL,f([[ParamName,_ParamType]],_ReturnType,XBody)),
    z3_function_to_b_pred(XBody,Pred,IdParam,IdReturn,AVL).
z3_function_to_b_pred(['-',Int],Pred,_IdParam,_IdReturn,_AVL) :- !,
    translate_z3_tree(Int,integer,int(Value)),
    MValue is -Value,
    Pred = b(value(int(MValue)),integer,[]).
z3_function_to_b_pred([X|Further],Pred,IdParam,IdReturn,AVL) :- !,
    get_texpr_type(IdParam,SubParamType),
    unique_typed_id("tmp",SubParamType,IdB),
    z3_function_to_b_pred(Further,FurtherPred,IdParam,IdB,AVL),

    avl_fetch(X,AVL,f([[_XParamName,_ParamType]],_ReturnType,XBody)),
    z3_function_to_b_pred(XBody,XPred,IdB,IdReturn,AVL),

    get_texpr_id(IdB,IdToReplace),
    replace_id_by_expr(XPred,IdToReplace,FurtherPred,Pred).
z3_function_to_b_pred(SingleVal,Pred,_,IdReturn,_) :-
    get_texpr_type(IdReturn,ReturnType),
    translate_z3_tree(SingleVal,ReturnType,Value),
    Pred = b(value(Value),ReturnType,[]).

z3_replace_let_id_by_value(Id,Val,Function,NewFunction) :-
    is_list(Function), !,
    maplist(z3_replace_let_id_by_value(Id,Val),Function,NewFunction).
z3_replace_let_id_by_value(Id,Val,Function,NewFunction) :-
    atomic(Function), !,
    Id = Function -> NewFunction = Val ; NewFunction = Function.

z3_model_to_avl(FullModel,AVL) :-
    atom_codes(FullModel,Codes),
    lisp_style_expressions(E,Codes,[]),
    z3_lisp_to_avl(E,AVL).

z3_lisp_to_avl(E,AVL) :-
    empty_avl(In),
    z3_lisp_to_avl(E,In,AVL).

z3_lisp_to_avl([],A,A).
z3_lisp_to_avl([['define-fun',Name,Param,Return,Tree]|T],AIn,AOut) :-
    avl_store(Name,AIn,f(Param,Return,Tree),ATemp),
    z3_lisp_to_avl(T,ATemp,AOut).
% currently I think I do not need these facts...
% as far as I understand they represent enumerated set Elements
z3_lisp_to_avl([['declare-fun',_Name,_Params,_Type0]|T],AIn,AOut) :-
    z3_lisp_to_avl(T,AIn,AOut).
% these occur for cardinality constraints
z3_lisp_to_avl([['forall',_Vars,_Pred]|T],AIn,AOut) :-
    z3_lisp_to_avl(T,AIn,AOut).

translate_cvc4_tree(['CONST_RATIONAL',Arg],integer,int(Val)) :-
    arg_is_number(Arg,Val).
translate_cvc4_tree(['CONST_BOOLEAN',1],boolean,pred_true).
translate_cvc4_tree(['CONST_BOOLEAN',0],boolean,pred_false).
translate_cvc4_tree(['CONST_STRING',Content],string,string(Content)).
translate_cvc4_tree(['SINGLETON',ValIn],set(X),[ValOut]) :-
    translate_cvc4_tree(ValIn,X,ValOut).
translate_cvc4_tree(['EMPTYSET',emptyset,_Type],set(_),[]).
translate_cvc4_tree(['TUPLE',V1,V2],couple(Type1,Type2),(TV1,TV2)) :-
    translate_cvc4_tree(V1,Type1,TV1),
    translate_cvc4_tree(V2,Type2,TV2).
translate_cvc4_tree(['APPLY_CONSTRUCTOR','__cvc4_tuple_ctor',V1,V2],couple(Type1,Type2),(TV1,TV2)) :-
    translate_cvc4_tree(V1,Type1,TV1),
    translate_cvc4_tree(V2,Type2,TV2).
translate_cvc4_tree(['UNION',V1,V2],set(X),Val) :-
    translate_cvc4_tree(V1,set(X),TV1),
    translate_cvc4_tree(V2,set(X),TV2),
    append(TV1,TV2,Val).
translate_cvc4_tree(['UNINTERPRETED_CONSTANT',C],global(SId),fd(PBCId,SId)) :-
    atom_codes(C,CCodes),
    parse_uc(CId,CCodes,[]),
    PBCId is CId + 1.
translate_cvc4_tree(['RECORD',_CVCRecordType|Fields],record(RFields),rec(FieldsOut)) :-
    maplist(translate_cvc4_record,RFields,Fields,FieldsOut).

translate_cvc4_record(field(Name,BType),CVCVal,field(Name,BVal)) :-
    translate_cvc4_tree(CVCVal,BType,BVal).

parse_uc(Id) -->
    "uc__SORT_TYPE_var_", number(_),
    "__", number(N), {number_codes(Id,N)}.

lisp_style_expressions([E|Es]) -->
    ws, lisp_style_expression(E), ws, !,
    lisp_style_expressions(Es).
lisp_style_expressions([]) --> ws, [].
lisp_style_expressions(_,In,_) :-
    format("lisp-style parser failed on: ~s~n",[In]),fail.

ws --> [W], {W = 32 ; W = 10 ; W = 13}, ws.
ws --> ";;", any_but_newline.
ws --> [].
any_but_newline --> [C], {C \= 10}, any_but_newline ; "\n", ws.

lisp_style_expression(A) --> non_empty_symbol(Cs), {atom_codes(A, Cs)}.
lisp_style_expression(N) --> number(Cs), {number_codes(N, Cs)}.
lisp_style_expression(N) --> "-", number(Cs), {number_codes(N2,Cs), N is -N2}.
lisp_style_expression(A) --> string(Codes), {atom_codes(A,Codes)}.
lisp_style_expression(List) --> "(", lisp_style_expressions(List), ")".
lisp_style_expression(record_type(Fields)) -->
    "[#", ws, record_type_fields(Fields), ws, "#]".

record_type_fields([F|Fields]) -->
    record_type_field(F), ws, ",", ws, record_type_fields(Fields).
record_type_fields([F]) --> record_type_field(F).

record_type_field(field(NameA,TypeA)) -->
    non_empty_symbol(Name), ":(TYPE_CONSTANT ", non_empty_symbol(Type), " type)",
    {atom_codes(NameA,Name),atom_codes(TypeA,Type)}.

number([D|Ds]) --> digit(D), number(Ds).
number([D]) --> digit(D).

digit(D) --> [D], {48 =< D, D =< 57}.

string(Codes) --> "\"", non_empty_symbol(Codes), "\"".

non_empty_symbol([A|As]) -->
    [A],
    { 65 =< A, A =< 90
    ; 97 =< A, A =< 122
    % ; 48 =< A, A =< 57
    ; A = 124 % "|"
    ; A = 95 % "_"
    ; A = 45 % "-"
    ; A = 39 % "'"
    ; A = 33 % "!"
    ; 60 =< A, A =< 62 % "<", "=", ">"
    },
    symbol(As).

symbol([A|As]) -->
    [A],
    { 65 =< A, A =< 90
    ; 97 =< A, A =< 122
    ; 48 =< A, A =< 57
    ; A = 124 % "|"
    ; A = 95 % "_"
    ; A = 45 % "-"
    ; A = 39 % "'"
    ; A = 33 % "!"
    ; 60 =< A, A =< 62 % "<", "=", ">"
    },
    symbol(As).
symbol([]) --> "".
