% (c) 2014-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(smtlib2_parser,[parse_smtlib2/2]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smtlib).
:- module_info(description,'The SMT-LIB 2 Interpreter - Parser').

:- use_module(library(lists),[append/2]).
:- use_module(probsrc(debug),[debug_println/2]).

parse_smtlib2(Source, ListOfCommands) :-
  debug_println(5,parsing_smtlib2(Source)),
  reset_parser,
  commands(ListOfCommands,Source,[]),
  debug_println(5,finished_parsing).

% meta magic
% at_least_once consumes rule A at least once and returns a list
% multiple_times consumes rule A zero or more times

:- meta_predicate multiple_times(3,-,-,-).
:- meta_predicate at_least_once(3,-,-,-).
at_least_once(A,[Partial|List]) -->
  call(A,Partial), ws,
  !,
  multiple_times(A,List).
at_least_once(A,_) -->  err_message(A).

err_message(Expected) -->  {line_counter(N), format_with_colour(user_error,[red],"! Unrecognized ~w on line ~w.~n",[Expected,N]),fail}.
err_message(_) --> any_chars_but_newline(S), !, {format_with_colour(user_error,[red],"! '~s'~n",[S]),fail}.

multiple_times(A,[Partial|List]) -->
  call(A,Partial), ws,
  multiple_times(A,List).
multiple_times(_A,[],In,In).

:- meta_predicate multiple_times_no_whitespace(3,-,-,-).
:- meta_predicate at_least_once_no_whitespace(3,-,-,-).
at_least_once_no_whitespace(A,[Partial|List]) -->
  call(A,Partial),
  multiple_times_no_whitespace(A,List).

multiple_times_no_whitespace(A,List) -->
  at_least_once_no_whitespace(A,List).
multiple_times_no_whitespace(_A,[],In,In).

:- use_module(probsrc(tools_printing),[format_with_colour/4]).

% grammar rules
commands(Cmds) --> real_ws, !, commands(Cmds).
commands([Command|Commands]) -->
         command(Command), {debug_println(5,command(Command))}, !,
         commands(Commands).
commands(Commands) --> ";", any_chars_but_newline(_), newline, !, commands(Commands).
commands([]) --> "".
commands(_) --> err_message(command).

reset_parser :- retractall(line_counter(_)), assert(line_counter(1)).
:- dynamic line_counter/1.
line_counter(1).
inc_line_counter :-
   retract(line_counter(N)), debug_println(5,processed_line(N)),
   N1 is N+1, assert(line_counter(N1)).
newline --> "\n", {inc_line_counter}.

command(set_logic(Logic)) --> "(set-logic", ws, symbol(Logic), ")".
command(set_option(Option)) --> "(set-option", ws, attribute(Option), ")".
command(set_info(Info)) --> "(set-info", ws, attribute(Info), ")".
command(set_info(Info)) --> "(meta-info", ws, attribute(Info), ")".
command(reset) --> "(reset)".
command(declare_sort(Name,Arrity)) --> "(declare-sort", ws, !,
        symbol(Name), ws, numeral(Arrity), ")".
command(define_sort(Name,Parameters,Sort)) --> "(define-sort", ws, !,
        symbol(Name), ws, "(", multiple_times(symbol,(Parameters)), ")", ws,
        !,
        must_be_smt_sort(Sort), ")".
command(declare_fun(Name,Parameters,Result)) --> "(declare-fun", ws, !,
       must_be_symbol(Name), ws,
       "(", multiple_times(smt_sort,Parameters), ws, ")", ws,
       must_be_smt_sort(Result), ws, ")".
command(define_fun(Name,Parameters,Result,Term)) --> "(define-fun", ws, !,
       symbol(Name), ws, "(", multiple_times(sorted_var,Parameters), ")", ws,
       !,
       must_be_smt_sort(Result), ws, must_be_term(Term), ")".
command(push(Num)) --> "(push", ws, numeral(Num), ")".
command(pop(Num)) --> "(pop", ws, numeral(Num), ")".
command(assert(Term)) --> "(assert", ws, !,
        must_be_term(Term), ws, ")".
command(check_sat) --> "(check-sat", ")".
command(get_assert) --> "(get-assert", ")".
command(get_proof) --> "(get-proof", ")".
command(get_model) --> "(get-model", ")".
command(get_unsat_core) --> "(get-unsat-core", ")".
command(get_value(Terms)) --> "(get-value", ws, !, "(", at_least_once(term,Terms), ")", ")".
command(get_assign) --> "(get-assign", ")".
command(get_option(Name)) --> "(get-option", ws, !, keyword(Name), ")".
command(get_info(Name)) --> "(get-info", ws, !, keyword(Name), ")".
command(exit) --> "(exit", ")".
command(exit) --> ":x". % for ProB REPL compatibility

%command(_,In,_) :- !, atom_codes(A,In), format('Parsing failed on: ~w~n',[A]), !, fail.

symbol(Symbol) --> (simplesymbol(Symbol) -> [] ; asciicomment(Symbol)).
simplesymbol(Symbol) --> single_symbol(Start), symbols_or_digits(Rest), {atom_codes(Symbol,[Start|Rest])}.
single_symbol(S) --> [S], {member(S,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+-/*=%?!.$_~&^<>@")}.
single_symbol_or_digit(S) --> single_symbol(S) ; digit(S).
symbols_or_digits([H|T]) --> single_symbol_or_digit(H), symbols_or_digits(T).
symbols_or_digits([]) --> "".

asciicomment(string(Symbol)) --> "|", any_chars_but_pipe(Codes), "|", {atom_codes(Symbol,Codes)}.

any_chars_but_pipe([C|T]) --> [C], {C \= 124}, any_chars_but_pipe(T).
any_chars_but_pipe([]) --> "".

any_chars_but_newline([C|T]) --> [C], {C \= 10}, any_chars_but_newline(T).
any_chars_but_newline([]) --> "".

attribute(attribute(A,Value)) --> keyword(A), ws, attribute_value(Value).
attribute(attribute(A)) --> keyword(A).

attribute_value(Value) --> symbol(Value).
attribute_value(Value) --> spec_constant(Value).
attribute_value(Value) --> "(", at_least_once(sexpr,Value), ")".

numeral(0) --> "0".
numeral(N) --> digit_no_zero(D1), multiple_times_no_whitespace(digit,D), {number_codes(N,[D1|D])}.

decimal_numeral(D) --> numeral(D1), ".", numeral(D2),
                       {number_codes(D1,DC1), number_codes(D2,DC2),
                        append([DC1,".",DC2],Codes),
                        number_codes(D,Codes)}.

digit(D) --> [D], {[D]="0"} ; digit_no_zero(D).
digit_no_zero(D) --> [D], {member(D,"123456789")}.

smt_sort(sort(S)) --> identifier(S).
smt_sort(sort(S)) --> numeral(S). % not yet supported in Translation, what does it mean?
smt_sort(sort('BitVec',[Length])) --> "(_ BitVec ", numeral(Length), ")".
smt_sort(sort(S,Params)) --> "(", identifier(S), ws, !,
                              at_least_once(smt_sort,Params), ws, ")".
% like smt_sort, but throw error message:
must_be_smt_sort(T) --> smt_sort(T),!.
must_be_smt_sort(_) --> err_message(sort).

sorted_var(svar(Name,Sort)) --> "(", symbol(Name), ws, smt_sort(Sort), ")".

term(let(Bindings,Term)) -->
  "(let", ws, !, "(", at_least_once(binding,Bindings), ")", ws, !, must_be_term(Term),  !, ws, ")".
term(forall(Vars,Term)) -->
   "(forall", ws, !, "(", at_least_once(sorted_var,Vars), ")", ws, !, must_be_term(Term), !, ws, ")".
term(exists(Vars,Term)) -->
   "(exists", ws, !, "(", at_least_once(sorted_var,Vars), ")", ws, !, must_be_term(Term), !,  ws, ")".
term(T) --> spec_constant(T),!.
term(T) --> qual_identifier(T),!.
term(id_term(Id,Terms)) --> "(", qual_identifier(Id), ws, !, at_least_once(term,Terms), ws, ")".
term(attributed_term(Term,Attributes)) --> "(", "!", ws, !,
     must_be_term(Term), ws, !,
     at_least_once(attribute,Attributes), ws, ")".

% term which raises error if it cannot be found
must_be_term(T) --> term(T),!.
must_be_term(_) --> err_message(term).

binding(bind(Symbol,Term)) --> "(", must_be_symbol(Symbol), ws, !, must_be_term(Term), ws, ")".
% binding which raises error if it cannot be found
must_be_symbol(T) --> symbol(T),!.
must_be_symbol(_) --> err_message(symbol).

qual_identifier(id(Id)) --> identifier(Id).
qual_identifier(id(Id,Sort)) --> "(as ", ws, !, identifier(Id), ws, smt_sort(Sort), ")".
qual_identifier(id(sign_extend(Parameter))) --> "(_ sign_extend ", !, numeral(Parameter), ")".
qual_identifier(id(zero_extend(Parameter))) --> "(_ zero_extend ", !, numeral(Parameter), ")".
qual_identifier(id(extract(P1,P2))) --> "(_ extract ", !, numeral(P1), ws, numeral(P2), ")".


identifier(Id) --> symbol(Id).
%identifier(id_num(Id,Num)) --> "(", "_", symbol(Id), numeral(Num), ")".

keyword(KW) --> ":", symbols_or_digits(L), {atom_codes(KW,L)}.

sexpr(E) --> spec_constant(E).
sexpr(E) --> symbol(E).
sexpr(E) --> keyword(E).
sexpr(E) --> "(", multiple_times(sexpr,E), ")".

spec_constant(integer(I)) --> numeral(I).
spec_constant(decimal(D)) --> decimal_numeral(D).
spec_constant(string(S)) --> "\"", any_chars_but_quotations(SCodes), "\"", {atom_codes(S,SCodes)}.
spec_constant(bitvector(Val,Length)) --> "(_ bv", numeral(Val), " ", numeral(Length), ")".
spec_constant(bitvector(Binary)) --> "#b", binary_numbers(Binary).

binary_numbers([H|T]) --> binary_number(H), binary_numbers(T).
binary_numbers([N]) --> binary_number(N).
binary_number(0) --> "0".
binary_number(1) --> "1".

any_chars_but_quotations([C|T]) --> [C], {C \= 34}, any_chars_but_quotations(T).
% two quotations are allowed in smtlib 2.5
any_chars_but_quotations([C,C|T]) --> [C,C], {C = 34}, any_chars_but_quotations(T).
any_chars_but_quotations([]) --> "".

% whitespaces
ws --> " ", !, ws.
ws --> newline, !, ws.
ws --> "".

real_ws --> " ", !, ws.
real_ws --> newline, ws.
