% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(state_custom_dot_graph,[tcltk_generate_state_custom_dot_graph/1, state_custom_dot_graph_available/0]).

:- use_module(module_information).
:- module_info(group,dot).
:- module_info(description,'This module provides a way to generate custom state graphs (using info from DEFINITIONS).').

:- use_module(dot_graph_generator).

%:- use_module(self_check).
:- use_module(error_manager).

:- use_module(state_space,[current_expression/2]).
:- use_module(bsyntaxtree, [get_texpr_id/2]).
:- use_module(specfile, [state_corresponds_to_fully_setup_b_machine/2]).
:- use_module(debug, [debug_println/2]).
:- use_module(bmachine, [b_get_machine_custom_nodes_function/2,b_get_machine_custom_edges_function/2]).

state_custom_dot_graph_available :-
   b_get_machine_custom_nodes_function(_,_),
   b_get_machine_custom_edges_function(_,_).

tcltk_generate_state_custom_dot_graph(File) :-
       current_expression(_CurID,State),
       state_corresponds_to_fully_setup_b_machine(State,BState),!,
       (b_get_machine_custom_nodes_function(_,_)
        -> true
        ; add_error_and_fail(tcltk_generate_state_custom_dot_graph,'No CUSTOM_GRAPH_NODES Function in DEFINITIONS')),
       (b_get_machine_custom_edges_function(_,_)
        -> true
        ; add_error_and_fail(tcltk_generate_state_custom_dot_graph,'No CUSTOM_GRAPH_EDGES Function in DEFINITIONS')),
       retractall(custom_nodes(_,_,_)),
       retractall(custom_edges(_,_,_)),
       retractall(custom_node_id(_,_)), retractall(nodectr(_)),
       retractall(custom_trans_label(_,_)), retractall(trans_ctr(_)), assert(trans_ctr(1)),
       eval_defs(BState),
       gen_dot_graph(File,state_custom_dot_graph,node_predicate,trans_predicate,none,none),
       print(finished),nl.
tcltk_generate_state_custom_dot_graph(_) :-
       add_error_and_fail(tcltk_generate_state_custom_dot_graph,'Only possible for initialised B machine.').


eval_defs(BState) :-
       b_get_machine_custom_nodes_function(NodesFunction,Nr),
       eval_definition_fun(NodesFunction, BState, NodesRes),
       debug_println(19,custom_nodes(Nr,NodesRes)),
       (get_texpr_id(NodesFunction,EFID)
         -> assert(custom_nodes(EFID,Nr,NodesRes))
          ; number_codes(Nr,NC), append("custom",NC,CC), atom_codes(Custom,CC),
            assert(custom_nodes(Custom,Nr,NodesRes))),
       fail.
eval_defs(BState) :-
       b_get_machine_custom_edges_function(EdgesFunction,Nr),
       eval_definition_fun(EdgesFunction, BState, EdgesRes),
       debug_println(19,custom_edges(Nr,EdgesRes)),
       (get_texpr_id(EdgesFunction,Label) -> true
        ; b_get_machine_custom_edges_function(_,Nr2), Nr2 \= Nr % other relation exists; makes sense to use label
         -> number_codes(Nr,NC), append("custom",NC,CC),
            atom_codes(Label,CC)
        ; otherwise -> Label=''
       ),
       assert(custom_edges(Label,Nr,EdgesRes)),
       fail.
eval_defs(_).

:- dynamic custom_nodes/3, custom_edges/3.


:- use_module(tools,[string_escape/2]).
:- public node_predicate/6.
node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color) :- SubGraph=none,
     custom_nodes(_ID,_Nr,Nodes),
     (var(Nodes) -> add_error(node_predicate,'Variable custom_nodes: ',Nodes),fail ; true),
     get_node_and_colour(Nodes,NodeVal,ColVal,Shape,Style),
     (translate_value_and_escape(NodeVal,NodeDesc) -> true ; NodeDesc = '???'),
     translate_bvalue_to_colour(ColVal,Color),
     %tools:print_message(node(NodeVal,ColVal,Color,NodeID)),
     gen_id(NodeVal,NodeID). %tools:print_message(nodeid(NodeID)).

get_node_and_colour(CustomNodes,NodeVal,ColVal,Shape,Style) :-
   member(Node,CustomNodes),
   deconstruct_node(Node,NodeVal,ColVal,Shape,Style).

deconstruct_node(((NodeVal,ShapeVal),ColVal),ResNodeVal,ResColVal,Shape,Style) :-
   valid_dot_shape(ShapeVal,Shape),
   !, % we have a triple (Node,"box","colour") = ((Node,"box"),"colour")
   ResNodeVal=NodeVal,ResColVal=ColVal, Style=filled.
deconstruct_node((NodeVal,ColVal),ResNodeVal,ResColVal,Shape,Style) :- !, % we have a pair (Node,"colour")
   ResNodeVal=NodeVal, ResColVal=ColVal, Shape=box, Style=filled.
deconstruct_node(NodeVal,NodeVal,NodeVal,box,filled).

:- use_module(preferences, [is_of_type/2]).
valid_dot_shape(string(Str),Str) :- is_of_type(Str,dot_shape).
   % triangle,ellipse,box,diamond,hexagon,octagon,house,invtriangle,invhouse,invtrapez,doubleoctagon,egg,parallelogram,pentagon,trapezium...

:- dynamic custom_node_id/2, nodectr/1.
nodectr(0).

get_ctr(Res) :- retract(nodectr(C)),!, C1 is C+1, assert(nodectr(C1)), Res=C.
get_ctr(0) :- assert(nodectr(1)).

% generate or lookup ID for node
gen_id(NodeVal,ID) :-
 (custom_node_id(NodeVal,ID) -> true
   ; get_ctr(C),
     assert(custom_node_id(NodeVal,C)),ID=C).

% lookup ID and generate message if it does not exist
lookup_id(NodeVal,ID) :-
 (custom_node_id(NodeVal,ID) -> true
   ; format(user_error,'Node not specified in CUSTOM_GRAPH_NODES: ~w~n',[NodeVal]),
     gen_id(NodeVal,ID)
 ).

:- dynamic custom_trans_label/2, trans_ctr/1.
trans_ctr(1).
gen_trans_color(Label,Col,IsColor) :- custom_trans_label(Label,C),!,Col=C,IsColor=false.
gen_trans_color(Label,Col,IsColor) :- try_translate_bvalue_to_colour(Label,C),!,Col=C,IsColor=true.
gen_trans_color(Label,Col,IsColor) :-  retract(trans_ctr(Ctr)), C1 is Ctr+1, assert(trans_ctr(C1)),
   translate_bvalue_to_colour(int(Ctr),C),
   assert(custom_trans_label(Label,C)), Col=C,IsColor=false.


:- public trans_predicate/5.
trans_predicate(NodeID,Label,SuccID,Color,Style) :-
   custom_edges(DefaultLabel,_,Edges),
   (var(Edges) -> add_error(trans_predicate,'Variable custom_edges: ',Edges),fail ; true),
   member((From,To),Edges),
   trans_pred_aux(From,DefaultLabel,To,NodeID,SuccID,Label,Color,Style).

% we have a transition where the colour is specified (e.g., CUSTOM_GRAPH_EDGES == {n1,col,n2 | ... }
% we assume trans_predicate has been called first
trans_pred_aux(FromValue,_DefaultLabel,To,NodeID,SuccID,ELabel,Color,solid) :-
   get_trans_label_and_color(FromValue,From,Color,ELabel), % the From Value contains color and label
   !,
   lookup_id(From,NodeID),
   lookup_id(To,SuccID).
trans_pred_aux(From,_DefaultLabel,(LabelVal,ToVal),NodeID,SuccID,ELabel,Color,solid) :-
   get_trans_label_and_color((ToVal,LabelVal),To,Color,ELabel),  % the To Value contains color and label
   !,
   lookup_id(From,NodeID),
   lookup_id(To,SuccID).
trans_pred_aux(From,DefaultLabel,To,NodeID,SuccID,Label,blue,solid) :- Label=DefaultLabel,
   % no label or color specified in transition pair
   lookup_id(From,NodeID),
   lookup_id(To,SuccID).

% try and decompose a from value, detecting color and label string  ((From,"label"),"color")
get_trans_label_and_color((FromVal,LabelVal),From,Color,ELabel) :-
    gen_trans_color(LabelVal,Color,IsColor),
    (IsColor
     -> get_trans_label(FromVal,From,ELabel)
     ; translate_value_and_escape(LabelVal,ELabel),
       From=FromVal
    ).

get_trans_label((From,string(LabelVal)),From,ELabel) :- string_escape(LabelVal,ELabel).
get_trans_label(From,From,'').  % or should we use DefaultLabel

translate_value_and_escape(string(S),ELabel) :- !, string_escape(S,ELabel).
translate_value_and_escape(LabelVal,ELabel) :-
    translate:translate_bvalue(LabelVal,Label),string_escape(Label,ELabel).

:- use_module(b_interpreter,[b_compute_expression_nowf/4]).
:- use_module(custom_explicit_sets,[try_expand_custom_set/2]).

eval_definition_fun(AnimFunction,BState, FunctionRes) :- %trace,
       b_compute_expression_nowf(AnimFunction,[],BState,FunctionResCl),
       %nl, print('FunctionResult'(FunctionResCl)),nl,
       try_expand_custom_set(FunctionResCl, FunctionRes).
