% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(store, [ %store_values_for_existing_ids/4,
                  store_value_for_existing_id/4,
                  variable_is_defined/2,
                  store_value/4,
                  store_updates_and_normalise/3,
                  store_updates/3,
                  store_intermediate_updates/4,
                  save_updates_in_existing_store/2,
                  %lazy_store_updates_and_normalise/3,
                  merge_updates/3,
                  get_id_value/3,
                  lookup_value/3,
                  lookup_value/4,lookup_value_with_span/5,
                  lookup_value_without_check/4,
                  lookup_value_for_existing_id/3, lookup_value_for_existing_id/4,
                  lookup_and_delete_value/7,
                  no_value_for_variable/2,
                  empty_state/1,
                  add_var_to_localstate/4,
                  set_up_localstate/4,
                  set_up_undefined_localstate/3, set_up_undefined_localstate_with_predefined/4,
                  delete_variables_from_state/3,

                  normalise_value_for_var/3, normalise_value_for_var/4,
                  normalise_value/2,
                  l_normalise_values/2,
                  l_expand_and_normalise_values/2, l_expand_and_normalise_values/3,
                  normalise_store/2,
                  copy_variable_store/4,
                  check_valid_store/2
                  ]).

:- use_module(library(lists)).
:- use_module(self_check).
:- use_module(debug).

:- use_module(typechecker).
:- use_module(preferences).
:- use_module(tools).

:- use_module(bsyntaxtree).

:- use_module(error_manager).

:- use_module(bmachine, [b_is_constant/1]).

:- type variable_id +--> (lambda_res(integer) ; atomic).
:- type binding +--> bind(variable_id,any).
:- type store +--> list(binding).

:- use_module(tools).
:- use_module(module_information).
:- module_info(group,interpreter).
:- module_info(description,'This module takes care of storing B values in an environment and normalising values.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

/* --------------------------------------------------------- */

/* store and lookup of values */



variable_is_defined(Var,Store) :- safe_member(bind(Var,_),Store,variable_is_defined).

store_value(Var,Value,Old,New) :- store_value2(Old,Var,Value,New).
store_value2([],Var,Value,[bind(Var,Value)]).
store_value2([bind(V,VV)|T],Var,Value,[bind(V,NValue)|UT]) :-
  ( Var=V ->
      NValue=Value, UT=T
  ; otherwise ->
      NValue=VV, store_value2(T,Var,Value,UT)).


store_value_for_existing_id(L,ID,V,L2) :- var(ID),!,
    add_internal_error('VarID is var !!',store_value_for_existing_id(ID,V,L,L2)),fail.
store_value_for_existing_id(L,ID,V,L2) :- store_value_for_existing_id_aux(L,ID,V,L2).
store_value_for_existing_id_aux([],Var,Value,[bind(Var,Value)]) :-
    add_error(store_value_for_existing_id,'Variable does not exist in store: ',Var).
store_value_for_existing_id_aux([bind(V,VV)|T],Var,Value,[bind(V,VV2)|UT]) :-
    %we could use: arg(1,BIND,V), arg(2,BIND,VV),
    (Var=V -> VV2=Value, UT=T
     ;  VV2=VV,store_value_for_existing_id_aux(T,Var,Value,UT)).





/* store_intermediate_updates(InState,Updates,OutStates) */
/* rewrote it so that it copies state skeleton even if updates are not yet known */

:- assert_must_succeed((store:store_intermediate_updates([bind(y,1),bind(x,2),bind(a,22)],[bind(x,33)],R,unit_test1),R==[bind(y,1),bind(x,33),bind(a,22)])).
:- assert_must_succeed((store:store_intermediate_updates([bind(Y,1),bind(X,2),bind(A,22)],[bind(x,33)],R,unit_test2),X=x,A=a,Y=y,R==[bind(y,1),bind(x,33),bind(a,22)])).



store_intermediate_updates(InStore,Update,OutStore,_PP) :- Update==[], !,
  OutStore=InStore.
store_intermediate_updates(InStore,Update,OutStore,PP) :-
   %print(store_intermediate_updates(InStore,Update,OutStore,PP)),nl,
   store_intermediate_updates2(InStore,Update,OutStore,ctxt(PP,InStore,Update)).

:- block store_intermediate_updates2(-,?,?,?), store_intermediate_updates2(?,-,?,?).
store_intermediate_updates2([],Updates,[],PP) :-
    store_intermediate_updates3(Updates,PP).
store_intermediate_updates2([bind(Var,OldVal)|InT],Updates,[bind(Var,NewVal)|OutT],PP) :-
   %we could do: arg(1,BIND,Var), arg(2,BIND,OldVal),
   % TO DO: maybe it would be better to copy Update variables as they come in rather than forcing the order as in the store
   search_and_delete_value(Var,OldVal,NewVal,Updates,NUpdates),
   (NUpdates == [] -> OutT=InT
   ;  store_intermediate_updates2(InT,NUpdates,OutT,PP)).

:- use_module(library(ordsets),[ord_intersection/3]).
:- block store_intermediate_updates3(-,?).
store_intermediate_updates3(Updates,PP) :-
   ( Updates = [] -> true
   %; Updates = [bind(abort__flag__info,_)] -> true
   ;  PP = ctxt(SpanInfo,InStore,FullUpdate),
      % TO DO: determine if illegal assignments, which do not occur in InStore
      get_ids_from_store(InStore,_,LegalIds),
      get_ids_from_store(FullUpdate,WrIDs,_WrittenIds),
      get_ids_from_store(Updates,_,IllegalIds),
      ord_intersection(IllegalIds,LegalIds,DoubleIds),
      (DoubleIds \= [] % at least some of the assignments are legal, hence must be double
       ->   add_error(store_intermediate_updates,'Double assignments:',
                  DoubleIds:(illegal_updates(Updates),written_ids(WrIDs)),SpanInfo)
        ;   add_error(store_intermediate_updates,'Illegal assignments:',
                  IllegalIds:(illegal_updates(Updates),legal_ids(LegalIds),written_ids(WrIDs)),SpanInfo)
      )
   ).

get_ids_from_store(Store,Ids,SIds) :- findall(Id,member(bind(Id,_),Store),Ids), sort(Ids,SIds).

:- assert_pre(store:store_updates_and_normalise(A,B,_C),(type_check(A,store),type_check(B,store))).
:- assert_post(store:store_updates_and_normalise(_A,_B,C),(type_check(C,store))).

% TO DO: optimize; probably better to traverse InState once, rather than for each update binding
%:- block store_updates_and_normalise(-,?,?). % block normally not necessary
store_updates_and_normalise(Updates,InState,OutState) :- var(Updates), !,
   add_internal_error('Variable update binding list:',store_updates_and_normalise(Updates,InState,OutState)),
   when(ground(Updates), store_updates_and_normalise(Updates,InState,OutState)).
store_updates_and_normalise(Updates,InState,OutState) :- %print(store_updates_and_normalise),nl,
   preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,EXPAND),
   sort(Updates,SUpdates), % only necessary for checking for multiple variables; we could remove this call
   store_updates_and_normalise1(SUpdates,EXPAND,'$$NONE$$',InState,OutState).

store_updates_and_normalise1([],_EXPAND,_Last,InState,InState).
store_updates_and_normalise1([bind(Var,NewVal)|RestUpdates],EXPAND,LastVar,InState,OutState) :-
   store_updates_and_normalise2(Var,EXPAND,LastVar,NewVal,RestUpdates,InState,OutState).
%:- block store_updates_and_normalise2(-,?,?,?,?).
store_updates_and_normalise2(Var,EXPAND,LastVar,NewVal,RestUpdates,InState,OutState) :-
  var(Var),!,
  add_internal_error('Prolog variable for Var: ',store_updates_and_normalise2(Var,EXPAND,LastVar,NewVal,RestUpdates,InState,OutState)),
  when(nonvar(Var),
        store_updates_and_normalise2(Var,EXPAND,LastVar,NewVal,RestUpdates,InState,OutState)).
store_updates_and_normalise2(Var,EXPAND,LastVar,NewVal,RestUpdates,InState,OutState) :-
   normalise_value_for_var(Var,EXPAND,NewVal,NormalisedNewVal),
   %NormalisedNewVal=NewVal, % comment in to avoid normalising values
   store_value_for_existing_id(InState,Var,NormalisedNewVal,InState2),
   % following is in principle no longer necessary:
   (Var==LastVar ->
      add_error(store_updates_and_normalise,'Variable has been assigned twice in parallel: ',Var:NewVal)
    ; true % no longer necessary; bmachine checks this: (b_is_constant(Var) -> add_error(store_updates_and_normalise,'Assigned value to constant: ',Var) ; true)
   ),
   store_updates_and_normalise1(RestUpdates,EXPAND,Var,InState2,OutState).


% store already normalised updates, e.g., for operation_reuse
% does not check for duplicate assignments in parallel
store_updates([],S,S).
store_updates([bind(Var,NewVal)|RestUpdates],InState,OutState) :-
   store_value_for_existing_id(InState,Var,NewVal,InState2),
   store_updates(RestUpdates,InState2,OutState).



/* use to merge updates, e.g. for sequential composition LHS ; RHS */
:- assert_pre(store:merge_updates(_A,_B,_C),true). % (type_check(_A,store),type_check(_B,store))).
:- assert_post(store:merge_updates(_A,_B,_C),true). %(type_check(_C,store))).

:- assert_must_succeed((store:merge_updates([bind(x,1)],[bind(x,2)],R),R==[bind(x,2)])).
:- assert_must_succeed((store:merge_updates([bind(x,1)],[bind(X,2),bind(y,3)],R),X=x,R==[bind(x,2),bind(y,3)])).
:- assert_must_succeed((store:merge_updates([bind(x,1)],[bind(X,2),bind(y,3)],R),X=z,R==[bind(z,2),bind(y,3),bind(x,1)])).
:- assert_must_succeed((store:merge_updates([bind(y,1),bind(x,1)],[bind(x,2)],R),R==[bind(x,2),bind(y,1)])).

:- block merge_updates(-,?,?).
merge_updates([],RHSUpdates,RHSUpdates).
merge_updates([Bind|LHT],RHSUpdates,Rest) :-
    merge_updates2( Bind,LHT,RHSUpdates,Rest).
:- block merge_updates2(?,?,-,?).
merge_updates2(bind(Var,NewVal),LHT,RHSUpdates,Rest) :-
    merge_remove(RHSUpdates,Var,NewVal,RHSUpdates2),
    merge_updates(LHT,RHSUpdates2,Rest).

:- block merge_remove(-,?,?,?).
merge_remove([],Var,NewVal,[bind(Var,NewVal)]).
merge_remove([bind(Var2,Val2)|T2],Var1,Val1,[bind(Var2,Val2)|NewT2]) :-
    merge_remove_aux(Var2,T2,Var1,Val1,NewT2).

:- block merge_remove_aux(-,?,?,?,?).
merge_remove_aux(Var2,T2,Var1,Val1,NewT2) :-
    (Var1=Var2
        -> NewT2=T2  /* binding in second updates overrides first update */
         ; merge_remove(T2,Var1,Val1,NewT2)
      ).
/* ----------------------- */

:- block save_updates_in_existing_store(-,?).
save_updates_in_existing_store([],_Store).
save_updates_in_existing_store([bind(Var,Value)|Urest],Store) :-
    save_updates_in_existing_store2(Var,Value,Store),
    save_updates_in_existing_store(Urest,Store).
:- block save_updates_in_existing_store2(-,?,?).
save_updates_in_existing_store2(Var,Value,Store) :-
    lookup_value_for_existing_id(Var,Store,Value).


:- assert_must_succeed(store:l_normalise_values([int(1)],[int(1)])).

l_normalise_values([],[]).
l_normalise_values([H|T],[NH|NT]) :- normalise_value(H,NH), l_normalise_values(T,NT).

:- block l_expand_and_normalise_values(-,?).
l_expand_and_normalise_values([],[]).
l_expand_and_normalise_values([H|T],[NH|NT]) :-
    normalise_value_for_var(unkown,true,H,NH),
    l_expand_and_normalise_values(T,NT).
:- block l_expand_and_normalise_values(-,?,?).
l_expand_and_normalise_values([],[],_).
l_expand_and_normalise_values([H|T],[NH|NT],[VarName|VT]) :-
    %% print(normalising(VarName)),nl, %%
    normalise_value_for_var(VarName,true,H,NH),
    l_expand_and_normalise_values(T,NT,VT).



:- assert_must_succeed(store:normalise_value(int(1),int(1))).
:- assert_must_succeed((store:normalise_value([int(1)],R), R==avl_set(node(int(1),true,0,empty,empty)))).
:- assert_must_succeed((store:normalise_value([int(1),int(2)],R1),
                        store:normalise_value([int(2),int(1)],R2),R1==R2)).
%:- assert_must_succeed((store:normalise_value(X,[int(1),int(2)]),X=[int(2),int(1)])).
:- assert_must_succeed((store:normalise_value(X,_R),X=[Y,Z],
             Y=[int(2),int(1)],Z=[int(1)])).
:- assert_must_succeed(store:normalise_value(_X,[int(1),int(3)])).
:- assert_must_succeed((store:normalise_value(X,[int(3),int(1)]),
                        kernel_objects:equal_object(X,[int(1),int(3)]))).
:- assert_must_fail(store:normalise_value(int(2),int(1))).


normalise_value_for_var(Var,Val,NormVal) :-
    preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,EXPAND),
    normalise_value_for_var(Var,EXPAND,Val,NormVal).

:- block normalise_value_for_var(?,?,-,-).
normalise_value_for_var(Var,EXPAND,Val,NormVal) :-  % print(norm(Var,EXPAND,Val,NormVal)),nl, trace,
   expand_closure_value_for_normalisation(Val,EXPAND,Var,ENewVal), % also normalises values inside closure if not expanded
   normalise_value(ENewVal,NormVal). %  , tools_printing:print_term_summary(normalised(ENewVal,NormVal)).

:- block normalise_value(-,-).
normalise_value(V,Res) :-
    (nonvar(V),var(Res) -> ~~mnf(normalise2(V,ResR)),Res=ResR ; normalise3(V,Res)).
    % (nonvar(V),var(Res) -> normalise2(V,Res) ; normalise3(V,Res)).



:- use_module(b_global_sets,[b_integer_or_string_set/1,all_elements_of_type/2]).
:- use_module(library(avl)).

%:- use_module(debug,[time/1]).

% a normalise which does not need to call equal_object on the result argument:
normalise2([H|T],Res) :- !,
   normalise_list([H|T],Res).
normalise2((A,B),(NA,NB)) :- !, normalise_value(A,NA), normalise_value(B,NB).
normalise2(global_set(GS),N) :- !,
   (b_integer_or_string_set(GS)
     -> N = global_set(GS) /* can be very large or infinite, so don't expand */
     ; all_elements_of_type(GS,Res),
       normalise2(Res,N)).
normalise2(avl_set(A),R) :- !,
  (%empty_avl(A)
    A=empty
     -> R=[] ;
   %A = node(_,_,_,empty,_) -> R = avl_set(A) ;  % avoid normalising certain one and two element sets; does not seem to bring any improvement
   avl_to_list(A,L), ord_list_to_avl(L,B), R=avl_set(B)).
normalise2(string(S),R) :- !, R=string(S).
normalise2(int(S),R) :- !, R=int(S).
normalise2(fd(N,T),R) :- !, R=fd(N,T).
normalise2(rec(Fields),R) :- !, R=rec(NFields), normalise_fields(Fields,NFields).
normalise2(V,V).

:- block normalise_fields(-,-).
normalise_fields([],[]).
normalise_fields([field(Name,Val)|T], [field(Name,NVal)|NT]) :- normalise_value(Val,NVal),
    normalise_fields(T,NT).

/* version to be called if for some reason second argument already instantiated;
  e.g. when operation arguments and results are already instantiated */
%%normalise3(X,R) :- var(R),var(X),!,R=X. /* just in case the when is unfrozen by hypercall */
:- use_module(kernel_objects,[equal_object/3]).
%normalise3(R,[H|T]) :- !,kernel_objects:equal_object(R,[H|T]).
/* don't insist on normalising if a set is  already partially given*/
normalise3((A,B),(NA,NB)) :- !, normalise_value(A,NA), normalise_value(B,NB).
normalise3(V,Res) :- equal_object(V,Res,normalise3).


:- use_module(custom_explicit_sets,[construct_avl_from_lists/2]).
:- use_module(kernel_tools,[ground_value_check/2]).
normalise_list(List,Res) :-
        ground_value_check(List,ListGround), % this is slightly slower than ground() for long lists not containing avl_set values
        normalise_list_aux(ListGround,List,Res).
:- block normalise_list_aux(-,?,?).
normalise_list_aux(_,List,Res) :-
        construct_avl_from_lists(List,Avl), %print(done),nl,
        kernel_objects:equal_object(Res,Avl,normalise_list).


expand_closure_value_for_normalisation(Var,_,_,R) :- var(Var),!,R=Var.
expand_closure_value_for_normalisation(closure(P,T,B),EXPAND,VarName,ExpandedValue) :- !, %print(expand_closure_value_for_normalisation(P,T,B)),nl,
   expand_closure_value2(P,T,B,EXPAND,VarName,ExpandedValue).
%expand_closure_value_for_normalisation(closure_x(P,T,B,Exp),EXPAND,VarName,ExpandedValue) :- !,
%   (ground(Exp) -> ExpandedValue=Exp ; expand_closure_value_for_normalisation(closure(P,T,B),EXPAND,VarName,ExpandedValue)).
expand_closure_value_for_normalisation(Val,_Expand,_VarName,Val).

:- use_module(custom_explicit_sets,[dont_expand_this_explicit_set/1,dont_expand_this_explicit_set/2, is_infinite_or_very_large_explicit_set/1]).
:- use_module(clpfd_interface,[catch_clpfd_overflow_call2/2]).
:- use_module(tools_timeout, [time_out_with_factor_call/3]).

:- block expand_closure_value2(-,?,?,?,?,?), expand_closure_value2(?,-,?,?,?,?), expand_closure_value2(?,?,-,?,?,?).
expand_closure_value2(P,T,B,EXPAND,VarName,ExpandedValue) :- !,
   % print(expand_closure_value_for_normalisation(P,T,B)),nl,
   (dont_expand_for_norm(EXPAND,closure(P,T,B))
      ->   construct_normalised_closure(P,T,B,ExpandedValue)   %, print(not_expanding(closure(P,T,B))),nl
      ;    %print(expanding(closure(P,T,B))),nl,
      catch_clpfd_overflow_call2(
           time_out_with_factor_call(
               call_with_enumeration_warning( /* ensure we generate enumeration_warning exceptions */
                 custom_explicit_sets:expand_closure_to_avl_or_list(P,T,B,ExpandedValue,no_check,no_wf_available)
                ),
                0.3,store:time_out_expand_backup(VarName,P,T,B,ExpandedValue))
             % , print(expanded_closure(ExpandedValue)),nl
             , (format('% CLP(FD) overflow during closure expansion: ~w~n',[VarName]),
                construct_normalised_closure(P,T,B,ExpandedValue)))
    ).

dont_expand_for_norm(false,_).
dont_expand_for_norm(true,C) :-
   dont_normalise_this_explicit_set(C) ;
   dont_expand_this_explicit_set(C).
dont_expand_for_norm(limit(Limit),C) :-
   dont_normalise_this_explicit_set(C) ;
   dont_expand_this_explicit_set(C,Limit).
dont_expand_for_norm(force,C) :- is_infinite_or_very_large_explicit_set(C).

:- use_module(memoization,[is_memoization_closure/4]).
% we usually do not want to expand memoized closures when storing in the state !
dont_normalise_this_explicit_set(closure(P,T,B)) :- is_memoization_closure(P,T,B,_).

/*
:- use_module(custom_explicit_sets,[dont_expand_this_explicit_set/1, is_interval_closure/5]).
% maybe systematically expand small intervals and global_set's for normalization purposes ??
% (alternatively, we could re-convert certain avl-sets to symbolic form here)
dont_expand_closure_value(_,_,_) :-
   preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,false),!.
%dont_expand_closure_value(P,T,B) :-
% is_interval_closure(P,T,B,Low,Up), number(Low), number(Up),!,
%    % interval closures are quite efficient for certain manipulations
%   Size is Up+1-Low, Size>100. % another magic constant ; which value to choose ??
%dont_expand_closure_value(_,_,_) :-
%   preferences:get_preference(convert_comprehension_sets_into_closures,false),!.
%   % in the case above: we have already decided not to expand the closure earlier, dont repeat check !? TO DO: check that this is indeed ok; that no closures are constructed otherwise; maybe we should just check for small interval closures ?
dont_expand_closure_value(P,T,B) :- dont_expand_this_explicit_set(closure(P,T,B)).
   */

time_out_expand_backup(VarName,P,T,B,ExpandedValue) :-
   add_message(expand_closure_value,'Time-out while trying to expand value of: ',VarName,B),
   construct_normalised_closure(P,T,B,ExpandedValue).

construct_normalised_closure(P,T,B1,closure(P,T,NB)) :-
   %print(normalising(P)),nl,
   %b_compile(B1,P,[],[],B2,no_wf_available), % probably much more expensive than simple code with can_be_computed below
   normalise_closure_body(B1,NB).
  % print(norm),nl,print(B),nl,print(NB),nl,nl.

% normalise values that are compiled inside closures
normalise_closure_body(b(OP,T,I),b(NOP,T,FI)) :-
   (var(I) -> print(b(OP,T,I)),nl , when(nonvar(OP),(print(b(OP,T,I)),nl)) ; true),
   filter_info(I,FI),
   ncb(OP,OP2),
   (can_be_computed(OP2) % when compiling not all information was available for computation, now it is
    -> b_interpreter:b_compute_expression_nowf(b(OP,T,I),[],[],Val),
       %print(comp(Val)),nl,
       ncb(value(Val),NOP) % TO DO: better apply the can_be_computed check after ncb, if performance ok
    ; NOP=OP2).

filter_info(V,R) :- var(V),!,R=V.
filter_info([],[]).
filter_info([H|T],Res) :-
    (important_info(H) -> Res=[H|TR] ; Res=TR), filter_info(T,TR).

important_info(contains_wd_condition).
important_info(prob_annotation(_)).
important_info(used_ids(_)).
important_info(was(_)).
important_info(nodeid(_)).
important_info(allow_to_lift_exists).

:- use_module(kernel_mappings).
% ncb stands for normalise closure body
ncb(V,R) :- var(V),!, R=V.
ncb(bool_set,R) :- !, R=bool_set.
ncb(truth,R) :- !, R=truth.
ncb(event_b_identity,R) :- !, R=event_b_identity.
ncb(falsity,R) :- !, R=falsity.
ncb(max_int,R) :- !, R=max_int.
ncb(min_int,R) :- !, R=min_int.
ncb(value(V),value(NV)) :- !,normalise_value_for_var(closure_body,V,NV).
ncb(integer(X),integer(NX)) :- !, X=NX.
ncb(identifier(X),identifier(NX)) :- !, X=NX.
ncb(string(X),string(NX)) :- !, X=NX.
ncb(string_set,R) :- !, R=string_set.
ncb(integer_set(X),integer_set(NX)) :- !, X=NX.
ncb(set_extension(X),set_extension(NX)) :- !, X=NX.
ncb(partition(S,L),partition(NS,NL)) :- !, normalise_closure_body(S,NS),
   maplist(normalise_closure_body,L,NL).
ncb(general_sum(Ids,Condition,Expression),general_sum(Ids,NC,NE)) :- !,
   normalise_closure_body(Condition,NC),normalise_closure_body(Expression,NE).
ncb(general_product(Ids,Condition,Expression),general_product(Ids,NC,NE)) :- !,
   normalise_closure_body(Condition,NC),normalise_closure_body(Expression,NE).
ncb(sequence_extension(X),sequence_extension(NX)) :- !, X=NX.
ncb(external_function_call(F,L),external_function_call(F,LL)) :- !, L=LL.
ncb(external_pred_call(F,L),external_pred_call(F,LL)) :- !, L=LL.
ncb(equal(A,B),equal(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(couple(A,B),couple(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(interval(A,B),interval(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(function(A,B),function(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(operation_call_in_expr(A,B),operation_call_in_expr(NA,B)) :- !,
    normalise_closure_body(A,NA).
ncb(relations(A,B),relations(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(member(A,B),member(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(not_member(A,B),not_member(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(negation(B),negation(NB)) :- !, normalise_closure_body(B,NB).
ncb(conjunct(A,B),conjunct(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(disjunct(A,B),disjunct(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(equivalence(A,B),equivalence(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(if_then_else(A,B,C),if_then_else(NA,NB,NC)) :- !,
   normalise_closure_body(A,NA), normalise_closure_body(B,NB), normalise_closure_body(C,NC).
ncb(implication(A,B),implication(NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(comprehension_set(V,B),comprehension_set(V,NB)) :- !, normalise_closure_body(B,NB).
ncb(convert_bool(B),convert_bool(NB)) :- !, normalise_closure_body(B,NB).
ncb(exists(IDs,B),exists(IDs,NB)) :- !, normalise_closure_body(B,NB).
ncb(forall(IDs,A,B),forall(IDs,NA,NB)) :- !, normalise_closure_body(A,NA), normalise_closure_body(B,NB).
ncb(assertion_expression(Cond,ErrMsg,Expr),assertion_expression(NA,ErrMsg,NB)) :- !,
  normalise_closure_body(Cond,NA), normalise_closure_body(Expr,NB).
ncb(freetype_case(FT,IsCase,Expr),freetype_case(FT,IsCase,NB)) :- !,
  normalise_closure_body(Expr,NB).
ncb(freetype_constructor(Id,Case,Expr),freetype_constructor(Id,Case,NB)) :- !,
  normalise_closure_body(Expr,NB).
ncb(freetype_destructor(Id,Case,Expr),freetype_destructor(Id,Case,NB)) :- !,
  normalise_closure_body(Expr,NB).
ncb(record_field(R,Name),record_field(NR,Name)) :- !,
  normalise_closure_body(R,NR).
ncb(rec(Fields),rec(NFields)) :- !,
  ncb_fields(Fields,NFields).
ncb(kodkod(Id,Ids),R) :- !, R=kodkod(Id,Ids).
ncb(lazy_let_expr(Id,E,Expr),lazy_let_expr(Id,NE,NExpr)) :- !, normalise_closure_body(E,NE),
  normalise_closure_body(Expr,NExpr).
ncb(lazy_let_pred(Id,E,Pred),lazy_let_pred(Id,NE,NPred)) :- !, normalise_closure_body(E,NE),
  normalise_closure_body(Pred,NPred).
% lazy_let_subst should not occur in closure bodies
ncb(lazy_lookup_expr(X),lazy_lookup_expr(NX)) :- !, X=NX.
ncb(lazy_lookup_pred(X),lazy_lookup_pred(NX)) :- !, X=NX.
ncb(let_predicate(Ids,AssignmentExprs,Pred),let_predicate(Ids,AssignmentExprs,NPred)) :- !,
  normalise_closure_body(Pred,NPred).
ncb(let_expression(Ids,AssignmentExprs,Expr),let_expression(Ids,AssignmentExprs,NExpr)) :- !,
  normalise_closure_body(Expr,NExpr).
ncb(let_expression_global(Ids,AssignmentExprs,Expr),let_expression_global(Ids,AssignmentExprs,NExpr)) :- !,
  normalise_closure_body(Expr,NExpr).
ncb(F,NF) :- functor(F,Op,1),
  (unary_function(Op,_,_) ; unary_in_boolean_type(Op,_)), !,
  arg(1,F,Arg1), functor(NF,Op,1), arg(1,NF,NormArg1),
  normalise_closure_body(Arg1,NormArg1).
ncb(F,NF) :- functor(F,Op,2),
  (binary_function(Op,_,_) ; kernel_mappings:binary_boolean_operator(Op,_,_) ; binary_in_boolean_type(Op,_)), !,
  arg(1,F,Arg1),arg(2,F,Arg2), functor(NF,Op,2), arg(1,NF,NormArg1), arg(2,NF,NormArg2),
  normalise_closure_body(Arg1,NormArg1),
  normalise_closure_body(Arg2,NormArg2).
ncb(X,X) :- print(ncb(X)),nl.


can_be_computed(domain(X)) :- is_known_set_or_int_value(X).
can_be_computed(range(X)) :- is_known_set_or_int_value(X).
can_be_computed(plus(X,Y)) :- is_known_set_or_int_value(X),is_known_set_or_int_value(Y).
can_be_computed(minus(X,Y)) :- is_known_set_or_int_value(X),is_known_set_or_int_value(Y).
can_be_computed(times(X,Y)) :- is_known_set_or_int_value(X),is_known_set_or_int_value(Y).
% check for more possible partial evaluations; check for common code with b_compiler, CSE or other parts of ProB

is_known_set_or_int_value(b(value(V),_,_)) :- is_known_aux(V).
is_known_aux(X) :- var(X),!,fail.
is_known_aux([]).
is_known_aux(avl_set(_)).
is_known_aux(int(Nr)) :- number(Nr).

ncb_fields([],[]).
ncb_fields([field(Name,Expr)|T], [field(Name,NExpr)|NT]) :-
   normalise_closure_body(Expr,NExpr),
   ncb_fields(T,NT).

%ncb(set_extension([b(couple(b(add(b(identifier(idx),integer,[nodeid(pos(53,1,11,91,11,93))]),b(integer(1),integer,[nodeid(pos(54,1,11,95,11,95))])),integer,[nodeid(pos(52,1,11,91,11,95))]),b(identifier(bit),integer,[nodeid(pos(55,1,11,101,11,103))])),couple(integer,integer),[nodeid(pos(51,1,11,91,11,103))])]))

normalise_store(T,NT) :-
    preferences:get_computed_preference(convert_closures_into_explicit_form_for_store,EXPAND),
    normalise_store1(T,EXPAND,NT).
:- block normalise_store1(-,?,?).
normalise_store1([],_Expand,Res) :- !, Res=[].
normalise_store1([bind(Var,Val)|T],EXPAND,Res) :- !, Res = [bind(Var,NVal)|NT],
    %print(normalise(Val)),nl,print(Var),nl,
    normalise_value_for_var(Var,EXPAND,Val,NVal),
    normalise_store1(T,EXPAND,NT).
normalise_store1(Store,E,Res) :-
    add_internal_error('Illegal store: ',normalise_store1(Store,E,Res)),
    Res=Store.

/* ----------------------- */

:- assert_pre(store:lookup_value_for_existing_id(Id,State,_Val),
                (type_check(Id,variable_id),list_skeleton(State),type_check(State,store))).
:- assert_post(store:lookup_value_for_existing_id(_Id,_State,_Val), true).

lookup_value_for_existing_id(Id,State,Res) :-
 (member_bind(Id,Val,State) % safe_member(bind(Id,Val),State,lookup_value_for_existing_id)
   -> check_val(Val,Id,State),Res=Val
    ; add_internal_error('Call Failed: ',lookup_value_for_existing_id(Id,State,Res)),
      Res=fail
  ).

:- assert_pre(store:lookup_value_for_existing_id(Id,LS,State,_Val),
                (type_check(Id,variable_id),type_check(LS,store),type_check(State,store),list_skeleton(LS),list_skeleton(State))).
:- assert_post(store:lookup_value_for_existing_id(_Id,_LS,_State,_Val), true).

lookup_value_for_existing_id(Id,LocalState,State,Val) :-
 (member_bind(Id,Val,LocalState) %safe_member(bind(Id,Val),LocalState,lookup_value_for_existing_id__4)
   -> check_val(Val,Id,LocalState)
    ; lookup_value_for_existing_id(Id,State,Val)).

lookup_value(Id,State,Val) :-
 (member_bind(Id,Val,State) %safe_member(bind(Id,Val),State,lookup_value)
   -> check_val(Val,Id,State)
  ).
lookup_value(Id,LocalState,State,Val) :-
 (member_bind(Id,Val,LocalState) %safe_member(bind(Id,Val),LocalState,lookup_value__4)
   -> check_val(Val,Id,LocalState)
    ; member_bind(Id,Val,State) -> check_val(Val,Id,State)).
lookup_value_with_span(Id,LocalState,State,Val,Span) :-
 (member_bind(Id,Val,LocalState) %safe_member(bind(Id,Val),LocalState,lookup_value__4)
   -> check_val_with_span(Val,Id,LocalState,Span)
    ; member_bind(Id,Val,State) -> check_val_with_span(Val,Id,State,Span)).
lookup_value_without_check(Id,LocalState,State,Val) :- % does not check if value is undefined
 (member_bind(Id,Val,LocalState) %safe_member(bind(Id,Val),LocalState,lookup_value__4)
   -> true
    ; member_bind(Id,Val,State) -> true).

% a faster version of member(bind(Id,Val),List) with only one solution
member_bind(Id,Val,[H|T]) :-
      arg(1,H,Id) -> arg(2,H,Val) ; member_bind(Id,Val,T).

get_id_value(Id,State,Val) :- safe_member(bind(Id,Val),State,get_id_value).  /* can be backtracked over */

%safe_member(X,T) :- safe_member(X,T,safe_member).
safe_member(_,[],_) :- !,fail.
safe_member(X,[H|T],Loc) :- !, (X=H ; safe_member(X,T,Loc)).
safe_member(X,T,Loc) :- add_internal_error('Argument not a list: ',safe_member(X,T,Loc)),fail.

check_val_with_span(X,V,_State,Span) :-
   (X==term(undefined) ->
     add_error(reading_undefined_variable,'Trying to read undefined variable: ',V,Span) % do not fail; otherwise we will look elsewhere and generate another error message, see test 1660
     ; true).

check_val(X,V,_State) :-
   (X==term(undefined) ->
     add_error(reading_undefined_variable,'Trying to read undefined variable: ',V) % do not fail
     ; true).

:- assert_must_succeed(( store:lookup_and_delete_value(x,X,[bind(y,2),bind(x,3)],R,true,_WF,F),
  X==3, R==[bind(y,2)],F==true )).
:- assert_must_succeed(( store:lookup_and_delete_value(x,X,[bind(y,2)|T],R,true,_WF,F),
  T = [bind(x,3)],
  X==3, R==[bind(y,2)],F==true )).

lookup_and_delete_value(Var,Value,T,UT,ReportError,WF,Finished) :-
   lookup_and_delete_value1(T,Var,Value,UT,ReportError,WF,Finished).
:- use_module(kernel_waitflags,[add_abort_error_span/5]).
:- block lookup_and_delete_value1(-,?,?,?,?,?,?).
lookup_and_delete_value1([],Var,Value,[],ReportError,WF,Finished) :-
   (ReportError=report_error(WF)
     ->  add_abort_error_span(missing_assignment_error,'Statement did not assign to variable: ',Var,unknown,WF)
       /* lookup_and_delete is called only from get_results */
     ;   true
   ), no_value_for_variable(Value,Var),
   Finished=true.
lookup_and_delete_value1([bind(Var2,Value2)|T],Var,Value,Res,ReportError,WF,Finished) :-
     lookup_and_delete_value2(Var2,Value2,T,Var,Value,Res,ReportError,WF,Finished).
:- block lookup_and_delete_value2(-,?,?,?,?,?,?,?,?).
lookup_and_delete_value2(Var2,Value2,T,Var,Value,Res,ReportError,WF,Finished) :-
   (Var=Var2
      -> Value=Value2, T=Res, Finished=true
      ;  Res=[bind(Var2,Value2)|UT],
          lookup_and_delete_value1(T,Var,Value,UT,ReportError,WF,Finished)
    ).

no_value_for_variable(term(no_value_for(Var)),Var).

:- assert_must_succeed(( store:search_and_delete_value(x,22,X,[bind(y,2),bind(x,3)],R),
  X==3, R==[bind(y,2)] )).
:- assert_must_succeed(( store:search_and_delete_value(x,22,X,[bind(y,2)],R),
  X==22, R==[bind(y,2)] )).
:- assert_must_succeed(( store:search_and_delete_value(x,22,X,[bind(y,2)|T],R),
  T = [bind(x,3)],
  X==3, R==[bind(y,2)] )).

:- block search_and_delete_value(-,?,?,?,?).
search_and_delete_value(Var,OldVal,Value,InStore,OutStore) :-
  search_and_delete_value2(InStore,Var,OldVal,Value,OutStore).

:- block search_and_delete_value2(-,?,?,?,?).
search_and_delete_value2([],_Var,OldVal,Value,[]) :- eq_obj(Value,OldVal).
search_and_delete_value2([BIND|T],Var,OldVal,Value,Res) :- % BIND=bind(Var2,Value2)
   arg(1,BIND,Var2), arg(2,BIND,Value2),
   search_and_delete_value3(Var2,Value2,T,Var,OldVal,Value,Res).

:- block search_and_delete_value3(-,?,?,?,?,?,?).
search_and_delete_value3(Var2,Value2,T,Var,OldVal,Value,Res) :-
    (Var=Var2
      -> eq_obj(Value,Value2), T=Res
      ;  Res=[bind(Var2,Value2)|UT],
         search_and_delete_value2(T,Var,OldVal,Value,UT)
    ).

eq_obj(Val1,Val2) :- (var(Val1);var(Val2)),!,Val1=Val2.
eq_obj(Val1,Val2) :- equal_object(Val1,Val2,search_and_delete_value).

empty_state([]).
add_var_to_localstate(Var,Val,InState,[bind(Var,Val)|InState]).

:- assert_pre(store:set_up_localstate(Vars,Vals,In,_),
                (type_check(Vars,list(variable_id)),type_check(Vals,vlist(any)),type_check(In,store))).
:- assert_post(store:set_up_localstate(_,_,_,Out), type_check(Out,store)).

set_up_localstate([],[],S,S).
set_up_localstate([Var|T],[Val|FT],InS,[bind(VarID,Val)|OutS]) :-
   /* add Var,Val to Local state at the beginning to properly handle nested
      blocks with same variable names */
   (Var=b(identifier(Id),_,_) -> %print(typed_identifier_for_set_up_localstate(Var)),nl,
                                 VarID = Id
                               ; VarID = Var),
   set_up_localstate(T,FT,InS,OutS).

set_up_undefined_localstate([],S,S).
set_up_undefined_localstate([Var|T],InS,[bind(Id,term(undefined))|OutS]) :-
   def_get_texpr_id(Var,Id),
   /* add Var,Val to Local state at the beginning to properly handle nested
      blocks with same variable names */
   set_up_undefined_localstate(T,InS,OutS).

:- use_module(tools_lists,[delete_first/3]).
% sets up undefined values for variables, except for a list of predefined variables (all of which must be used)
set_up_undefined_localstate_with_predefined([],PredefinedValues,S,S) :-
   (PredefinedValues=[] -> true
     ; add_internal_error('Unknown predefined values: ',set_up_undefined_localstate_with_predefined([],PredefinedValues,S,S))).
set_up_undefined_localstate_with_predefined([Var|T],PredefinedValues,InS,[bind(Id,Val)|OutS]) :-
   def_get_texpr_id(Var,Id),
   delete_first(InS,bind(Id,_),InS2), % in case of name clash: remove the old binding from the state
   (select(bind(Id,Val),PredefinedValues,Rest)
     -> set_up_undefined_localstate_with_predefined(T,Rest,InS2,OutS)
     ;  Val=term(undefined),
        set_up_undefined_localstate_with_predefined(T,PredefinedValues,InS2,OutS)).


:- assert_pre(store:delete_variables_from_state(Vars,In,_),
                (type_check(Vars,list(variable_id)),type_check(In,store))).
:- assert_post(store:delete_variables_from_state(_,_,Out), type_check(Out,store)).

delete_variables_from_state([],S,S).
delete_variables_from_state([Var|T],InState,OutState) :-
   delete_variable(InState, Var, S2),
   delete_variables_from_state(T,S2,OutState).

delete_variable([],_,[]).
delete_variable([bind(V,Val)|T],Var,Res) :-
  (Var==V
    ->  Res = T
    ;   Res = [bind(V,Val)|DT], delete_variable(T,Var,DT)
  ).
/* --------------------------------- */

% unused at the moment:
%:- assert_pre(store:copy_store_template(In,_),
%                (type_check(In,store))).
%:- assert_post(store:copy_store_template(_,Out), type_check(Out,store)).
%copy_store_template([],[]).
%copy_store_template([bind(Var,_)|T],[bind(Var,_)|CT]) :- copy_store_template(T,CT).

:- use_module(library(ordsets)).
% copy_store_template copies a store and its values except some changing variables
% Two new stores are created: constants are not copied into the first new update state, but
% to the second new full state
copy_variable_store(InState,SortedChanges,Updates,NewFullState) :-
   (SortedChanges=[] -> Updates = [],NewFullState=InState % Query Operation/Event ,print(no_upd),nl
    ; copy_variable_store2(InState,SortedChanges,Updates,NewFullState)).

copy_variable_store2([],_,[],[]).
copy_variable_store2([bind(Var,Old)|T],SortedChanges,Updates,NewFullState) :-
    ( b_is_constant(Var) ->
        Updates = CT, NewFullState = [bind(Var,Old)|FT]
    %; safe_member(Var,SortedChanges,copy_variable_store) ->
    ; ord_member(Var,SortedChanges) ->
    %; ord_delete(SortedChanges,Var,SortedChanges2) -> % in principle better than ord_member; but seems slightly slower on examples/EventBPrologPackages/Advance_WP2/v6_Sep2014/ex_mch.eventb
        Updates = [bind(Var,New)|CT], NewFullState = [bind(Var,New)|FT]
    ; otherwise ->
        Updates = CT, NewFullState = [bind(Var,Old)|FT]
    ), copy_variable_store2(T,SortedChanges,CT,FT).

%ord_delete([H|T],El,Res) :- El @>= H,
%  (H==El -> Res=T ; Res = [H|RR], ord_delete(T,El,RR)).


% -------------------------------

:- assert_must_succeed(store:check_valid_store([bind(x,int(1)), bind(y,int(2))],unit_test)).

:- block check_valid_store(-,?).
check_valid_store([],_PP).
check_valid_store([bind(Var,Value)|T],PP) :- member(bind(Var,Val2),T),
  add_internal_error('Multiple occurences of variable: ',bind(Var,Value,Val2,PP)),
  fail.
check_valid_store([_|T],PP) :- check_valid_store(T,PP).


