% (c) 2014-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(predicate_handling,[prime_predicate/2,
                              create_state_predicate/2,
                              create_succ_state_predicate/2,
                              create_constants_state_predicate/2,
                              get_initial_state_predicate/1,
                              get_transition_predicate/1,
                              get_single_transition_predicate/2,
                              create_negation_and_cleanup/2,
                              create_primary_inputs_predicate/2,
                              deprime_state/2,
                              prime_n_times/3,
                              unprimed/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'Handling and creation of predicates used in the symbolic model checker.').

:- use_module(library(lists)).

:- use_module(probsrc(b_global_sets), [b_global_set/1,
                                       is_b_global_constant/3]).
:- use_module(probsrc(bmachine),[b_get_machine_operation/4,
                                 b_get_machine_constants/1,
                                 b_get_machine_variables/1,
                                 b_get_initialisation_from_machine/2,
                                 b_get_properties_from_machine/1,
                                 get_proven_invariant/2]).
:- use_module(probsrc(before_after_predicates),[before_after_predicate_with_equalities/3,
                                 before_after_predicate_list_disjunct_with_equalities/3,
                                 before_after_predicate_for_operation/2]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2,
                                     disjunct_predicates/2,
                                     disjunction_to_list/2,
                                     transform_bexpr/3,
                                     syntaxtransformation/5,
                                     create_texpr/4,
                                     safe_create_texpr/3,
                                     create_negation/2,
                                     get_texpr_expr/2,
                                     get_texpr_type/2,
                                     unique_typed_id/3]).

:- use_module(probsrc(preferences), [get_preference/2]).

get_initial_state_predicate(Init) :-
    % the initial state contains both the constants defined by the properties
    % and the variables set by the initialisation
    b_get_properties_from_machine(Properties),
    b_get_initialisation_from_machine(Initialisation,_),
    before_after_predicate_with_equalities(Initialisation,[],InitialisationPred),
    conjunct_predicates([Properties,InitialisationPred],PredWithPrimes),
    transform_bexpr(predicate_handling:remove_prime_from_expression,PredWithPrimes,Init).

get_transition_predicate(TransitionPredicate) :-
    get_preference(use_po,false), !,
    findall(Body,b_get_machine_operation(_Name,_Results,_Parameters,Body),ListOfBodies),
    b_get_machine_variables(Variables),
    b_get_machine_constants(Constants),
    append(Variables,Constants,VarsAndConsts),
    before_after_predicate_list_disjunct_with_equalities(ListOfBodies,VarsAndConsts,TransitionPredicateNoInputs),
    add_primary_inputs(TransitionPredicateNoInputs,TransitionPredicate).
get_transition_predicate(TransitionPredicate) :-
    get_preference(use_po,true), !,
    findall(op(OpName,Pred),(before_after_predicate_for_operation(OpName,Pred)),Ops),
    maplist(op_to_transition,Ops,SingleSteps),
    disjunct_predicates(SingleSteps,TransitionPredicateNoInputs),
    add_primary_inputs(TransitionPredicateNoInputs,TransitionPredicate).

op_to_transition(op(OpName,BaPred),StepConstraint) :-
    (get_proven_invariant(OpName,ProvenInvariant)
     -> true ; ProvenInvariant = b(truth,pred,[])),
    prime_predicate(ProvenInvariant,PrimedProvenInvariant),
    conjunct_predicates([BaPred,PrimedProvenInvariant],StepConstraint).

get_single_transition_predicate(OpName,TP) :-
    b_get_machine_operation(OpName,_Results,_Parameters,Body),
    b_get_machine_variables(Variables),
    b_get_machine_constants(Constants),
    append(Variables,Constants,VarsAndConsts),
    before_after_predicate_with_equalities(Body,VarsAndConsts,TransitionPredicateNoInputs),
    add_primary_inputs(TransitionPredicateNoInputs,TP).

:- meta_predicate maplist(4,-,-,-,-).
maplist(Pred, As, Bs, Cs, Ds) :-
    (foreach(A,As),
     foreach(B,Bs),
     foreach(C,Cs),
     foreach(D,Ds),
     param(Pred)
     do  call(Pred, A, B, C, D)
    ).

add_primary_inputs(TIn,TOut) :-
    % transition is a disjunction
    % as a first step add an integer variable to each disjunct
    % + a constraint stating that the sum is 1
    disjunction_to_list(TIn,Disjuncts),
    maplist(add_input_to_op,Disjuncts,TopLevelIdentifiers,_InnerIdentifiers,DisjunctsOut),
    sum_is_one(TopLevelIdentifiers,Constraint),
    disjunct_predicates(DisjunctsOut,Disjunction),
    conjunct_predicates([Disjunction,Constraint],TOut).

add_input_to_op(Op,IId,InnerIIds,OpOut) :-
    unique_typed_id("_ic3_pi",integer,IId),
    safe_create_texpr(integer(1),integer,Int1),
    safe_create_texpr(equal(IId,Int1),pred,EqualToOne),
    add_inner_inputs_to_op(Op,InnerIIds,Op2),
    conjunct_predicates([Op2,EqualToOne],OpOut).

% add variables to substitutions like ANY, etc.
% these variables should make the substitution deterministic
% once they are fixed to a value
add_inner_inputs_to_op(Subst,Ids,SubstOut) :-
    get_texpr_expr(Subst,Expr),
    add_inner_inputs_to_op_aux(Expr,Ids,ExprOut), !,
    safe_create_texpr(ExprOut,pred,SubstOut).
%add_inner_inputs_to_op(Subst,[],Subst) :-
%    translate:translate_bexpression(Subst,PPSubst),
%    format('not implemented in add_inner_inputs_to_op: ~w~n',[PPSubst]).
add_inner_inputs_to_op(Subst,[],Subst).
% these might contain a non-deterministic choice somewhere
add_inner_inputs_to_op_aux(conjunct(A,B),Ids,conjunct(NA,NB)) :- !,
    add_inner_inputs_to_op(A,AIds,NA),
    add_inner_inputs_to_op(B,BIds,NB),
    append(AIds,BIds,Ids).
add_inner_inputs_to_op_aux(disjunct(A,B),Ids,disjunct(NA,NB)) :- !,
    add_inner_inputs_to_op(A,AIds,NA),
    add_inner_inputs_to_op(B,BIds,NB),
    append(AIds,BIds,Ids).
% these have a non-deterministic choice. add a variable.
add_inner_inputs_to_op_aux(member(A,B),[InputId],conjunct(b(member(A,B),pred,[]),Eq)) :- !,
    get_texpr_type(A,Type),
    unique_typed_id("_ic3_pi",Type,InputId),
    safe_create_texpr(equal(A,InputId),pred,Eq).
% these should always be deterministic
add_inner_inputs_to_op_aux(equal(A,B),[],equal(A,B)) :- !.

sum_is_one(Identifiers,Constraint) :-
    build_sum(Identifiers,Sum),
    safe_create_texpr(integer(1),integer,Int1),
    safe_create_texpr(equal(Sum,Int1),pred,Constraint).

build_sum([Id],Id).
build_sum([Id|Ids],Sum) :-
    build_sum(Ids,SubSum),
    safe_create_texpr(add(Id,SubSum),integer,Sum).

remove_prime_from_binding(binding(Id,Res,PP),binding(NId,Res,PP)) :-
    remove_prime(Id,NId).

remove_prime_from_expression(b(identifier(Id),Type,Infos),b(identifier(NId),Type,Infos)) :- !,
    remove_prime(Id,NId).
remove_prime_from_expression(Expr,Expr).

remove_prime(Id,NoPrimeId) :-
    atom_codes(Id,Codes),
    append(NoPrimeIdCodes,[39],Codes), !, % last character is 39 = prime
    atom_codes(NoPrimeId,NoPrimeIdCodes).
remove_prime(Id,Id).

create_state_predicate(Solution,StatePredicate) :-
   include(unprimed,Solution,UPSolution),
   create_state_predicate_aux(UPSolution,StatePredicate).

create_succ_state_predicate(Solution,SStatePredicate) :-
    deprime_state(Solution,DePrimedState),
    create_state_predicate_aux(DePrimedState,SStatePredicate).

deprime_state(State,DeprimedState) :-
    include(primed,State,PSolution),
    maplist(remove_prime_from_binding,PSolution,DeprimedState).

create_constants_state_predicate(Solution,ConstantsStatePred) :-
    include(unprimed,Solution,UPSolution),
    b_get_machine_constants(Csts),
    include(filter_vars_and_constants(Csts),UPSolution,UPSolutionOnlyConstants),
    UPSolutionOnlyConstants \= [],
    create_state_predicate_aux(UPSolutionOnlyConstants,ConstantsStatePred).

create_primary_inputs_predicate(Solution,InputsPred) :-
    exclude(primed,Solution,NoPrimes),
    include(primary_input,NoPrimes,NoPrimesNoState),
    maplist(create_single_assignment_predicate(NoPrimesNoState),NoPrimesNoState,SolutionPreds),
    conjunct_predicates(SolutionPreds,InputsPred).

create_state_predicate_aux(Solution,StatePredicate) :-
    b_get_machine_variables(Vars), b_get_machine_constants(Csts),
    append(Vars,Csts,VandC),
    include(filter_vars_and_constants(VandC),Solution,SolutionOnlyStateVars),
    maplist(create_single_assignment_predicate(VandC),SolutionOnlyStateVars,SolutionPreds),
    conjunct_predicates(SolutionPreds,StatePredicate).

create_single_assignment_predicate(VandC,binding(Name,Value,_PP),Pred) :-
    % fetch the type for the particular variable / constant
    member(b(identifier(Name),Type,Infos),VandC),
    Pred = b(equal(b(identifier(Name),Type,Infos),b(value(Value),Type,[])),pred,[]).

filter_vars_and_constants(VandC,binding(Id,_,_)) :-
    member(b(identifier(Id),_,_),VandC).

unprimed(binding(Id,_,_)) :- atom_codes(Id,Codes), \+last(Codes,39).
primed(binding(Id,_,_)) :- atom_codes(Id,Codes), last(Codes,39).

primary_input(binding(Id,_,_)) :- atom_codes(Id,Codes), prefix("_ic3_pi", Codes).

prime_predicate(A,B) :-
    create_texpr(OExpr,Type,Info,A),
    create_texpr(NExpr,Type,Info,B),
    prime_texpr(OExpr,Type,NExpr).

% certain identifiers should not be primed
% global sets
do_not_prime_identifier(Id,set(global(Id))) :-
    b_global_set(Id).
% members of global sets
do_not_prime_identifier(MId,global(SId)) :-
    is_b_global_constant(SId,_,MId).

prime_texpr(identifier(Old),Type,identifier(New)) :-
    % write(pt(Old,Type)), nl,
    do_not_prime_identifier(Old,Type), !,
    Old = New.
prime_texpr(identifier(Old),_Type,identifier(New)) :- !,
    atom_concat(Old,'\'',New).
prime_texpr(lazy_lookup_expr(Old),_,lazy_lookup_expr(New)) :-
    !, atom_concat(Old,'\'',New).
prime_texpr(lazy_lookup_pred(Old),_,lazy_lookup_pred(New)) :-
    !, atom_concat(Old,'\'',New).
prime_texpr(OExpr,_,NExpr) :-
    syntaxtransformation(OExpr,Subs,_TNames,NSubs,NExpr),
    prime_texpr_l(Subs,NSubs).
prime_texpr_l([],[]).

prime_texpr_l([Old|ORest],[New|NRest]) :-
    prime_predicate(Old,New),
    prime_texpr_l(ORest,NRest).

% involves specialised cleanup rules for atb provers
create_negation_and_cleanup(P,NP) :-
    get_texpr_expr(P,implication(A,B)), !,
    create_negation_and_cleanup(B,NegB),
    safe_create_texpr(conjunct(A,NegB),pred,NP).
create_negation_and_cleanup(P,NP) :-
    create_negation(P,NP).

prime_n_times(0,P,P) :- !.
prime_n_times(N,P,Primed) :-
    prime_predicate(P,T),
    N2 is N - 1,
    prime_n_times(N2,T,Primed).
