% (c) 2020-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(tools_fastread, [ read_term_from_stream/2,
                            read_term_from_string/2,
                            read_test/1, tok_test/1
                   ]).

:- use_module(module_information,[module_info/2]).

:- module_info(group,infrastructure).
:- module_info(description,'This module provides a more efficient way to parse and read Prolog terms.').

:- use_module(self_check).
:- use_module(error_manager,[add_error/3]).

% Limitations:
% - only prefix notation and lists supported
% - supports atoms with and without quotes, in quotes one can use octal and hex literals and backslash escaping
% - no support for variables (yet?)
% - no support for operator declarations
% - currently no floats
% - no real support for (a,b) terms using , as infix


% --------------------

% Tokeniser


:- use_module(tools,[print_memory_used_wo_gc/0]).
tok_read(File) :- statistics(walltime,_),
    %open(File,read,Stream1),reset_line_nr, repeat, get_code(Stream1,C), C = -1,!, close(Stream1),
    %statistics(walltime,[_,W1]), format('~n% Walltime ~w ms to get_codes ~w~n',[W1,File]),
    open(File,read,Stream),reset_line_nr,
    read_tokens(Stream),close(Stream),
    statistics(walltime,[_,W2]), format('~n% Walltime ~w ms to tokenise ~w~n',[W2,File]),
    print_memory_used_wo_gc.

% set_prolog_flag(profiling,on)
tok_test(1) :-
    tok_read('/Users/leuschel/git_root/prob_examples/public_examples/B/Demo/Bakery0.prob').
tok_test(2) :-
    tok_read('/Users/leuschel/git_root/prob_examples/public_examples/B/Benchmarks/CAN_BUS_tlc.prob').
tok_test(3) :-
    tok_read('/Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated4000.prob').
tok_test(4) :-
    tok_read('/Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated10000.prob').
tok_test(7) :-
    tok_read('/Users/leuschel/git_root/private_examples/ClearSy/2019_Nov/rule_perf4/rule_split.prob').

read_tokens(Stream) :-
   get_next_token(Stream,Tok), %format('Token: ~w~n',[Tok]),
   !,
   (Tok = eof -> true
     ; read_tokens(Stream)).
read_tokens(Stream) :-
   get_code(Stream,C),
   format('Could not process: ~w~n',[C]).


% get next Prolog token from a stream
% TO DO: produce a version without peek, which returns first code following token
get_next_token(Stream,Token) :- get_code(Stream,C), !,
   get_next_token1(C,Stream,Token).

get_next_token1(-1,_,eof).
get_next_token1(32,Stream,Token) :- !, get_next_token(Stream,Token).
get_next_token1(10,Stream,Token) :- !, inc_line_nr, get_next_token(Stream,Token).
get_next_token1(13,Stream,Token) :- !, get_next_token(Stream,Token).
get_next_token1(0'(,_,open_paren). % adding a cut here seems to degrade performance, but leads to backtracking
get_next_token1(0'),_,close_paren).
get_next_token1(0'[,_,open_bracket).
get_next_token1(0'],_,close_bracket).
get_next_token1(0',,_,comma).
get_next_token1(0'.,_,dot).
get_next_token1(39,Stream,atom(T)) :- !, get_quoted_atom_codes(Stream,T).
get_next_token1(X,Stream,number([X|T])) :- is_digit(X),!, get_number_codes(Stream,T).
get_next_token1(X,Stream,atom([X|T])) :- is_lowcase(X),!, get_atom_codes(Stream,T).
get_next_token1(X,_,_) :- prolog_code_error(X),fail.

get_number_codes(Stream,[H|T]) :- peek_code(Stream,H), is_digit(H),
     !, get_code(Stream,_), get_number_codes(Stream,T).
get_number_codes(_,[]).
% TO DO: also accept floats

get_atom_codes(Stream,[H|T]) :-
   peek_code(Stream,H), is_unquoted_atom_letter(H),!,
   get_code(Stream,_), get_atom_codes(Stream,T).
get_atom_codes(_,[]).

get_quoted_atom_codes(Stream,Res) :- get_code(Stream,H),
   (  H = 92 -> get_code(Stream,H2), % backslash: read next code and treat it specially
                treat_escaped_code(H2,Stream,Res)
    ; H = 39 -> Res=[]
    ; H = -1 -> format('Unterminated quoted atom: ~w~n',[H]),fail
    ; Res=[H|T], get_quoted_atom_codes(Stream,T)).

% see page 60, Section 4.1.7.6 Escape Sequences in SICStus Prolog manual
treat_escaped_code(Code,Stream,[NewCode|T]) :- escape_conversion(Code,NewCode),!,
    get_quoted_atom_codes(Stream,T).
treat_escaped_code(Octal,Stream,[OctResult|T]) :- is_octal_digit(Octal,Nr),!, % treat chars like \374\
   get_code(Stream,NxtCode),
   get_octal_codes(NxtCode,Stream,Nr,OctResult),
   get_quoted_atom_codes(Stream,T).
treat_escaped_code(0'x,Stream,[HexResult|T]) :- !, % treat chars like \x21D4\
   get_code(Stream,NxtCode),
   get_hex_codes(NxtCode,Stream,0,HexResult),
   get_quoted_atom_codes(Stream,T).
treat_escaped_code(10,Stream,T) :- !, %\LFD ignored ; TO DO: also support 13
   get_quoted_atom_codes(Stream,T).
treat_escaped_code(-1,_,_) :- !, format('Unterminated quoted atom ending with backslash at: ~w~n',[end_of_file]),fail.
treat_escaped_code(X,_,_) :- prolog_code_error('escape sequence',X),fail.

escape_conversion(39,39). % single quote '
escape_conversion(34,34). % double quote "
escape_conversion(92,92).
escape_conversion(0'n,10).
escape_conversion(0't,9).
escape_conversion(0'b,8).
escape_conversion(0'r,13).


get_octal_codes(H,Stream,Acc,Res) :- is_octal_digit(H,HNr),
     !, A2 is Acc*8+HNr,
     get_code(Stream,NxtCode), get_octal_codes(NxtCode,Stream,A2,Res).
get_octal_codes(92,_,Acc,Res) :- !, Res=Acc.
get_octal_codes(X,_,_,_) :- prolog_code_error('octal literal',X),fail.

get_hex_codes(H,Stream,Acc,Res) :- is_hex_digit(H,HNr),
     !, A2 is Acc*16+HNr,
     get_code(Stream,NxtCode), get_hex_codes(NxtCode,Stream,A2,Res).
get_hex_codes(92,_,Acc,Res) :- !, Res=Acc.
get_hex_codes(X,_,_,_) :- prolog_code_error('hex literal',X),fail.



is_digit(X) :- X>=48, X=<57.
:- assert_must_succeed(tools_fastread:is_octal_digit(0'7,7)).
is_octal_digit(X,Nr) :- X>=48, X=<55, Nr is X-48.
:- assert_must_succeed(tools_fastread:is_hex_digit(0'f,15)).
:- assert_must_succeed(tools_fastread:is_hex_digit(0'F,15)).
:- assert_must_succeed(tools_fastread:is_hex_digit(0'9,9)).
is_hex_digit(X,Nr) :- X>=48, X=<57, Nr is X-48.
is_hex_digit(X,Nr) :- X>=65, X=<70, Nr is X-55. % upper-case A-F
is_hex_digit(X,Nr) :- X>=97, X=<102, Nr is X-87. % lower-case a-f
is_lowcase(X) :- X>=97, X=<122.

is_unquoted_atom_letter(X) :- X>=97, X=<122.
is_unquoted_atom_letter(X) :- X>=65, X=<90.
is_unquoted_atom_letter(X) :- X=95.
is_unquoted_atom_letter(X) :- X>=48, X=<57.

%is_quoted_atom_letter(X) :- X \= 39. %, X \=10, X \=13.

% skip over whitespace and return code after whitspace (peeked)
skip_ws(Stream,NextCode) :- peek_code(Stream,C),
  (is_skip_whitespace(C) ->  get_code(Stream,_),skip_ws(Stream,NextCode)
   ; NextCode=C).

is_skip_whitespace(32).
is_skip_whitespace(10) :- inc_line_nr.
is_skip_whitespace(13).

% ------------------------

expect_token(Stream,ETok) :- get_next_token(Stream,RTok),!,
  (RTok = ETok -> true
    ; cur_line(L),
      format(user_error,'Unexpected token on line ~w: ~w, expected ~w~n',[L,RTok,ETok]),
      fail).

prolog_token_error(Tok) :- cur_line(L),
   %format(user_error,'Unexpected token on line ~w: ~w~n',[L,Tok]),
   add_error(fastread,'Unexpected token on line:',Tok:L).
prolog_code_error(Code) :- cur_line(L),
   format(user_error,'Unexpected character on line ~w: ~w "~s"~n',[L,Code,[Code]]),
   add_error(fastread,'Unexpected character code on line:',Code:line(L)).
prolog_code_error(Kind,Code) :- cur_line(L),
   format(user_error,'Unexpected character for ~w on line ~w: ~w "~s"~n',[Kind,L,Code,[Code]]),
   add_error(fastread,'Unexpected character code on line:',Kind:Code:line(L)).

:- dynamic cur_line/1.
cur_line(1).
inc_line_nr :- retract(cur_line(N)), N1 is N+1, assert(cur_line(N1)).
reset_line_nr :- retract(cur_line(_)), assert(cur_line(1)).

% ------------------------

:- assert_must_succeed(tools_fastread:read_term_from_string("atom('a',1,2).",atom(a,1,2))).
:- assert_must_succeed(tools_fastread:read_term_from_string("'\\''.",'\'')).
:- assert_must_succeed(tools_fastread:read_term_from_string("'\\374\\'.",'\374\')).

:- use_module(library(codesio),[open_codes_stream/2]).
read_term_from_string(Codes,Term) :-
   open_codes_stream(Codes,Stream),
   read_term_from_stream(Stream,Term),
   close(Stream).


% read next term on stream, term has to be followed by '.'
read_term_from_stream(Stream,Term) :-
   get_next_token(Stream,Token),!,
   get_next_term(Token,Stream,Term),!,
   (Term='end_of_file' -> true
     ; expect_token(Stream,dot)).

% get one term, without dot at end:
get_next_term(eof,_,'end_of_file').
get_next_term(number(C),_,Term) :- number_codes(Term,C).
get_next_term(atom(C),Stream,Term) :- atom_codes(Atom,C), complete_atom(Atom,Stream,Term).
get_next_term(open_bracket,Stream,List) :- get_term_list(Stream,close_bracket,List).
get_next_term(open_paren,Stream,'$pair'(List)) :- get_term_list(Stream,close_paren,List).
get_next_term(close_paren,_,_) :- prolog_token_error(dot),fail.
get_next_term(close_bracket,_,_) :- prolog_token_error(dot),fail.
get_next_term(comma,_,_) :- prolog_token_error(dot),fail.
get_next_term(dot,_,_) :- prolog_token_error(dot),fail.

% check if an atom is followed by a list of arguments (Term1,...)
complete_atom(Atom,Stream,Term) :- skip_ws(Stream,C), C==0'(, !,
   get_code(Stream,_),
   % TO DO: special case for binary operators like conjunct, transform into flat list (associativity holds) meaning we can get tail-recursion
   get_term_list(Stream,close_paren,List),
   Term =.. [Atom|List].
complete_atom(Atom,_,Atom). % atom without parentheses, i.e., on its own

% get list of terms separated by , and terminated by provided token
get_term_list(Stream,EndToken,List) :- get_next_token(Stream,Token), get_term_list_aux(Token,Stream,EndToken,List).

get_term_list_aux(EndToken,_,EndToken,List) :- !, List=[].
get_term_list_aux(Token,Stream,EndToken,[Term1|TList]) :- get_next_term(Token,Stream,Term1),
   get_next_token(Stream,Token2),!,
   get_term_list_aux2(Token2,Stream,EndToken,TList).

get_term_list_aux2(EndToken,_,EndToken,List) :- !, List=[].
get_term_list_aux2(comma,Stream,EndToken,[Term1|TList]) :-
   get_next_token(Stream,Token1),
   get_next_term(Token1,Stream,Term1),!,
   get_next_token(Stream,Token2),
   get_term_list_aux2(Token2,Stream,EndToken,TList).
get_term_list_aux2(Token2,_,EndToken,_) :-
   format(user_error,'Unexpected token: ~w, expected ~w~n',[Token2,EndToken]),fail.

term_read(File) :- statistics(walltime,_),
    open(File,read,Stream),read_terms(Stream),close(Stream),
    statistics(walltime,[_,W1]), format('~n% Walltime ~w ms to tokenise and read ~w~n',[W1,File]),
    print_memory_used_wo_gc.

read_terms(Stream) :-
   read_term_from_stream(Stream,Term),
   %read(Stream,Term), % Prolog's read
   write_term(Term,[max_depth(5),numbervars(true)]),nl,
   !,
   (Term = 'end_of_file' -> true
     ; read_terms(Stream)).
read_terms(Stream) :-
   get_code(Stream,C),
   format('Could not process: ~w~n',[C]).

read_test(1) :-
    term_read('/Users/leuschel/git_root/prob_examples/public_examples/B/Demo/Bakery0.prob').
read_test(3) :-
    term_read('/Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated4000.prob').
read_test(4) :-
    term_read('/Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated10000.prob').

read_test(7) :-
    term_read('/Users/leuschel/git_root/private_examples/ClearSy/2019_Nov/rule_perf4/rule_split.prob').

read_test(8) :-
    term_read('/Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/SATLIB/sudoku.prob').



/*

with new read_term_from_stream:

| ?- read_test(4).
parser_version(2019-03-05 17:22:26.861)
classical_b(Generated10000,[/Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated10000.mch])
machine(abstract_machine(pos(1,1,1,1,50007,4),machine(pos(2,1,1,1,1,8)),machine_header(pos(3,1,1,9,1,23),Generated10000,[]),[constants(pos(...),[...|...]),properties(pos(...),conjunct(...)),variables(...)|...]))
end_of_file

% Walltime 774 ms to tokenise /Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated10000.prob
 108.476 MB

with fastreadwrite:
% Walltime 89 ms to fastread from ./fastrw_out.probz
 109.004 MB


with Prolog read
| ?- rtest4.
parser_version(2019-03-05 17:22:26.861)
classical_b(Generated10000,[/Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated10000.mch])
machine(abstract_machine(pos(1,1,1,1,50007,4),machine(pos(2,1,1,1,1,8)),machine_header(pos(3,1,1,9,1,23),Generated10000,[]),[constants(pos(...),[...|...]),properties(pos(...),conjunct(...)),variables(...)|...]))
end_of_file

% Walltime 4193 ms to tokenise /Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated10000.prob
 286.429 MB


% Walltime 102 ms to get_codes /Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated4000.prob

% Walltime 233 ms to tokenise /Users/leuschel/git_root/prob_examples/public_examples/B/PerformanceTests/Generated/Generated4000.prob
 116.813 MB ( 11.956 MB program)



 | ?- rtest7.
parser_version(23928fd10a7a580ccd609bf614ed4552107bbd6a)
classical_b(rule,[/Users/leuschel/git_root/private_examples/ClearSy/2019_Nov/rule_perf4/rule_split.mch,/Users/leuschel/git_root/prob_prolog/stdlib/CHOOSE.def,/Users/leuschel/git_root/prob_prolog/stdlib/LibraryBits.def,/Users/leuschel/git_root/prob_prolog/stdlib/LibraryCSV.def|...])
machine(abstract_machine(pos(4687857,17,1,1,79,4),machine(pos(4687858,17,1,1,1,8)),machine_header(pos(4687859,17,1,9,1,21),LibraryFiles,[]),[abstract_constants(pos(...),[...|...]),properties(pos(...),conjunct(...)),definitions(...)|...]))
machine(abstract_machine(pos(4688547,18,1,1,94,4),machine(pos(4688548,18,1,1,1,8)),machine_header(pos(4688549,18,1,9,1,20),LibraryMath,[]),[abstract_constants(pos(...),[...|...]),definitions(pos(...),[...|...]),properties(...)|...]))
machine(abstract_machine(pos(4689324,19,1,1,176,4),machine(pos(4689325,19,1,1,1,8)),machine_header(pos(4689326,19,1,9,1,23),LibraryStrings,[]),[abstract_constants(pos(...),[...|...]),properties(pos(...),conjunct(...)),definitions(...)|...]))
machine(abstract_machine(pos(1,1,1,1,1929826,4),machine(pos(2,1,1,1,1,8)),machine_header(pos(3,1,2,5,2,9),rule,[]),[definitions(pos(...),[...|...]),sees(pos(...),[...|...]),constants(...)|...]))
end_of_file

% Walltime 31964 ms to tokenise /Users/leuschel/git_root/private_examples/ClearSy/2019_Nov/rule_perf4/rule_split.prob
 659.087 MB
yes

% Walltime 6243 ms to fastread from ./fastrw_out.probz
 659.025 MB


% Walltime 5676 ms to get_codes /Users/leuschel/git_root/private_examples/ClearSy/2019_Nov/rule_perf4/rule_split.prob

% Walltime 15616 ms to tokenise /Users/leuschel/git_root/private_examples/ClearSy/2019_Nov/rule_perf4/rule_split.prob
 116.840 MB ( 11.982 MB program)

 */

