% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)
% Flow analysis
% Filename: flow.pl

:- module(flow, [create_enable_graph/1, create_enable_graphs/1, reset_flow/0, print_base_guide/1,print_event_guide/2]).

:- use_module(self_check).

:- use_module(module_information,[module_info/2]).
:- module_info(group,experimental).
:- module_info(description,'Flow Analysis for Event-B models. Extracts a flow graph from a model.').

:-use_module(bmachine,[b_top_level_operation/1,b_get_invariant_from_machine/1,wp_untyped/3,clear_wp/0,nonchanging_guard/2]).
:-use_module(bsyntaxtree).

:- use_module(translate, [translate_bexpression/2]).

:- dynamic eventnumber/2.
:- dynamic guide_wp/3.
:- dynamic wp_typed/3, event_wp/3.

:- use_module(error_manager).
:- use_module(bmachine,[typecheck_predicates/4]).

add_wp(S,D,TypedPred) :-
    assert( wp_typed(S,D,TypedPred) ).
update_typed_wp :-
    retract( wp_untyped(S,D,P) ),
    bmachine:typecheck_predicates(P,TypedPred,S,D),
    add_wp(S,D,TypedPred),
    fail.
update_typed_wp.

wp(S,D,TypedPred) :-
    update_typed_wp,
    wp_typed(S,D,TypedPred).

% ---------------------------------- Interface

:- meta_predicate print_single_enable_graph(*,3).
:- meta_predicate print_single_enable_graph(*,3,*).

create_enable_graph(X):- build_enable_graph,
                         print_single_enable_graph(X,wp).

create_enable_graphs(X):- build_enable_graph,
                          print_multiple_enable_graphs(X).

print_base_guide(X) :- build_enable_graph,
                       print_single_enable_graph(X,guide_wp).

print_event_guide(X,E) :- build_enable_graph,
                          retractall(event_wp(_,_,_)),
                          build_guide(E,N),
                          print_single_enable_graph(X,event_wp,N).


reset_flow :- retractall(done_flow(_)), retractall(guide_wp(_,_,_)).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_flow,
                           'Reset flow analysis.').

:- public backward_analysis/1. % still useful ??
backward_analysis(Event) :- build_enable_graph,
                            print_flow_automaton(Event).

% -------------------------------- Flow Automaton

print_flow_automaton(E) :- calculate_flow_automaton([E],[],X), print(X), nl.

calculate_flow_automaton([],A,A).
calculate_flow_automaton([H|T],A,R) :- findall((S,P),(wp(S,H,P), \+false_predicate(P)),L), append(L,A,NA), calculate_flow_automaton(T,NA,R).


% -------------------------------- Enable Graph



:- dynamic done_flow/1.

build_enable_graph :- (done_flow(enable_graph) -> true;
                       ((print('Computing enable_graph'),nl,
                         enumerate_events,
                         prepare_input,
                         find_contradictions,
                         add_init_edges,
                         law_of_motion,
                         build_base_guide,
                         print('Done'),nl)
                         -> assert(done_flow(enable_graph))
                         ; add_error(flow,'Could not build enable_graph'),fail)).

print_single_enable_graph(File,NameEdges) :- events(Nodes), print_single_enable_graph(File,NameEdges,Nodes).

print_single_enable_graph(File,NameEdges,Nodes) :-
                    print(writing_to_dot_file(File)),nl,
                    tell(File),
                    dot_header('EnableGraph'),
                    dot_node(Nodes),
                    findall((S,D,P), call(NameEdges,S,D,P),WL),
                    dot_edge(WL),
                    dot_footer,
                    told,
                    print(closing(File)),nl.

print_multiple_enable_graphs(File) :-
                    print(writing_to_dot_file(File)),nl,
                    tell(File),
                    dot_header('EnableGraphs'), print('rankdir=LR'),nl, events(L), dot_node(L), print_subgraph(L),
                    dot_footer, told,
                    print(closing(File)),nl.

print_subgraph([]).
print_subgraph([H|T]) :-  findall((H,D,P), wp(H,D,P),WL),
  print_edges(WL),
  print_subgraph(T).

% -------------------------------- Prepare Input / Normalize

prepare_input :- statistics(runtime,_),
                  normalize_inv,
              normalize_wp,
                  statistics(runtime,[_,T2]), print(normalizing(T2)),nl.

normalize_wp :- findall((S,D,P), wp(S,D,P), L), clear_wp, normalize_wps(L).

normalize_wps([]).
normalize_wps([(S,D,P)|T]) :- conjunct_predicates(P,P3),conjunction_to_list(P3,P2),normalize_list(P2,R), add_wp(S,D,R), normalize_wps(T).


normalize_inv :- b_get_invariant_from_machine(Invariant),
                 conjunction_to_list(Invariant,L),
                 normalize_invs(L).
normalize_invs([]).
normalize_invs([H|T]) :- translate_bexpression(H,HP),
     print(normalizing_inv(HP)),nl, normalize(H,HX), get_texpr_expr(HX,HR),
     ((HR =.. [Op,X,Y] ,keep(Op)) -> assert_pred_inv(X,Op,Y);true),
     normalize_invs(T).

normalize_guards(E) :- findall((E,X),nonchanging_guard(E,X), L),
                       print(mormalizing_grd(E,L)),nl,
                       normalize_guards2(L).

normalize_guards2([]).
normalize_guards2([(E,P)|T]) :-  normalize_list(P,R), assert_guards(R,E), normalize_guards2(T).

assert_guards([],_).
assert_guards([H|T],E) :- get_texpr_expr(H,truth),!, assert_guards(T,E).
assert_guards([H|T],E) :- get_texpr_expr(H,HR),
   HR =.. [Op,X,Y],!, (keep(Op) -> assert_pred_guard(Op,X,Y);true), assert_guards(T,E).
assert_guards([H|_],_) :- add_error(flow,'Cannot normalize: ',H),fail.

% ------------------------------ Normalizer

normalize(TExpr,TNExpr) :- remove_bt(TExpr,Expr,NExpr,TNExpr), normalize2(Expr,NExpr).

normalize2(not_equal(X,boolean_true),equal(X,boolean_false)) :- !.
normalize2(not_equal(X,boolean_false),equal(X,boolean_true)) :- !.
normalize2(negation(TExpr),R) :- !,
   remove_bt(TExpr,Expr,NExpr,_), negate_boolean_expression(Expr,NExpr), normalize2(NExpr,R).
normalize2(TExpr, R) :-
   syntaxtransformation(TExpr,TSubExpr,_,TNSubExpr,R),!, normalize_list(TSubExpr, TNSubExpr).
normalize2(TExpr,_) :- add_error(flow,'Unknown expression in normalize2: ',TExpr),fail.

normalize_list([],R) :- !,R=[].
normalize_list([H|T],[HR|TR]) :- !, normalize(H,HR),normalize_list(T,TR).
normalize_list(A,B) :- add_error(flow,'Illegal arguments in normalize_list: ',(A,B)),fail.

/** keep is used to specify which invariants shall be kept
    (we keep only those parts that can be handled by the builtin prover) */
keep(boolean_false).
keep(boolean_true).
keep(equal).
keep(not_equal).
keep(member).
keep(not_member).
keep(less).
keep(greater).
keep(less_equal).
keep(greater_equal).

negate_boolean_expression(A,B) :-
   if(neg(A,B),true, add_error_fail(flow,'Unknown operator in negate: ',neg(A,B))).


neg(boolean_true,boolean_false).
neg(boolean_false,boolean_true).
neg(equal(X,Y),not_equal(X,Y)).
neg(not_equal(X,Y),equal(X,Y)).
neg(equal(X,Y),not_equal(X,Y)).
neg(member(X,Y),not_member(X,Y)).
neg(not_member(X,Y),member(X,Y)).
neg(less(X,Y),greater_equal(X,Y)).
neg(less_equal(X,Y),greater(X,Y)).
neg(greater_equal(X,Y),less(X,Y)).
neg(greater(X,Y),less_equal(X,Y)).
neg(subset_strict(X,Y),not_subset_strict(X,Y)).
neg(not_subset_strict(X,Y),subset_strict(X,Y)).
neg(subset(X,Y),not_subset(X,Y)).
neg(not_subset(X,Y),subset(X,Y)).


:- assert_must_succeed((flow:neg(boolean_true, boolean_false))).
:- assert_must_succeed((flow:neg(boolean_false, boolean_true))).

:- assert_must_succeed((flow:neg(equal(1,3), not_equal(1,3)))).
:- assert_must_succeed((flow:neg(not_equal(1,3), equal(1,3)))).

:- assert_must_succeed((flow:neg(member(1,{2}), not_member(1,{2})))).
:- assert_must_succeed((flow:neg(not_member(1,{2}), member(1,{2})))).

:- assert_must_succeed((flow:neg(greater(1,2), less_equal(1,2)))).
:- assert_must_succeed((flow:neg(greater_equal(1,2), less(1,2)))).
:- assert_must_succeed((flow:neg(less(1,2), greater_equal(1,2)))).
:- assert_must_succeed((flow:neg(less_equal(1,2), greater(1,2)))).

:- assert_must_succeed((flow:neg(not_subset_strict(foo,bar), subset_strict(foo,bar)))).
:- assert_must_succeed((flow:neg(subset_strict(foo,bar), not_subset_strict(foo,bar)))).
:- assert_must_succeed((flow:neg(subset(foo,bar), not_subset(foo,bar)))).
:- assert_must_succeed((flow:neg(not_subset(foo,bar), subset(foo,bar)))).


% -------------------------------- Init Edges
add_init_edges :- events(L), add_init_edges(L).
add_init_edges([]).
add_init_edges(['INITIALISATION'|T]) :- !, add_init_edges(T).
add_init_edges([H|T]) :- H\='INITIALISATION',(wp('INITIALISATION',H,_)->true;add_wp('INITIALISATION',H,[])),add_init_edges(T).

% -------------------------------- Law of motion
law_of_motion :- events(L), law_of_motion(L).
law_of_motion([]).
law_of_motion(['INITIALISATION'|T]) :- !, law_of_motion(T).
law_of_motion([H|T]) :- H\='INITIALISATION',(wp(H,H,_)->true;add_wp(H,H,[])),law_of_motion(T).


% -------------------------------- Base Guiding Automaton

build_base_guide :- findall((S,D,P),wp(S,D,P),L), build_base_guide(L).
build_base_guide([]).
build_base_guide([(S,D,P)|T]) :- (false_predicate(P) -> true; assert(guide_wp(S,D,P))), build_base_guide(T).

build_guide(E,RV) :-  guide_dfs([E],RV), retractall(event_wp(_,_,_)),filter_from_guide(RV).

guide_dfs(Stack, Visited) :- guide_dfs(Stack,[],Visited).

guide_dfs([],SeenSoFar,SeenSoFar).
guide_dfs([Event|Tail],SeenSoFar,Visited) :- member(Event,SeenSoFar),!, guide_dfs(Tail,SeenSoFar,Visited). % ignore!
guide_dfs([Event|Tail],SeenSoFar,Visited) :- findall(S,guide_wp(S,Event,_),L), append(L,Tail,Todo), guide_dfs(Todo,[Event|SeenSoFar],Visited).


filter_from_guide(V) :- guide_wp(S,D,P), member(S,V), member(D,V), assert(event_wp(S,D,P)), fail.
filter_from_guide(_).


% -------------------------------- Flow Graph





% --------------------------------- Prover
:- dynamic pred_grd/3, pred_new/3.
:- dynamic pred_guard/3.
:- dynamic pred_inv/3.
:- dynamic change/0.
:- dynamic inconsistent/0.
%:- dynamic trans/3.

false_predicate([b(boolean_false,boolean,[nodeid(none)])]).

find_contradictions :- findall((S,D,P), wp(S,D,P), L), clear_wp, find_contradictions(L).

find_contradictions([]).
find_contradictions([(S,D,P)|T]) :- retractall(inconsistent),
                                    retractall(pred_grd(_,_,_)),
                                    retractall(pred_new(_,_,_)),
                                    retractall(pred_guard(_,_,_)),
                                    normalize_guards(S),
                                    find_contradictions2(P),
                                    (inconsistent -> false_predicate(F), add_wp(S,D,F)
                                    ;   otherwise -> (simplify(P,PR), add_wp(S,D,PR))),
                                    find_contradictions(T).

find_contradictions2([]).
find_contradictions2([H|_]) :- get_texpr_expr(H,boolean_false), !, assert(inconsistent).
find_contradictions2([H|T]) :- get_texpr_expr(H,HR), (HR =.. [Op,X,Y] -> assert_pred(Op,X,Y); true), (\+inconsistent -> huelle, find_contradictions2(T);true).


huelle :- pred(Op,X,Y),pred(Op2,Y,Z), combine(Op,Op2,Op3),
         assert_pred(Op3,X,Z),fail.
huelle :- (retract(change) -> huelle2 ; true).

huelle2 :- retract(pred_new(X,Op,Y)), huelle2(X,Op,Y).
huelle2 :- (retract(change) -> huelle2 ; true).

huelle2(X,Op,Y) :- pred(Op2,Y,Z), combine(Op,Op2,Op3),  assert_pred(Op3,X,Z),fail.
huelle2(Y,Op,Z) :- pred(Op2,X,Y), combine(Op2,Op,Op3),  assert_pred(Op3,X,Z),fail.

assert_pred(Op,X,Y) :- %print(try(Op,X,Y)),nl,
                        pred(Op,X,Y) -> true
                      ; (assert(pred_grd(X,Op,Y)), %print(pred_grd(Op,X,Y)),nl,
                         assert(pred_new(X,Op,Y)),
                         (change -> true ; assert(change)),

                         %nl,print(check_inconsistent(Op,X,Y)),nl,

                         (inconsistent_fact(Op,X,Y) -> (assert(inconsistent), print('INCONSISTENT'), nl) ; true),
                         ((sym(Op),X\=Y)
                          -> assert(pred_grd(Y,Op,X)), assert(pred_new(X,Op,Y)) ; true)).

assert_pred_inv(Op,Y,Z) :-
    pred_inv(Y,Op,Z) -> true
    ; assert(pred_inv(Y,Op,Z)),
      ((sym(Op),Z\=Y) -> assert(pred_inv(Z,Op,Y)) ; true).
assert_pred_guard(Op,Y,Z) :-
    pred_guard(Y,Op,Z) -> true
    ; assert(pred_guard(Y,Op,Z)),
      ((sym(Op),Z\=Y) -> assert(pred_guard(Z,Op,Y)) ; true).

pred(Op,Y,Z) :- pred_grd(Y,Op,Z).
pred(Op,Y,Z) :- pred_inv(Y,Op,Z).
pred(Op,Y,Z) :- pred_guard(Y,Op,Z).

inconsistent_fact(not_equal,X,X).
inconsistent_fact(less,X,X).
inconsistent_fact(greater,X,X).
inconsistent_fact(equal,X,Y) :-  value(X), value(Y), X \= Y.
inconsistent_fact(subset_strict,X,X).
inconsistent_fact(member,_,empty_set).

:- assert_must_succeed((flow:inconsistent_fact(not_equal,4,4))).
:- assert_must_succeed((flow:inconsistent_fact(less,4,4))).
:- assert_must_succeed((flow:inconsistent_fact(subset_strict,{4},{4}))).
:- assert_must_succeed((flow:inconsistent_fact(member,4,empty_set))).

:- assert_must_succeed((flow:value2(4))).
:- assert_must_succeed((flow:value2(boolean_false))).
:- assert_must_succeed((flow:value2(boolean_true))).
:- assert_must_succeed((flow:value2(integer(_,7)))).

value(X) :- get_texpr_expr(X,Y), value2(Y).

value2(X) :- number(X).
value2(boolean_false).
value2(boolean_true).
value2(integer(_,_)).

sym(equal).
sym(not_equal).

combine(equal,X,X).
combine(less_equal,less,less).
combine(less_equal,less_equal,less_equal).
combine(less,less,less).
combine(greater,greater,greater).
combine(greater_equal,greater,greater).
combine(greater_equal,greater_equal,greater_equal).
combine(subset_strict,subset_strict,subset_strict).

% ------------------------------ Simplifier

simplify([],[]).
simplify([H|T],R) :- (member(H,T) -> R=RT; %remove duplicates
                         otherwise -> (tautology(H) -> R=RT; %remove tautologies
                                          otherwise -> R=[H|RT])),
                      simplify(T,RT).

tautology(X) :- get_texpr_expr(X,Y), tautology2(Y).

tautology2(boolean_true).
tautology2(equal(A,A)).
tautology2(subset(A,A)).
tautology2(less_equal(A,A)).
tautology2(greater_equal(A,A)).


% -------------------------------- Helper
 :- dynamic count/1.
reset_count :- retractall(count(_)), assert(count(0)).
inc :- count(X), retract(count(_)), X1 is X+1, assert(count(X1)).

print_edges([]).
print_edges([(S,D,P)|T]) :- print_node(S,D,C,X),  format_edge(C,X,P), print_edges(T).

print_node(S,D,C,X) :- eventnumber(S,C), count(X), inc, format_node(X,D).

enumerate_events :- reset_count, events(L), enumerate_events(L).
enumerate_events([]).
enumerate_events([S|T]) :- count(C), assert(eventnumber(S,C)), inc, enumerate_events(T).

events(['INITIALISATION'|L]) :- findall(OperationName, b_top_level_operation(OperationName), L).

dot_header(Name) :- print('digraph '), print(Name),print(' {'),nl.

dot_node([]).
dot_node([S|T]) :- eventnumber(S,C), format_node(C,S), dot_node(T).

format_node(Number,Text) :- print(Number), print(' '), print('[label="'), print(Text), print('"]'),nl.

dot_edge([]).
dot_edge([(S,D,P)|T]) :- eventnumber(S,SN), eventnumber(D,DN),  format_edge(SN,DN,P), dot_edge(T).

format_edge(FromNumber,ToNumber,[]) :- format_edge2(FromNumber,ToNumber, green).
format_edge(FromNumber,ToNumber,F) :-  false_predicate(F),!,format_edge2(FromNumber,ToNumber, red).
format_edge(FromNumber, ToNumber, LP) :- conjunct_predicates(LP,P),translate_bexpression(P,Label), print(FromNumber), print('->'), print(ToNumber), print(' '), print('[label="'), print(Label), print('"]') ,nl.

format_edge2(FromNumber, ToNumber, green) :- print(FromNumber), print('->'), print(ToNumber), print(' '), print('[color="forestgreen" label=true]') ,nl.
format_edge2(FromNumber, ToNumber, red) :- print(FromNumber), print('->'), print(ToNumber), print(' '), print('[color="crimson" label=false]') ,nl.

dot_footer :- print('}').


% ---------------------------------- Main Program


% build_flow_graph:-  (open(State) -> (process(State,State), retract(open(State)), count(C), inc, assert(node(State,C)), build_flow_graph);
%                      otherwise -> true).

% :- use_module(library(ordsets)).
% process([],_).
% process([H|T],State) :- findall((D,P),trans(H,D,P),L), add_edges(L,State),
%                        process(T,State).

%join_nonchanging(X,OS,NS) :- list_to_ord_set(X,OX),no_effect(X,NE),list_to_ord_set(NE,ONE), ord_intersection(OX,ONE,NonChanging),list_to_ord_set(OS,Changed),ord_union(Changed,NonChanging,NS).

% add_edges([],_).
% add_edges([(Dest,Pred)|T],Src) :- assert(edge(Src,Dest,Pred)), assert_open(Dest), add_edges(T,Src).

% process([],A,A).
% process([H|T],A,R) :- trans(H,_,S), append(S,A,NS), process(T,NS,R).

% assert_open(R) :- ( open(R) -> true; (node(R,_) -> true ; assert(open(R)))).

% no_effect(X,R) :- events(All),list_to_ord_set(All,OA),no_effect(X,OA,R).
% no_effect([],A,A).
% no_effect([H|T],A,R) :- findall(Y,wp(H,Y,_),Z), list_to_ord_set(Z,OE), ord_subtract(A,OE,NA), no_effect(T,NA,R).


% ---------------------------------- Output


% print_dot2(Name) :- atom_codes(Name,NameString), append(NameString,"_components.dot",FileName),atom_codes(File,FileName),tell(File),
%                     dot_header(Name),findall((S,D,P), trans(S,D,P),WL), print_transitions(WL),
%                   dot_footer, told.


% print_flow(Name) :- atom_codes(Name,NameString), append(NameString,"_flow.dot",FileName),atom_codes(File,FileName),tell(File),dot_header(Name),
%                    flow_nodes, flow_edges,
%                    dot_footer, told.

% flow_nodes :- findall((S,C),node(S,C),L), flow_node(L).
% flow_node([]).
% flow_node([(S,C)|T]) :- print(C), print(' '), print('[label="'), print(S), print('"]'),nl,flow_node(T).
% flow_edges :- findall((S,D,P), edge(S,P,D), L), flow_edge(L).
% flow_edge([]).
% flow_edge([(S,P,D)|T]) :- node(S,SN), node(D,DN), print(SN), print('->'), print(DN), print(' '), print('[label="'),
%                           print_pred(P),
%                           print('"]') ,nl,flow_edge(T).

% :- dynamic graph2/2.

% print_transitions([]).
% print_transitions([(S,D,P)|T]) :- put_node(S,SN), put_node(D,DN),
%                                  print(SN), print('->'), print(DN), print(' '), print('[label="'), print(P), print('"]') ,nl, print_transitions(T).

%put_node(X,XN) :- graph2(X,XN),!.
% put_node(X,XN) :- count(XN),inc, assert(graph2(X,XN)), print(XN), print(' '), print('[label="'), print(X), print('"]'),nl.

