% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(units_interpreter_helpers, [units_build_initial_state/2,
                                      assign_vars/6,
                                      assign_var/6,
                                      state_store_list/3,
                                      state_store/3,
                                      merge_non_overriding_states/3,
                                      assign_vals_to_vars/3,
                                      merge_expression_list/2,
                                      execute_statements_in_parallel/5,
                                      execute_else_list/4,
                                      execute_select_list/4,
                                      execute_cases_list/5,
                                      execute_choice_list/4,
                                      execute_sequence/6,
                                      execute_while/5,
                                      execute_eventb_list/5,
                                      units_lookup_value/4,
                                      units_set_up_localstate/3,
                                      units_set_up_undefined_localstate/3,
                                      b_type_to_internal_type/3
                                     ]).

:- use_module(library(codesio)).
:- use_module(library(lists)).
:- use_module(library(samsort)).

:- use_module(probsrc(bmachine)).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(b_global_sets)).

:- use_module(units_interpreter).
:- use_module(units_domain).
:- use_module(units_alias).
:- use_module(unit_parser).
:- use_module(units_tools,[convert_to_atom/2]).

:- use_module(probsrc(module_information)).
:- module_info(group,plugin_units).
:- module_info(description,'Units Plugin: Helper Predicates').

:- dynamic cached_global_type/2.

% value lookup: Id to search for, local state, global state and return variable
units_lookup_value(Id,S1,S2,Val) :-
    statistics(walltime,_),
    statistics(runtime,_),
    units_lookup_value2(Id,S1,S2,Val),
    statistics(walltime,[_,Duration]),
    statistics(runtime,[_,RDuration]),
    (Duration < 20 -> true ; length(S1,LS1), length(S2,LS2),
      format('Warning: lookup of ~w took ~w (~w) ms (Val=~w,LS1=~w,LS2=~w)~n', [Id, Duration,RDuration,Val,LS1,LS2])).
units_lookup_value2(Id,LocalState,GlobalState,Value) :-
    (member(bind(Id,Val),LocalState) -> Value=Val
     ; member(bind(Id,Val),GlobalState) -> Value=Val
     ; units_lookup_global_value(Id,Value)).

units_lookup_global_value(Id,Val) :-
    lookup_global_constant(Id,fd(_,Set)), !,
    b_type_to_internal_type(global(Set),[],Val).
units_lookup_global_value(Id,Val) :-
    b_global_set(Id), !,
    b_type_to_internal_type(set(global(Id)), [], Val).
%units_lookup_global_value(Id,Val) :-
%    b_get_machine_all_constants(Csts),
%    member(b(identifier(Id),Type,Infos), Csts), !,
%    b_type_to_internal_type(Type,Infos,Val).
units_lookup_global_value(Id,_Val) :- add_error_fail(lookup_value, 'Variable could not be found: ', Id).

execute_statements_in_parallel([],_Local,State, State, []).
execute_statements_in_parallel([Stmt|TS],Local,StateIn, StateOut, [PathH|PathT]) :-
    units_execute_statement(Stmt, Local,StateIn, StateOut1, PathH),
    execute_statements_in_parallel(TS, Local,StateIn, StateOut2, PathT),
    merge_non_overriding_states(StateOut1,StateIn,TempStateOut),
    merge_non_overriding_states(StateOut2,TempStateOut,StateOut).

execute_else_list([],_,_State,[]).
execute_else_list([TExpr|Rest],LocalState,InState,OutState) :-
    get_texpr_expr(TExpr,if_elsif(Test,Body)),
    units_check_boolean_expression(Test,LocalState,InState),
    execute_else_list2(Body,Rest,LocalState,InState,OutState).
execute_else_list2(Body,Rest,LocalState,InState,OutState) :-
    units_execute_statement(Body,LocalState,InState,OutState1,_Path),
    execute_else_list(Rest,LocalState,InState,OutState2),
    merge_non_overriding_states(OutState1,OutState2,OutState).

execute_select_list([],_,_State,[]).
execute_select_list([TExpr|Rest],LocalState,InState,OutState) :-
    get_texpr_expr(TExpr,select_when(Test,Body)),
    units_check_boolean_expression(Test,LocalState,InState),
    execute_select_list2(Body,Rest,LocalState,InState,OutState).
execute_select_list2(Body,Rest,LocalState,InState,OutState) :-
    units_execute_statement(Body,LocalState,InState,OutState1,_Path),
    execute_select_list(Rest,LocalState,InState,OutState2),
    merge_non_overriding_states(OutState1,OutState2,OutState).

execute_cases_list(_ToCompare,[],_,_State,[]).
execute_cases_list(ToCompare,[TExpr|Rest],LocalState,InState,OutState) :-
    get_texpr_expr(TExpr,case_or(Expr,Subst)),
    units_compute_expressions(Expr,LocalState,InState,Results),
    merge_expression_list(Results,MergedResult),
    same_units(ToCompare,MergedResult),
    units_execute_statement(Subst,LocalState,InState,OutState1,_Path),
    execute_cases_list(ToCompare,Rest,LocalState,InState,OutState2),
    merge_non_overriding_states(OutState1,OutState2,OutState).

execute_choice_list([],_,_State,[]).
execute_choice_list([Stmt|Rest],LocalState,InState,OutState) :-
    units_execute_statement(Stmt,LocalState,InState,OutState1,_Path),
    execute_choice_list(Rest,LocalState,InState,OutState2),
    merge_non_overriding_states(OutState1,OutState2,OutState).

execute_sequence(First,UnwrappedRest,LocalState,InState,OutUpdates,Path) :-
    Path = sequence(Path1,RestPath),
    units_execute_statement(First,LocalState,InState,LHSUpdates,Path1),
    merge_non_overriding_states(InState,LHSUpdates,IntermediateState),
    execute_statement_when_state_set(UnwrappedRest,LocalState,IntermediateState,LHSUpdates,OutUpdates,RestPath).

execute_eventb_list([], _LocalState, InState, InState, []).
execute_eventb_list([Stmt|Stmts],LocalState,InState,OutState, [Path1|Paths]) :-
    units_execute_statement(Stmt,LocalState,InState,LHSUpdates,Path1),
    merge_non_overriding_states(InState,LHSUpdates,IntermediateState),
    execute_eventb_list(Stmts,LocalState,IntermediateState,OutState,Paths).

execute_while(Condition,Statement,LocalState,InState,OutState) :-
    units_check_boolean_expression(Condition,LocalState,InState),
    units_execute_statement(Statement,LocalState,InState,PossibleOutState,_Path),
    ( PossibleOutState = InState ->
         OutState = PossibleOutState
      ;  execute_while(Condition,Statement,LocalState,PossibleOutState,OutState)
    ).

execute_statement_when_state_set(Stmt,LS,State,UpdatesSofar,Res,Path) :-
   units_execute_statement2(Stmt,[],LS,State,StmtUpdates,Path),
   merge_non_overriding_states(UpdatesSofar,StmtUpdates,Res).

merge_expression_list([U], U).
merge_expression_list([U1|T],U) :-
    merge_expression_list(T,UList),
    lub(U1,UList,U).

merge_non_overriding_states([],State,State).
merge_non_overriding_states([bind(Key,Val)|T], State, [bind(Key,Result)|OutTail]) :-
    (
      selectchk(bind(Key,Val2),State,SOut)
      -> lub(Val,Val2,Result)
      ;  Result=Val, SOut = State
    ),
    merge_non_overriding_states(T,SOut,OutTail).

units_set_up_localstate(Parameters,ParaValues,LocalState) :-
    units_build_initial_state(Parameters,TypedLocalState),
    maplist(bind_values,TypedLocalState,ParaValues,LocalState).

bind_values(bind(Id,Val),Val2,bind(Id,Val3)) :-
    lub(Val,Val2,Val3).

units_build_initial_state(BTypeList,InitialState) :-
    units_build_initial_state2(BTypeList, UnsortedInitialState),
    samsort(UnsortedInitialState,InitialState).
units_build_initial_state2([],[]) :- !.
units_build_initial_state2([b(identifier(Name),Type,Args)|T],[bind(Name,InternalType)|StateT]) :-
    %length(T,Len), format('Still ~w vars to go to build initial state~n', [Len]),
    b_type_to_internal_type(Type,Args,InternalType), !,
    units_build_initial_state(T,StateT).

units_set_up_undefined_localstate([],S,S).
units_set_up_undefined_localstate([Var|T],InS,[bind(Id,_)|OutS]) :-
   def_get_texpr_id(Var,Id),
   units_set_up_undefined_localstate(T,InS,OutS).

get_unit_from_pragma(Args,Res) :-
    member(unit(Unit),Args), !,
    unit_args_to_list(Unit, Res).
get_unit_from_pragma(Args,[[0,AtomUnit,1]]) :-
    member(new_unit(Unit),Args), !,
    convert_to_atom(Unit,AtomUnit).
get_unit_from_pragma(_,_).

b_type_to_internal_type(boolean, Args, boolean(Res)) :-
    !, get_unit_from_pragma(Args,Res).
b_type_to_internal_type(string, Args, string(Res)) :-
    !, get_unit_from_pragma(Args,Res).
b_type_to_internal_type(integer, Args, integer(Res)) :-
    !, get_unit_from_pragma(Args,Res).
b_type_to_internal_type(global(X), _Args, Result) :-
    cached_global_type(X,Res), !, Result=global(Res,X).
b_type_to_internal_type(global(X), Args, global(Res,X)) :-
    !, full_b_machine(machine(_Name,Parts)),
    member(deferred_sets/DeferredSetsList, Parts),
    member(enumerated_sets/EnumeratedSetsList, Parts),
    append(DeferredSetsList,EnumeratedSetsList, AllGlobalSetsList),
    find_global_unit_in_list(X,AllGlobalSetsList, UnitFromSet),
    get_unit_from_pragma(Args,UnitFromPragma),
    (UnitFromSet = UnitFromPragma -> Res = UnitFromPragma ; fail),
    assert(cached_global_type(X,Res)).
b_type_to_internal_type(set(X), Args, set(Res)) :-
    !, b_type_to_internal_type(X,Args,Res).
b_type_to_internal_type(seq(X), Args, set(couple(integer(_),Res))) :-
    !, b_type_to_internal_type(X,Args,Res).
b_type_to_internal_type(couple(X,Y),Args,couple(XRes,YRes)) :-
    !, b_type_to_internal_type(X,Args,XRes),
    b_type_to_internal_type(Y,Args,YRes).
b_type_to_internal_type(record(Fields), Args, rec(UnitFields)) :-
    !, maplist(b_type_to_internal_type_maplist(Args),Fields,UnitFields).
b_type_to_internal_type(field(Name,Type),Args,field(Name,UType)) :-
    !, b_type_to_internal_type(Type,Args,UType).
b_type_to_internal_type(X, _Args, _Res) :- !,
    add_error_fail(b_type_to_internal_type, 'Can not construct internal type for: ', [X]).

b_type_to_internal_type_maplist(Args,A,B) :-
    b_type_to_internal_type(A,Args,B).

find_global_unit_in_list(_Global,[],_Res).
find_global_unit_in_list(Global,[H|_T],Res) :-
    H = b(identifier(Global),set(global(Global)),Infos), !,
    (member(unit(Unit),Infos) -> unit_args_to_list(Unit, Res) ; true).
find_global_unit_in_list(Global,[_H|T],Res) :-
    find_global_unit_in_list(Global,T,Res).

assign_vars([], [], _LocalState, _StateIn, [], []).
assign_vars([TVarH|TVarT], [TValH|TValT], LocalState, StateIn, [NewVarH|NewVarT], [NewValH|NewValT]) :-
    get_texpr_expr(TVarH,VarH),
    assign_var(VarH, TValH, LocalState,StateIn, NewVarH, NewValH),
    assign_vars(TVarT, TValT, LocalState,StateIn, NewVarT, NewValT).

assign_var(identifier(Id),Value,_LS,_S,Id,Value) :- !.
% dead code?
%assign_var(function(Fun,Arg),Value,LocalState,InState,Id,FullValue) :-
%    !, units_compute_expression(Fun,LocalState,InState,FunVal),
%    units_compute_expression(Arg,LocalState,InState,ArgVal),
%    lub(FunVal,set(couple(ArgVal,Value)),NewFun),
%    get_texpr_expr(Fun,Lhs),
%    assign_var(Lhs,NewFun,LocalState,InState,Id,FullValue).

state_store_list([], State, State).
state_store_list([H|T], StateIn, StateOut) :-
    state_store(H, StateIn, StateIn1),
    state_store_list(T, StateIn1, StateOut).

state_store(bind(Key,Value), [], []) :-
    !, add_error_fail(units_state_store, 'Variable could not be found in state: ', [Key,Value]).
state_store(bind(Key,Value), [bind(Key,X)|Rest], [bind(Key,Value2)|Rest]) :- var(X), !, Value2 = Value.
state_store(bind(Key,integer(Value1)), [bind(Key,integer(Value2))|Rest], [bind(Key,integer(Value3))|Rest]) :- !, lub(Value1,Value2,Value3).
state_store(bind(Key,set(Value1)), [bind(Key,set(Value2))|Rest], [bind(Key,set(Value3))|Rest]) :- !, lub(Value1,Value2,Value3).
state_store(bind(Key,couple(Value11,Value12)), [bind(Key,couple(Value21,Value22))|Rest],
           [bind(Key,couple(Value31,Value32))|Rest]) :- !, lub(Value11,Value21,Value31), lub(Value12,Value22,Value32).
state_store(bind(Key,global(Value1,Global)), [bind(Key,global(Value2,Global))|Rest],
            [bind(Key,global(Value3,Global))|Rest]) :- !, lub(Value1,Value2,Value3).
state_store(bind(Key,rec(Fields1)), [bind(Key,rec(Fields2))|Rest], [bind(Key,rec(Fields3))|Rest]) :- !,
    maplist(lub,Fields1,Fields2,Fields3).
state_store(bind(Key,boolean(_)), [bind(Key,boolean(_))|Rest], [bind(Key,boolean(_))|Rest]) :- !.
state_store(bind(Key,string(_)), [bind(Key,string(_))|Rest], [bind(Key,string(_))|Rest]) :- !.
state_store(Pair, [Pair2|Rest], [Pair2|StateOut]) :-
    Pair \= Pair2, !,
    state_store(Pair, Rest, StateOut).

assign_vals_to_vars([], [], []).
assign_vals_to_vars([H1|T1], [H2|T2], [bind(H1,H2)|T]) :-
    assign_vals_to_vars(T1, T2, T).
